!> Computations of coefficients for the DLP and HLL-DLP flux
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_coeff_DLP
#if !defined(_TP_FLUX_)

  use mod_numerics, only: PR
  use type_mesh, only: dim_
  use type_mesh, only: pmesh, BndInterfaces
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_IntInterfaces
#endif /* _FC_PAR_OMP_ */
#if defined(_SRC_ENABLE_)
  use mod_asymptotic, only: g_u
#endif /* _SRC_ENABLE_ */
  use mod_physics_var, only: NbEq
  use mod_flux_1D, only: flux_1D
  use mod_bnd_values, only: compute_value_bnd
  use type_mesh_DLP, only: IntInterfaces, max_Card_S, NbM
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
  use type_IO, only: StdErr
  use mod_tools, only: plus1
  use type_mesh, only: nodes
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_quad, only: quad_seg
  use mod_HO_MOOD_poly, only: compute_polynomial_reconstruction
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

  implicit none

  private
  public compute_nu_all

  public compute_nu
  public compute_nu_inter
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
  public compute_nu_tilde_inter
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

#if defined(_SRC_ENABLE_)
  public compute_nu_bar
  public compute_nu_bar_inter
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
  public compute_nu_bar_tilde_inter
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
#endif /* _SRC_ENABLE_ */

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute all the \f$\nu\f$ coefficients for the scheme of order one,
  !> and the for the high-order scheme (if needed)
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_coeff_dlp_mp_compute_nu_inter
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_coeff_dlp_mp_compute_nu_tilde_inter
  subroutine compute_nu_all(U_n,t_n)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n

      real(PR), dimension(NbEq,max_Card_S,NbM) :: nu
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
      real(PR), dimension(NbEq,max_Card_S,quad_seg(MaxDegree)%nb_points,NbM) :: nu_tilde
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
      integer :: IdxIntInt
      integer :: Card_S

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_IntInterfaces)
      do IdxIntInt = 1, pmesh%NbIntInterfaces
        Card_S = IntInterfaces(IdxIntInt)%Card_S

        call compute_nu_inter(IdxIntInt,U_n,t_n,nu)
        IntInterfaces(IdxIntInt)%nu(:,1:Card_S,:) = nu(:,1:Card_S,:)

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        call compute_nu_tilde_inter(IdxIntInt,U_n,t_n,MaxDegree,nu_tilde)
        IntInterfaces(IdxIntInt)%nu_tilde(:,1:Card_S,:,:) = nu_tilde(:,1:Card_S,:,:)
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

      end do
      !$OMP END DO

  end subroutine compute_nu_all

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the \f$\nu\f$ coefficients for the scheme of order one
  !>
  !> @param[in] IdxIntInt Index of the interface
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[out] nu Coefficients
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_coeff_dlp_mp_compute_nu
  subroutine compute_nu_inter(IdxIntInt,U_n,t_n,nu)
      integer, intent(in) :: IdxIntInt
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), dimension(:,:,:), intent(out) :: nu

      real(PR), dimension(NbEq,max_Card_S,NbM) :: U_DLP
      integer :: CellJ
      integer :: IdxBndInt
      integer :: idxM, idxJ

      ! Get the mean values over the stencil of DLP
      loop_M: do idxM = 1, NbM
        loop_S_Ki: do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
          CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)
          test_bnd: if (CellJ > 0) then
            U_DLP(:,idxJ,idxM) = U_n(:,CellJ)
          else
            IdxBndInt = - CellJ
            U_DLP(:,idxJ,idxM) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
          end if test_bnd
        end do loop_S_Ki
      end do loop_M

      call compute_nu(IdxIntInt,U_DLP,nu)

  end subroutine compute_nu_inter

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the \f$\tilde{\nu}\f$ coefficients for the high-order scheme
  !>
  !> @param[in] IdxIntInt Index of the interface
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[in] degree Degree of the polynomial reconstruction
  !> @param[out] nu_tilde Coefficients
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_coeff_dlp_mp_compute_nu
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  subroutine compute_nu_tilde_inter(IdxIntInt,U_n,t_n,degree,nu_tilde)
      integer, intent(in) :: IdxIntInt
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      integer, intent(in) :: degree
      real(PR), dimension(:,:,:,:), intent(out) :: nu_tilde

      real(PR), dimension(NbEq,max_Card_S,NbM) :: nu
      real(PR), dimension(NbEq,max_Card_S,NbM) :: U_tilde
      real(PR), dimension(dim_) :: xQ, xA, xB
      integer :: Card_S
      integer :: i_K, i_K_p1
      integer :: CellK, CellJ
      integer :: IdxBndInt
      integer :: idxM, idxJ
      integer :: IdxQuad

      Card_S = IntInterfaces(IdxIntInt)%Card_S

      test_degree: if (degree >= 1 .and. degree <= 5) then
        CellK = IntInterfaces(IdxIntInt)%nghb(1)
        i_K = IntInterfaces(IdxIntInt)%i_Cell(1)
        i_K_p1 = plus1(i_K,pmesh%elt(CellK)%NbInterfaces)

        xA(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K))%x(:)
        xB(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K_p1))%x(:)

        loop_quad: do IdxQuad = 1, quad_seg(degree)%nb_points
          xQ(:) = quad_seg(degree)%bar_coord(1,IdxQuad)*xA(:) &
              + quad_seg(degree)%bar_coord(2,IdxQuad)*xB(:)

          ! Get the reconstructed values over the stencil of DLP
          loop_M: do idxM = 1, NbM
            loop_S_Ki: do idxJ = 1, Card_S
              CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)
              test_bnd: if (CellJ > 0) then
                call compute_polynomial_reconstruction(U_n,CellJ,degree,pmesh%elt(CellJ)%CoeffPolyMOOD,xQ,t_n,U_tilde(:,idxJ,idxM))
              else
                IdxBndInt = - CellJ
                CellJ = BndInterfaces(IdxBndInt)%IdxCell
                U_tilde(:,idxJ,idxM) = compute_value_bnd(IdxBndInt,U_n,xQ,t_n)
              end if test_bnd
            end do loop_S_Ki
          end do loop_M

          call compute_nu(IdxIntInt,U_tilde,nu)

          nu_tilde(:,1:Card_S,idxQuad,:) = nu(:,1:Card_S,:)

        end do loop_quad

      else
        write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong degree: ", degree
        flush(StdErr)
        STOP
      end if test_degree

  end subroutine compute_nu_tilde_inter
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$\nu\f$ coefficients for both sides of an interior interface
  !> using the values over the stencil
  !>
  !> @param[in] IdxIntInt Index of the interface
  !> @param[in] U_tilde Solution (reconstructed or not) at the points used for DLP
  !> @param[out] nu Coefficients
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_mp_flux_1d
  subroutine compute_nu(IdxIntInt,U_tilde,nu)
      integer, intent(in) :: IdxIntInt
      real(PR), dimension(:,:,:), intent(in)  :: U_tilde
      real(PR), dimension(:,:,:), intent(out) :: nu

      real(PR), dimension(NbEq) :: F_LJ, F_KJ
      real(PR), dimension(NbEq) :: U_L_tilde, U_K_tilde
      real(PR), dimension(NbEq) :: G_L, G_K
      real(PR), dimension(NbEq) :: den
      real(PR), dimension(dim_) :: eta, tau
      real(PR), dimension(NbM)  :: gamma_
      real(PR) :: beta
      integer :: CellK, CellL
      integer :: idxJ, CellJ, eq
      integer :: idxM
      integer :: Card_S

      beta = IntInterfaces(IdxIntInt)%beta
      Card_S = IntInterfaces(IdxIntInt)%Card_S

      do idxM = 1, NbM
        nu(:,1,idxM) = beta
        do idxJ = 2, Card_S
          nu(:,idxJ,idxM) = 0._PR
        end do
      end do

      test_adm: if (.not. IntInterfaces(IdxIntInt)%admissible) then
        idxJ = 1

        CellK = IntInterfaces(IdxIntInt)%nghb(1)
        CellL = IntInterfaces(IdxIntInt)%nghb(2)
        U_L_tilde(:) = U_tilde(:,idxJ,1)
        U_K_tilde(:) = U_tilde(:,idxJ,2)

        idxM = 1
        eta(:) = IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
        tau(:) = IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM)
        F_KJ(:) = flux_1D(U_K_tilde,U_L_tilde,eta,tau)
        G_K(:) = (IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM) - beta)*F_KJ(:)

        idxM = 2
        !eta(:) = IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
        !tau(:) = IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM)
        !F_LJ(:) = flux_1D(U_L_tilde,U_K_tilde,eta,tau)
        F_LJ(:) = - F_KJ(:)
        G_L(:) = (IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM) - beta)*F_LJ(:)


        do idxJ = 2, Card_S
          idxM = 1
          CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)
          eta(:) = IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
          tau(:) = IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM)
          F_KJ(:) = flux_1D(U_K_tilde,U_tilde(:,idxJ,idxM),eta,tau)
          G_K(:) = G_K(:) + IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM)*F_KJ(:)

          idxM = 2
          CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)
          eta(:) = IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
          tau(:) = IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM)
          F_LJ(:) = flux_1D(U_L_tilde,U_tilde(:,idxJ,idxM),eta,tau)
          G_L(:) = G_L(:) + IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM)*F_LJ(:)
        end do

        den(:) = abs(G_L) + abs(G_K)

        loop_eq: do eq = 1, NbEq
          test_neg: if (G_L(eq)*G_K(eq) < 0._PR) then
            gamma_(1) = abs(G_L(eq))/den(eq) ! gamma_K
            gamma_(2) = abs(G_K(eq))/den(eq) ! gamma_L

            loop_M: do idxM = 1, NbM
              nu(eq,1,idxM) = nu(eq,1,idxM) - 2._PR*beta*gamma_(idxM)
              loop_J: do idxJ = 1, Card_S
                nu(eq,idxJ,idxM) = nu(eq,idxJ,idxM) &
                    + 2._PR*gamma_(idxM)*IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM)
              end do loop_J
            end do loop_M

          end if test_neg
        end do loop_eq

      end if test_adm

  end subroutine compute_nu

#if defined(_SRC_ENABLE_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the \f$\bar{\nu}\f$ coefficients for the scheme of order one
  !>
  !> On the equation:
  !> \f[\partial_t(f(u)) - \text{div}(\alpha(u) \nabla g(u)) = 0 \f]
  !>
  !> @param[in] IdxIntInt Index of the interface
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[out] nu_bar Coefficients
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_coeff_dlp_mp_compute_nu_bar
  subroutine compute_nu_bar_inter(IdxIntInt,U_n,t_n,nu_bar)
      integer, intent(in) :: IdxIntInt
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), dimension(:,:), intent(out) :: nu_bar

      real(PR), dimension(NbEq,max_Card_S,NbM) :: U_DLP
      integer :: CellJ
      integer :: IdxBndInt
      integer :: idxM, idxJ

      ! Get the mean values over the stencil of DLP
      loop_M: do idxM = 1, NbM
        loop_S_Ki: do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
          CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)
          test_bnd: if (CellJ > 0) then
            U_DLP(:,idxJ,idxM) = U_n(:,CellJ)
          else
            IdxBndInt = - CellJ
            U_DLP(:,idxJ,idxM) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
          end if test_bnd
        end do loop_S_Ki
      end do loop_M

      call compute_nu_bar(IdxIntInt,U_DLP,nu_bar)

  end subroutine compute_nu_bar_inter

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the \f$\tilde{\bar{\nu}}\f$ coefficients for the HO scheme
  !>
  !> On the equation:
  !> \f[\partial_t(f(u)) - \text{div}(\alpha(u) \nabla g(u)) = 0 \f]
  !>
  !> @param[in] IdxIntInt Index of the interface
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[in] degree Degree of the polynomial reconstruction
  !> @param[out] nu_bar_tilde Coefficients
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_coeff_dlp_mp_compute_nu_bar
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  subroutine compute_nu_bar_tilde_inter(IdxIntInt,U_n,t_n,degree,nu_bar_tilde)
      integer, intent(in) :: IdxIntInt
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      integer, intent(in) :: degree
      real(PR), dimension(:,:,:), intent(out) :: nu_bar_tilde

      real(PR), dimension(max_Card_S,NbM) :: nu_bar
      real(PR), dimension(NbEq,max_Card_S,NbM) :: U_tilde
      real(PR), dimension(dim_) :: xQ, xA, xB
      integer :: Card_S
      integer :: i_K, i_K_p1
      integer :: CellK, CellJ
      integer :: IdxBndInt
      integer :: idxM, idxJ
      integer :: IdxQuad

      Card_S = IntInterfaces(IdxIntInt)%Card_S

      test_degree: if (degree >= 1 .and. degree <= 5) then
        CellK = IntInterfaces(IdxIntInt)%nghb(1)
        i_K = IntInterfaces(IdxIntInt)%i_Cell(1)
        i_K_p1 = plus1(i_K,pmesh%elt(CellK)%NbInterfaces)

        xA(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K))%x(:)
        xB(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K_p1))%x(:)

        loop_quad: do IdxQuad = 1, quad_seg(degree)%nb_points
          xQ(:) = quad_seg(degree)%bar_coord(1,IdxQuad)*xA(:) &
              + quad_seg(degree)%bar_coord(2,IdxQuad)*xB(:)

          ! Get the reconstructed values over the stencil of DLP
          loop_M: do idxM = 1, NbM
            loop_S_Ki: do idxJ = 1, Card_S
              CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)
              test_bnd: if (CellJ > 0) then
                call compute_polynomial_reconstruction(U_n,CellJ,degree,pmesh%elt(CellJ)%CoeffPolyMOOD,xQ,t_n,U_tilde(:,idxJ,idxM))
              else
                IdxBndInt = - CellJ
                U_tilde(:,idxJ,idxM) = compute_value_bnd(IdxBndInt,U_n,xQ,t_n)
              end if test_bnd
            end do loop_S_Ki
          end do loop_M

          call compute_nu_bar(IdxIntInt,U_tilde,nu_bar)

          nu_bar_tilde(1:Card_S,idxQuad,:) = nu_bar(1:Card_S,:)

        end do loop_quad

      else
        write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong degree: ", degree
        flush(StdErr)
        STOP
      end if test_degree

  end subroutine compute_nu_bar_tilde_inter
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the \f$\bar{\nu}\f$ coefficients of an interior interface
  !> using the values over the stencil
  !>
  !> On the equation:
  !> \f[\partial_t(f(u)) - \text{div}(\alpha(u) \nabla g(u)) = 0 \f]
  !>
  !> @param[in] IdxIntInt Index of the interface
  !> @param[in] U_tilde Solution (reconstructed or not) at the points used for DLP
  !> @param[out] nu_bar Coefficients
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_euler_mp_g_u
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_g_u
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1_mp_g_u
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_g_u
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_g_u
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_sw_mp_g_u
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_g_u
  subroutine compute_nu_bar(IdxIntInt,U_tilde,nu_bar)
      integer, intent(in) :: IdxIntInt
      real(PR), dimension(:,:,:), intent(in) :: U_tilde
      real(PR), dimension(:,:), intent(out) :: nu_bar

      real(PR), dimension(NbM)  :: gamma_
      real(PR) :: g_uK, g_uL
      real(PR) :: G_L, G_K
      real(PR) :: den
      real(PR) :: beta_bar
      integer :: CellK, CellL
      integer :: idxJ
      integer :: Card_S
      integer :: idxM

      beta_bar = IntInterfaces(IdxIntInt)%beta_bar
      Card_S = IntInterfaces(IdxIntInt)%Card_S

      do idxM  = 1, NbM
        nu_bar(1,idxM) = beta_bar
        do idxJ = 2, Card_S
          nu_bar(idxJ,idxM) = 0._PR
        end do
      end do

      test_adm: if (.not. IntInterfaces(IdxIntInt)%admissible) then
        idxJ = 1

        CellK = IntInterfaces(IdxIntInt)%nghb(1)
        CellL = IntInterfaces(IdxIntInt)%nghb(2)

        g_uL = g_u(U_tilde(:,idxJ,1))
        g_uK = g_u(U_tilde(:,idxJ,2))

        idxM = 1
        G_K = (IntInterfaces(IdxIntInt)%coeff_n_KJ(idxJ,idxM) - beta_bar)*(g_uK - g_uL)

        idxM = 2
        G_L = (IntInterfaces(IdxIntInt)%coeff_n_KJ(idxJ,idxM) - beta_bar)*(g_uL - g_uK)

        do idxJ = 2, Card_S
          idxM = 1
          G_K = G_K + IntInterfaces(IdxIntInt)%coeff_n_KJ(idxJ,idxM)&
              *(g_uK - g_u(U_tilde(:,idxJ,idxM)))

          idxM = 2
          G_L = G_L + IntInterfaces(IdxIntInt)%coeff_n_KJ(idxJ,idxM)&
              *(g_uL - g_u(U_tilde(:,idxJ,idxM)))
        end do

        den = abs(G_L) + abs(G_K)

        test_neg: if (G_L*G_K < 0._PR) then
          gamma_(1) = abs(G_L)/den ! gamma_K
          gamma_(2) = abs(G_K)/den ! gamma_L

          loop_M: do idxM = 1, NbM
            nu_bar(1,idxM) = nu_bar(1,idxM) - 2._PR*beta_bar*gamma_(idxM)
            loop_J: do idxJ = 1, Card_S
              nu_bar(idxJ,idxM) = nu_bar(idxJ,idxM) + 2._PR*gamma_(idxM)*IntInterfaces(IdxIntInt)%coeff_n_KJ(idxJ,idxM)
            end do loop_J
          end do loop_M

        end if test_neg

      end if test_adm

  end subroutine compute_nu_bar
#endif /* _SRC_ENABLE_ */

#endif /* .not. _TP_FLUX_ */
end module mod_coeff_DLP
