!> MOOD routines relative to the degree of the reconstruction
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_HO_MOOD_degree
#if defined(_HO_MOOD_)

  use mod_numerics, only: PR
  use type_IO, only: StdOut, StdErr
  use type_mesh, only: pmesh, BndInterfaces
#if defined(_MESH_PERIODIC_)
  use type_mesh, only: Periodic_BND
#endif /* _MESH_PERIODIC_ */
  use type_mesh_DLP, only: IntInterfaces
  use type_HO_MOOD, only: MaxDegree, MinDegree
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */

  use type_recompute, only: NbCellsInvalid, NbCellsToUpdate
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
  use type_recompute, only: NbCellsInvalidTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
  use type_recompute, only: CellsToUpdate
  use type_recompute, only: FlagCells

  use mod_detection, only: IsNumericallyAdmissible, DoesNotOscillate
#if defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_))
  use mod_detection, only: IsEntropic
#endif /* defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_)) */

  use mod_physics, only: IsPhysicallyAdmissible

  use mod_tools_sort, only: qsort

  implicit none
  private
  public computeIntFacePD
  public computeBndFacePD
  public GetCellsToUpdate

!#define _IO_EXTENSIVE_PRINT_

contains

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute degree of the polynomial reconstruction on an interior interface \c FacePD
  !>
  !> @param[in] IdxIntInt Index of an interior interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine computeIntFacePD(IdxIntInt)
      integer, intent(in) :: IdxIntInt

      integer :: CellK, CellL

      CellK = IntInterfaces(IdxIntInt)%nghb(1)
      CellL = IntInterfaces(IdxIntInt)%nghb(2)

      IntInterfaces(IdxIntInt)%FacePD = min(pmesh%elt(CellK)%CellPD,pmesh%elt(CellL)%CellPD)

  end subroutine computeIntFacePD

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute degree of the polynomial reconstruction on an boundary interface \c FacePD
  !>
  !> @param[in] IdxBndInt Index of an boundary interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine computeBndFacePD(IdxBndInt)
      integer, intent(in) :: IdxBndInt

      integer :: CellK, CellL

      CellK = BndInterfaces(IdxBndInt)%IdxCell

#if defined(_MESH_PERIODIC_)
      if (BndInterfaces(IdxBndInt)%BndTag(1) == Periodic_BND) then
        CellL = BndInterfaces(IdxBndInt)%IdxCellPeriodic
        BndInterfaces(IdxBndInt)%FacePD = min(pmesh%elt(CellK)%CellPD,pmesh%elt(CellL)%CellPD)
      else
        BndInterfaces(IdxBndInt)%FacePD = pmesh%elt(CellK)%CellPD
      endif
#else
      BndInterfaces(IdxBndInt)%FacePD = pmesh%elt(CellK)%CellPD
#endif

  end subroutine computeBndFacePD

  !--------------------------------------------------------------------!
  !> @brief
  !> Get the cells the not admissible cells (using the
  !> a posteriori criterion) and decrement their degree.
  !>
  !> @param[in] U_n, U_n_p1 Solution at previous time and current time
  !> @param[in] t_n, dt Current time and time step
  !> @param[in] NbReComp Number of re-computations done
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_isphysicallyadmissible

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_detection_mp_isnumericallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_detection_mp_isentropic
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_detection_mp_doesnotoscillate
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_degree_mp_decreasedegree
  subroutine GetCellsToUpdate(U_n,U_n_p1,t_n,dt,NbReComp)
      real(PR), dimension(:,:), intent(in) :: U_n, U_n_p1
      real(PR), intent(in) :: t_n, dt
      integer, intent(in) :: NbReComp

      integer :: CellK, idxK, i_K, CellL

#if defined(_HO_MOOD_FULL_DEC_)
      if (NbReComp == 0) then
        ! To decrement all the cells degree to the minimal degree
        ! (for debugging purpose)
        !$OMP SINGLE
        NbCellsInvalid = pmesh%NbCells
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        NbCellsInvalidTP = pmesh%NbCells
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        NbCellsToUpdate = pmesh%NbCells
        !$OMP END SINGLE NOWAIT
        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%Nbcells
          FlagCells(CellK) = 1
          CellsToUpdate(CellK) = CellK
          pmesh%elt(CellK)%CellPD = MinDegree
        end do
        !$OMP END DO
        return
      else
        !$OMP SINGLE
        NbCellsInvalid = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        NbCellsInvalidTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        NbCellsToUpdate = 0
        !$OMP END SINGLE
        return
      end if

#else /* _HO_MOOD_FULL_DEC_ */

      !$OMP SINGLE
      NbCellsInvalid = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      NbCellsInvalidTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      !$OMP END SINGLE

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt: do CellK = 1, pmesh%NbCells

        FlagCells(CellK) = 0

        test_not_already_min: if (pmesh%elt(CellK)%CellPD > MinDegree) then

          test_ND: if (.not. IsNumericallyAdmissible(U_n_p1(:,CellK))) then
            FlagCells(CellK) = 1
            call DecreaseDegree(CellK)
            cycle loop_elt
          endif test_ND

#if defined(_HO_MOOD_PAD_)
          test_PAD: if (.not. IsPhysicallyAdmissible(U_n_p1(:,CellK))) then
            FlagCells(CellK) = 1
            call DecreaseDegree(CellK)
            cycle loop_elt
          endif test_PAD
#endif /* _HO_MOOD_PAD_ */

#if defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_))
          test_entropic: if (.not. IsEntropic(CellK,U_n,U_n_p1,t_n,dt)) then
            FlagCells(CellK) = 1
            call DecreaseDegree(CellK)
            cycle loop_elt
          endif test_entropic
#endif /* defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_)) */

#if defined(_HO_MOOD_DMP_) || defined(_HO_MOOD_ED_)
          test_Oscillate: if (.not. DoesNotOscillate(CellK,U_n,U_n_p1)) then
            FlagCells(CellK) = 1
            call DecreaseDegree(CellK)
            cycle loop_elt
          end if test_Oscillate
#endif /* _HO_MOOD_DMP_ || _HO_MOOD_ED_ */

        end if test_not_already_min
      end do loop_elt
      !$OMP END DO

      !$OMP SINGLE
      test_has_invalid_cells: if (NbCellsInvalid > 0) then
        ! The neighbors of the invalid cells need to be updated
        NbCellsToUpdate = NbCellsInvalid
        loop_CellsInvalid: do idxK = 1, NbCellsInvalid
          CellK = CellsToUpdate(idxK)

          loop_nghb: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
            CellL = pmesh%elt(CellK)%nghb(i_K)
            test_not_bnd: if (CellL /= 0) then
              test_not_flagged: if (FlagCells(CellL) == 0) then
                ! Add the neighbor of the invalid cell
                FlagCells(CellL) = 1
                NbCellsToUpdate = NbCellsToUpdate + 1
                CellsToUpdate(NbCellsToUpdate) = CellL
              end if test_not_flagged
            end if test_not_bnd
          end do loop_nghb

        end do loop_CellsInvalid

        ! Sort the array CellsToUpdate to decrease memory jumps in the re-computations
        ! WARNING: do not do the same thing for the TP flux correction,
        ! as the routine assumed that the first cells are the invalid cells
        ! and then the cells to update.
        call qsort(NbCellsToUpdate,CellsToUpdate)

#if defined(_IO_EXTENSIVE_PRINT_)
        write(StdOut,'(A,I9)') 'Nb cells invalid         : ', NbCellsInvalid
        write(StdOut,'(A)')    'CellsToUpdate            : '
        write(Stdout,'(I9)') CellsToUpdate(1:NbCellsInvalid)
        write(StdOut,'(A)')    'CellsToRecompute         : '
        write(Stdout,'(I9)') CellsToUpdate(NbCellsInvalid+1:NbCellsToUpdate)
#endif /* _IO_EXTENSIVE_PRINT_ */
      endif test_has_invalid_cells
      !$OMP END SINGLE
#endif /* _HO_MOOD_FULL_DEC_ */

  end subroutine GetCellsToUpdate

  !--------------------------------------------------------------------!
  !> @brief
  !> Decrease the degree of a cell
  !>
  !> @param[in] CellK Index of the cell
  !>
  !> @warning For now the routine directly decrease to the degree 0
  !> (or -1 if the correction with the two points flux is activated)
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine DecreaseDegree(CellK)
      integer, intent(in) :: CellK

#if defined(_TP_CORREC_)
      if (pmesh%elt(CellK)%CellPD <= 0) then
        ! The scheme of order one failed, activate the TP flux correction
        pmesh%elt(CellK)%CellPD = -1
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        !$OMP ATOMIC
        NbCellsInvalidTP = NbCellsInvalidTP + 1
#endif /* (_HO_MOOD_MAX_DEGREE_ >= 1) */
      else !if (pmesh%elt(CellK)%CellPD > 0) then
        pmesh%elt(CellK)%CellPD = 0
      endif
#else /* _TP_CORREC_ */
      pmesh%elt(CellK)%CellPD = 0
#endif /* _TP_CORREC_ */

      !$OMP CRITICAL (CRITICAL_DecreaseDegree)
      NbCellsInvalid = NbCellsInvalid + 1
      CellsToUpdate(NbCellsInvalid) = CellK
      !$OMP END CRITICAL (CRITICAL_DecreaseDegree)

  end subroutine DecreaseDegree

#endif /* _HO_MOOD_ */
end module mod_HO_MOOD_degree
