!> MOOD computations for polynomial reconstructions
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_HO_MOOD_poly
#if defined(_HO_MOOD_)

  use mod_numerics, only: PR, zero, one
  use type_IO, only: StdErr
  use mod_tools, only: x_to_alpha
  use type_mesh, only: dim_
  use type_mesh, only: pmesh
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
  use type_HO_MOOD, only: alphas
  use mod_HO_MOOD_int, only: compute_I_KL
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
  use type_HO_MOOD, only: Nd_max
  use type_HO_MOOD, only: compute_Nd
  use mod_physics_var, only: NbEq
  use mod_physics, only: IsPhysicallyAdmissible
#if defined(_SRC_ENABLE_)
  use mod_asymptotic, only: compute_beta
#endif /* _SRC_ENABLE_ */

  implicit none
  private
  public compute_polynomial_coefficients
  public compute_polynomial_reconstruction
  public check_polynomial_reconstruction

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the polynomial reconstruction \f$\widetilde{\bf U}\f$ of \f${\bf U}\f$ on \f$K\f$
  !>
  !> @param[in] U_n Mean values at current time
  !> @param[in] CellK Current cell
  !> @param[in] degree Degree of the reconstruction
  !> @param[in] R_K Coefficients of the polynomial reconstruction
  !> @param[in] x Coordinate of reconstruction point
  !> @param[in] t_n Current time
  !> @param[out] U_K_tilde Polynomial reconstruction on K
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_x_to_alpha
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_int_mp_compute_i_kl
  !DIR$ ATTRIBUTES FORCEINLINE :: type_ho_mood_mp_compute_nd
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_check_polynomial_reconstruction
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_beta
  subroutine compute_polynomial_reconstruction(U_n,CellK,degree,R_K,x,t_n,U_K_tilde)
      real(PR), dimension(:,:), intent(in) :: U_n
      integer, intent(in) :: CellK
      integer, intent(in) :: degree
      real(PR), dimension(:,:), intent(in) :: R_K
      real(PR), dimension(:), intent(in) :: x
      real(PR), intent(in) :: t_n
      real(PR), dimension(:), intent(out) :: U_K_tilde

      real(PR), dimension(dim_) :: xK, xQ
      integer :: idx_alpha, Nd
#if defined(_SRC_ENABLE_)
      real(PR), dimension(NbEq) :: U_K
      real(PR) :: beta_K
#endif /* _SRC_ENABLE_ */

      U_K_tilde(:) = U_n(:,CellK)

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
      xK(:) = pmesh%elt(CellK)%x(:)
      Nd = compute_Nd(degree)

      do idx_alpha = 1, Nd
        xQ(:) = x(:) - xK(:)
        U_K_tilde(:) = U_K_tilde(:) + &
            R_K(:,idx_alpha)*(x_to_alpha(xQ, alphas(:,idx_alpha)) - compute_I_KL(CellK,CellK,alphas(:,idx_alpha)))
      end do

#if defined(_HO_MOOD_CHECK_POLY_REC_)
      call check_polynomial_reconstruction(U_n,CellK,U_K_tilde)
#endif /* _HO_MOOD_CHECK_POLY_REC_ */

#if defined(_SRC_ENABLE_)
      U_K(:) = U_n(:,CellK)
      beta_K = compute_beta(CellK,U_K,t_n)
      U_K_tilde(:) = beta_K*U_K_tilde(:) + (1._PR - beta_K)*U_K(:)
#endif /* _SRC_ENABLE_ */

#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

  end subroutine compute_polynomial_reconstruction

  !--------------------------------------------------------------------------
  !> @brief
  !> Check the polynomial reconstruction \f$\widetilde{\bf U}\f$ of \f${\bf U}\f$ on \f$K\f$
  !>
  !> @param[in] U_n Mean values at current time
  !> @param[in] CellK Current cell
  !> @param[inout] U_K_tilde Polynomial reconstruction on K, modified if not admissible
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_isphysicallyadmissible
  subroutine check_polynomial_reconstruction(U_n,CellK,U_K_tilde)
      real(PR), dimension(:,:), intent(in) :: U_n
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(inout) :: U_K_tilde

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
#if defined(_HO_MOOD_PAD_)
      if (.not. IsPhysicallyAdmissible(U_K_tilde)) then
        U_K_tilde(:) = U_n(:,CellK)
      end if
#endif /* _HO_MOOD_PAD_ */
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

  end subroutine check_polynomial_reconstruction

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the polynomial coefficients of the polynomial reconstruction
  !>
  !> @param[in] U_n Mean values at current time
  !> @param[in] CellK Current cell
  !> @param[out] R_K Coefficients of the polynomial reconstruction
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine compute_polynomial_coefficients(U_n,CellK,R_K)
      real(PR), dimension(:,:), intent(in) :: U_n
      integer, intent(in) :: CellK
      real(PR), dimension(:,:), intent(out) :: R_K

      real(PR), dimension(NbEq,pmesh%elt(CellK)%NbS_MOOD) :: U_S
      integer :: idxJ, CellJ
#if defined(_HO_MOOD_MATRIX_SCALING_)
      integer :: idx_alpha
      real(PR) :: scaling
#endif /* _HO_MOOD_MATRIX_SCALING_ */

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
      do idxJ = 1, pmesh%elt(CellK)%NbS_MOOD
        CellJ = pmesh%elt(CellK)%stencil_MOOD(idxJ)
        U_S(:,idxJ) = U_n(:,CellJ) - U_n(:,CellK)
      end do

#if defined(_ENABLE_LAPACK_)
      call DGEMM('N', 'N', NbEq, Nd_max, pmesh%elt(CellK)%NbS_MOOD, one, U_S, NbEq, pmesh%elt(CellK)%T_PI_X, pmesh%elt(CellK)%NbS_MOOD, zero, R_K, NbEq)
#else
      R_K(:,:) = MATMUL(U_S,pmesh%elt(CellK)%T_PI_X)
#endif /* _ENABLE_LAPACK_ */

#if defined(_HO_MOOD_MATRIX_SCALING_)
      scaling = 1._PR/sqrt(pmesh%elt(CellK)%area)
      do idx_alpha = 1, Nd_max
        R_K(:,idx_alpha) = R_K(:,idx_alpha)*scaling**sum(alphas(:,idx_alpha))
      end do
#endif /* _HO_MOOD_MATRIX_SCALING_ */

#else
      write(StdErr,'(2A,I4,A,A)')__FILE__,':', __LINE__, " -> Routine not defined with _HO_MOOD_MAX_DEGREE_ =", _HO_MOOD_MAX_DEGREE_
      flush(StdErr)
      STOP
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

  end subroutine compute_polynomial_coefficients

#endif /* _HO_MOOD_ */
end module mod_HO_MOOD_poly
