!> Parameters for the high order scheme, here MOOD \cite Dio_2012
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module type_HO_MOOD
#if defined(_HO_MOOD_)

  use type_mesh, only: dim_

  implicit none

  private
  public MaxDegree, MaxOrder
  public MinDegree
  public Nd_max, NbS_MOOD
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
  public alphas
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
  public compute_Nd

  !> Maximal degree of the polynomial reconstruction
  integer, parameter :: MaxDegree = _HO_MOOD_MAX_DEGREE_

  !> Minimal degree of the polynomial reconstruction
#if defined(_TP_CORREC_)
  integer, parameter :: MinDegree = - 1
#else
  integer, parameter :: MinDegree = 0
#endif /* _TP_CORREC_ */

  !> Maximal order of the scheme
  integer, parameter :: MaxOrder = MaxDegree + 1

  !> Maximal number of unknows \f$\mathcal{N}(d) = \frac{\prod_{i=1}^n (d+i)}{n!} - 1\f$,
  !> with \f$n\f$ the dimension.
  !> @warning The value coded is valid for a 2D code.
  integer, parameter :: Nd_max = ISHFT((MaxDegree + 1)*(MaxDegree + 2),-1) - 1

  !> Size of the stencil used for the reconstruction, it should be at least:
  !> \f$N_S = 1.5 \times \mathcal{N}(d)\f$ (for a 2D code)
  !integer, parameter :: NbS_MOOD = Nd_max
  !integer, parameter :: NbS_MOOD = Nd_max + ISHFT(Nd_max,-1)
  integer, parameter :: NbS_MOOD = ISHFT(Nd_max,1)

  !> Multi-index \f$\boldsymbol{\alpha}\f$
#if (_HO_MOOD_MAX_DEGREE_ == 1)
  integer, dimension(dim_,Nd_max), parameter :: alphas = reshape(SOURCE=(/1, 0, 0, 1/), &
                                                                          SHAPE=(/dim_, Nd_max/))
#elif (_HO_MOOD_MAX_DEGREE_ == 2)
  integer, dimension(dim_,Nd_max), parameter :: alphas = reshape(SOURCE=(/1, 0, 0, 1, &
                                                                          2, 0, 1, 1, 0, 2/), &
                                                                          SHAPE=(/dim_, Nd_max/))
#elif (_HO_MOOD_MAX_DEGREE_ == 3)
  integer, dimension(dim_,Nd_max), parameter :: alphas = reshape(SOURCE=(/1, 0, 0, 1, &
                                                                          2, 0, 1, 1, 0, 2, &
                                                                          3, 0, 2, 1, 1, 2, 0, 3/),&
                                                                          SHAPE=(/dim_, Nd_max/))
#elif (_HO_MOOD_MAX_DEGREE_ == 4)
  integer, dimension(dim_,Nd_max), parameter :: alphas = reshape(SOURCE=(/1, 0, 0, 1, &
                                                                          2, 0, 1, 1, 0, 2, &
                                                                          3, 0, 2, 1, 1, 2, 0, 3, &
                                                                          4, 0, 3, 1, 2, 2, 1, 3, 0, 4/), &
                                                                          SHAPE=(/dim_, Nd_max/))
#elif (_HO_MOOD_MAX_DEGREE_ == 5)
  integer, dimension(dim_,Nd_max), parameter :: alphas = reshape(SOURCE=(/1, 0, 0, 1, &
                                                                          2, 0, 1, 1, 0, 2, &
                                                                          3, 0, 2, 1, 1, 2, 0, 3, &
                                                                          4, 0, 3, 1, 2, 2, 1, 3, 0, 4, &
                                                                          5, 0, 4, 1, 3, 2, 2, 3, 1, 4, 0, 5/), &
                                                                          SHAPE=(/dim_, Nd_max/))
#endif /* _HO_MOOD_MAX_DEGREE_ */


contains

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the number of unknowns for the polynomial reconstruction
  !>
  !> @param[in] degree Degree of the polynomial reconstruction
  !> @return \f$\mathcal{N}(d) = \frac{(d+1)(d+2)}{2} - 1\f$,
  !>
  !> @warning Routine only valid in 2D
  !> @remark For a degree of \f$0\f$ (order one) or \f$-1\f$ (TP flux),
  !> this routine gives \f$\mathcal{N}(d) < 1\f$ and the loops begining at one
  !> will not be evaluated
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function compute_Nd(degree) result(Nd)
      integer, intent(in) :: degree
      integer :: Nd

      Nd = ISHFT((degree + 1)*(degree + 2),-1) - 1

  end function compute_Nd

#endif /* _HO_MOOD_ */
end module type_HO_MOOD
