!> Read file containing solution readed (for instance to use in a restart)
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_read

  use type_mesh, only: pmesh
  use mod_numerics, only: PR, MaxLenStr
  use mod_physics_var, only: NbEq, conservative_name
  use mod_open_file, only: OpenFile

#if defined(_ENABLE_HDF5_)
  use hdf5
#elif defined(_ENABLE_NC_)
  use netcdf
#endif /* _BINARY_FORMAT_ */

  implicit none

  private
  public read_solution
  public RestartFile, Restarting

  !> File to read for the restart
  character(len=MaxLenStr), save :: RestartFile
  !> Restart?
  logical, save :: Restarting

contains

  !--------------------------------------------------------------------!
  !> @brief
  !> Wrapper to read solution from plain or binary file (for instance to use in a restart)
  !>
  !> @param[out] U0 Solution at time \f$t^0\f$
  !> @param[out] NbIter  Number of iterations
  !> @param[out] T_init Initial time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine read_solution(U0,NbIter,T_init)
      real(PR), dimension(:,:), intent(out) :: U0
      integer, intent(out) :: NbIter
      real(PR), intent(out) :: T_init

#if defined(_ENABLE_HDF5_)
      call read_solution_h5(U0,NbIter,T_init)
#elif defined(_ENABLE_NC_)
      call read_solution_nc(U0,NbIter,T_init)
#else
      call read_solution_plain(U0,NbIter,T_init)
#endif /* _BINARY_FORMAT_ */
  end subroutine read_solution

  !--------------------------------------------------------------------!
  !> @brief
  !> Read solution from a plain file (for instance to use in a restart)
  !>
  !> @param[out] U0 Solution at time \f$t^0\f$
  !> @param[out] NbIter  Number of iterations
  !> @param[out] T_init Initial time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine read_solution_plain(U0,NbIter,T_init)
      real(PR), dimension(:,:), intent(out) :: U0
      integer, intent(out) :: NbIter
      real(PR), intent(out) :: T_init

      integer :: read_id
      integer :: CellK, eq

      call OpenFile(unit=read_id,file=RestartFile,action='read',status='old')
      !=================================================================!
      ! Three header lines
      do CellK = 1, 3
        read(read_id,*)
      end do
      read(read_id,'(I9,1X,ES23.16)') NbIter, T_init
      !=================================================================!
      read(read_id,*) ((U0(eq,CellK),eq=1,NbEq),CellK=1,pmesh%NbCells)
      CLOSE(read_id)

  end subroutine read_solution_plain

#if defined(_ENABLE_NC_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Read solution from a NetCDF \cite NetCDF, \cite NetCDF90
  !> file (for instance to use in a restart)
  !>
  !> @param[out] U0 Solution at time \f$t^0\f$
  !> @param[out] NbIter  Number of iterations
  !> @param[out] T_init Initial time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine read_solution_nc(U0,NbIter,T_init)
      real(PR), dimension(:,:), intent(out) :: U0
      integer, intent(out) :: NbIter
      real(PR), intent(out) :: T_init

      integer :: eq
      integer :: ncErr, ncId, ncVarId

      ! Open the file. NF90_NOwrite tells netCDF we want read-only access to
      ! the file.
      ncErr = nf90_open(RestartFile, NF90_NOwrite, ncId)

      ! Get the varid of the data variable, based on its name.
      ncErr = nf90_inq_varid(ncId, "Iterations", ncVarId)
      ! Read the data.
      ncErr = nf90_get_var(ncId, ncVarId, NbIter)

      ncErr = nf90_inq_varid(ncId, "Time", ncVarId)
      ncErr = nf90_get_var(ncId, ncVarId, T_init)

      do eq=1, NbEq
        ncErr = nf90_inq_varid(ncId, trim(conservative_name(eq)), ncVarId)
        ncErr = nf90_get_var(ncId, ncVarId, U0(eq,:))
      end do

      ! Close the file, freeing all resources.
      ncErr = nf90_close(ncId)
      !==================================================================!
  end subroutine read_solution_nc
#endif /* _ENABLE_NC_ */


#if defined(_ENABLE_HDF5_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Read solution from an HDF5 \cite HDF5 file (for instance to use in a restart)
  !>
  !> @param[out] U0 Solution at time \f$t^0\f$
  !> @param[out] NbIter  Number of iterations
  !> @param[out] T_init Initial time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine read_solution_h5(U0,NbIter,T_init)
      real(PR), dimension(:,:), intent(out) :: U0
      integer, intent(out) :: NbIter
      real(PR), intent(out) :: T_init

      integer :: eq

      integer :: h5Err
      integer(HSIZE_T), dimension(1) :: dims

      integer(HID_T) :: dset_id    ! Dataset identifier
      integer(HID_T) :: h5Id       ! File identifier

      !!==================================================================!
      ! Initialize FORTRAN interface.
      call h5open_f(h5Err)

      ! Open an existing file.
      call h5fopen_f(RestartFile, H5F_ACC_RDONLY_F, h5Id, h5Err)

      call h5dopen_f(h5Id, 'Iterations', dset_id, h5Err)
      call h5dread_f(dset_id, H5T_NATIVE_INTEGER, NbIter, dims, h5Err)
      call h5dclose_f(dset_id, h5Err)

      call h5dopen_f(h5Id, 'Time', dset_id, h5Err)
      call h5dread_f(dset_id, H5T_NATIVE_DOUBLE, T_init, dims, h5Err)
      call h5dclose_f(dset_id, h5Err)

      do eq = 1, NbEq
        call h5dopen_f(h5Id, trim(conservative_name(eq)), dset_id, h5Err)
        call h5dread_f(dset_id, H5T_NATIVE_DOUBLE, U0(eq,:), dims, h5Err)
        call h5dclose_f(dset_id, h5Err)
      end do

      ! Close the file.
      call h5fclose_f(h5Id, h5Err)

      ! Close FORTRAN interface.
      call h5close_f(h5Err)

      !==================================================================!
  end subroutine read_solution_h5
#endif /* _ENABLE_HDF5_ */

end module mod_read
