!> Outputs to screen and disk
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_write_output

  ! If the following line is uncommented the code will
  ! write a file for each time out_disk
!#define _TIME_NO_TIME_STEP_OUTPUTS_
#if !defined(_TIME_NO_TIME_STEP_OUTPUTS_)
#define _TIME_STEP_OUTPUTS_
#endif /* _TIME_NO_TIME_STEP_OUTPUTS_ */

  use mod_numerics, only: PR, MaxLenStr
  use type_IO, only: StdOut
  use mod_tools_time, only: GetTime, Print_Time_Stats
  use mod_tools_time, only: TimeCounters
  use mod_tools_error, only: PrintError
  use mod_physics_var, only: T_init, T_end

#if defined(_IO_ENABLE_VTK_)
  use mod_write_VTK, only: write_VTK_primary
  use type_IO, only: out_disk, n_out_disk
#endif /* _IO_ENABLE_VTK_ */
  use type_IO, only: out_screen, n_out_screen
#if defined(_IO_ENABLE_DUMP_)
  use mod_dump, only: dump_solution
  use type_IO, only: dump_disk, n_dump_disk
#endif /* _IO_ENABLE_DUMP_ */
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
  use type_recompute, only: PrintRecompStats
#endif /* _TP_CORREC_ || _HO_MOOD_ */

  implicit none

  private
  public write_output_init
  public write_output

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Initialize counters for writing
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine write_output_init()
#if defined(_IO_ENABLE_DUMP_)
    n_dump_disk = 0
#endif /* _IO_ENABLE_DUMP_ */
#if defined(_IO_ENABLE_VTK_)
    n_out_disk = 0
#endif /* _IO_ENABLE_VTK_ */
    n_out_screen = 0
  end subroutine write_output_init

  !--------------------------------------------------------------------------
  !> @brief
  !> Outputs to screen and disk
  !>
  !> @param[in] U_n_p1 Solution at time \f$t^{n+1}\f$
  !> @param[in] dt Time step
  !> @param[in] t_n_p1 Physical time
  !> @param[in] NbIter Number of iterations made
  !> @param[in] info Supplementary informations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine write_output(U_n_p1,dt,t_n_p1,NbIter,info)
    real(PR), dimension(:,:), intent(in) :: U_n_p1
    real(PR), intent(in) :: dt, t_n_p1
    integer, intent(in) :: NbIter
    character(len=*), intent(in) :: info

    real(PR) :: completion
    character(len=9) :: NbIter_str
    character(len=MaxLenStr) :: info_time

    completion = 100._PR*(t_n_p1 - T_init)/(T_end - T_init)
    write(NbIter_str,'(I9.9)') NbIter
    info_time = trim(info)//'_'//NbIter_str
    !$OMP SINGLE
    !--------------------------------------------------------------------!
    !                        Output on disk
    !--------------------------------------------------------------------!
#if defined(_IO_ENABLE_VTK_)
#if defined(_TIME_STEP_OUTPUTS_)
    if (floor(completion) >= n_out_disk) then
#endif /* _TIME_STEP_OUTPUTS_ */
      TimeCounters%IO(1) = GetTime()
      if (completion > 0._PR) n_out_disk = n_out_disk + out_disk
      call write_VTK_primary(U_n_p1,info_time,completion,NbIter,t_n_p1)
      TimeCounters%IO(2) = TimeCounters%IO(2) + GetTime() - TimeCounters%IO(1)
#if defined(_TIME_STEP_OUTPUTS_)
    end if
#endif /* _TIME_STEP_OUTPUTS_ */
#endif /* _IO_ENABLE_VTK_ */
    !--------------------------------------------------------------------!

    !--------------------------------------------------------------------!
    !              Dump solution for a (possible) restart
    !--------------------------------------------------------------------!
#if defined(_IO_ENABLE_DUMP_)
#if defined(_TIME_STEP_OUTPUTS_)
    if (floor(completion) >= n_dump_disk) then
#endif /* _TIME_STEP_OUTPUTS_ */
      TimeCounters%Dump(1) = GetTime()
      if (completion > 0._PR) n_dump_disk = n_dump_disk + dump_disk
      call dump_solution(U_n_p1,info_time,NbIter,t_n_p1)
      TimeCounters%Dump(2) = TimeCounters%Dump(2) + GetTime() - TimeCounters%Dump(1)
#if defined(_TIME_STEP_OUTPUTS_)
    end if
#endif /* _TIME_STEP_OUTPUTS_ */
#endif /* _IO_ENABLE_DUMP_ */
    !--------------------------------------------------------------------!

    !--------------------------------------------------------------------!
    !                        Output on screen
    !--------------------------------------------------------------------!
#if defined(_TIME_STEP_OUTPUTS_)
    if (floor(completion) >= n_out_screen) then
#endif /* _TIME_STEP_OUTPUTS_ */
      TimeCounters%Printing(1) = GetTime()
      if (completion > 0._PR) n_out_screen = n_out_screen + out_screen
      call PrintError(NbIter,dt,t_n_p1,U_n_p1)
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
      call PrintRecompStats(NbIter)
#endif /* _TP_CORREC_ || _HO_MOOD_ */
      call Print_Time_Stats(completion)
      flush(StdOut)
      TimeCounters%Printing(2) = TimeCounters%Printing(2) + GetTime() - TimeCounters%Printing(1)
#if defined(_TIME_STEP_OUTPUTS_)
    end if
#endif /* _TIME_STEP_OUTPUTS_ */
    !--------------------------------------------------------------------!
    !$OMP END SINGLE

  end subroutine write_output
end module mod_write_output
