!> Write solution and meshes in silo (https://wci.llnl.gov/simulation/computer-codes/silo) files for
!> VisIt (https://wci.llnl.gov/simulation/computer-codes/visit) \cite VisIt_2012
!> or ParaView (http://www.paraview.org/) \cite ParaView_2007 .
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_write_silo
#if defined(_ENABLE_SILO_) && defined(_MESH_P2D_)

  use mod_numerics, only: PR, MaxLenStr
  use type_IO, only: StdOut, StdErr
  use type_IO, only: Output_Directory
  use type_mesh, only: pmesh, nodes, BndInterfaces, dim_
  use type_mesh, only: basename_mesh_file
  use mod_physics_var, only: NbEq, primitive_name, test_case
  use mod_physics, only: conservative_to_primitive
#if defined(_HO_MOOD_) && defined(_SRC_ENABLE_)
  use mod_asymptotic, only: compute_beta
#endif /* _HO_MOOD_ && _SRC_ENABLE_ */

  implicit none

  include 'silo_f9x.inc'

  private
  public write_silo_primary

  !> Number of vertices in a segment
  integer, parameter :: NbVertEdge = 2
  !> Number of vertices in a triangle
  integer, parameter :: NbVertTri = 3
  !> Entity are count from 1
  integer, parameter :: origin = 1

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Write the primal mesh and the solution in a silo file
  !>
  !> @param[in] U_n_p1 Solution at time \f$t^{n+1}\f$
  !> @param[in] info Supplementary informations
  !> @param[in] completion Completion of the run
  !> @param[in] NbIter Number of iterations
  !> @param[in] t_n_p1 Physical time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_conservative_to_primitive

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_beta
  subroutine write_silo_primary(U_n_p1,info,completion,NbIter,t_n_p1)
    real(PR), dimension(:,:), intent(in) :: U_n_p1
    character(len=*), intent(in)         :: info
    integer, intent(in) :: NbIter
    real(PR), intent(in)                 :: completion, t_n_p1

    real(PR), dimension(NbEq) :: U_K
#if defined(_PHY_EULER_) || defined(_PHY_SW_)
    real(PR), dimension(NbEq) :: V_K ! To get back primitive variable
#endif /* _PHY_ */
    integer :: NbVer, CellK, i_K, eq, IdxBndInt, IdxVert
    character(len=2) :: NbVer_str

    ! Silo stuff
    integer :: dbid, flid, zlid, meshid, varid, optlistid
    integer :: err, ierr
    character (len=MaxLenStr) :: SiloFile
    integer :: len_SiloFile
    integer, dimension(:), allocatable :: fnodelist, znodelist
    real(PR), dimension(:), allocatable :: x, y, sol

    SiloFile='test.silo'!trim(Output_Directory)//'/'//trim(info)//'_primary.silo'
    len_SiloFile=len_trim(SiloFile)

    ! Create the silo file
    err = dbcreate(SiloFile, len_SiloFile, &
        DB_CLOBBER, DB_LOCAL, &
        trim(info), len_trim(info), &
        DB_PDB, &
        dbid)
    if (err /= 0) then
      write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, " -> silo error: ", err
      flush(StdErr)
      STOP
    end if

    allocate(fnodelist(NbVertEdge*pmesh%NbBndInterfaces))
    do IdxBndInt = 1, pmesh%NbBndInterfaces
      do IdxVert = 1, NbVertEdge
        fnodelist(NbVertEdge*(IdxBndInt - 1) + IdxVert) = BndInterfaces(IdxBndInt)%vertices(IdxVert)
      end do
    end do

    allocate(znodelist(3*pmesh%NbCells))
    do CellK = 1, pmesh%NbCells
      do i_K = 1, pmesh%elt(CellK)%NbInterfaces
        znodelist(pmesh%elt(CellK)%NbInterfaces*(CellK-1) + i_K) = pmesh%elt(CellK)%vertices_glob(i_K)
      end do
    end do

    ! Write the boundary interfaces
    !err = dbcalcfl (znodelist, pmesh%NbVert, origin, NbVertTri, pmesh%NbCells, 1, &
        !DB_F77NULL, 0, flid)
    !err = dbwrtfl (dbid, "Facelist", len('Facelist'), flid, ierr)

    err = dbputfl(dbid, 'Facelist', len('Facelist'), &
        pmesh%NbBndInterfaces, dim_, &
        fnodelist, NbVertEdge*pmesh%NbBndInterfaces, origin, &
        DB_F77NULL, NbVertEdge, pmesh%NbBndInterfaces, 1, &
        DB_F77NULL, DB_F77NULL, 0, flid)
    if (err /= 0) then
      write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, " -> silo error: ", err
      flush(StdErr)
      STOP
    end if

    ! Write the cells
    err = dbputzl2(dbid, 'Zonelist', len('Zonelist'), &
        pmesh%NbCells, dim_, &
        znodelist, NbVertTri*pmesh%NbCells, origin, &
        0, 0, &
        DB_ZONETYPE_TRIANGLE, NbVertTri*pmesh%NbCells, pmesh%NbCells, 1, &
        DB_F77NULL, zlid)
    if (err /= 0) then
      write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, " -> silo error: ", err
      flush(StdErr)
      STOP
    end if

    ! Add option
    err = dbmkoptlist(2, optlistid)
    err = dbaddiopt(optlistid, DBOPT_CYCLE, NbIter)
    err = dbaddropt(optlistid, DBOPT_TIME, t_n_p1)

    ! Write the mesh
    allocate(x(pmesh%NbVert))
    allocate(y(pmesh%NbVert))
    do IdxVert = 1, pmesh%NbVert
      x(IdxVert) = nodes(IdxVert)%x(1)
      y(IdxVert) = nodes(IdxVert)%x(2)
    end do
    !err = dbputum(dbid, trim(basename_mesh_file), len_trim(basename_mesh_file), dim_, &
    err = dbputum(dbid, 'tri', 3, dim_, &
        x, y, DB_F77NULL, &
        'X', len('X'), 'Y', len('Y'), DB_F77NULLSTRING, 0, &
        DB_DOUBLE, &
        pmesh%NbVert, pmesh%NbCells, &
        'Zonelist', len('Zonelist'),DB_F77NULLSTRING, 0, &
        !optlistid, meshid)
        DB_F77NULL, meshid)
    if (err /= 0) then
      write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, " -> silo error: ", err
      flush(StdErr)
      STOP
    end if

    ! Write the solutions
    allocate(sol(pmesh%NbCells))
    do eq = 1, NbEq
      do CellK = 1, pmesh%NbCells
        call conservative_to_primitive(U_n_p1(:,CellK),V_K)
        sol(CellK) = V_K(eq)
      end do
      err = dbputuv1 (dbid, trim(primitive_name(eq)), len_trim(primitive_name(eq)), &
          trim(basename_mesh_file), len_trim(basename_mesh_file), &
          sol, pmesh%NbCells, &
          DB_F77NULL, 0, &
          DB_DOUBLE, DB_ZONECENT, &
          DB_F77NULL, varid)
    end do

    ! Close the silo file
    err = dbclose(dbid)
    if (err /= 0) then
      write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, " -> silo error: ", err
      flush(StdErr)
      STOP
    end if

  end subroutine write_silo_primary

#endif /* _ENABLE_SILO_ && _MESH_P2D_*/
end module mod_write_silo
