!> Write solution and meshes in VTK files for
!> VisIt (https://wci.llnl.gov/simulation/computer-codes/visit) \cite VisIt_2012
!> or ParaView (http://www.paraview.org/) \cite ParaView_2007 .
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_write_VTK
#if defined(_IO_ENABLE_VTK_)

  use type_mesh, only: pmesh, nodes, dim_
#if defined(_MESH_D2P_)
  use type_mesh, only: dmesh, smesh
  use mod_mesh, only: MESH_D2P_ENABLE_DMESH, MESH_D2P_ENABLE_SMESH
#endif /* _MESH_ */
  use mod_numerics, only: PR
  use type_IO, only: FMT_D, FMT_I
  use type_IO, only: Output_Directory
  use mod_physics_var, only: NbEq, primitive_name, test_case
  use mod_physics, only: conservative_to_primitive
#if defined(_HO_MOOD_) && defined(_SRC_ENABLE_)
  use mod_asymptotic, only: compute_beta
#endif /* _HO_MOOD_ && _SRC_ENABLE_ */
#if defined(_PHY_M1MG_)
  use mod_physics_M1MG_var, only: NbGroup
#endif /* defined(_PHY_M1MG_) */
  use mod_open_file, only: OpenFile

  implicit none

  private
#if defined(_MESH_D2P_)
  public write_VTK_diamond, write_VTK_secondary
#endif /* _MESH_ */
  public write_VTK_primary

  !> VTK type for a quadrangle/diamond
  integer, parameter :: VTK_QUAD = 9
  !> VTK type for a polygon
  integer, parameter :: VTK_POLYGON = 7

  integer :: dmesh_id
  integer :: pmesh_id
  integer :: smesh_id

!#define _IO_WRITE_DX_
!#define _IO_WRITE_AREA_
!#define _IO_WRITE_PERIMETER_
!#define _IO_WRITE_VECTORS_

contains
#if defined(_MESH_D2P_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Write the diamond mesh in a VTK file
  !>
  !> @param[in] info Supplementary informations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine write_VTK_diamond(info)
      character(len=*), intent(in) :: info

      integer, parameter :: NbVer  = 4 ! Number of vertices in a diamond
      integer :: NbInterfaces          ! Number of interfaces in the diamond mesh
      integer :: CellK, diam

      ! Check if we really use the diamond mesh
      if (.not. MESH_D2P_ENABLE_DMESH) return

      NbInterfaces = 5*dmesh%NbDiams

      ! Diamonds mesh
      call OpenFile(unit=dmesh_id,file=trim(Output_Directory)//'/'//trim(info)//'_diamond.vtk',action='write',status='replace')
      !=================================================================!
      write(dmesh_id,'(A)')'# vtk DataFile Version 3.0.'
      write(dmesh_id,'(A)')'# Data'
      write(dmesh_id,'(A)')'ASCII'
      !=================================================================!
      write(dmesh_id,'(A)')'DATASET UNSTRUCTURED_GRID'
      !-----------------------------------------------------------------!
      write(dmesh_id,'(A,'//FMT_I//',A)') 'POINTS ', dmesh%NbVert,' double'
      DO CellK = 1, dmesh%NbVert
        write(dmesh_id,'(3('//FMT_D//',1X))') nodes(CellK)%x(:), 0._PR
      END DO
      !-----------------------------------------------------------------!
      write(dmesh_id,'(A,2(1X,'//FMT_I//'))') 'CELLS', dmesh%NbDiams, NbInterfaces
      DO diam = 1, dmesh%NbDiams
        write(dmesh_id,'(I1,4(1X,'//FMT_I//'))') NbVer, dmesh%elt(:,diam)-1
      END DO
      !-----------------------------------------------------------------!
      write(dmesh_id,'(A,'//FMT_I//')') 'CELL_TYPES ',dmesh%NbDiams
      DO diam = 1, dmesh%NbDiams
        write(dmesh_id,'(I1)') VTK_QUAD
      END DO
      !=================================================================!
#if defined(_IO_WRITE_AREA_)
      write(dmesh_id,'(A,'//FMT_I//')') 'CELL_DATA ',dmesh%NbDiams
      !-----------------------------------------------------------------!
      write(dmesh_id,'(A)')'SCALARS Area double'
      write(dmesh_id,'(A)')'LOOKUP_TABLE default'
      do diam = 1, dmesh%NbDiams
        write(dmesh_id,'('//FMT_D//')') dmesh%area(diam)
      end do
#endif /* _IO_WRITE_AREA_ */
      !=================================================================!
      CLOSE(dmesh_id)

  end subroutine write_VTK_diamond
#endif /* _MESH_ */

  !--------------------------------------------------------------------!
  !> @brief
  !> Write the primal mesh and the solution in a VTK file
  !>
  !> @param[in] U_n_p1 Solution at time \f$t^{n+1}\f$
  !> @param[in] info Supplementary informations
  !> @param[in] completion Completion of the run
  !> @param[in] NbIter Number of iterations
  !> @param[in] t_n_p1 Physical time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_conservative_to_primitive
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_conservative_to_primitive

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_beta
  subroutine write_VTK_primary(U_n_p1,info,completion,NbIter,t_n_p1)
      real(PR), dimension(:,:), intent(in) :: U_n_p1
      character(len=*), intent(in)         :: info
      integer, intent(in) :: NbIter
      real(PR), intent(in)                 :: completion, t_n_p1

      real(PR), dimension(NbEq) :: U_K
      real(PR), dimension(NbEq) :: V_K ! To get back primitive variable
#if defined(_PHY_M1MG_)
      real(PR) :: tmp
      integer :: group, eq0
#endif /* defined(_PHY_M1MG_) */
      integer :: NbVer, CellK,  eq
      character(len=2) :: NbVer_str

      ! Primary mesh
      call OpenFile(unit=pmesh_id,file=trim(Output_Directory)//'/'//trim(info)//'_primary.vtk',action='write',status='replace')
      !=================================================================!
      write(pmesh_id,'(A)')'# vtk DataFile Version 3.0.'
      write(pmesh_id,'(A,F6.2,A,'//FMT_D//',A,'//FMT_I//')')'# Data at: ', completion, '%; time: ', t_n_p1, 's; iterations: ', NbIter
      write(pmesh_id,'(A)')'ASCII'
      !=================================================================!
      write(pmesh_id,'(A)')'DATASET UNSTRUCTURED_GRID'
      !-----------------------------------------------------------------!
      write(pmesh_id,'(A)') 'FIELD FieldData 2'
      write(pmesh_id,'(A)') 'TIME 1 1 double'
      write(pmesh_id, '('//FMT_D//')') t_n_p1
      write(pmesh_id,'(A)') 'CYCLE 1 1 int'
      write(pmesh_id, '('//FMT_I//')') NbIter
      !-----------------------------------------------------------------!
#if defined(_MESH_D2P_)
      write(pmesh_id,'(A,'//FMT_I//',A)')'POINTS ', dmesh%NbVert, ' double'
      do CellK = 1, dmesh%NbVert
#elif defined(_MESH_P2D_)
      write(pmesh_id,'(A,'//FMT_I//',A)')'POINTS ', pmesh%NbVert, ' double'
      do CellK = 1, pmesh%NbVert
#endif /* _MESH_ */
        write(pmesh_id,'(3('//FMT_D//',1X))') nodes(CellK)%x(:), 0._PR
      end do
      !-----------------------------------------------------------------!
      write(pmesh_id,'(A,2(1X,'//FMT_I//'))')'CELLS', pmesh%NbCells, pmesh%NbInterfaces
      do CellK = 1, pmesh%NbCells
        NbVer = pmesh%elt(CellK)%NbInterfaces
        write(NbVer_str,'(I2)') NbVer
        write(pmesh_id,'(I2,1X,'//trim(NbVer_str)//'('//FMT_I//',1X))') pmesh%elt(CellK)%NbInterfaces, pmesh%elt(CellK)%vertices_glob(:)-1
      end do
      !-----------------------------------------------------------------!
      write(pmesh_id,'(A,'//FMT_I//')')'CELL_TYPES ', pmesh%NbCells
      do CellK = 1, pmesh%NbCells
        write(pmesh_id,'(I1)') VTK_POLYGON
      end do
      !=================================================================!
      write(pmesh_id,'(A,I9)')'CELL_DATA ',pmesh%NbCells
      !-----------------------------------------------------------------!
#if defined(_IO_WRITE_AREA_)
      write(pmesh_id,'(A)')'SCALARS Area double'
      write(pmesh_id,'(A)')'LOOKUP_TABLE default'
      do CellK = 1, pmesh%NbCells
        write(pmesh_id,'('//FMT_D//')') pmesh%elt(CellK)%area
      end do
#endif /* _IO_WRITE_AREA_ */
#if defined(_MESH_SAVE_PERIMETER_) && defined(_IO_WRITE_PERIMETER_)
      write(pmesh_id,'(A)')'SCALARS Area double'
      write(pmesh_id,'(A)')'LOOKUP_TABLE default'
      do CellK = 1, pmesh%NbCells
        write(pmesh_id,'('//FMT_D//')') pmesh%elt(CellK)%perimeter
      end do
#endif /* _MESH_SAVE_PERIMETER_ && _IO_WRITE_PERIMETER_ */
#if defined(_IO_WRITE_DX_)
      write(pmesh_id,'(A)')'SCALARS Area double'
      write(pmesh_id,'(A)')'LOOKUP_TABLE default'
      do CellK = 1, pmesh%NbCells
        write(pmesh_id,'('//FMT_D//')') pmesh%elt(CellK)%Delta_x
      end do
#endif /* _IO_WRITE_DX_ */
      !-----------------------------------------------------------------!
      do eq = 1,NbEq
        write(pmesh_id,'(3(A))')'SCALARS ', trim(primitive_name(eq)) ,' double'
        write(pmesh_id,'(A)')'LOOKUP_TABLE default'
        do CellK = 1,pmesh%NbCells
          U_K(:) = U_n_p1(:,CellK)
          call conservative_to_primitive(U_K,V_K)
          write(pmesh_id,'('//FMT_D//')') V_K(eq)
        end do
      end do
      !-----------------------------------------------------------------!
#if defined(_PHY_M1MG_)
      !-----------------------------------------------------------------!
      write(pmesh_id,'(3(A))')'SCALARS ', 'E_R' ,' double'
      write(pmesh_id,'(A)')'LOOKUP_TABLE default'
      do CellK = 1,pmesh%NbCells
        tmp = 0._PR
        do group = 1, NbGroup
          eq0 = (1 + dim_)*(group-1)
          tmp = tmp + U_n_p1(eq0+1,CellK)
        end do
        write(pmesh_id,'('//FMT_D//')') tmp
      end do
      write(pmesh_id,'(3(A))')'SCALARS ', 'F_Rx' ,' double'
      write(pmesh_id,'(A)')'LOOKUP_TABLE default'
      do CellK = 1,pmesh%NbCells
        tmp = 0._PR
        do group = 1, NbGroup
          eq0 = (1 + dim_)*(group-1)
          tmp = tmp + U_n_p1(eq0+2,CellK)
        end do
        write(pmesh_id,'('//FMT_D//')') tmp
      end do
      write(pmesh_id,'(3(A))')'SCALARS ', 'F_Ry' ,' double'
      write(pmesh_id,'(A)')'LOOKUP_TABLE default'
      do CellK = 1,pmesh%NbCells
        tmp = 0._PR
        do group = 1, NbGroup
          eq0 = (1 + dim_)*(group-1)
          tmp = tmp + U_n_p1(eq0+3,CellK)
        end do
        write(pmesh_id,'('//FMT_D//')') tmp
      end do
      !-----------------------------------------------------------------!
#endif /* defined(_PHY_M1MG_) */

#if defined(_IO_WRITE_VECTORS_)
      eq=2
      write(pmesh_id,'(3(A,1X))')'VECTORS ', trim(primitive_name(eq))//trim(primitive_name(eq+1)),' double'
      do CellK = 1, pmesh%NbCells
        U_K(:) = U_n_p1(:,CellK)
#if defined(_PHY_M1_)
        write(pmesh_id,'(3('//FMT_D//',1X))') U_K(eq), U_K(eq+1), 0._PR
#elif defined(_PHY_EULER_) || defined(_PHY_SW_)
        call conservative_to_primitive(U_K,V_K)
        write(pmesh_id,'(3('//FMT_D//',1X))') V_K(eq), V_K(eq+1), 0._PR
#endif /* _PHY_ */
      end do
#endif /* _IO_WRITE_VECTORS_ */
#if defined(_HO_MOOD_)
      write(pmesh_id,'(A)')'SCALARS degree int'
      write(pmesh_id,'(A)')'LOOKUP_TABLE default'
      do CellK = 1, pmesh%NbCells
        write(pmesh_id,'(I2)') pmesh%elt(CellK)%CellPD
      end do
#if defined(_SRC_ENABLE_)
      write(pmesh_id,'(A)')'SCALARS beta double'
      write(pmesh_id,'(A)')'LOOKUP_TABLE default'
      do CellK = 1, pmesh%NbCells
        write(pmesh_id,'('//FMT_D//')') compute_beta(CellK,U_n_p1(:,CellK),t_n_p1)
      end do
#endif /* _SRC_ENABLE_ */
#endif /* _HO_MOOD_ */
      !=================================================================!
      CLOSE(pmesh_id)

  end subroutine write_VTK_primary

#if defined(_MESH_D2P_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Write the dual mesh in a VTK file
  !>
  !> @param[in] info supplementary informations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine write_VTK_secondary(info)
      character(len=*), intent(in) :: info

      integer :: CellK, NbVer
      character(len=2) :: NbVer_str

      ! Check if we really use the secondary mesh
      if (.not. MESH_D2P_ENABLE_SMESH) return

      ! Secondary/Dual mesh
      call OpenFile(unit=smesh_id,file=trim(Output_Directory)//'/'//trim(info)//'_secondary.vtk',action='write',status='replace')
      !=================================================================!
      write(smesh_id,'(A)')'# vtk DataFile Version 3.0.'
      write(smesh_id,'(A)')'# DATA'
      write(smesh_id,'(A)')'ASCII'
      !=================================================================!
      write(smesh_id,'(A)')'DATASET UNSTRUCTURED_GRID'
      !-----------------------------------------------------------------!
      write(smesh_id,'(A,'//FMT_I//',A)')'POINTS ',dmesh%NbVert,' double'
      DO CellK = 1, dmesh%NbVert
        write(smesh_id,'(3('//FMT_D//',1X))') nodes(CellK)%x(:), 0._PR
      END DO
      !-----------------------------------------------------------------!
      write(smesh_id,'(A,2(1X,'//FMT_I//'))')'CELLS', smesh%NbCells, smesh%NbInterfaces
      DO CellK = 1, smesh%NbCells
        NbVer = smesh%elt(CellK)%NbInterfaces
        write(NbVer_str,'(I2)') NbVer
        write(smesh_id,'(I2,1X,'//trim(NbVer_str)//'('//FMT_I//',1X))') smesh%elt(CellK)%NbInterfaces, smesh%elt(CellK)%vertices_glob(:)-1
      END DO
      !-----------------------------------------------------------------!
      write(smesh_id,'(A,'//FMT_I//')')'CELL_TYPES ', smesh%NbCells
      DO CellK = 1, smesh%NbCells
        write(smesh_id,'(I1)') VTK_POLYGON
      END DO
      !=================================================================!
#if defined(_IO_WRITE_AREA_)
      write(smesh_id,'(A,'//FMT_I//')')'CELL_DATA ', smesh%NbCells
      !-----------------------------------------------------------------!
      write(smesh_id,'(A)')'SCALARS Area double'
      write(smesh_id,'(A)')'LOOKUP_TABLE default'
      DO CellK = 1, smesh%NbCells
        write(smesh_id,'('//FMT_D//')') smesh%elt(CellK)%area
      END DO
      !-----------------------------------------------------------------!
#endif /* _IO_WRITE_AREA_ */
      CLOSE(smesh_id)

  end subroutine write_VTK_secondary
#endif /* _MESH_ */

#endif /* _IO_ENABLE_VTK_ */
end module mod_write_vtk
