!> Tools for the re-computations done with a two point flux, when the cells are not admissible
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_TPCorrec
#if defined(_TP_CORREC_) && !defined(_HO_MOOD_)

!#define _IO_EXTENSIVE_PRINT_

  use mod_numerics, only: PR
  use type_IO, only: StdOut
  use type_mesh, only: pmesh
  use type_mesh_DLP, only: IntInterfaces
  use mod_detection, only: IsNumericallyAdmissible
  use mod_physics, only: IsPhysicallyAdmissible
#if defined(_IO_EXTENSIVE_PRINT_)
  use mod_tools_print, only: PrintVector
#endif /* _IO_EXTENSIVE_PRINT_ */
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells, chunk_IntInterfaces
#endif /* _FC_PAR_OMP_ */
  use type_recompute, only: NbCellsInvalid, NbCellsToUpdate
  use type_recompute, only: CellsToUpdate
  use type_recompute, only: FlagCells, FlagIntInterfaces

  implicit none

  private
  public GetCellsToCorrec

contains

  !--------------------------------------------------------------------!
  !> @brief
  !> Check which cells need to be corrected, this check is base on two criterion,
  !> the physical and the numerical admissibility.
  !>
  !> @param[in] U_n, U_n_p1 Solution at previous time and current time
  !> @param[in] NbCorrec Number of times that this subroutines was called
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_isphysicallyadmissible
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_isphysicallyadmissible

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_detection_mp_isnumericallyadmissible
  subroutine GetCellsToCorrec(U_n,U_n_p1,NbCorrec)
      real(PR), dimension(:,:), intent(in)    :: U_n, U_n_p1
      integer, intent(in)                     :: NbCorrec

      integer :: CellK, idxK, i_K, CellL, IdxIntInt

#if !defined(_TP_CORREC_FULL_)
      !$OMP SINGLE
      NbCellsInvalid  = 0
      !$OMP END SINGLE NOWAIT

      test_first_correc: if (NbCorrec == 0) then
        ! Only initialise the FlagIntInterfaces to zero (interface between to valid cells)
        ! at the first iterations, in order to not "promote" interfaces at the next iteration
        ! of the correction.
        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_IntInterfaces)
        do IdxIntInt = 1, pmesh%NbIntInterfaces
          FlagIntInterfaces(IdxIntInt) = 0
        end do
        !$OMP END DO
      end if test_first_correc

      ! First loop to get the not admissible cells
      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      do CellK = 1, pmesh%NbCells
        test_adm: if (.not. IsPhysicallyAdmissible(U_n_p1(:,CellK)) .or. .not. IsNumericallyAdmissible(U_n_p1(:,CellK))) then
          FlagCells(CellK) = 1
          !$OMP CRITICAL (GetCellsToCorrec_CRIT01)
          NbCellsInvalid = NbCellsInvalid + 1
          CellsToUpdate(NbCellsInvalid) = CellK
          !$OMP END CRITICAL (GetCellsToCorrec_CRIT01)
        else
          FlagCells(CellK) = 0
        end if test_adm
      end do
      !$OMP END DO


      !$OMP SINGLE
      test_has_invalid_cells: if (NbCellsInvalid > 0) then
        ! Now check if some not-admissible cells are nghb
        ! and add the neighbors to the cells to update
        NbCellsToUpdate = NbCellsInvalid
        loop_CellsInvalid: do idxK = 1, NbCellsInvalid
          CellK = CellsToUpdate(idxK)
          loop_nghb: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
            CellL = pmesh%elt(CellK)%nghb(i_K)
            test_not_bnd: if (CellL /= 0) then
              IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)

              test_not_flagged: if (FlagCells(CellL) == 0) then
                ! L is an admissible cell
                FlagIntInterfaces(IdxIntInt) = 1

                NbCellsToUpdate = NbCellsToUpdate + 1
                CellsToUpdate(NbCellsToUpdate) = CellL
                FlagCells(CellL) = 2

              else if (FlagCells(CellL) == 1) then
                ! L is a not admissible cell
                FlagIntInterfaces(IdxIntInt) = 2

              else if (FlagCells(CellL) == 2) then
                ! L is a nghb of an other not admissible cell
                FlagIntInterfaces(IdxIntInt) = 1

              end if test_not_flagged

            end if test_not_bnd
          end do loop_nghb
        end do loop_CellsInvalid


#if defined(_IO_EXTENSIVE_PRINT_)
        write(StdOut,'(A,I9)') 'Nb cells not admissible  : ', NbCellsInvalid
        write(StdOut,'(A,I9)') 'Nb correction(s)         : ', NbCorrec
        write(StdOut,'(A)')    'CellsToUpdate            : '
        call PrintVector(NbCellsInvalid, CellsToUpdate(1:NbCellsInvalid))
        write(StdOut,'(A)')    'CellsToRecompute         : '
        call PrintVector(NbCellsInvalid - NbCellsToUpdate, &
            CellsToUpdate(NbCellsInvalid+1:NbCellsToUpdate))
#endif /* _IO_EXTENSIVE_PRINT_ */
      endif test_has_invalid_cells
      !$OMP END SINGLE

#else
      if (NbCorrec == 0) then
        !$OMP SINGLE
        NbCellsInvalid  = pmesh%NbCells
        NbCellsToUpdate = pmesh%NbCells
        !$OMP END SINGLE NOWAIT

        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_IntInterfaces)
        do IdxIntInt = 1, pmesh%NbIntInterfaces
          FlagIntInterfaces(IdxIntInt) = 2

          CellK = IntInterfaces(IdxIntInt)%nghb(1)
          CellsToUpdate(CellK) = CellK
          FlagCells(CellK) = 1

          CellL = IntInterfaces(IdxIntInt)%nghb(2)
          CellsToUpdate(CellL) = CellL
          FlagCells(CellL) = 1
        end do
        !$OMP END DO

      else
        !$OMP SINGLE
        NbCellsInvalid  = 0
        !$OMP END SINGLE
      end if

#endif /* _TP_CORREC_FULL_ */

  end subroutine GetCellsToCorrec

#endif /* _TP_CORREC_ && .not. _HO_MOOD_ */
end module mod_TPCorrec
