!> Wrapper used to choose the asymptotic behaviour of the problem
!>
!> To choose between asymptotic compile with:
!> * \c -D_PHY_EULER_ for the isentropic Euler model: asymptotic_Euler.f90
!> * \c -D_PHY_HYDRORAD_ for the hyperbolic heat equation: asymptotic_HydroRad.f90
!> * \c -D_PHY_HYPHEAT_ for the hyperbolic heat equation: asymptotic_HypHeat.f90
!> * \c -D_PHY_M1_ for the M1 model: asymptotic_M1.f90
!> * \c -D_PHY_P1_ for the P1 model: asymptotic_P1.f90
!> * \c -D_PHY_SW_ for the SW equations: asymptotic_SW.f90
!> * \c -D_PHY_TELE_ for the Telegraph equations: asymptotic_Telegraph.f90
!>
!> Each problem could be rewritten as conservation laws:
!> \f[ \partial_t \textbf{U} + \text{div}(\textbf{F}(\textbf{U}))=\gamma(\textbf{U})(\textbf{R}(\textbf{U})-\textbf{U}) \f]
!> where \f$\textbf{U} \in \mathbb{R}^N\f$, \f$\textbf{F}\f$ is a smooth function
!> whose Jacobian has real eigenvalues, \f$\gamma \geq 0\f$ and \f$\textbf{R}\f$
!> is a smooth function which fulfills the compatibility conditions
!> stated in \cite BerLeFTur_2013 .
!>
!> The main characteristic of such systems is to degenerate when \f$\gamma\ t \to \infty \f$ into smaller
!> parabolic systems of the form:
!> \f[ \partial_{t} \textbf{u} - \textrm{div}\bigl({\bf D}(\textbf{u})\nabla\textbf{u}\bigr) = 0, \f]
!>
!> where \f$\textbf{u} \in \mathbb{R}^n\f$ is related to \f$\textbf{U}\f$ and \f${\bf D}\f$ is a positive function or a symmetric,
!> positive and definite matrix.
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.

module mod_asymptotic
#if defined(_SRC_ENABLE_)

  use mod_numerics, only: PR, eps
  use type_IO, only: StdErr
  use type_mesh, only: pmesh
  use mod_physics_var, only: T_end, NbEq
  use mod_physics, only: compute_greatest_wave_speed, compute_wave_speed

#if defined (_PHY_EULER_)
  use mod_asymptotic_Euler, only: gamma_, gamma_bar_KJ, gamma_bar_KK
  use mod_asymptotic_Euler, only: g_u, R

#elif defined(_PHY_HYDRORAD_)
  use mod_asymptotic_HydroRad, only: gamma_, gamma_bar_KJ, gamma_bar_KK
  use mod_asymptotic_HydroRad, only: g_u, R

#elif defined(_PHY_HYPHEAT_)
  use mod_asymptotic_HypHeat, only: gamma_, gamma_bar_KJ, gamma_bar_KK
  use mod_asymptotic_HypHeat, only: g_u, R

#elif defined(_PHY_M1_)
  use mod_asymptotic_M1, only: gamma_, gamma_bar_KJ, gamma_bar_KK
  use mod_asymptotic_M1, only: g_u, R

#elif defined(_PHY_M1MG_)
  use mod_asymptotic_M1MG, only: gamma_, gamma_bar_KJ, gamma_bar_KK
  use mod_asymptotic_M1MG, only: g_u, R

#elif defined(_PHY_P1_)
  use mod_asymptotic_P1, only: gamma_, gamma_bar_KJ, gamma_bar_KK
  use mod_asymptotic_P1, only: g_u, R

#elif defined(_PHY_SW_)
  use mod_asymptotic_SW, only: gamma_, gamma_bar_KJ, gamma_bar_KK
  use mod_asymptotic_SW, only: g_u, R

#elif defined(_PHY_TELE_)
  use mod_asymptotic_Telegraph, only: gamma_, gamma_bar_KJ, gamma_bar_KK
  use mod_asymptotic_Telegraph, only: g_u, R
#endif /* _PHY_ */

  implicit none

  private
#if defined(_HO_MOOD_)
  public compute_beta
#endif /* _HO_MOOD_ */
  public g_u
  public compute_alpha_KJ, compute_alpha_KK, R_bar
  public compute_alpha_KJ_NoAP, compute_alpha_KK_NoAP, R
  public gamma_

contains

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\alpha_{KJ}\f$.
  !>
  !> * For Rusanov:
  !>   \f[\alpha_{KJ} = \frac{b_{KJ}}{b_{KJ} + (\gamma + \bar{\gamma}) \delta_{KJ}}\f]
  !> * For HLL:
  !>   \f[\alpha_{KJ} = \frac{b^+_{KJ} - b^-_{KJ}}
  !>     {(b^+_{KJ} - b^-_{KJ}) + 2(\gamma + \bar{\gamma}) \delta_{KJ}}\f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] nu_J, nu_bar_J, delta_K Coefficients
  !> @param[in] U_K, U_J Solution on cells \f$K\f$ and \f$J\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal \f$KJ\f$
  !> @param[in] U_K_mean, U_J_mean Mean values on cells \f$K\f$ and \f$J\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_greatest_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_euler_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_sw_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_gamma_

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_euler_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_sw_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_gamma_bar_kj
  function compute_alpha_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,U_K_mean,U_J_mean)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: nu_J
      real(PR), intent(in)               :: nu_bar_J
      real(PR), dimension(:), intent(in) :: delta_K
      real(PR), dimension(:), intent(in) :: U_K, U_J
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), dimension(:), optional, intent(in) :: U_K_mean, U_J_mean
      real(PR), dimension(NbEq) :: compute_alpha_KJ

#if defined(_FLUX_RUS_)
      real(PR) :: b_KJ
#elif defined(_FLUX_HLL_)
      real(PR) :: wave_speed_K, wave_speed_J
#elif defined(_FLUX_HLLC_)
      real(PR), dimension(NbEq) :: U_K_star, U_J_star
      real(PR) :: wave_speed_K, wave_speed_J, wave_speed_star
#endif /* _FLUX */
      real(PR) :: gamma_K, gamma_bar

      gamma_K = gamma_(CellK,U_K)
#if defined(_HO_MOOD_)
      gamma_bar = gamma_bar_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K_mean,U_J_mean,eta,tau,gamma_K)
#else
      gamma_bar = gamma_bar_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,gamma_K)
#endif /* _HO_MOOD_ */

#if defined(_FLUX_RUS_)
      b_KJ = compute_greatest_wave_speed(U_K, U_J, eta, tau)
      compute_alpha_KJ(:) = b_KJ/(b_KJ + &
          (gamma_K + gamma_bar)*delta_K(:))

#elif defined(_FLUX_HLL_)
      call compute_wave_speed(U_K, U_J, eta, tau, wave_speed_K, wave_speed_J)
      compute_alpha_KJ(:) = (wave_speed_J - wave_speed_K)/((wave_speed_J - wave_speed_K) &
          + 2._PR*(gamma_K + gamma_bar)*delta_K(:))

#elif defined(_FLUX_HLLC_)
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> alpha for HLLC not yet implemented"
      flush(StdErr)
      STOP
#endif /* _FLUX_ */

  end function compute_alpha_KJ

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\alpha_{KJ}\f$ without AP correction.
  !>
  !> * For Rusanov:
  !>   \f[\alpha_{KJ} = \frac{b_{KJ}}{b_{KJ} + \gamma \delta_{KJ}}\f]
  !> * For HLL:
  !>   \f[\alpha_{KJ} = \frac{b^+_{KJ} - b^-_{KJ}}
  !>     {(b^+_{KJ} - b^-_{KJ}) + 2\gamma \delta_{KJ}}\f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] delta_K Coefficient
  !> @param[in] U_K, U_J Solution on cells \f$K\f$ and \f$J\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal \f$KJ\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_greatest_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_euler_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_sw_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_gamma_
  function compute_alpha_KJ_NoAP(CellK,delta_K,U_K,U_J,eta,tau)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: delta_K
      real(PR), dimension(:), intent(in) :: U_K, U_J
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), dimension(NbEq) :: compute_alpha_KJ_NoAP

#if defined(_FLUX_RUS_)
      real(PR) :: b_KJ
#elif defined(_FLUX_HLL_)
      real(PR) :: wave_speed_K, wave_speed_J
#elif defined(_FLUX_HLLC_)
      real(PR), dimension(NbEq) :: U_K_star, U_J_star
      real(PR) :: wave_speed_K, wave_speed_J, wave_speed_star
#endif /* _FLUX_ */

#if defined(_FLUX_RUS_)
      b_KJ = compute_greatest_wave_speed(U_K, U_J, eta, tau)
      compute_alpha_KJ_NoAP(:) = b_KJ/(b_KJ + gamma_(CellK,U_K)*delta_K(:))

#elif defined(_FLUX_HLL_)
      call compute_wave_speed(U_K, U_J, eta, tau, wave_speed_K, wave_speed_J)
      compute_alpha_KJ_NoAP(:) = (wave_speed_J - wave_speed_K)/((wave_speed_J - wave_speed_K) &
          + 2._PR*gamma_(CellK,U_K)*delta_K(:))

#elif defined(_FLUX_HLLC_)
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> alpha for HLLC not yet implemented"
      flush(StdErr)
      STOP
#endif /* _FLUX_ */

  end function compute_alpha_KJ_NoAP

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\alpha_{KK}\f$.
  !>
  !> * For Rusanov:
  !>   \f[\alpha_{KK} = \frac{b_{KK}}{b_{KK} + (\gamma + \bar{\gamma}) \delta_{KK}}\f]
  !> * For HLL:
  !>   \f[\alpha_{KK} = \frac{b^+_{KK} - b^-_{KK}}
  !>     {(b^+_{KK} - b^-_{KK}) + 2(\gamma + \bar{\gamma}) \delta_{KK}}\f]
  !> with \f$\delta_{KK} = \frac{|K|}{p_K}\f$
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal \f$KJ\f$
  !> @param[in] U_K_mean Mean value on cells \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_greatest_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_euler_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_sw_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_gamma_

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_euler_mp_gamma_bar_kk
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_gamma_bar_kk
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_gamma_bar_kk
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1_mp_gamma_bar_kk
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_gamma_bar_kk
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_gamma_bar_kk
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_sw_mp_gamma_bar_kk
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_gamma_bar_kk
  function compute_alpha_KK(CellK,U_K,eta,tau,U_K_mean)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), dimension(:), optional, intent(in) :: U_K_mean
      real(PR) :: compute_alpha_KK

#if defined(_FLUX_RUS_)
      real(PR) :: b_KK
#elif defined(_FLUX_HLL_)
      real(PR) :: wave_speed_Km, wave_speed_Kp
#elif defined(_FLUX_HLLC_)
      real(PR), dimension(NbEq) :: U_K_star, U_J_star
      real(PR) :: wave_speed_K, wave_speed_J, wave_speed_star
#endif /* _FLUX_ */
      real(PR) :: gamma_K, gamma_bar

      gamma_K = gamma_(CellK,U_K)
#if defined(_HO_MOOD_)
      gamma_bar = gamma_bar_KK(CellK,U_K_mean,eta,tau,gamma_K)
#else
      gamma_bar = gamma_bar_KK(CellK,U_K,eta,tau,gamma_K)
#endif /* _HO_MOOD_ */

#if defined(_FLUX_RUS_)
      b_KK = compute_greatest_wave_speed(U_K, U_K, eta, tau)
      compute_alpha_KK = b_KK/(b_KK + (gamma_K + gamma_bar)*pmesh%elt(CellK)%Delta_x)

#elif defined(_FLUX_HLL_)
      call compute_wave_speed(U_K, U_K, eta, tau, wave_speed_Km, wave_speed_Kp)
      compute_alpha_KK = (wave_speed_Kp - wave_speed_Km)/((wave_speed_Kp - wave_speed_Km) &
          + 2._PR*(gamma_K + gamma_bar)*pmesh%elt(CellK)%Delta_x)

#elif defined(_FLUX_HLLC_)
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> alpha for HLLC not yet implemented"
      flush(StdErr)
      STOP
#endif /* _FLUX_ */

  end function compute_alpha_KK

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\alpha_{KK}\f$ without AP correction.
  !>
  !> * For Rusanov:
  !>   \f[\alpha_{KK} = \frac{b_{KK}}{b_{KK} + (\gamma + \bar{\gamma}) \delta_{KK}}\f]
  !> * For HLL:
  !>   \f[\alpha_{KK} = \frac{b^+_{KK} - b^-_{KK}}
  !>     {(b^+_{KK} - b^-_{KK}) + 2(\gamma + \bar{\gamma}) \delta_{KK}}\f]
  !> with \f$\delta_{KK} = \frac{|K|}{p_K}\f$
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal \f$KJ\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_greatest_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_euler_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_sw_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_gamma_
  function compute_alpha_KK_NoAP(CellK,U_K,eta,tau)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR) :: compute_alpha_KK_NoAP

#if defined(_FLUX_RUS_)
      real(PR) :: b_KK
#elif defined(_FLUX_HLL_)
      real(PR) :: wave_speed_Km, wave_speed_Kp
#elif defined(_FLUX_HLLC_)
      real(PR), dimension(NbEq) :: U_K_star, U_J_star
      real(PR) :: wave_speed_K, wave_speed_J, wave_speed_star
#endif /* _FLUX_ */

#if defined(_FLUX_RUS_)
      b_KK = compute_greatest_wave_speed(U_K, U_K, eta, tau)
      compute_alpha_KK_NoAP = b_KK/(b_KK + gamma_(CellK,U_K)*pmesh%elt(CellK)%Delta_x)

#elif defined(_FLUX_HLL_)
      call compute_wave_speed(U_K, U_K, eta, tau, wave_speed_Km, wave_speed_Kp)
      compute_alpha_KK_NoAP = (wave_speed_Kp - wave_speed_Km)/((wave_speed_Kp - wave_speed_Km) &
          + 2._PR*gamma_(CellK,U_K)*pmesh%elt(CellK)%Delta_x)

#elif defined(_FLUX_HLLC_)
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> alpha for HLLC not yet implemented"
      flush(StdErr)
      STOP
#endif /* _FLUX_ */

  end function compute_alpha_KK_NoAP

#if defined(_HO_MOOD_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\beta_{K}\f$.
  !>
  !> \f[\beta_{KJ} = \frac{\Delta_{\lim}}{\Delta_{\lim} + \gamma_K*t*\Delta x_K}\f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !> @param[in] t_n Current time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_euler_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_sw_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_gamma_
  function compute_beta(CellK,U_K,t_n)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), intent(in) :: t_n
      real(PR) :: compute_beta

      ! Characteristic length, if \gamma*t*Delta_x is smaller
      ! then this lenght then the AP is not "needed"
      real(PR), parameter :: Delta_lim = 5.E-2_PR

      compute_beta = Delta_lim/(Delta_lim + gamma_(CellK,U_K)*t_n*pmesh%elt(CellK)%Delta_x)

  end function compute_beta
#endif /* _HO_MOOD_ */

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\bar{R}(U)\f$.
  !>
  !> \f[\bar{R}(U) = \frac{\gamma(U)R(U) + \bar{\gamma}(U) U}{\gamma(U) + \bar{\gamma}(U)}\f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] nu_J, nu_bar_J, delta_K Coefficients
  !> @param[in] U_K, U_J Solution on cells \f$K\f$ and \f$J\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal \f$KJ\f$
  !> @param[in] U_K_mean, U_J_mean Mean values on cells \f$K\f$ and \f$J\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_euler_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_sw_mp_gamma_
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_gamma_

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_euler_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_sw_mp_gamma_bar_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_gamma_bar_kj

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_euler_mp_r
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_r
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_r
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1_mp_r
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_r
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_r
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_sw_mp_r
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_r
  function R_bar(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,U_K_mean,U_J_mean)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: nu_J
      real(PR), intent(in)               :: nu_bar_J
      real(PR), dimension(:), intent(in) :: delta_K
      real(PR), dimension(:), intent(in) :: U_J, U_K
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), dimension(:), optional, intent(in) :: U_K_mean, U_J_mean
      real(PR), dimension(NbEq) :: R_bar

      real(PR) :: gamma_K, gamma_bar

      gamma_K   = gamma_(CellK,U_K)

#if defined(_HO_MOOD_)
      gamma_bar = gamma_bar_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K_mean,U_J_mean,eta,tau,gamma_K)
#else
      gamma_bar = gamma_bar_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,gamma_K)
#endif /* _HO_MOOD_ */
      if (.not. (abs(gamma_K + gamma_bar) < eps)) then
        R_bar(:) = (gamma_K*R(CellK,U_K) + gamma_bar*U_K(:))/(gamma_K + gamma_bar)
      else
        R_bar(:) = 0._PR
      end if

  end function R_bar

#endif /* _SRC_ENABLE_ */
end module mod_asymptotic
