!> Compute things related to source term and AP for the radiation hydrodynamics model.
!>
!> \f[
!>   \left\{
!>     \begin{array}{rcl}
!>       \partial_t \rho         + \text{div}(\rho {\bf u})                                      & = &  0 \\
!>       \partial_t \rho {\bf u} + \text{div}(\rho {\bf u}\otimes {\bf u}+ {\bf P}_R) + \nabla p & = &  \frac{1}{c} \sigma^f {\bf F}_R \\
!>       \partial_t E            + \text{div}((E+p){\bf u})                                      & = &  c \left(\sigma^a E_R - \sigma^e a T^4 \right)\\
!>       \partial_t E_R          + \text{div}({\bf F}_R)                                         & = &  c \left(\sigma^e a T^4 - \sigma^a E_R \right)\\
!>       \partial_t {\bf F}_R    + c^2 \text{div}(({\bf P}_R)                                    & = & -c \sigma^f {\bf F}_R
!>     \end{array}
!>   \right..
!> \f]
!>
!> with:
!> * \f$ \rho \f$: density of the fluid,
!> * \f$ {\bf u} \f$: speed of the fluid,
!> * \f$ E = \rho(\frac{u^2}{2} + e) \f$: total energy per unit volume, \f$e\f$ specific internal energy,
!> * \f$ \kappa > 0 \f$,
!> * the equation of state (EOS): \f$e=\frac{p}{\rho(\gamma-1)}\f$,
!> * the speed of the sound in the fluid is defined as: \f$c = \sqrt{\frac{\gamma p}{\rho}}\f$,
!> * \f$ E_R \f$ the radiative energy,
!> * \f$ {\bf F}_R \f$ the radiative flux,
!> * \f$ T \f$ the temperature,
!> * \f$ {\bf P}_R = \frac{E_R}{2}\left( (1-\chi)  + (3\chi-1)\frac{{\bf F}_R \otimes {\bf F}}{\Vert {\bf F} \Vert_2^2}\right) \f$ the radiative pressure,
!> * \f$\chi = \frac{3 + 4f^2}{5+ \sqrt{4-3f^2}} \f$ the Eddington factor, in order that \f$\chi\f$ is well defined:
!>     \f$ f=\frac{\Vert {\bf F} \Vert}{c E} \leq 1 \f$,
!> * \f$ c \f$ the speed of light.!>
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_asymptotic_HydroRad
#if defined(_SRC_ENABLE_) && defined(_PHY_HYDRORAD_)

  use mod_numerics, only: PR, eps
  use type_mesh, only: pmesh, dim_
  use mod_tools, only: min_NaN, max_NaN
  use mod_physics_HydroRad, only: compute_greatest_wave_speed, compute_wave_speed
  use mod_physics_HydroRad, only: compute_temperature, compute_pressure
  use mod_physics_HydroRad_var, only: NbEq, test_case
  use mod_physics_HydroRad_var, only: c, sigma, a, C_v
  use mod_physics_HydroRad_var, only: sigma_a, sigma_e, sigma_f

  implicit none

  private
  public gamma_, gamma_bar_KJ, gamma_bar_KK
  public g_u
  public R

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\gamma(U)=c \sigma^m\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_m
  function gamma_(CellK,U_K)
    integer, intent(in) :: CellK
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: gamma_

    gamma_ = c*compute_sigma_m(CellK,U_K)

  end function gamma_

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\bar{\gamma}(U) = c\bar{\sigma}^m\f$.
  !>
  !> * For Rusanov:
  !> \f[\bar{\sigma}_m = 3\sigma^f \frac{\nu^J_{K,i}b_{KJ}^2}
  !>   {2c^2\bar{\nu}^J_{K,i} \delta_{KJ}} - \sigma^m \f]
  !>
  !> * For HLL:
  !> \f[\bar{\sigma}_m = 3\sigma^f \frac{\nu^J_{K,i}(-b^+_{KJ}b^-_{KJ})}
  !>   {2c^2\bar{\nu}^J_{K,i} \delta_{KJ}} - \sigma^m \f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] nu_J, nu_bar_J, delta_K Coefficients
  !> @param[in] U_K, U_J Solution on cells \f$K\f$ and \f$J\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal to \f$KJ\f$
  !> @param[in] gamma_K \f$\gamma(U_K)\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_f
  function gamma_bar_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,gamma_K)
    integer, intent(in) :: CellK
    real(PR), dimension(:), intent(in) :: nu_J
    real(PR), intent(in)               :: nu_bar_J
    real(PR), dimension(:), intent(in) :: delta_K
    real(PR), dimension(:), intent(in) :: U_K, U_J
    real(PR), dimension(:), intent(in) :: eta, tau
    real(PR), intent(in) :: gamma_K
    real(PR) :: gamma_bar_KJ

#if defined(_SRC_DISABLE_AP_CORREC_)
    gamma_bar_KJ = 0._PR
#else
    real(PR) :: speed_square
    integer, parameter  :: eq_lim = 1

#if defined(_FLUX_RUS_)
    real(PR) :: b_KJ

    b_KJ = compute_greatest_wave_speed(U_K, U_J, eta, tau)
    speed_square = b_KJ**2

#elif defined(_FLUX_HLL_)
    real(PR) :: wave_speed_J, wave_speed_K

    call compute_wave_speed(U_K, U_J, eta, tau, wave_speed_K, wave_speed_J)
    speed_square = -wave_speed_J*wave_speed_K
#endif /* _FLUX_ */

    if (abs(nu_J(eq_lim)) < eps .or. abs(nu_bar_J) < eps) then
      gamma_bar_KJ = 0._PR
    else
      gamma_bar_KJ = 3._PR*compute_sigma_f(U_K)*nu_J(eq_lim)*speed_square&
          /(2._PR*c*nu_bar_J*delta_K(eq_lim)) - gamma_K

    end if
#endif /* _SRC_DISABLE_AP_CORREC_ */

  end function gamma_bar_KJ

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\bar{\gamma}(U) = c\bar{\sigma}^m\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal to \f$KJ\f$
  !> @param[in] gamma_K \f$\gamma(U_K)\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_f
  function gamma_bar_KK(CellK,U_K,eta,tau,gamma_K)
    integer, intent(in) :: CellK
    real(PR), dimension(:), intent(in) :: U_K
    real(PR), dimension(:), intent(in) :: eta, tau
    real(PR), intent(in) :: gamma_K
    real(PR) :: gamma_bar_KK

#if defined(_SRC_DISABLE_AP_CORREC_)
    gamma_bar_KK = 0._PR
#else

    real(PR) :: speed_square
#if defined(_FLUX_RUS_)
    real(PR) :: b_KK

    b_KK = compute_greatest_wave_speed(U_K, U_K, eta, tau)
    speed_square = b_KK**2

#elif defined(_FLUX_HLL_)
    real(PR) :: wave_speed_Kp, wave_speed_Km

    call compute_wave_speed(U_K, U_K, eta, tau, wave_speed_Km, wave_speed_Kp)
    speed_square = -wave_speed_Km*wave_speed_Kp
#endif /* _FLUX_ */

    gamma_bar_KK = 6._PR*compute_sigma_f(U_K)*speed_square/c - gamma_K
#endif /* _SRC_DISABLE_AP_CORREC_ */

  end function gamma_bar_KK

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\sigma^m(U)\f$.
  !>
  !> \f[\sigma^m = \max\left(\sigma^a,\sigma^f\right)\f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_a
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_f
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  function compute_sigma_m(CellK,U_K)
    integer, intent(in) :: CellK
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_m

    compute_sigma_m = max_NaN(compute_sigma_a(U_K),compute_sigma_f(U_K))
  end function compute_sigma_m

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\sigma_1(U) = \sigma^m - \sigma^a\f$.
  !>
  !> @param[in] sigma_m \f$\sigma^m\f$
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_a
  function compute_sigma_1(sigma_m,U_K)
    real(PR), intent(in)               :: sigma_m
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_1

    compute_sigma_1 = sigma_m - compute_sigma_a(U_K)
  end function compute_sigma_1

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\sigma_2(U) = \sigma^m - \sigma^f\f$.
  !>
  !> @param[in] sigma_m \f$\sigma^m\f$
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_f
  function compute_sigma_2(sigma_m,U_K)
    real(PR), intent(in)               :: sigma_m
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_2

    compute_sigma_2 = sigma_m - compute_sigma_f(U_K)
  end function compute_sigma_2

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$R(U)\f$.
  !>
  !> \f[\textbf{R}(\textbf{U}) = \frac{1}{\sigma^m}\left(\sigma^e a T^4 + \sigma_1 E,
  !>                  \sigma_2 F_x, \sigma_2 F_y,
  !>                  \frac{\sigma^a E}{\rho C_v} + \sigma_3 T\right)^T \f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_m
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_1
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_2
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_3
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_a
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hydrorad_mp_compute_sigma_e
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_temperature
  function R(CellK,U_K)
    integer, intent(in) :: CellK
    real(PR), dimension(:), intent(in) :: U_K
    real(PR), dimension(NbEq) :: R

    real(PR) :: rho, E, E_R, T
    real(PR), dimension(dim_) :: rho_u, F_R
    real(PR) :: sigma_m, sigma_1, sigma_2

    rho      = U_K(1)
    rho_u(:) = U_K(2:3)
    E        = U_K(4)
    E_R      = U_K(5)
    F_R(:)   = U_K(6:7)
    T        = compute_temperature(U_K)

    sigma_m = compute_sigma_m(CellK,U_K)
    sigma_1 = compute_sigma_1(sigma_m,U_K)
    sigma_2 = compute_sigma_2(sigma_m,U_K)

    R(1)   = rho
    R(2:3) = compute_sigma_f(U_K)*F_R(:)/(c**2*sigma_m) + rho_u(:)
    R(4)   = (compute_sigma_a(U_K)*E_R - compute_sigma_e(U_K)*a*T**4)/sigma_m + E
    R(5)   = (compute_sigma_e(U_K)*a*T**4 + sigma_1*E_R)/sigma_m
    R(6:7) = sigma_2/sigma_m*F_R(:)
  end function R

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$g(u)\f$ of the limit equation.
  !>
  !> Generic limit equation:
  !> \f[\partial_t(f(u)) - \text{div}(\alpha(u) \nabla g(u)) = 0 \f]
  !>
  !> Limit equation for \f$M_1\f$:
  !> \f[\partial_t(\rho C_v + aT^4) - \text{div}(\frac{c}{3\sigma^f}\nabla (a T^4)) = 0 \f]
  !>
  !> @param[in] U Conservative variable of the hyperbolic system
  !> @return \f$g(u)\f$, with \f$u\f$ the variable of the limit equation
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function g_u(U)
    real(PR), dimension(:), intent(in) :: U
    real(PR) :: g_u

    g_u = a*U(4)**4
  end function g_u

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the value of \f$\sigma^a\f$
  !>
  !> @param[in] U_K Conservative variable
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_min_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_temperature
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_pressure
  function compute_sigma_a(U_K)
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_a

    select case(test_case)
    case ('sonde', 'sonde_80km')
      compute_sigma_a = (compute_temperature(U_K)/1.E4_PR)**3*sigma_a
    case ('ball')
      compute_sigma_a = min_NaN(1000._PR,(compute_pressure(U_K)/1.E5_PR)*(compute_temperature(U_K)/1.E4_PR)**3)
    case  default
      compute_sigma_a = sigma_a
    end select

  end function compute_sigma_a

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the value of \f$\sigma^e\f$
  !>
  !> @param[in] U_K Conservative variable
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_min_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_temperature
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_pressure
  function compute_sigma_e(U_K)
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_e

    select case(test_case)
    case ('sonde', 'sonde_80km')
      compute_sigma_e = (compute_temperature(U_K)/1.E4_PR)**3*sigma_e
    case ('ball')
      compute_sigma_e = min_NaN(1000._PR,(compute_pressure(U_K)/1.E5_PR)*(compute_temperature(U_K)/1.E4_PR)**3)
    case  default
      compute_sigma_e = sigma_e
    end select

  end function compute_sigma_e

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the value of \f$\sigma^f\f$
  !>
  !> @param[in] U_K Conservative variable
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_min_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_temperature
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_pressure
  function compute_sigma_f(U_K)
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_f

    select case(test_case)
    case ('sonde', 'sonde_80km')
      compute_sigma_f = (compute_temperature(U_K)/1.E4_PR)**3*sigma_f
    case ('ball')
      compute_sigma_f = min_NaN(1000._PR,(compute_pressure(U_K)/1.E5_PR)*(compute_temperature(U_K)/1.E4_PR)**3)
    case  default
      compute_sigma_f = sigma_f
    end select

  end function compute_sigma_f

#endif /* _PHY_HYDRORAD_ */
end module mod_asymptotic_HydroRad
