!> Compute things related to source term and AP for the P1 model.
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t E   +               \partial_x F_x  +               \partial_y F_y  & = & 0 \\
!>   \partial_t F_x + \frac{c^2}{3} \partial_x E    +  \frac{c^2}{3}\partial_y E    & = & -c \sigma^f F_x \\
!>   \partial_t F_y + \frac{c^2}{3} \partial_x E    +  \frac{c^2}{3}\partial_y E    & = & -c \sigma^f F_y
!> \end{array}\right.\f]
!>
!> \f[\textbf{U} = \left(E, F_x, F_y\right)^T \f]
!> \f[\textbf{R}(\textbf{U}) = \left(E, 0, 0\right)^T \f]
!> \f[\gamma(\textbf{U}) = c \sigma^f(\textbf{U}) \f]
!>
!> Limit equation:
!> \f[\partial_t E - \text{div}\left(\frac{c}{3\sigma^f} \nabla E \right) = 0 \f]
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_asymptotic_P1
#if defined(_SRC_ENABLE_) && defined(_PHY_P1_)

  use mod_numerics, only: PR, eps, zero

  use type_mesh,  only: pmesh, dim_
  use mod_physics_P1, only: compute_greatest_wave_speed, compute_wave_speed
  use mod_physics_P1_var, only: NbEq, test_case
  use mod_physics_P1_var, only: c, sigma_f

  implicit none

  private
  public gamma_, gamma_bar_KJ, gamma_bar_KK
  public g_u
  public R

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\gamma(U)=c \sigma^f\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_compute_sigma_f
  function gamma_(CellK,U_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K
      real(PR) :: gamma_

      gamma_ = c*compute_sigma_f(U_K)

  end function gamma_

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\bar{\gamma}(U) = c\bar{\sigma}^m\f$.
  !>
  !> * For Rusanov:
  !> \f[\bar{\sigma}_m = 3\sigma^f \frac{\nu^J_{K,i}b_{KJ}^2}
  !>   {2c^2\bar{\nu}^J_{K,i} \delta_{KJ}} - \sigma^m \f]
  !>
  !> * For HLL:
  !> \f[\bar{\sigma}_m = 3\sigma^f \frac{\nu^J_{K,i}(-b^+_{KJ}b^-_{KJ})}
  !>   {2c^2\bar{\nu}^J_{K,i} \delta_{KJ}} - \sigma^m \f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] nu_J, nu_bar_J, delta_K Coefficients
  !> @param[in] U_K, U_J Solution on cells \f$K\f$ and \f$J\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal to \f$KJ\f$
  !> @param[in] gamma_K \f$\gamma(U_K)\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_compute_sigma_f
  function gamma_bar_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,gamma_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: nu_J
      real(PR), intent(in)               :: nu_bar_J
      real(PR), dimension(:), intent(in) :: delta_K
      real(PR), dimension(:), intent(in) :: U_K, U_J
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), intent(in) :: gamma_K
      real(PR) :: gamma_bar_KJ

#if defined(_SRC_DISABLE_AP_CORREC_)
      gamma_bar_KJ = 0._PR
#else
      real(PR) :: speed_square
      integer, parameter  :: eq_lim = 1

#if defined(_FLUX_RUS_)
      real(PR) :: b_KJ

      b_KJ = compute_greatest_wave_speed(U_K, U_J, eta, tau)
      speed_square = b_KJ**2

#elif defined(_FLUX_HLL_)
      real(PR) :: wave_speed_J, wave_speed_K

      call compute_wave_speed(U_K, U_J, eta, tau, wave_speed_K, wave_speed_J)
      speed_square = -wave_speed_J*wave_speed_K
#endif /* _FLUX_ */

      if (abs(nu_J(eq_lim)) < eps .or. abs(nu_bar_J) < eps) then
        gamma_bar_KJ = 0._PR
      else
        gamma_bar_KJ = 3._PR*compute_sigma_f(U_K)*nu_J(eq_lim)*speed_square&
            /(2._PR*c*nu_bar_J*delta_K(eq_lim)) - gamma_K
      end if
#endif /* _SRC_DISABLE_AP_CORREC_ */

  end function gamma_bar_KJ

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\bar{\gamma}(U) = c\bar{\sigma}^m\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal to \f$KJ\f$
  !> @param[in] gamma_K \f$\gamma(U_K)\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_p1_mp_compute_sigma_f
  function gamma_bar_KK(CellK,U_K,eta,tau,gamma_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), intent(in) :: gamma_K
      real(PR) :: gamma_bar_KK

#if defined(_SRC_DISABLE_AP_CORREC_)
      gamma_bar_KK = 0._PR
#else

      real(PR) :: speed_square
#if defined(_FLUX_RUS_)
      real(PR) :: b_KK

      b_KK = compute_greatest_wave_speed(U_K, U_K, eta, tau)
      speed_square = b_KK**2

#elif defined(_FLUX_HLL_)
      real(PR) :: wave_speed_Kp, wave_speed_Km

      call compute_wave_speed(U_K, U_K, eta, tau, wave_speed_Km, wave_speed_Kp)
      speed_square = -wave_speed_Km*wave_speed_Kp
#endif /* _FLUX_ */

      gamma_bar_KK = 6._PR*compute_sigma_f(U_K)*speed_square/c - gamma_K
#endif /* _SRC_DISABLE_AP_CORREC_ */

  end function gamma_bar_KK

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$R(U)\f$.
  !>
  !> \f[\textbf{R}(\textbf{U}) = \left(E, 0, 0\right)^T \f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function R(CellK,U_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(NbEq) :: R


      R(1)   = U_K(1)
      R(2:3) = zero

  end function R

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$g(u)\f$ of the limit equation.
  !>
  !> Generic limit equation:
  !> \f[\partial_t(f(u)) - \text{div}(\alpha(u) \nabla g(u)) = 0 \f]
  !>
  !> Limit equation for \f$M_1\f$:
  !> \f[\partial_t E  - \text{div}\left(\frac{c}{3\sigma^f}\nabla E\right) = 0 \f]
  !>
  !> @param[in] U Conservative variable of the hyperbolic system
  !> @return \f$g(u)\f$, with \f$u\f$ the variable of the limit equation
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function g_u(U)
      real(PR), dimension(:), intent(in) :: U
      real(PR) :: g_u

      g_u = U(1)
  end function g_u

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the value of \f$\sigma^f\f$
  !>
  !> @param[in] U Conservative variable
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_sigma_f(U)
      real(PR), dimension(:), intent(in) :: U
      real(PR) :: compute_sigma_f

      select case(test_case)
      case  default
        compute_sigma_f = sigma_f
      end select

  end function compute_sigma_f

#endif /* _PHY_P1_ */
end module mod_asymptotic_P1
