!> Compute things related to source term and AP for the Telegraph equations.
!>
!> \f[ \left\{\begin{array}{rcl}
!>   \partial_t u + \text{div}(\vec{a} u) &=& \sigma (v - u) \\
!>   \partial_t v + \text{div}(\vec{a} v) &=& \sigma (u - v)
!> \end{array} \right. \f]
!>
!> Limit equation:
!> \f[\partial_t V - \text{div}\left(\frac{a^2}{2\sigma} \nabla V \right) = 0, \f]
!> with \f$ V = u + v \f$.
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_asymptotic_Telegraph
#if defined(_SRC_ENABLE_) && defined(_PHY_TELE_)

  use mod_numerics, only: PR, eps
  use type_mesh,  only: pmesh
  use mod_physics_Telegraph_var, only: sigma
  use mod_physics_Telegraph_var, only: NbEq
  use mod_physics_Telegraph, only: compute_greatest_wave_speed

  implicit none

  private
  public gamma_, gamma_bar_KJ, gamma_bar_KK
  public g_u
  public R

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\gamma(U)=\sigma\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function gamma_(CellK,U_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K
      real(PR) :: gamma_

      gamma_ = sigma

  end function gamma_

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\bar{\gamma}(U) = \bar{\sigma}\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] nu_J, nu_bar_J, delta_K Coefficients
  !> @param[in] U_K, U_J Solution on cells \f$K\f$ and \f$J\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal to \f$KJ\f$
  !> @param[in] gamma_K \f$\gamma(U_K)\f$
  !>
  !> @warning not properly defined
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function gamma_bar_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,gamma_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: nu_J
      real(PR), intent(in)               :: nu_bar_J
      real(PR), dimension(:), intent(in) :: delta_K
      real(PR), dimension(:), intent(in) :: U_K, U_J
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), intent(in) :: gamma_K
      real(PR) :: gamma_bar_KJ

#if defined(_SRC_DISABLE_AP_CORREC_)
      gamma_bar_KJ = 0._PR
#else
      gamma_bar_KJ = gamma_K
#endif /* _SRC_DISABLE_AP_CORREC_ */

  end function gamma_bar_KJ

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\bar{\gamma}(U) = \bar{\sigma}\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal to \f$KJ\f$
  !> @param[in] gamma_K \f$\gamma(U_K)\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_telegraph_mp_gamma_
  function gamma_bar_KK(CellK,U_K,eta,tau,gamma_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K, eta, tau
      real(PR), intent(in) :: gamma_K
      real(PR) :: gamma_bar_KK

#if defined(_SRC_DISABLE_AP_CORREC_)
      gamma_bar_KK = 0._PR
#else
      gamma_bar_KK = gamma_K
#endif /* _SRC_DISABLE_AP_CORREC_ */

  end function gamma_bar_KK

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$R(U)\f$.
  !>
  !> \f[\textbf{R}(\textbf{U}) = \left(v, u\right)^T \f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function R(CellK,U_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(NbEq) :: R

      R(1) = U_K(2)
      R(2) = U_K(1)

  end function R

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$g(u)\f$ of the limit equation.
  !>
  !> Generic limit equation:
  !> \f[\partial_t(f(u)) - \text{div}(\alpha(u) \nabla g(u)) = 0 \f]
  !>
  !> Limit equation for the Telegraph equations (with \f$V = u + v\f$):
  !> \f[\partial_t(V) - \text{div}(\frac{a^2}{2\sigma} \nabla p(V)) = 0 \f]
  !>
  !> @param[in] U Conservative variable of the hyperbolic system
  !> @return \f$g(u)\f$, with \f$u\f$ the variable of the limit equation
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function g_u(U)
      real(PR), dimension(:), intent(in) :: U
      real(PR) :: g_u

      g_u = U(1) + U(2)
  end function g_u

#endif /* _PHY_TELE_ */
end module mod_asymptotic_Telegraph
