!> Detection routines
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_detection
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)

  use mod_numerics, only: PR
  use mod_physics_var, only: NbEq
#if defined(_HO_MOOD_)
  use type_mesh, only: dim_, pmesh, BndInterfaces
  !use type_mesh, only: nodes
  !use mod_tools, only: plus1
  !use type_mesh_DLP, only: IntInterfaces
  use mod_bnd_values, only: compute_value_bnd
#if defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_))
  use mod_flux_1D_entropy_HO_MOOD, only: flux_1D_entropy_HO_MOOD, flux_1D_entropy_bnd_HO_MOOD
#if defined(_PHY_EULER_)
  use mod_physics_Euler_entropy, only: compute_entropy_math, compute_entropy_phy, compute_entropy_r
  use mod_physics_Euler_var, only: adiabatic_index
#elif defined(_PHY_SW_)
  use mod_physics_SW_entropy, only: compute_entropy_math
#endif /* _PHY_ */
#endif /* defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_)) */
#if defined(_PHY_EULER_)
  use mod_physics_Euler, only: compute_pressure
#elif defined(_PHY_HYDRORAD_)
  use mod_physics_HydroRad, only: compute_pressure
#endif /* defined(_PHY_EULER_) || defined(_PHY_HYDRORAD_) */
#if defined(_PHY_M1MG_)
  use mod_physics_M1MG_var, only: NbGroup
#endif /* defined(_PHY_M1MG_) */
#if (_HO_MOOD_MAX_DEGREE_ >= 2)
  use type_mesh, only: dim_
#endif /* _HO_MOOD_MAX_DEGREE_ */
  use type_HO_MOOD, only: Nd_max
#endif /* _HO_MOOD_ */

#if defined(_FC_INTEL_) \
  || (defined(_FC_GNU_) && ((__GNUC__>=5 && _KIND_ < 16) || __GNUC__>=6 )) \
  || defined(_FC_PGI_)
  use, INTRINSIC :: IEEE_ARITHMETIC, only: IEEE_IS_NAN
#endif /* _FC_ */

  implicit none

  private
  public IsNumericallyAdmissible
#if defined(_HO_MOOD_)
  public DoesNotOscillate
#if defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_))
  public IsEntropic
#endif /* defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_)) */
#endif /* _HO_MOOD_ */

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the value of the solution on a cell is numerically admissible
  !> \cite ClaDioLou_2011 \cite DioClaLou_2012 .
  !>
  !> @param[in] U_K Mean value on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function IsNumericallyAdmissible(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      logical :: IsNumericallyAdmissible

      integer :: eq

      IsNumericallyAdmissible = .TRUE.

      loop_eq: do eq = 1, NbEq
#if defined(_FC_INTEL_) \
  || (defined(_FC_GNU_) && ((__GNUC__>=5 && _KIND_ < 16) || __GNUC__>=6 )) \
  || defined(_FC_PGI_)
        if (IEEE_IS_NAN(U_K(eq)) ) then
          IsNumericallyAdmissible = .FALSE.
          exit loop_eq
        end if
#else
        if (ISNAN(U_K(eq))) then
          IsNumericallyAdmissible = .FALSE.
          exit loop_eq
        end if
#endif /* _FC_ */
      end do loop_eq

  end function IsNumericallyAdmissible

#if defined(_HO_MOOD_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the mean value does not oscillate
  !>
  !> For that check if the solution respect the DMP (or the ED \cite ClaFig_2014)
  !> and the u2 criterion \cite Dio_2012
  !>
  !> @param[in] CellK Index of the current cell
  !> @param[in] U_n Solution at previous step
  !> @param[in] U_n_p1 Solution at current step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function DoesNotOscillate(CellK,U_n,U_n_p1)
      integer, intent(in) :: CellK
      real(PR), dimension(:,:), intent(in) :: U_n, U_n_p1
      logical :: DoesNotOscillate

      ! Variable on which we test for the DMP/ED and u2
      integer, parameter :: eq_check = 1
#if (_HO_MOOD_MAX_DEGREE_ >= 2)
      real(PR), dimension(NbEq,Nd_max) :: R_K, R_L
      real(PR), dimension(dim_) :: curvature_min, curvature_max
#endif /* _HO_MOOD_MAX_DEGREE_ */
      real(PR) :: U_min, U_max
      real(PR) :: delta
      integer :: idxL, CellL
#if defined(_PHY_HYPHEAT_) || defined(_PHY_HYDRORAD_)
      integer :: eq
#endif /* defined(_PHY_HYPHEAT_) || defined(_PHY_HYDRORAD_) */
#if defined(_PHY_M1MG_)
      integer :: eq0, group
#endif /* defined(_PHY_M1MG_) */

      DoesNotOscillate = .TRUE.

      delta = pmesh%elt(CellK)%Delta_x

#if defined(_HO_MOOD_DMP_)
      ! First get the minimum and the maximum of the solution
      ! at previous time step over the neighborhood
      U_min = U_n(eq_check,CellK)
      U_max = U_n(eq_check,CellK)
      loop_nghbhood_MinMax: do idxL = 1, pmesh%elt(CellK)%NbNghbhood
        CellL = pmesh%elt(CellK)%nghbhood(idxL)
        U_min = min(U_min,U_n(eq_check,CellL))
        U_max = max(U_max,U_n(eq_check,CellL))
      end do loop_nghbhood_MinMax

      ! Then, check for the Discrete Maximum Principle (see (2.3) of \cite Dio_2012 )
      test_DMP: if (U_n_p1(eq_check,CellK) < U_min - 1.E-14_PR .or. U_n_p1(eq_check,CellK) > U_max + 1.E-14_PR) then
        ! Besides, check if we can relax the DMP (see 2.3.2 in \cite Dio_2012 ) on flat areas
        test_DMP_relax: if ((U_max - U_min) > delta**3) then
          DoesNotOscillate = .FALSE.
        end if test_DMP_relax
      end if test_DMP

#if defined(_PHY_EULER_) && defined(_PHY_EULER_IDEAL_GAS_)
      if (DoesNotOscillate) then
        U_min = compute_pressure(U_n(:,CellK))
        U_max = compute_pressure(U_n(:,CellK))
        loop_nghbhood_MinMax_p: do idxL = 1, pmesh%elt(CellK)%NbNghbhood
          CellL = pmesh%elt(CellK)%nghbhood(idxL)
          U_min = min(U_min,compute_pressure(U_n(:,CellL)))
          U_max = max(U_max,compute_pressure(U_n(:,CellL)))
        end do loop_nghbhood_MinMax_p

        test_DMP_p: if (compute_pressure(U_n_p1(:,CellK)) < U_min - 1.E-14_PR .or. compute_pressure(U_n_p1(:,CellK)) > U_max + 1.E-14_PR) then
          test_DMP_relax_p: if ((U_max - U_min) > delta**3) then
            DoesNotOscillate = .FALSE.
          end if test_DMP_relax_p
        end if test_DMP_p
      endif
#elif defined(_PHY_M1_)
      if (DoesNotOscillate) then
        U_min = U_n(NbEq,CellK)
        U_max = U_n(NbEq,CellK)
        loop_nghbhood_MinMax_T: do idxL = 1, pmesh%elt(CellK)%NbNghbhood
          CellL = pmesh%elt(CellK)%nghbhood(idxL)
          U_min = min(U_min,U_n(NbEq,CellL))
          U_max = max(U_max,U_n(NbEq,CellL))
        end do loop_nghbhood_MinMax_T

        test_DMP_T: if (U_n_p1(NbEq,CellK) < U_min - 1.E-14_PR .or. U_n_p1(NbEq,CellK) > U_max + 1.E-14_PR) then
          test_DMP_relax_T: if ((U_max - U_min) > delta**3) then
            DoesNotOscillate = .FALSE.
          end if test_DMP_relax_T
        end if test_DMP_T
      endif
#elif defined(_PHY_M1MG_)
      loop_group: do group = 2, NbGroup
        eq0 = (1 + dim_)*(group-1)
        if (DoesNotOscillate) then
          U_min = U_n(eq0+1,CellK)
          U_max = U_n(eq0+1,CellK)
          loop_nghbhood_MinMax_E_R: do idxL = 1, pmesh%elt(CellK)%NbNghbhood
            CellL = pmesh%elt(CellK)%nghbhood(idxL)
            U_min = min(U_min,U_n(eq0+1,CellL))
            U_max = max(U_max,U_n(eq0+1,CellL))
          end do loop_nghbhood_MinMax_E_R

          test_DMP_E_R: if (U_n_p1(eq0+1,CellK) < U_min - 1.E-14_PR .or. U_n_p1(eq0+1,CellK) > U_max + 1.E-14_PR) then
            test_DMP_relax_E_R: if ((U_max - U_min) > delta**3) then
              DoesNotOscillate = .FALSE.
              exit loop_group
            end if test_DMP_relax_E_R
          end if test_DMP_E_R
        endif
      end do loop_group

      if (DoesNotOscillate) then
        U_min = U_n(NbEq,CellK)
        U_max = U_n(NbEq,CellK)
        loop_nghbhood_MinMax_T: do idxL = 1, pmesh%elt(CellK)%NbNghbhood
          CellL = pmesh%elt(CellK)%nghbhood(idxL)
          U_min = min(U_min,U_n(NbEq,CellL))
          U_max = max(U_max,U_n(NbEq,CellL))
        end do loop_nghbhood_MinMax_T

        test_DMP_T: if (U_n_p1(NbEq,CellK) < U_min - 1.E-14_PR .or. U_n_p1(NbEq,CellK) > U_max + 1.E-14_PR) then
          test_DMP_relax_T: if ((U_max - U_min) > delta**3) then
            DoesNotOscillate = .FALSE.
          end if test_DMP_relax_T
        end if test_DMP_T
      endif
#elif defined(_PHY_HYPHEAT_)
      if (DoesNotOscillate) then
        loop_eq : do eq = 2, NbEq
          U_min = U_n(eq,CellK)
          U_max = U_n(eq,CellK)
          loop_nghbhood_MinMax_eq: do idxL = 1, pmesh%elt(CellK)%NbNghbhood
            CellL = pmesh%elt(CellK)%nghbhood(idxL)
            U_min = min(U_min,U_n(eq,CellL))
            U_max = max(U_max,U_n(eq,CellL))
          end do loop_nghbhood_MinMax_eq

          test_DMP_eq: if (U_n_p1(eq,CellK) < U_min - 1.E-14_PR .or. U_n_p1(eq,CellK) > U_max + 1.E-14_PR) then
            test_DMP_relax_eq: if ((U_max - U_min) > delta**3) then
              DoesNotOscillate = .FALSE.
              exit loop_eq
            end if test_DMP_relax_eq
          end if test_DMP_eq
        end do loop_eq
      endif
#elif defined(_PHY_HYDRORAD_)
      if (DoesNotOscillate) then
        U_min = compute_pressure(U_n(:,CellK))
        U_max = compute_pressure(U_n(:,CellK))
        loop_nghbhood_MinMax_p: do idxL = 1, pmesh%elt(CellK)%NbNghbhood
          CellL = pmesh%elt(CellK)%nghbhood(idxL)
          U_min = min(U_min,compute_pressure(U_n(:,CellL)))
          U_max = max(U_max,compute_pressure(U_n(:,CellL)))
        end do loop_nghbhood_MinMax_p

        test_DMP_p: if (compute_pressure(U_n_p1(:,CellK)) < U_min - 1.E-14_PR .or. compute_pressure(U_n_p1(:,CellK)) > U_max + 1.E-14_PR) then
          test_DMP_relax_p: if ((U_max - U_min) > delta**3) then
            DoesNotOscillate = .FALSE.
          end if test_DMP_relax_p
        end if test_DMP_p
      endif
      if (DoesNotOscillate) then
        eq = 5
        U_min = U_n(eq,CellK)
        U_max = U_n(eq,CellK)
        loop_nghbhood_MinMax_ER: do idxL = 1, pmesh%elt(CellK)%NbNghbhood
          CellL = pmesh%elt(CellK)%nghbhood(idxL)
          U_min = min(U_min,U_n(eq,CellL))
          U_max = max(U_max,U_n(eq,CellL))
        end do loop_nghbhood_MinMax_ER

        test_DMP_ER: if (U_n_p1(eq,CellK) < U_min - 1.E-14_PR .or. U_n_p1(eq,CellK) > U_max + 1.E-14_PR) then
          test_DMP_relax_ER: if ((U_max - U_min) > delta**3) then
            DoesNotOscillate = .FALSE.
          end if test_DMP_relax_ER
        end if test_DMP_ER
      endif
#endif /* _PHY_ */

#elif defined(_HO_MOOD_ED_)
      ! First get the minimum and the maximum of the solution
      ! at current time step over the neighborhood
      U_min = U_n_p1(eq_check,CellK)
      U_max = U_n_p1(eq_check,CellK)
      loop_nghbhood_MinMax: do idxL = 1, pmesh%elt(CellK)%NbNghbhood
        CellL = pmesh%elt(CellK)%nghbhood(idxL)
        U_min = min(U_min,U_n_p1(eq_check,CellL))
        U_max = max(U_max,U_n_p1(eq_check,CellL))
      end do loop_nghbhood_MinMax

      ! Then, check for the Extrema Detector (see 3.2.3 in \cite ClaFig_2014 )
      test_ED: if (U_n_p1(eq_check,CellK) < U_min - delta**3 .or. U_n_p1(eq_check,CellK) > U_max + delta**3) then
        DoesNotOscillate = .FALSE.
      end if test_ED
#endif /* _HO_MOOD_DETECTOR_ */


#if (defined(_HO_MOOD_DMP_) || defined(_HO_MOOD_ED_)) && defined(_HO_MOOD_U2_)
#if (_HO_MOOD_MAX_DEGREE_ >= 2)
      ! Next, if the solution does not respect the DMP/ED check if it is
      ! a smooth extrema with the u2 criterion
      test_not_Oscillate: if (.not. DoesNotOscillate) then
        DoesNotOscillate = .TRUE.

        ! Get the coefficients of the polynomial reconstruction
        R_K(:,:) = pmesh%elt(CellK)%CoeffPolyMOOD(:,:)

        ! Initialise the curvatures
        curvature_min(1) = R_K(eq_check,3)
        curvature_max(1) = R_K(eq_check,3)
        curvature_min(2) = R_K(eq_check,5)
        curvature_max(2) = R_K(eq_check,5)

        ! Get the minimum and the maximum of the curvatures over the neighborhood
        loop_nghbhood_curv: do idxL = 1, pmesh%elt(CellK)%NbNghbhood
          CellL = pmesh%elt(CellK)%nghbhood(idxL)
          R_L(:,:) = pmesh%elt(CellL)%CoeffPolyMOOD(:,:)
          curvature_min(1) = min(curvature_min(1),R_L(eq_check,3))
          curvature_max(1) = max(curvature_max(1),R_L(eq_check,3))

          curvature_min(2) = min(curvature_min(2),R_L(eq_check,5))
          curvature_max(2) = max(curvature_max(2),R_L(eq_check,5))
        end do loop_nghbhood_curv

        curvature_min(:) = 2._PR*curvature_min(:)
        curvature_max(:) = 2._PR*curvature_max(:)

        test_u2: if (.not. (&
            curvature_max(1)*curvature_min(1) > - delta .and. &
            (max(abs(curvature_min(1)),abs(curvature_max(1))) < delta .or. &
            abs(curvature_min(1)/curvature_max(1)) > 0.5_PR)&
            .and. &
            curvature_max(2)*curvature_min(2) > - delta .and. &
            (max(abs(curvature_min(2)),abs(curvature_max(2))) < delta .or. &
            abs(curvature_min(2)/curvature_max(2)) > 0.5_PR)&
            )) then
          DoesNotOscillate = .FALSE.
        end if test_u2

      end if test_not_Oscillate
#endif /* _HO_MOOD_MAX_DEGREE_ */
#endif /* (_HO_MOOD_DMP_ || _HO_MOOD_ED_) && _HO_MOOD_U2_ */

  end function DoesNotOscillate

#if defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_))
  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the value of the solution on a cell is entropic (see for instance \cite BerDes_2014 )
  !>
  !> The solution is entropic if it satisfies entropy inequalities:
  !> \f[\partial_t \eta(U) + \text{div}G(U) \leq 0, \f]
  !> with \f$\eta\f$ a convex function and \f$G\f$ the entropy flux.
  !>
  !> @param[in] CellK Cell to check
  !> @param[in] U_n Mean values at current time
  !> @param[in] U_n_p1 Mean values at next time step
  !> @param[in] t_n, dt Current time and time step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_entropy_mp_compute_entropy_math
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_entropy_mp_compute_entropy_phy
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_entropy_mp_compute_entropy_r

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_entropy_mp_compute_entropy_math

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_entropy_mp_flux_1d_entropy_ho_mood
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_entropy_mp_flux_1d_entropy_bnd_ho_mood
  function IsEntropic(CellK,U_n,U_n_p1,t_n,dt)
      integer, intent(in) :: CellK
      real(PR), dimension(:,:), intent(in) :: U_n, U_n_p1
      real(PR), intent(in) :: t_n, dt
      logical :: IsEntropic

      real(PR), dimension(NbEq) :: U_K
      real(PR), dimension(dim_) :: normal, tangent
      real(PR) :: flux_num_entropy
      real(PR) :: LHS
      integer :: i_K, CellL
      integer :: IdxIntInt, IdxBndInt

      U_K(:) = U_n(:,CellK)

      LHS = 0._PR
      loop_nghb: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
        CellL = pmesh%elt(CellK)%nghb(i_K)

        normal(:) = pmesh%elt(CellK)%normal(:,i_K)
        tangent(:) = pmesh%elt(CellK)%tangent(:,i_K)

        test_bnd: if (CellL /= 0) then
          IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)
          flux_num_entropy = flux_1D_entropy_HO_MOOD(U_n,t_n,CellK,CellL,IdxIntInt,i_K,normal,tangent)
        else
          IdxBndInt = pmesh%elt(CellK)%BndInterfaces(i_K)
          flux_num_entropy = flux_1D_entropy_bnd_HO_MOOD(U_n,t_n,CellK,IdxBndInt,i_K,normal,tangent)
        endif test_bnd

          LHS = LHS + pmesh%elt(CellK)%le(i_K)*flux_num_entropy

      end do loop_nghb

#if defined(_FLUX_HLLC_) && defined(_PHY_EULER_)
      LHS = LHS*dt/pmesh%elt(CellK)%area + &
          U_n_p1(1,CellK)*compute_entropy_r(U_n_p1(:,CellK)) - &
          U_K(1)*compute_entropy_r(U_K)
#else
      LHS = LHS*dt/pmesh%elt(CellK)%area + &
          compute_entropy_math(U_n_p1(:,CellK)) - &
          compute_entropy_math(U_K)
#endif /* _FLUX_HLLC_ && _PHY_EULER_ */

      !test_entropic: if (LHS > pmesh%elt(CellK)%Delta_x) then
      test_entropic: if (LHS > 1.E-14_PR) then
      !test_entropic: if (LHS > 0._PR) then
        IsEntropic = .FALSE.
      else
        IsEntropic = .TRUE.
      end if test_entropic

  end function IsEntropic
#endif /* defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_)) */

#endif /* _HO_MOOD */

#endif /* _TP_CORREC_ || _HO_MOOD_ */
end module mod_detection
