!> Routines for to compute the numerical flux of entropy
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.

module mod_flux_1D_entropy_HO_MOOD
#if defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_))

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: dim_, pmesh, nodes, BndInterfaces
  use type_mesh_DLP, only: IntInterfaces
  use mod_tools, only: plus1
#if defined(_FLUX_RUS_)
  use mod_physics, only: compute_greatest_wave_speed
#elif defined(_FLUX_HLL_)
  use mod_physics, only: compute_wave_speed, physic_flux
#elif defined(_FLUX_HLLC_) && defined(_PHY_EULER_)
  use mod_physics_Euler, only: compute_HLLC_values
  use mod_physics_Euler_entropy, only: compute_entropy_phy, compute_entropy_r
  use mod_physics_Euler, only: compute_pressure
  use mod_physics_Euler_var, only: adiabatic_index
  use mod_flux_1D, only: flux_1D
#endif /* _FLUX_ */
#if defined(_PHY_EULER_)
  use mod_physics_Euler_entropy, only: compute_entropy_math, entropy_flux
#elif defined(_PHY_SW_)
  use mod_physics_SW_entropy, only: compute_entropy_math, entropy_flux
#endif /* _PHY_ */
  use mod_physics_var, only: NbEq
  use mod_bnd_values, only: compute_value_bnd
  use type_HO_MOOD, only: Nd_max
  use mod_HO_MOOD_poly, only: compute_polynomial_reconstruction
  use mod_HO_MOOD_quad, only: quad_seg

  implicit none

  private
  public flux_1D_entropy_HO_MOOD
  public flux_1D_entropy_bnd_HO_MOOD

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the numerical entropy flux
  !>
  !> @param[in] U_K, U_L States at each sides of the interface
  !> @param[in] normal, tangent Normal and tangent vector to the interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_entropy_mp_compute_entropy_r
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_entropy_mp_compute_entropy_math
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_entropy_mp_entropy_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_hllc_values
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_physic_flux

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_entropy_mp_compute_entropy_math
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_entropy_mp_entropy_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_physic_flux
  function flux_1D_entropy(U_K,U_L,normal,tangent)
      real(PR), dimension(:), intent(in) :: U_K, U_L
      real(PR), dimension(:), intent(in) :: normal, tangent
      real(PR) :: flux_1D_entropy

      real(PR), dimension(dim_) :: G_K, G_L
#if defined(_FLUX_HLLC_) && defined(_PHY_EULER_)
      real(PR), dimension(NbEq) :: flux_num
#endif /* defined(_FLUX_HLLC_) && defined(_PHY_EULER_) */
      real(PR) :: eta_K, eta_L
#if defined(_FLUX_RUS_)
      real(PR) :: b_KL
#elif defined(_FLUX_HLL_)
      real(PR), dimension(NbEq,dim_) :: F_K, F_L
      real(PR), dimension(NbEq) :: U_KL
      real(PR) :: eta_KL
      real(PR) :: wave_speed_K, wave_speed_L
#elif defined(_FLUX_HLLC_)
      real(PR), dimension(NbEq) :: U_K_star, U_L_star
      real(PR) :: eta_K_star, eta_L_star
      real(PR) :: wave_speed_K, wave_speed_L, wave_speed_star
#endif /* _FLUX_ */

      G_K(:) = entropy_flux(U_K)
      eta_K = compute_entropy_math(U_K)

      G_L(:) = entropy_flux(U_L)
      eta_L = compute_entropy_math(U_L)

#if defined(_FLUX_RUS_)
      b_KL = compute_greatest_wave_speed(U_K, U_L, normal, tangent)

      flux_1D_entropy = 0.5_PR*(dot_product(G_K + G_L,normal) &
          - b_KL*(eta_L - eta_K))

#elif defined(_FLUX_HLL_)
      call compute_wave_speed(U_K, U_L, normal, tangent, &
          wave_speed_K, wave_speed_L)

      F_K(:,:) = physic_flux(U_K)
      F_L(:,:) = physic_flux(U_L)

      U_KL(:) =((wave_speed_L*U_L(:) - wave_speed_K*U_K(:)) &
          - MATMUL(F_L - F_K,normal)) &
          /(wave_speed_L - wave_speed_K)
      eta_KL = compute_entropy_math(U_KL)

      wave_speed_K = abs(wave_speed_K)
      wave_speed_L = abs(wave_speed_L)

      flux_1D_entropy = 0.5_PR*(dot_product(G_K + G_L,normal) &
          + (wave_speed_L - wave_speed_K)*eta_KL &
          - (wave_speed_L*eta_L - wave_speed_K*eta_K))

#elif defined(_FLUX_HLLC_) && defined(_PHY_EULER_)
      !call compute_HLLC_values(U_K, U_L, normal, tangent, &
          !wave_speed_K, wave_speed_star, wave_speed_L, &
          !U_K_star, U_L_star)

      !eta_K_star = compute_entropy_math(U_K_star)
      !eta_L_star = compute_entropy_math(U_L_star)

      !wave_speed_K = abs(wave_speed_K)
      !wave_speed_star = abs(wave_speed_star)
      !wave_speed_L = abs(wave_speed_L)

      !flux_1D_entropy = 0.5_PR*(dot_product(G_K + G_L,normal) &
          !+ (wave_speed_star - wave_speed_K)*eta_K_star &
          !+ (wave_speed_L - wave_speed_star)*eta_L_star &
          !- (wave_speed_L*eta_L - wave_speed_K*eta_K))

      flux_num(:) = flux_1D(U_K,U_L,normal,tangent)
      if (flux_num(1) > 0._PR) then
        flux_1D_entropy = flux_num(1)*compute_entropy_r(U_K)
      else
        flux_1D_entropy = flux_num(1)*compute_entropy_r(U_L)
      end if
#endif /* _FLUX_ */

  end function flux_1D_entropy

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the numerical entropy flux using the polynomial reconstruction
  !> (see Appendix A of \cite Dio_2012 for the quadrature rules)
  !>
  !> @param[in] U_n Mean values at current time
  !> @param[in] t_n Current time
  !> @param[in] CellK, CellJ Index of the cell at each side of the interface
  !> @param[in] IdxIntInt Index of the interior interface
  !> @param[in] i_K Index of the interface in CellK
  !> @param[in] eta Direction to compute the flux
  !> @param[in] tau Tangent direction
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_plus1
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_entropy_mp_flux_1d_entropy
  function flux_1D_entropy_HO_MOOD(U_n,t_n,CellK,CellJ,IdxIntInt,i_K,eta,tau)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      integer, intent(in) :: CellK, CellJ, IdxIntInt, i_K
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR) :: flux_1D_entropy_HO_MOOD

      real(PR), dimension(NbEq) :: U_K_tilde, U_J_tilde
      real(PR), dimension(NbEq,Nd_max) :: R_K, R_J
      real(PR), dimension(dim_) :: xQ, xA, xB
      integer :: IdxQuad, i_K_p1, degree

      degree = IntInterfaces(IdxIntInt)%FacePD

      test_degree: if (degree >= 1 .and. degree <= 5) then
        R_K(:,:) = pmesh%elt(CellK)%CoeffPolyMOOD(:,:)
        R_J(:,:) = pmesh%elt(CellJ)%CoeffPolyMOOD(:,:)

        i_K_p1 = plus1(i_K,pmesh%elt(CellK)%NbInterfaces)

        xA(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K))%x(:)
        xB(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K_p1))%x(:)

        flux_1D_entropy_HO_MOOD = 0._PR

        loop_quad: do IdxQuad = 1, quad_seg(degree)%nb_points
          xQ(:) = quad_seg(degree)%bar_coord(1,IdxQuad)*xA(:) &
              + quad_seg(degree)%bar_coord(2,IdxQuad)*xB(:)

          call compute_polynomial_reconstruction(U_n,CellK,degree,R_K,xQ,t_n,U_K_tilde)
          call compute_polynomial_reconstruction(U_n,CellJ,degree,R_J,xQ,t_n,U_J_tilde)

          flux_1D_entropy_HO_MOOD = flux_1D_entropy_HO_MOOD &
              + quad_seg(degree)%weights(IdxQuad)*flux_1D_entropy(U_K_tilde,U_J_tilde,eta,tau)
        end do loop_quad

#if defined(_TP_CORREC_)
      else if (degree == 0 .or. degree == -1) then
#else
      else if (degree == 0) then
#endif /* _TP_CORREC_ */
        flux_1D_entropy_HO_MOOD = flux_1D_entropy(U_n(:,CellK),U_n(:,CellJ),eta,tau)

      else
        write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong degree: ", degree
        flush(StdErr)
        STOP
      end if test_degree

  end function flux_1D_entropy_HO_MOOD

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the numerical entropy flux using the polynomial reconstruction
  !> (see Appendix A of \cite Dio_2012 for the quadrature rules)
  !> (routine for flux which involves boundary values)
  !>
  !> @param[in] U_n Mean values at current time
  !> @param[in] t_n Current time
  !> @param[in] CellK Current cell
  !> @param[in] IdxBndInt Index of the boundary interface
  !> @param[in] i_K Index of the interface in CellK
  !> @param[in] eta Direction to compute the flux
  !> @param[in] tau Tangent direction
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_plus1
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_entropy_mp_flux_1d_entropy
  function flux_1D_entropy_bnd_HO_MOOD(U_n,t_n,CellK,IdxBndInt,i_K,eta,tau)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      integer, intent(in) :: CellK, IdxBndInt, i_K
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR) :: flux_1D_entropy_bnd_HO_MOOD

      real(PR), dimension(NbEq) :: U_K_tilde, U_J_tilde
      real(PR), dimension(NbEq,Nd_max) :: R_K
      real(PR), dimension(dim_) :: xQ, xA, xB
      integer :: IdxQuad, i_K_p1, degree

      degree = BndInterfaces(IdxBndInt)%FacePD

      test_degree: if (degree >= 1 .and. degree <= 5) then
        R_K(:,:) = pmesh%elt(CellK)%CoeffPolyMOOD(:,:)

        i_K_p1 = plus1(i_K,pmesh%elt(CellK)%NbInterfaces)

        xA(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K))%x(:)
        xB(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K_p1))%x(:)

        flux_1D_entropy_bnd_HO_MOOD = 0._PR

        loop_quad: do IdxQuad = 1, quad_seg(degree)%nb_points
          xQ(:) = quad_seg(degree)%bar_coord(1,IdxQuad)*xA(:) &
              + quad_seg(degree)%bar_coord(2,IdxQuad)*xB(:)

          call compute_polynomial_reconstruction(U_n,CellK,degree,R_K,xQ,t_n,U_K_tilde)
          U_J_tilde(:) = compute_value_bnd(IdxBndInt,U_n,xQ,t_n)

          flux_1D_entropy_bnd_HO_MOOD = flux_1D_entropy_bnd_HO_MOOD &
              + quad_seg(degree)%weights(IdxQuad)*flux_1D_entropy(U_K_tilde,U_J_tilde,eta,tau)
        end do loop_quad

#if defined(_TP_CORREC_)
      else if (degree == 0 .or. degree == -1) then
#else
      else if (degree == 0) then
#endif /* _TP_CORREC_ */
        U_J_tilde(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
        flux_1D_entropy_bnd_HO_MOOD = flux_1D_entropy(U_n(:,CellK),U_J_tilde,eta,tau)

      else
        write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong degree: ", degree
        flush(StdErr)
        STOP
      end if test_degree

  end function flux_1D_entropy_bnd_HO_MOOD

#endif /* defined(_HO_MOOD_ENTROPIC_) && (defined(_PHY_EULER_) || defined(_PHY_SW_)) */
end module mod_flux_1D_entropy_HO_MOOD
