!> Computations of the numerical fluxes with source term in one direction (HO version).
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_flux_1D_source_HO_MOOD
#if defined(_HO_MOOD_) && defined(_SRC_ENABLE_)

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: pmesh, BndInterfaces, dim_
  use type_mesh_DLP, only: IntInterfaces, nu_one
  use mod_physics_var, only: NbEq
  use mod_flux_1D_source, only: flux_1D_source, flux_1D_source_NoAP
  use mod_bnd_values, only: compute_value_bnd
  use mod_tools, only: plus1
  use type_mesh, only: nodes
  use type_HO_MOOD, only: Nd_max
  use mod_HO_MOOD_quad, only: quad_seg
  use mod_HO_MOOD_poly, only: compute_polynomial_reconstruction
#if defined(_TP_CORREC_)
  use mod_asymptotic, only: compute_alpha_KK_NoAP
#endif /* _TP_CORREC_ */

  implicit none

  private
  public flux_1D_source_bnd_HO_MOOD

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the flux using the polynomial reconstruction (see Appendix A of \cite Dio_2012
  !> for the quadrature rules)
  !>
  !> @param[in] U_n Mean values at current time
  !> @param[in] CellK, CellJ Index of the cell at each side of the interface
  !> @param[in] IdxIntInt Index of the interior interface
  !> @param[in] i_K Index of the interface in CellK
  !> @param[in] idxM Side of the interface
  !> @param[in] eta Direction to compute the flux
  !> @param[in] tau Tangent direction
  !> @param[in] t_n Current time
  !> @return \f$\sum\limits_{r=1}^R \xi_{K,i,r} \ \overline{\boldsymbol{\mathcal{F}}}\left({\bf U}_{K,r}^n, {\bf U}_{J,r}^n\right)\cdot\boldsymbol{\eta}_{KJ}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_plus1
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_source_mp_flux_1d_source
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_source_mp_flux_1d_source_noap
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_alpha_kk_noap
  function flux_1D_source_HO_MOOD(U_n,CellK,CellJ,IdxIntInt,i_K,idxM,eta,tau,t_n)
      real(PR), dimension(:,:), intent(in) :: U_n
      integer, intent(in) :: CellK, CellJ, IdxIntInt, i_K, idxM
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), intent(in) :: t_n
      real(PR), dimension(NbEq) :: flux_1D_source_HO_MOOD

      real(PR), dimension(NbEq) :: U_K_tilde, U_J_tilde
      real(PR), dimension(NbEq,Nd_max) :: R_K, R_J
      real(PR), dimension(dim_) :: xQ, xA, xB
      integer ::  i_K_p1, degree, IdxQuad, idxJ
#if defined(_TP_CORREC_)
      real(PR), dimension(NbEq) :: delta_K
      real(PR) :: alpha_KK
#endif /* _TP_CORREC_ */

      degree = IntInterfaces(IdxIntInt)%FacePD
      idxJ = 1

      test_degree: if (degree >= 1 .and. degree <= 5) then
        R_K(:,:) = pmesh%elt(CellK)%CoeffPolyMOOD(:,:)
        R_J(:,:) = pmesh%elt(CellJ)%CoeffPolyMOOD(:,:)

        i_K_p1 = plus1(i_K,pmesh%elt(CellK)%NbInterfaces)

        xA(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K))%x(:)
        xB(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K_p1))%x(:)

        flux_1D_source_HO_MOOD(:) = 0._PR

        loop_quad: do IdxQuad = 1, quad_seg(degree)%nb_points
          xQ(:) = quad_seg(degree)%bar_coord(1,IdxQuad)*xA(:) &
              + quad_seg(degree)%bar_coord(2,IdxQuad)*xB(:)

          call compute_polynomial_reconstruction(U_n,CellK,degree,R_K,xQ,t_n,U_K_tilde)

          call compute_polynomial_reconstruction(U_n,CellJ,degree,R_J,xQ,t_n,U_J_tilde)

          flux_1D_source_HO_MOOD(:) = flux_1D_source_HO_MOOD(:) &
              + quad_seg(degree)%weights(IdxQuad)*flux_1D_source(CellK,&
              IntInterfaces(IdxIntInt)%nu_tilde(:,IdxQuad,idxJ,idxM),&
              IntInterfaces(IdxIntInt)%nu_bar(idxJ,idxM),&
              pmesh%elt(CellK)%delta_K_tilde(:,IdxQuad),&
              U_K_tilde,U_J_tilde,eta,tau,U_n(:,CellK),U_n(:,CellJ))

        end do loop_quad

      else if (degree == 0) then
        flux_1D_source_HO_MOOD(:) = flux_1D_source(CellK,&
            IntInterfaces(IdxIntInt)%nu(:,idxJ,idxM),&
            IntInterfaces(IdxIntInt)%nu_bar(idxJ,idxM),&
            pmesh%elt(CellK)%delta_K,&
            U_n(:,CellK),U_n(:,CellJ),eta,tau,U_n(:,CellK),U_n(:,CellJ))


#if defined(_TP_CORREC_)
      else if (degree == -1) then
        delta_K(:) = pmesh%elt(CellK)%Delta_x
        alpha_KK = compute_alpha_KK_NoAP(CellK,U_n(:,CellK),eta,tau)
        flux_1D_source_HO_MOOD(:) = flux_1D_source_NoAP(CellK,delta_K,U_n(:,CellK),U_n(:,CellJ),eta,tau,alpha_KK)
#endif /* _TP_CORREC_ */

      else
        write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong degree: ", degree
        flush(StdErr)
        STOP
      end if test_degree

  end function flux_1D_source_HO_MOOD

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the flux using the polynomial reconstruction (see Appendix A of \cite Dio_2012
  !> for the quadrature rules) (routine for flux which involves boundary values)
  !>
  !> @param[in] U_n Mean values at current time
  !> @param[in] t_n Current time
  !> @param[in] CellK Current cell
  !> @param[in] IdxBndInt Index of the boundary interface
  !> @param[in] i_K Index of the interface in CellK
  !> @param[in] eta Direction to compute the flux
  !> @param[in] tau Tangent direction
  !> @return \f$\sum\limits_{r=1}^R \xi_{K,i,r} \ \overline{\boldsymbol{\mathcal{F}}}\left({\bf U}_{K,r}^n, {\bf U}_{J,r}^n\right)\cdot\boldsymbol{\eta}_{KJ}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_plus1
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_source_mp_flux_1d_source
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_source_mp_flux_1d_source_noap
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_alpha_kk_noap
  function flux_1D_source_bnd_HO_MOOD(U_n,t_n,CellK,IdxBndInt,i_K,eta,tau)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      integer, intent(in) :: CellK, IdxBndInt, i_K
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), dimension(NbEq) :: flux_1D_source_bnd_HO_MOOD

      real(PR), dimension(NbEq) :: U_K_tilde, U_J_tilde
      real(PR), dimension(NbEq,Nd_max) :: R_K
      real(PR), dimension(dim_) :: xQ, xA, xB
      integer :: IdxQuad, i_K_p1, degree
#if defined(_TP_CORREC_)
      real(PR), dimension(NbEq) :: delta_K
      real(PR) :: alpha_KK
#endif /* _TP_CORREC_ */

      degree = BndInterfaces(IdxBndInt)%FacePD

      test_degree: if (degree >= 1 .and. degree <= 5) then
        R_K(:,:) = pmesh%elt(CellK)%CoeffPolyMOOD(:,:)

        i_K_p1 = plus1(i_K,pmesh%elt(CellK)%NbInterfaces)

        xA(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K))%x(:)
        xB(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K_p1))%x(:)

        flux_1D_source_bnd_HO_MOOD(:) = 0._PR

        loop_quad: do IdxQuad = 1, quad_seg(degree)%nb_points
          xQ(:) = quad_seg(degree)%bar_coord(1,IdxQuad)*xA(:) &
              + quad_seg(degree)%bar_coord(2,IdxQuad)*xB(:)

          call compute_polynomial_reconstruction(U_n,CellK,degree,R_K,xQ,t_n,U_K_tilde)

          U_J_tilde(:) = compute_value_bnd(IdxBndInt,U_n,xQ,t_n)

          flux_1D_source_bnd_HO_MOOD(:) = flux_1D_source_bnd_HO_MOOD(:) + &
              quad_seg(degree)%weights(IdxQuad)*flux_1D_source(CellK,&
              nu_one,1._PR/BndInterfaces(IdxBndInt)%dKL,&
              pmesh%elt(CellK)%delta_K_tilde(:,IdxQuad),&
              U_K_tilde,U_J_tilde,eta,tau,U_n(:,CellK),U_J_tilde)

        end do loop_quad

      else if (degree == 0) then
        U_J_tilde(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
        flux_1D_source_bnd_HO_MOOD(:) = flux_1D_source(CellK,&
            nu_one,1._PR/BndInterfaces(IdxBndInt)%dKL,&
            pmesh%elt(CellK)%delta_K,&
            U_n(:,CellK),U_J_tilde,eta,tau,U_n(:,CellK),U_J_tilde)

#if defined(_TP_CORREC_)
      else if (degree == -1) then
        U_J_tilde(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
        delta_K(:) = pmesh%elt(CellK)%Delta_x
        alpha_KK = compute_alpha_KK_NoAP(CellK,U_n(:,CellK),eta,tau)
        flux_1D_source_bnd_HO_MOOD(:) = flux_1D_source_NoAP(CellK,delta_K,U_n(:,CellK),U_J_tilde,eta,tau,alpha_KK)
#endif /* _TP_CORREC_ */

      else
        write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong degree: ", degree
        flush(StdErr)
        STOP
      end if test_degree
  end function flux_1D_source_bnd_HO_MOOD

#endif /* _HO_MOOD_ && _SRC_ENABLE_ */
end module mod_flux_1D_source_HO_MOOD
