!> Computations of the numerical fluxes for the acoustic-transport splitting
!>
!> \cite Gir_2014 , \cite ChaGirKok_2016 .
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_flux_AT
#if defined(_FLUX_AT_)

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: dim_
  use type_mesh, only: pmesh, BndInterfaces
  use type_mesh_DLP, only: IntInterfaces

  use mod_physics_var, only: NbEq
  use mod_bnd_values, only: compute_value_bnd

#if defined(_PHY_EULER_)
  use mod_physics_Euler_AT, only: compute_speed_relax
  use mod_physics_Euler_AT, only: compute_u_star, compute_pi_star
  use mod_physics_Euler_AT, only: conservative_to_relaxed
  use mod_physics_Euler_AT, only: lagrangian_to_conservative
  use mod_physics_Euler_AT, only: conservative_to_lagrangian
#endif /* defined(_PHY_EULER_) */

#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells, chunk_IntInterfaces, chunk_BndInterfaces
#endif /* _FC_PAR_OMP_ */

#if defined(_FLUX_AT_IMEX_) && defined(_ENABLE_PETSC_)
  use mod_petsc, only: petsc_init, petsc_create, petsc_solve, petsc_get_solution, petsc_end
#endif /* defined(_FLUX_AT_IMEX_) && defined(_ENABLE_PETSC_) */


  implicit none

  private
  public AT_init, AT_end
  public compute_U_acoustic
  public flux_AT
  public U_acoustic, W_alpha

  real(PR), dimension(:,:), allocatable, save :: W_alpha
  real(PR), dimension(:,:), allocatable, save :: U_acoustic

contains

  !--------------------------------------------------------------------------
  !> @brief
  !> Initialize arrays relative to the acoustic transport splitting
  !>
  !> @authors Florian Blachère
  !> @authors Sébastien Guisset
  !> @authors Maxime Stauffert
  !--------------------------------------------------------------------------
  subroutine AT_init()

    allocate(U_acoustic(NbEq,pmesh%NbCells))
    allocate(W_alpha(NbEq,pmesh%NbCells))
#if defined(_ENABLE_PETSC_) && defined(_FLUX_AT_IMEX_)
    call Petsc_init()
#endif /* defined(_ENABLE_PETSC_) && defined(_FLUX_AT_IMEX_) */

  end subroutine AT_init

  !--------------------------------------------------------------------------
  !> @brief
  !> Finalize arrays relative to the acoustic transport splitting
  !>
  !> @authors Florian Blachère
  !> @authors Sébastien Guisset
  !> @authors Maxime Stauffert
  !--------------------------------------------------------------------------
  subroutine AT_end()

    deallocate(U_acoustic)
    deallocate(W_alpha)
#if defined(_ENABLE_PETSC_) && defined(_FLUX_AT_IMEX_)
    call Petsc_end()
#endif /* defined(_ENABLE_PETSC_) && defined(_FLUX_AT_IMEX_) */

  end subroutine AT_end

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the numerical flux associated to the acoustic part
  !>
  !> @param[in] a_KJ Relaxation "speed"
  !> @param[in] W_alpha_K, W_alpha_L Primite relaxed variables
  !> @param[in] normal_i Direction to compute the flux
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function flux_acoustic(a_KJ,W_alpha_K,W_alpha_L,normal_i)
    real(PR), intent(in) :: a_KJ
    real(PR), dimension(:), intent(in) :: W_alpha_K, W_alpha_L
    real(PR), dimension(:), intent(in) :: normal_i
    real(PR), dimension(NbEq) :: flux_acoustic

#if defined(_PHY_EULER_)
    real(PR) :: pi_star, u_star

    u_star = compute_u_star(a_KJ,W_alpha_K,W_alpha_L,normal_i)
    pi_star = compute_pi_star(a_KJ,W_alpha_K,W_alpha_L,normal_i)
    flux_acoustic(1) = -u_star
    flux_acoustic(2:3) = pi_star*normal_i(:)
    flux_acoustic(4) = pi_star*u_star
#else
    write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> flux not defined."
    flush(StdErr)
    STOP
#endif
  end function flux_acoustic

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the conservative variable during the acoustic step
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] dt Time step
  !> @param[in] t_n Current time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_U_acoustic(U_n,dt,t_n)
    real(PR), dimension(:,:), intent(in) :: U_n
    real(PR), intent(in) :: dt, t_n

    real(PR), dimension(NbEq) :: flux_n, U_lagragian
    real(PR), dimension(NbEq) :: U_bnd, W_alpha_bnd
    real(PR), dimension(dim_) :: normal_i
    real(PR) :: le_i, a_i
    integer :: CellK, CellL
    integer :: i_K
    integer :: IdxIntInt, IdxBndInt

    !==================================================================!
    !      Initialisation and get the primitive relaxed variables
    !==================================================================!
    !$OMP DO &
    !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
    do CellK = 1, pmesh%NbCells
      call conservative_to_lagrangian(U_n(:,CellK),U_acoustic(:,CellK))
    end do
    !$OMP END DO
    call compute_W_alpha(U_n,dt,t_n)
    !==================================================================!

    !==================================================================!
    !         Loop over the interior interfaces to compute the flux
    !==================================================================!
    !$OMP DO &
    !$OMP REDUCTION(+:U_acoustic) &
    !$OMP SCHEDULE(STATIC,chunk_IntInterfaces)
    do IdxIntInt = 1, pmesh%NbIntInterfaces
      CellK = IntInterfaces(IdxIntInt)%nghb(1)
      CellL = IntInterfaces(IdxIntInt)%nghb(2)
      le_i = IntInterfaces(IdxIntInt)%le
      i_K = IntInterfaces(IdxIntInt)%i_Cell(1)
      normal_i = pmesh%elt(CellK)%normal(:,i_K)
      a_i = compute_speed_relax(U_n(:,CellK),U_n(:,CellL))

      flux_n(:) = flux_acoustic(a_i,W_alpha(:,CellK),W_alpha(:,CellL),normal_i)

      U_acoustic(:,CellK) = U_acoustic(:,CellK) - dt*le_i/(pmesh%elt(CellK)%area*U_n(1,CellK))*flux_n(:)
      U_acoustic(:,CellL) = U_acoustic(:,CellL) + dt*le_i/(pmesh%elt(CellL)%area*U_n(1,CellL))*flux_n(:)
    end do
    !$OMP END DO
    !==================================================================!


    !==================================================================!
    !                Loop over the boundary interfaces
    !==================================================================!
    !$OMP DO &
    !$OMP REDUCTION(+:U_acoustic) &
    !$OMP SCHEDULE(STATIC,chunk_BndInterfaces)
    do IdxBndInt = 1, pmesh%NbBndInterfaces
      CellK = BndInterfaces(IdxBndInt)%IdxCell
      i_K = BndInterfaces(IdxBndInt)%IdxInterface
      normal_i = pmesh%elt(CellK)%normal(:,i_K)
      le_i = pmesh%elt(CellK)%le(i_K)

      U_bnd(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
      a_i = compute_speed_relax(U_n(:,CellK),U_bnd)
      W_alpha_bnd(:) = compute_value_bnd(IdxBndInt,W_alpha,BndInterfaces(IdxBndInt)%x,t_n,'prim')
      flux_n(:) = flux_acoustic(a_i,W_alpha(:,CellK),W_alpha_bnd,normal_i)

      U_acoustic(:,CellK) = U_acoustic(:,CellK) - dt*le_i/(pmesh%elt(CellK)%area*U_n(1,CellK))*flux_n(:)
    end do
    !$OMP END DO
    !==================================================================!

    !==================================================================!
    !   Loop over the cells to get back to the conservative variables
    !==================================================================!
    !$OMP DO &
    !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
    do CellK = 1, pmesh%NbCells
      U_lagragian(:) = U_acoustic(:,CellK)
      call lagrangian_to_conservative(U_lagragian,U_acoustic(:,CellK))
    end do
    !$OMP END DO
    !==================================================================!

  end subroutine compute_U_acoustic

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the numerical flux associated to AT splitting
  !>
  !> @param[in] U_K, U_L States at each sides of the interface
  !> @param[in] U_acoustic_K, U_acoustic_L Conservative variables at the
  !>   end of the acoustic step
  !> @param[in] W_alpha_K, W_alpha_L Primitive relaxed variables
  !> @param[in] normal_i Direction to compute the flux
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function flux_AT(U_K,U_L,U_acoustic_K,U_acoustic_L,W_alpha_K,W_alpha_L,normal_i)
    real(PR), dimension(:), intent(in) :: U_K, U_L
    real(PR), dimension(:), intent(in) :: U_acoustic_K, U_acoustic_L
    real(PR), dimension(:), intent(in) :: W_alpha_K, W_alpha_L
    real(PR), dimension(:), intent(in) :: normal_i
    real(PR), dimension(NbEq) :: flux_AT

#if defined(_PHY_EULER_)
    real(PR) :: pi_star, u_star
    real(PR) :: a_KL
    real(PR) :: rho_star, E_star
    real(PR), dimension(dim_) :: rho_u_star

    a_KL = compute_speed_relax(U_K,U_L)
    u_star = compute_u_star(a_KL,W_alpha_K,W_alpha_L,normal_i)
    pi_star = compute_pi_star(a_KL,W_alpha_K,W_alpha_L,normal_i)

    if (u_star > 0._PR) then
      rho_star = U_acoustic_K(1)
      rho_u_star(:) = U_acoustic_K(2:3)
      E_star = U_acoustic_K(4)
    else
      rho_star = U_acoustic_L(1)
      rho_u_star(:) = U_acoustic_L(2:3)
      E_star = U_acoustic_L(4)
    end if

    flux_AT(1) = rho_star*u_star
    flux_AT(2:3) = rho_u_star(:)*u_star + pi_star*normal_i(:)
    flux_AT(4) = (E_star + pi_star)*u_star

#else
    write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> flux not defined."
    flush(StdErr)
    STOP
#endif
  end function flux_AT

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$W^\alpha\f$ ie calculus of the primitive relaxed variables
  !> \f$(\rho,u,v,\pi)\f$ from the eulerian conservative variables
  !> \f$(\rho,\rho u, \rho v, E)\f$
  !>
  !> @param[in] U_n eulerian conservative variables at time \f$t^n\f$
  !> @param[in] dt Time step
  !> @param[in] t_n Current time
  !>
  !> @authors Florian Blachère
  !> @authors Sébastien Guisset
  !> @authors Maxime Stauffert
  !--------------------------------------------------------------------!
  subroutine compute_W_alpha(U_n,dt,t_n)
    real(PR), dimension(:,:), intent(in) :: U_n
    real(PR), intent(in) :: dt, t_n

#if defined(_FLUX_AT_EXEX_)
    integer :: CellK

    !$OMP DO &
    !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
    do CellK = 1, pmesh%NbCells
      call conservative_to_relaxed(U_n(:,CellK),W_alpha(:,CellK))
    end  do
    !$OMP END DO

#elif defined(_ENABLE_PETSC_) && defined(_FLUX_AT_IMEX_)
    call petsc_create(U_n,dt,t_n)
    !$OMP SINGLE
    call petsc_solve()
    !$OMP END SINGLE
    call petsc_get_solution(U_n,W_alpha)

#else
    write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Method not defined"
    flush(StdErr)
    STOP
#endif /* defined(_FLUX_AT_) */
  end subroutine compute_W_alpha

#endif /* defined(_FLUX_AT_) */
end module mod_flux_AT
