!> Computations of the numerical fluxes.
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_flux_num

  use mod_numerics, only: PR
  use type_mesh, only: pmesh, BndInterfaces, dim_
  use mod_physics_var, only: NbEq
  use type_mesh_DLP, only: IntInterfaces, nu_one
  use mod_flux_1D, only: flux_1D
  use mod_bnd_values, only: compute_value_bnd
#if defined(_SRC_ENABLE_)
  use mod_flux_1D_source, only: flux_1D_source, flux_1D_source_NoAP
#endif /* _SRC_ENABLE_ */
#if defined(_FLUX_AT_)
  use mod_flux_AT, only: flux_AT
  use mod_flux_AT, only: U_acoustic, W_alpha
#endif /* _FLUX_AT_*/


  implicit none

  private
  public compute_flux_n
  public compute_flux_bnd_n
#if defined(_SRC_ENABLE_)
  public compute_flux_bnd_n_NoAP
#endif /* _SRC_ENABLE_ */

contains

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the flux on an interface using the scheme
  !>
  !> @param[in] CellK, IdxIntInt, idxM Indexes of the cell, the interface
  !>            and the point \f$M\f$ to use (1 or 2)
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[out] flux_n Flux
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_mp_flux_1d
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_src_mp_flux_1d_source
  subroutine compute_flux_n(CellK,IdxIntInt,idxM,U_n,t_n,flux_n)
      integer, intent(in) :: CellK, IdxIntInt, idxM
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), dimension(:),  intent(out) :: flux_n

      real(PR), dimension(dim_) :: eta, tau
      real(PR), dimension(NbEq) :: U_K, U_J, nu_J
      integer :: idxJ, CellJ
      integer :: IdxBndInt
#if defined(_FLUX_AT_)
      real(PR), dimension(NbEq) :: U_acoustic_J, W_alpha_J
#endif /* defined(_FLUX_AT_) */
#if defined(_SRC_ENABLE_)
      real(PR) :: nu_bar_J
#endif /* _SRC_ENABLE_ */

      U_K(:) = U_n(:,CellK)
      flux_n(:) = 0._PR

      loop_S_Ki: do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
        eta(:) = IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
        tau(:) = IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM)
        CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)
        nu_J(:) = IntInterfaces(IdxIntInt)%nu(:,idxJ,idxM)

        if (CellJ > 0) then
          U_J(:) = U_n(:,CellJ)
#if defined(_FLUX_AT_)
          U_acoustic_J(:) = U_acoustic(:,CellJ)
          W_alpha_J(:) = W_alpha(:,CellJ)
#endif /* defined(_FLUX_AT_) */
        else
          IdxBndInt = - CellJ
          U_J(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
#if defined(_FLUX_AT_)
          U_acoustic_J(:) = compute_value_bnd(IdxBndInt,U_acoustic,BndInterfaces(IdxBndInt)%x,t_n)
          W_alpha_J(:) = compute_value_bnd(IdxBndInt,W_alpha,BndInterfaces(IdxBndInt)%x,t_n,'prim')
#endif /* defined(_FLUX_AT_) */
        endif
#if defined(_SRC_ENABLE_)
        nu_bar_J = IntInterfaces(IdxIntInt)%nu_bar(idxJ,idxM)
        flux_n(:) = flux_n(:) + nu_J(:)*flux_1D_source(CellK,nu_J,nu_bar_J,pmesh%elt(CellK)%delta_K,U_K,U_J,eta,tau)
#else
#if defined(_FLUX_AT_)
        flux_n(:) = flux_n(:) + nu_J(:)*flux_AT(U_K(:),U_J(:),&
            U_acoustic(:,CellK),U_acoustic_J(:),&
            W_alpha(:,CellK),W_alpha_J(:),eta)
#else
        flux_n(:) = flux_n(:) + nu_J(:)*flux_1D(U_K,U_J,eta,tau)
#endif /* defined(_FLUX_AT_) */
#endif /* _SRC_ENABLE_ */
      end do loop_S_Ki

  end subroutine compute_flux_n

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the flux across boundary interfaces
  !>
  !> @param[in] CellK, IdxBndInt, i_K  Indexes of the cell and the boundary interfaces
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[out] flux_n Normal flux across the boundary interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_bnd_values_mp_compute_value_bnd
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_mp_flux_1d
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_src_mp_flux_1d_source
  subroutine compute_flux_bnd_n(CellK,IdxBndInt,i_K,U_n,t_n,flux_n)
      integer, intent(in) :: CellK, IdxBndInt, i_K
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), dimension(:), intent(out) :: flux_n

      real(PR), dimension(NbEq) :: U_J, U_K
#if defined(_FLUX_AT_)
      real(PR), dimension(NbEq) :: U_acoustic_bnd, W_alpha_bnd
#endif /* defined(_FLUX_AT_) */
      real(PR), dimension(dim_) :: normal, tangent
#if defined(_SRC_ENABLE_)
      real(PR) :: nu_bar_J
#endif /* _SRC_ENABLE_ */

      U_K(:) = U_n(:,CellK)

      normal(:)  = pmesh%elt(CellK)%normal(:,i_K)
      tangent(:) = pmesh%elt(CellK)%tangent(:,i_K)

      ! Get the boundary value
      U_J(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
#if defined(_FLUX_AT_)
      U_acoustic_bnd(:) = compute_value_bnd(IdxBndInt,U_acoustic,BndInterfaces(IdxBndInt)%x,t_n)
      W_alpha_bnd(:) = compute_value_bnd(IdxBndInt,W_alpha,BndInterfaces(IdxBndInt)%x,t_n,'prim')
      flux_n(:) = flux_AT(U_K(:),U_J,&
          U_acoustic(:,CellK),U_acoustic_bnd,&
          W_alpha(:,CellK),W_alpha_bnd,normal)
#else

      ! Compute the flux
#if defined(_SRC_ENABLE_)
      nu_bar_J =  1._PR/BndInterfaces(IdxBndInt)%dKL
      flux_n(:) = flux_1D_source(CellK,nu_one,nu_bar_J,pmesh%elt(CellK)%delta_K,U_K,U_J,normal,tangent)
#else
      flux_n(:) = flux_1D(U_K,U_J,normal,tangent)
#endif /* _SRC_ENABLE_ */
#endif /* defined(_FLUX_AT_) */

  end subroutine compute_flux_bnd_n

#if defined(_SRC_ENABLE_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the flux across boundary interfaces without AP correction.
  !>
  !> @param[in] CellK, IdxBndInt, i_K  Indexes of the cell and the boundary interfaces
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[in] alpha_KK Coefficient \f$\alpha_{KK}\f$ with or without AP corection
  !> @param[out] flux_n Normal flux across the boundary interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_bnd_values_mp_compute_value_bnd
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_src_mp_flux_1d_source_noap
  subroutine compute_flux_bnd_n_NoAP(CellK,IdxBndInt,i_K,U_n,t_n,alpha_KK,flux_n)
    integer, intent(in)                    :: CellK, IdxBndInt, i_K
    real(PR), dimension(:,:), intent(in) :: U_n
    real(PR), intent(in)                   :: t_n, alpha_KK
    real(PR), dimension(:), intent(out) :: flux_n

    real(PR), dimension(NbEq) :: U_J, U_K
    real(PR), dimension(dim_) :: normal, tangent

    U_K(:) = U_n(:,CellK)
    U_J(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)

    normal(:)  = pmesh%elt(CellK)%normal(:,i_K)
    tangent(:) = pmesh%elt(CellK)%tangent(:,i_K)

    ! Compute the flux
    flux_n(:) =  flux_1D_source_NoAP(CellK,pmesh%elt(CellK)%delta_K,U_K,U_J,normal,tangent,alpha_KK)

  end subroutine compute_flux_bnd_n_NoAP
#endif /* _SRC_ENABLE_ */

end module mod_flux_num
