!> Computations of the numerical fluxes (HO version).
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_flux_num_HO_MOOD
#if defined(_HO_MOOD_)

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: nodes, pmesh, BndInterfaces, dim_
  use mod_physics_var, only: NbEq
  use type_mesh_DLP, only: IntInterfaces, max_Card_S, NbM, nu_one
  use mod_flux_1D, only: flux_1D
#if defined(_SRC_ENABLE_)
  use mod_flux_1D_source, only: flux_1D_source, flux_1D_source_NoAP
  use mod_flux_1D_source_HO_MOOD, only: flux_1D_source_bnd_HO_MOOD
#if defined(_TP_CORREC_)
  use mod_asymptotic, only: compute_alpha_KK_NoAP
#endif /* _TP_CORREC_ */
#endif /* _SRC_ENABLE_ */
  use mod_bnd_values, only: compute_value_bnd
  use mod_flux_1D_HO_MOOD, only: flux_1D_bnd_HO_MOOD
  use type_HO_MOOD, only: Nd_max, MaxDegree
  use mod_tools, only: plus1
  use mod_HO_MOOD_poly, only: compute_polynomial_reconstruction
  use mod_HO_MOOD_quad, only: quad_seg

  implicit none

  private
  public compute_flux_n_HO_MOOD
  public compute_flux_bnd_n_HO_MOOD

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the flux on an interface using the HO scheme
  !>
  !> @param[in] CellK, IdxIntInt, idxM Indexes of the cell, the interface
  !>            and the point \f$M\f$ to use (1 or 2)
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[out] flux_n Flux
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_plus1

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_mp_flux_1d
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_ho_mood_mp_flux_1d_ho_mood
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_ho_mood_mp_flux_1d_bnd_ho_mood

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_source_mp_flux_1d_source
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_source_mp_flux_1d_source_noap
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_alpha_kk_noap
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_source_ho_mood_mp_flux_1d_source_bnd_ho_mood

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  subroutine compute_flux_n_HO_MOOD(CellK,IdxIntInt,idxM,U_n,t_n,flux_n)
      integer, intent(in)                  :: CellK, IdxIntInt, idxM
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), dimension(:),  intent(out) :: flux_n

      real(PR), dimension(NbEq) :: U_K, U_J
      real(PR), dimension(NbEq) :: U_K_tilde, U_J_tilde
      real(PR), dimension(dim_) :: eta, tau
      real(PR), dimension(dim_) :: xQ, xA, xB
      integer  :: idxJ, CellJ, i_K, i_K_p1
      integer :: IdxQuad
      integer :: IdxBndInt
      integer :: degree
#if defined(_SRC_ENABLE_) && defined(_TP_CORREC_)
      real(PR), dimension(NbEq) :: delta_K
      real(PR) :: alpha_KK
#endif /* _SRC_ENABLE_ && _TP_CORREC_*/

      real(PR), dimension(NbEq,Nd_max) :: R_K

      i_K = IntInterfaces(IdxIntInt)%i_Cell(idxM)
      degree = IntInterfaces(IdxIntInt)%FacePD
      flux_n(:) = 0._PR

      !====================================================================!
      test_degree: if (degree == 0) then
        U_K(:) = U_n(:,CellK)

       loop_S_Ki: do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
          eta(:) = IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
          tau(:) = IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM)
          CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)

          if (CellJ > 0) then
            U_J(:) = U_n(:,CellJ)
          else
            IdxBndInt = - CellJ
            U_J(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
          endif

#if defined(_SRC_ENABLE_)
          flux_n(:) = flux_n(:) + IntInterfaces(IdxIntInt)%nu(:,idxJ,idxM)*&
              flux_1D_source(CellK,IntInterfaces(IdxIntInt)%nu(:,idxJ,idxM),&
              IntInterfaces(IdxIntInt)%nu_bar(idxJ,idxM),pmesh%elt(CellK)%delta_K,&
              U_K,U_J,eta,tau,U_K,U_J)
#else
          flux_n(:) = flux_n(:) + IntInterfaces(IdxIntInt)%nu(:,idxJ,idxM)*flux_1D(U_K,U_J,eta,tau)
#endif /* _SRC_ENABLE_ */
        end do loop_S_Ki
        !====================================================================!

        !====================================================================!
#if defined(_TP_CORREC_)
      else if (degree == -1) then
        eta(:) = pmesh%elt(CellK)%normal(:,i_K)
        tau(:) = pmesh%elt(CellK)%tangent(:,i_K)
        CellJ = pmesh%elt(CellK)%nghb(i_K)
#if defined(_SRC_ENABLE_)
        alpha_KK = compute_alpha_KK_NoAP(CellK,U_n(:,CellK),eta,tau)
        delta_K(:) = pmesh%elt(CellK)%Delta_x
        flux_n(:) = flux_1D_source_NoAP(CellK,delta_K,U_n(:,CellK),U_n(:,CellJ),eta,tau,alpha_KK)
#else
        flux_n(:) = flux_1D(U_n(:,CellK),U_n(:,CellJ),eta,tau)
#endif /* _SRC_ENABLE_ */
#endif /* _TP_CORREC_ */
        !====================================================================!

        !====================================================================!
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
      else if (degree >= 1 .and. degree <= 5) then

        U_K(:) = U_n(:,CellK)

        R_K(:,:) = pmesh%elt(CellK)%CoeffPolyMOOD(:,:)
        i_K_p1 = plus1(i_K,pmesh%elt(CellK)%NbInterfaces)

        xA(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K))%x(:)
        xB(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K_p1))%x(:)

        loop_quad: do IdxQuad = 1, quad_seg(degree)%nb_points
          xQ(:) = quad_seg(degree)%bar_coord(1,IdxQuad)*xA(:) &
              + quad_seg(degree)%bar_coord(2,IdxQuad)*xB(:)

          call compute_polynomial_reconstruction(U_n,CellK,degree,R_K,xQ,t_n,U_K_tilde)

          loop_S_Ki_MOOD: do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
            CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)
            test_bnd: if (CellJ > 0) then
              call compute_polynomial_reconstruction(U_n,CellJ,degree,pmesh%elt(CellJ)%CoeffPolyMOOD,xQ,t_n,U_J_tilde)
#if defined(_SRC_ENABLE_)
              U_J(:) = U_n(:,CellJ)
#endif /* _SRC_ENABLE_ */
            else
              IdxBndInt = - CellJ
              CellJ = BndInterfaces(IdxBndInt)%IdxCell
              U_J_tilde(:) = compute_value_bnd(IdxBndInt,U_n,xQ,t_n)
#if defined(_SRC_ENABLE_)
              U_J(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
#endif /* _SRC_ENABLE_ */
            end if test_bnd

            eta(:) = IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
            tau(:) = IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM)
#if defined(_SRC_ENABLE_)
            flux_n(:) = flux_n(:) + &
                quad_seg(degree)%weights(IdxQuad)*IntInterfaces(IdxIntInt)%nu_tilde(:,idxJ,IdxQuad,idxM)*&
                flux_1D_source(CellK,IntInterfaces(IdxIntInt)%nu_tilde(:,idxJ,IdxQuad,idxM),&
                IntInterfaces(IdxIntInt)%nu_bar(idxJ,idxM),&
                pmesh%elt(CellK)%delta_K_tilde(:,idxQuad),&
                U_K_tilde,U_J_tilde,eta,tau,U_K,U_J)
#else
            flux_n(:) = flux_n(:) + &
                quad_seg(degree)%weights(IdxQuad)*IntInterfaces(IdxIntInt)%nu_tilde(:,idxJ,IdxQuad,idxM)*&
                flux_1D(U_K_tilde,U_J_tilde,eta,tau)
#endif /* _SRC_ENABLE_ */
          end do loop_S_Ki_MOOD

        end do loop_quad

#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
        !====================================================================!

        !====================================================================!
      else
        write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong degree: ", degree
        flush(StdErr)
        STOP
        !====================================================================!

      end if test_degree

  end subroutine compute_flux_n_HO_MOOD

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the flux across boundary interfaces
  !>
  !> @param[in] CellK, IdxBndInt, i_K  Indexes of the cell and the boundary interfaces
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[out] flux_n Normal flux across the boundary interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_ho_mood_mp_flux_1d_bnd_ho_mood
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_source_ho_mood_mp_flux_1d_source_bnd_ho_mood
  subroutine compute_flux_bnd_n_HO_MOOD(CellK,IdxBndInt,i_K,U_n,t_n,flux_n)
      integer, intent(in)                    :: CellK, IdxBndInt, i_K
      real(PR), dimension(:,:), intent(in)   :: U_n
      real(PR), intent(in)                   :: t_n
      real(PR), dimension(NbEq), intent(out) :: flux_n

#if defined(_SRC_ENABLE_)
      flux_n(:) = flux_1D_source_bnd_HO_MOOD(U_n,t_n,CellK,IdxBndInt,i_K,pmesh%elt(CellK)%normal(:,i_K),pmesh%elt(CellK)%tangent(:,i_K))
#else
      flux_n(:) = flux_1D_bnd_HO_MOOD(U_n,t_n,CellK,IdxBndInt,i_K,pmesh%elt(CellK)%normal(:,i_K),pmesh%elt(CellK)%tangent(:,i_K))
#endif /* _SRC_ENABLE_ */

  end subroutine compute_flux_bnd_n_HO_MOOD

#endif /* _HO_MOOD_ */
end module mod_flux_num_HO_MOOD
