!> Subroutines to read <tt>.msh</tt> files coming from Gmsh (http://geuz.org/gmsh/) \cite Gmsh_2009
!> and create primary and secondary meshes.
!>
!> @authors Florian Blachère
!> @authors Rodolphe Turpault
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.

module mod_mesh_D2P
#if defined(_MESH_D2P_)

  use mod_numerics, only: PR, pi, InitMin, InitMax, eps
  use type_IO, only: StdOut, StdErr
  use type_mesh, only: dim_, Max_Nb_Nghb
  use type_mesh, only: type_mesh_ps
  use type_mesh, only: dmesh, pmesh, smesh, nodes, BndInterfaces
  use type_mesh, only: Neumann_BND, Dirichlet_BND, Wall_BND
  use type_mesh_DLP, only: IntInterfaces
  use mod_tools, only: CrossProduct2D, plus1
  use mod_tools_array, only:  IsPresent
  use mod_tools_mesh, only: TriangleArea, compute_normal
#include "../include/gccversion.h"
#if defined(_FC_PGI_) \
  || (defined(_FC_GNU_) && _GNUC_VERSION_LT_(4,6,0)) \
  || (defined(_FC_GNU_) && _GNUC_VERSION_LT_(6,1,0) && _KIND_ == 16)
  use mod_tools_array, only: NORM2
#endif
  use mod_open_file, only: OpenFile

  implicit none

  private
  public read_msh, create_meshes, free_meshes
  public MESH_D2P_ENABLE_SMESH, MESH_D2P_ENABLE_DMESH

!#define _MESH_D2P_ENABLE_SMESH_
#if defined(_MESH_D2P_ENABLE_SMESH_)
  logical, parameter :: MESH_D2P_ENABLE_SMESH = .TRUE.
#else
  logical, parameter :: MESH_D2P_ENABLE_SMESH = .FALSE.
#endif /* _MESH_D2P_ENABLE_SMESH_*/

!#define _MESH_D2P_ENABLE_DMESH_
#if defined(_MESH_D2P_ENABLE_DMESH_)
  logical, parameter :: MESH_D2P_ENABLE_DMESH = .TRUE.
#else
  logical, parameter :: MESH_D2P_ENABLE_DMESH = .FALSE.
#endif /* _MESH_D2P_ENABLE_DMESH_ */

  integer, parameter :: MSH_LINE  = 1
  integer, parameter :: MSH_QUAD  = 3
  integer, parameter :: MSH_POINT = 15

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Read a diamond mesh created by Gmsh (http://geuz.org/gmsh/) \cite Gmsh_2009 .
  !>
  !> @param[in] msh_file Path to the \c .msh file
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  subroutine read_msh(msh_file)
      CHARACTER(len=*), intent(in) :: msh_file

      real(PR), dimension(3)    :: coord
      real(PR), dimension(dim_) :: x_min, x_max
      character(len=8)          :: readLine
      integer :: CellK, d
      integer :: IdxBndInt, NbElements, diam
      integer :: idx, ElementType, NbTag, PhysicalTag, GeometricalTag
      integer :: FileId

      x_min = InitMin
      x_max = InitMax

      write(StdOut,'(2A)')'Opening the following mesh file: ', TRIM(msh_file)
      call OpenFile(unit=FileId,file=TRIM(msh_file), action='read', status='old')

      ! Read the number of nodes in the mesh
      readLine = ''
      do while (.NOT. readLine(2:6)=='Nodes')
        read (FileId, '(A8)') readLine
        readLine = ADJUSTL(readLine)
      end do
      read (FileId, *) dmesh%NbVert
      write(StdOut,'(A,I9)')'Nb vertices: ',dmesh%NbVert
      if (dmesh%NbVert<2) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " ->  You need at least two nodes in the mesh."
        flush(StdErr)
        STOP
      end if

      allocate(nodes(dmesh%NbVert))
      nodes(:)%msh     = 0
      nodes(:)%onb     = .FALSE.

      ! Read the nodes
      do CellK = 1, dmesh%NbVert
        ! Trying to read a node
        read (FileId,*) idx, coord
        nodes(CellK)%x = coord(1:dim_)
        do d = 1,dim_
          x_min(d) = min(coord(d),x_min(d))
          x_max(d) = max(coord(d),x_max(d))
        end do
      end do

      pmesh%x_min(:) = x_min(:)
      pmesh%x_max(:) = x_max(:)
      pmesh%length(:) = pmesh%x_max(:) - pmesh%x_min(:)
      pmesh%x_center(:) = 0.5_PR*(pmesh%x_max(:) + pmesh%x_min(:))

      readLine = ''
      do while (.NOT.(readLine(2:5)=='Elem'))
        read (FileId, '(A8)') readLine
        readLine = ADJUSTL(readLine)
      end do
      read (FileId, *) NbElements

      dmesh%id     = 0
      dmesh%NbDiams = 0
      dmesh%NbBndInterfaces   = 0

      ! First loop to get the number of elements (diamonds and boundary interfaces)
      do CellK = 1, NbElements
        read (FileId, *) idx, ElementType, NbTag, PhysicalTag, GeometricalTag
        if (NbTag /= 2) then
          write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> more than two tags in the mesh file"
          flush(StdErr)
          STOP
        end if

        if (ElementType == MSH_QUAD) then ! Quadrangle
          dmesh%NbDiams = dmesh%NbDiams + 1
        elseif (ElementType == MSH_POINT) then ! Single point

        elseif (ElementType == MSH_LINE .and. PhysicalTag >= 100) then ! Interior line

        elseif (ElementType == MSH_LINE .and. PhysicalTag <  100) then ! Boundary line
          dmesh%NbBndInterfaces = dmesh%NbBndInterfaces + 1
        else
          write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, " -> unknown element: ", ElementType
          flush(StdErr)
          STOP
        end if
      end do

      if (dmesh%NbDiams<1) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> You nead at least one element in the mesh"
        flush(StdErr)
        STOP
      end if

      write(StdOut,'(A,I9)')'Nb diamonds: ',dmesh%NbDiams

      allocate(dmesh%elt(4,dmesh%NbDiams))
      allocate(dmesh%normals(2*dim_,dmesh%NbDiams))
      allocate(dmesh%diags(2,dmesh%NbDiams))
      allocate(dmesh%area(dmesh%NbDiams))

      pmesh%NbIntInterfaces = dmesh%NbDiams
#if defined (_MESH_D2P_ENABLE_SMESH_)
      smesh%NbIntInterfaces = dmesh%NbDiams
#endif /* _MESH_D2P_ENABLE_SMESH_ */
      allocate(IntInterfaces(pmesh%NbIntInterfaces))

      pmesh%NbBndInterfaces = dmesh%NbBndInterfaces
#if defined (_MESH_D2P_ENABLE_SMESH_)
      smesh%NbBndInterfaces = smesh%NbBndInterfaces
#endif /* _MESH_D2P_ENABLE_SMESH_ */
      allocate(BndInterfaces(dmesh%NbBndInterfaces))

      ! Second loop to get the elements, for that we need to restart the reading of the file
      close(FileId)
      readLine = ''
      call OpenFile(unit=FileId,file=TRIM(msh_file), action='read', status='old')
      do while (.NOT.(readLine(2:5)=='Elem'))
        read (FileId, '(A8)') readLine
        readLine = ADJUSTL(readLine)
      end do
      read (FileId, *) NbElements

      IdxBndInt = 0
      diam = 0
      do CellK = 1, NbElements
        read (FileId,*) idx, ElementType, NbTag, PhysicalTag, GeometricalTag
        if (ElementType == MSH_QUAD) then ! Quadrangle
          diam = diam + 1
          backspace(FileId)  ! Read one line back
          read (FileId, *) idx, ElementType, NbTag, PhysicalTag, GeometricalTag, dmesh%elt(:,diam)
        elseif (ElementType == MSH_LINE .and. PhysicalTag < 100) then ! Boundary line
          IdxBndInt = IdxBndInt + 1
          backspace(FileId)  ! Read one line back
          allocate(BndInterfaces(IdxBndInt)%BndTag(NbTag))
          read (FileId, *) idx, ElementType, NbTag, BndInterfaces(IdxBndInt)%BndTag(:), BndInterfaces(IdxBndInt)%vertices(:)

          nodes(BndInterfaces(IdxBndInt)%vertices(1))%onb = .TRUE.
          nodes(BndInterfaces(IdxBndInt)%vertices(2))%onb = .TRUE.
        end if

      end do

      close(FileId)

      write(StdOut,'(A)')'Computational area:'
      write(StdOut,'(A,ES23.16,A,ES23.16,A)') 'x in: [', x_min(1),' ;', x_max(1),']'
      write(StdOut,'(A,ES23.16,A,ES23.16,A)') 'y in: [', x_min(2),' ;', x_max(2),']'

      write(StdOut,'(A)')'End of the file reading'
      write(StdOut,'(A)')"-------------------------------------------------------"

  end subroutine read_msh

  !--------------------------------------------------------------------!
  !> @brief
  !> Create the primary and secondary meshes from the diamond mesh
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !>
  !--------------------------------------------------------------------!
  subroutine create_meshes()
      integer :: i, j, CellK, CellL
      integer :: a, b, c, d
      integer :: a_loc, b_loc, a_glob, b_glob
      integer :: next, NbInterfaces, NbNghb
      integer :: diam
      integer :: NbEltPrimary, NbEltSecondary
      integer :: i_K
      integer :: idxBndInt, idxIntInt
      integer, dimension(:,:,:), allocatable :: cellinterface
      integer, dimension(:,:),   allocatable :: infointerface
      integer, dimension(:),     allocatable :: nc
      integer, dimension(Max_Nb_Nghb)    :: nghb
      real(PR), dimension(dim_) :: xA,xB,xC,xD
      real(PR) :: TotalArea, perimeter
#if defined (_DEBUG_)
      real(PR), dimension(dim_) :: sum_ei_ni
#endif /* _DEBUG */

#if defined(_MESH_D2P_ENABLE_SMESH_)
      write(StdOut,'(A)')'Creating the primary and secondary meshes...'
#else
      write(StdOut,'(A)')'Creating the primary mesh...'
#endif /* _MESH_D2P_ENABLE_SMESH_ */

      allocate(cellinterface(2,Max_Nb_Nghb,dmesh%NbVert))
      allocate(infointerface(Max_Nb_Nghb,dmesh%NbVert))
      allocate(nc(dmesh%NbVert))

      cellinterface(:,:,:) = 0
      nc(:)           = 0

      pmesh%id = 1
      smesh%id = 2

      ! Loop on the boundary interfaces
      do idxBndInt = 1, dmesh%NbBndInterfaces
        nc(BndInterfaces(idxBndInt)%vertices(1)) =  nc(BndInterfaces(idxBndInt)%vertices(1))+1
        cellinterface(1:2,nc(BndInterfaces(idxBndInt)%vertices(1)),BndInterfaces(idxBndInt)%vertices(1)) = BndInterfaces(idxBndInt)%vertices(1:2)
        nc(BndInterfaces(idxBndInt)%vertices(2)) =  nc(BndInterfaces(idxBndInt)%vertices(2))+1
        cellinterface(1:2,nc(BndInterfaces(idxBndInt)%vertices(2)),BndInterfaces(idxBndInt)%vertices(2)) = BndInterfaces(idxBndInt)%vertices(1:2)
      end do

      ! Connect the interfaces
      do diam = 1, dmesh%NbDiams
        a = dmesh%elt(1,diam)
        b = dmesh%elt(2,diam)
        c = dmesh%elt(3,diam)
        d = dmesh%elt(4,diam)

        nc(a) = nc(a)+1
        cellinterface(1,nc(a),a)=d
        cellinterface(2,nc(a),a)=b

        nc(b) = nc(b)+1
        cellinterface(1,nc(b),b)=c
        cellinterface(2,nc(b),b)=a

        nc(c) = nc(c)+1
        cellinterface(1,nc(c),c)=b
        cellinterface(2,nc(c),c)=d

        nc(d) = nc(d)+1
        cellinterface(1,nc(d),d)=a
        cellinterface(2,nc(d),d)=c
      end do

      ! Save the neighbours of a point
      do CellK = 1, dmesh%NbVert
        NbNghb = 0
        nghb(:) = 0
        do j = 1, nc(CellK)
          do i = 1, 2
            CellL = cellinterface(i,j,CellK)
            if (.not. IsPresent(CellL,nghb) .and. CellL /= CellK) then
              NbNghb = NbNghb + 1
              nghb(NbNghb) = CellL
            end if
          end do
        end do
        allocate(nodes(CellK)%nghb(NbNghb))
        nodes(CellK)%NbNghb = NbNghb
        nodes(CellK)%nghb(:) = nghb(1:NbNghb)
      end do

      ! Mark the element for pmesh or smesh
      CALL mark(pmesh%NbCells, smesh%NbCells)

      ! Free the connectivity of nodes (we only need it to mark the nodes)
      do CellK = 1, dmesh%NbVert
        deallocate(nodes(CellK)%nghb)
      end do

      if (pmesh%NbCells + smesh%NbCells /= dmesh%NbVert) then
        write(StdErr,'(A,3(I8,1X))')'Error, two many points: ', dmesh%NbVert, pmesh%NbCells, smesh%NbCells
        flush(StdErr)
        STOP
      end If

      ! The cells of a mesh are the vertices of the other
      pmesh%NbVert = smesh%NbCells
      smesh%NbVert = pmesh%NbCells

      write(StdOut,'(A,I8)')'Nb primary cells:    ', pmesh%NbCells
      write(StdOut,'(A,I8)')'Nb secondaray cells: ', smesh%NbCells

      allocate(pmesh%elt(pmesh%NbCells))
#if defined(_MESH_D2P_ENABLE_SMESH_)
      allocate(smesh%elt(smesh%NbCells))
#endif /* _MESH_D2P_ENABLE_SMESH_ */


      ! Create pmesh and smesh from the cellinterface array
      NbEltPrimary = 1
      NbEltSecondary = 1
      do CellK = 1, dmesh%NbVert
        if (nodes(CellK)%msh == pmesh%id) then
          if (nodes(CellK)%onb) then
            pmesh%elt(NbEltPrimary)%onb = .TRUE.
          else
            pmesh%elt(NbEltPrimary)%onb = .FALSE.
          endif
          NbInterfaces = COUNT(cellinterface(1,:,CellK)/=0)
          pmesh%elt(NbEltPrimary)%GlobalIdx = CellK
          pmesh%elt(NbEltPrimary)%x(:) = nodes(CellK)%x(:)
          pmesh%elt(NbEltPrimary)%NbInterfaces = NbInterfaces
          nodes(CellK)%LocalIdx = NbEltPrimary
          allocate(pmesh%elt(NbEltPrimary)%e(2,NbInterfaces))
          allocate(pmesh%elt(NbEltPrimary)%vertices_glob(NbInterfaces))
          allocate(pmesh%elt(NbEltPrimary)%IntInterfaces(NbInterfaces))
          allocate(pmesh%elt(NbEltPrimary)%nghb_glob(NbInterfaces))
          allocate(pmesh%elt(NbEltPrimary)%nghb(NbInterfaces))
          allocate(pmesh%elt(NbEltPrimary)%normal(dim_,NbInterfaces))
          allocate(pmesh%elt(NbEltPrimary)%tangent(dim_,NbInterfaces))
          allocate(pmesh%elt(NbEltPrimary)%le(NbInterfaces))
          allocate(pmesh%elt(NbEltPrimary)%BndInterfaces(NbInterfaces))
          pmesh%elt(NbEltPrimary)%BndInterfaces(:) = 0
          do idxIntInt = 1, pmesh%elt(NbEltPrimary)%NbInterfaces
            pmesh%elt(NbEltPrimary)%e(:,idxIntInt) = cellinterface(:,idxIntInt,CellK)
          end do
          NbEltPrimary = NbEltPrimary + 1
        else if (nodes(CellK)%msh == smesh%id) then
#if defined(_MESH_D2P_ENABLE_SMESH_)
          if (nodes(CellK)%onb) then
            smesh%elt(NbEltSecondary)%onb = .TRUE.
          else
            smesh%elt(NbEltSecondary)%onb = .FALSE.
          endif
          NbInterfaces=COUNT(cellinterface(1,:,CellK)/=0)
          smesh%elt(NbEltSecondary)%GlobalIdx = CellK
          smesh%elt(NbEltSecondary)%x(:) = nodes(CellK)%x(:)
          smesh%elt(NbEltSecondary)%NbInterfaces = NbInterfaces
          nodes(CellK)%LocalIdx = NbEltSecondary
          allocate(smesh%elt(NbEltSecondary)%e(2,NbInterfaces))
          allocate(smesh%elt(NbEltSecondary)%vertices_glob(NbInterfaces))
          allocate(smesh%elt(NbEltSecondary)%IntInterfaces(NbInterfaces))
          allocate(smesh%elt(NbEltSecondary)%nghb_glob(NbInterfaces))
          allocate(smesh%elt(NbEltSecondary)%nghb(NbInterfaces))
          allocate(smesh%elt(NbEltSecondary)%normal(dim_,NbInterfaces))
          allocate(smesh%elt(NbEltSecondary)%tangent(dim_,NbInterfaces))
          allocate(smesh%elt(NbEltSecondary)%le(NbInterfaces))
          allocate(smesh%elt(NbEltSecondary)%BndInterfaces(NbInterfaces))
          smesh%elt(NbEltSecondary)%BndInterfaces(:) = 0
          do idxIntInt = 1, smesh%elt(NbEltSecondary)%NbInterfaces
            smesh%elt(NbEltSecondary)%e(:,idxIntInt) = cellinterface(:,idxIntInt,CellK)
          end do
          NbEltSecondary = NbEltSecondary + 1
#endif /* _MESH_D2P_ENABLE_SMESH_ */
        else
          write(StdErr,'(2A,I4,A,I8)')__FILE__,':', __LINE__, " -> one nodes was not marked: ", CellK
          flush(StdErr)
          STOP
        end if
      end do

      deallocate(cellinterface)

      write(StdOut,'(A)')'Connecting vertices...'
      call connect_vertices(pmesh)
#if defined(_MESH_D2P_ENABLE_SMESH_)
      call connect_vertices(smesh)
#endif /* _MESH_D2P_ENABLE_SMESH_ */
      write(StdOut,'(A)')'Computations of the normals, areas...'
      write(StdOut,'(A)')'For primary mesh: '
      call compute_geometrical(pmesh)
#if defined(_MESH_D2P_ENABLE_SMESH_)
      write(StdOut,'(A)')'For secondary mesh: '
      call compute_geometrical(smesh)
#endif /* _MESH_D2P_ENABLE_SMESH_ */

      ! Connect diamonds, and compute some stuff about diamonds
      TotalArea=0._PR
      infointerface(:,:) = 0
      nc(:) = 1
      dmesh%Delta_x = InitMin
      ! First loop to get the information
      do diam = 1, dmesh%NbDiams
        a = dmesh%elt(1,diam)
        b = dmesh%elt(2,diam)
        c = dmesh%elt(3,diam)
        d = dmesh%elt(4,diam)

        infointerface(nc(a),a) = diam ! Associated diamond
        nc(a) = nc(a)+1
        infointerface(nc(b),b) = diam
        nc(b) = nc(b)+1
        infointerface(nc(c),c) = diam
        nc(c) = nc(c)+1
        infointerface(nc(d),d) = diam
        nc(d) = nc(d)+1

        xA(:) = nodes(a)%x(:)
        xB(:) = nodes(b)%x(:)
        xC(:) = nodes(c)%x(:)
        xD(:) = nodes(d)%x(:)

        dmesh%area(diam) = TriangleArea(xA,xB,xC) + TriangleArea(xC,xD,xA)
        dmesh%normals(1:dim_,diam) = compute_normal(xB,xD,xA)
        dmesh%normals(dim_+1:2*dim_,diam) = compute_normal(xA,xC,xB)
        dmesh%diags(1,diam)=NORM2(xC-xA)
        dmesh%diags(2,diam)=NORM2(xD-xB)
        TotalArea = TotalArea+dmesh%area(diam)

        ! Space step
        perimeter = norm2(xA-xB) + norm2(xB-xC) + norm2(xC-xD) + norm2(xD-xA)
        dmesh%Delta_x = min(dmesh%Delta_x,dmesh%area(diam)/perimeter)

      end do

      write(StdOut,'(A)')'For diamond mesh: '
      write(StdOut,'(A,ES23.16)')'* Total area of the mesh:         ', TotalArea
      write(StdOut,'(A,ES23.16)')'* Space step of the mesh:         ', dmesh%Delta_x

      write(StdOut,'(A)')'Dispatching informations...'
      call dispatch(nc, infointerface, pmesh)
#if defined(_MESH_D2P_ENABLE_SMESH_)
      call dispatch(nc, infointerface, smesh)
#endif /* _MESH_D2P_ENABLE_SMESH_ */

      deallocate(infointerface)
      deallocate(nc)

      ! Loop on the boundary interfaces to link them to the elements
      do idxBndInt = 1, dmesh%NbBndInterfaces
        ! Global number of the two vertices of the interface
        a_glob = BndInterfaces(idxBndInt)%vertices(1)
        b_glob = BndInterfaces(idxBndInt)%vertices(2)
        ! Local number of the two vertices of the interface (in pmesh or smesh)
        a_loc = nodes(a_glob)%LocalIdx
        b_loc = nodes(b_glob)%LocalIdx

        if (nodes(a_glob)%msh == pmesh%id) then
          BndInterfaces(idxBndInt)%IdxCell = a_loc
          NbInterfaces = pmesh%elt(a_loc)%NbInterfaces
          do j=1, NbInterfaces
            a = pmesh%elt(a_loc)%vertices_glob(j)
            next = plus1(j,NbInterfaces)
            b = pmesh%elt(a_loc)%vertices_glob(next)
            if ( (a == a_glob .and. b == b_glob) &
                .or. (a == b_glob .and. b == a_glob)) exit
          end do
          if (j>NbInterfaces) then
            write(StdErr,'(2A,I4,A,I8)')__FILE__,':', __LINE__, " -> trouble to link the boundary interface: ", idxBndInt
            write(StdErr,'(A,I8)')"to the element: ", a_glob
            flush(StdErr)
            STOP
          end if
          BndInterfaces(idxBndInt)%IdxInterface = j
          pmesh%elt(a_loc)%BndInterfaces(j) = idxBndInt
          BndInterfaces(idxBndInt)%x(:) = pmesh%elt(a_loc)%x(:) + pmesh%Delta_x*pmesh%elt(a_loc)%normal(:,j)
          BndInterfaces(idxBndInt)%dKL = pmesh%Delta_x

#if defined(_MESH_D2P_ENABLE_SMESH_)
          BndInterfaces(i)%IdxSecCell = b_loc
          NbInterfaces = smesh%elt(b_loc)%NbInterfaces
          do j=1, NbInterfaces
            a = smesh%elt(b_loc)%vertices_glob(j)
            next = plus1(j,NbInterfaces)
            b = smesh%elt(b_loc)%vertices_glob(next)
            if ( (a == a_glob .and. b == b_glob) &
                .or. (a == b_glob .and. b == a_glob)) exit
          end do
          if (j>NbInterfaces) then
            write(StdErr,'(2A,I4,A,I8)')__FILE__,':', __LINE__, " -> trouble to link the boundary interface: ", idxBndInt
            write(StdErr,'(A,I8)')"to the element: ", b_glob
            flush(StdErr)
            STOP
          end if
          BndInterfaces(idxBndInt)%IdxSecInterface = j
          smesh%elt(b_loc)%BndInterfaces(j) = idxBndInt
          BndInterfaces(idxBndInt)%x(:) = smesh%elt(b_loc)%x(:) + smesh%Delta_x*smesh%elt(b_loc)%normal(:,j)
          BndInterfaces(idxBndInt)%dKL = smesh%Delta_x
#endif /* _MESH_D2P_ENABLE_SMESH_ */

        else
#if defined(_MESH_D2P_ENABLE_SMESH_)
          BndInterfaces(idxBndInt)%IdxSecCell = a_loc
          NbInterfaces = smesh%elt(a_loc)%NbInterfaces
          do j=1, NbInterfaces
            a = smesh%elt(a_loc)%vertices_glob(j)
            next = plus1(j,NbInterfaces)
            b = smesh%elt(a_loc)%vertices_glob(next)
            if ( (a == a_glob .and. b == b_glob) &
                .or. (a == b_glob .and. b == a_glob)) exit
          end do
          if (j>NbInterfaces) then
            write(StdErr,'(2A,I4,A,I8)')__FILE__,':', __LINE__, " -> trouble to link the boundary interface: ", idxBndInt
            write(StdErr,'(A,I8)')"to the element: ", a_glob
            flush(StdErr)
            STOP
          end if
          BndInterfaces(idxBndInt)%IdxSecInterface = j
          smesh%elt(a_loc)%BndInterfaces(j) = idxBndInt
          BndInterfaces(idxBndInt)%x(:) = smesh%elt(a_loc)%x(:) + smesh%Delta_x*smesh%elt(a_loc)%normal(:,j)
          BndInterfaces(idxBndInt)%dKL = smesh%Delta_x
#endif /* _MESH_D2P_ENABLE_SMESH_ */

          BndInterfaces(idxBndInt)%IdxCell = b_loc
          NbInterfaces = pmesh%elt(b_loc)%NbInterfaces
          do j=1,NbInterfaces
            a = pmesh%elt(b_loc)%vertices_glob(j)
            next = plus1(j,NbInterfaces)
            b = pmesh%elt(b_loc)%vertices_glob(next)
            if ( (a == a_glob .and. b == b_glob) &
                .or. (a == b_glob .and. b == a_glob)) exit
          end do
          if (j>NbInterfaces) then
            write(StdErr,'(2A,I4,A,I8)')__FILE__,':', __LINE__, " -> trouble to link the boundary interface: ", idxBndInt
            write(StdErr,'(A,I8)')"to the element: ", b_glob
            flush(StdErr)
            STOP
          end if
          BndInterfaces(idxBndInt)%IdxInterface = j
          pmesh%elt(b_loc)%BndInterfaces(j) = idxBndInt
          BndInterfaces(idxBndInt)%x(:) = pmesh%elt(b_loc)%x(:) + pmesh%Delta_x*pmesh%elt(b_loc)%normal(:,j)
          BndInterfaces(idxBndInt)%dKL = pmesh%Delta_x

        end if

      end do

#if defined(_HO_MOOD_)
      write(StdOut,'(A)')'Getting neighborhood...'
      call GetNghbhood(pmesh)
#if defined(_MESH_D2P_ENABLE_SMESH_)
      call GetNghbhood(smesh)
#endif /* _MESH_D2P_ENABLE_SMESH_ */
#endif /* _HO_MOOD_  */

#if defined (_DEBUG_)
      do CellK = 1, pmesh%NbCells
        sum_ei_ni(:) = 0._PR
        do i_K = 1, pmesh%elt(CellK)%NbInterfaces
         sum_ei_ni(:) = sum_ei_ni(:) + pmesh%elt(CellK)%le(i_K)*pmesh%elt(CellK)%normal(:,i_K)
        end do
        if (abs(sum_ei_ni(1)) > eps .or. abs(sum_ei_ni(2)) > eps) then
          write(StdOut,'(2A,I4,A)') __FILE__,':', __LINE__, " -> Cell not closed in pmesh:"
          write(StdOut,'(I8,1X,ES23.16,1X,ES23.16)') CellK, sum_ei_ni(1), sum_ei_ni(2)
          read(*,*)
        end if
      end do
#if defined(_MESH_D2P_ENABLE_SMESH_)
      do CellK = 1, smesh%NbCells
        sum_ei_ni(:) = 0._PR
        do i_K = 1, smesh%elt(CellK)%NbInterfaces
         sum_ei_ni(:) = sum_ei_ni(:) + smesh%elt(CellK)%le(i)*smesh%elt(CellK)%normal(:,i)
        end do
        if (abs(sum_ei_ni(1)) > eps .or. abs(sum_ei_ni(2)) > eps) then
          write(StdOut,'(2A,I4,A)') __FILE__,':', __LINE__, " -> Cell not closed in smesh:"
          write(StdOut,'(I8,1X,ES23.16,1X,ES23.16)') CellK, sum_ei_ni(1), sum_ei_ni(2)
        end if
      end do
#endif /* _MESH_D2P_ENABLE_SMESH_ */
#endif /* _DEBUG_ */


      ! Link the primary elements to the interior interfaces
      ! TODO move that somewhere for more efficiency
      do idxIntInt = 1, pmesh%NbIntInterfaces
        ! One diagonal of the diamond is the interface between K and L,
        ! the other one is the segment between them.

        ! Get the cell centers
        if (nodes(dmesh%elt(1,idxIntInt))%msh == pmesh%id) then
          CellK = nodes(dmesh%elt(1,idxIntInt))%LocalIdx
          CellL = nodes(dmesh%elt(3,idxIntInt))%LocalIdx
        else
          CellK = nodes(dmesh%elt(2,idxIntInt))%LocalIdx
          CellL = nodes(dmesh%elt(4,idxIntInt))%LocalIdx
        end if

        IntInterfaces(idxIntInt)%nghb = [CellK, CellL]
      end do

      ! We can free some memory because the significant
      ! informations are saved in interfaces
      if (.not. MESH_D2P_ENABLE_DMESH) then
        deallocate(dmesh%elt)
        deallocate(dmesh%normals)
        deallocate(dmesh%diags)
        deallocate(dmesh%area)
      endif

#if defined(_MESH_D2P_ENABLE_SMESH_)
      write(StdOut,'(A)')'End of the creation of the primary and secondary meshes'
#else
      write(StdOut,'(A)')'End of the creation of the primary mesh'
#endif /* _MESH_D2P_ENABLE_SMESH_ */

  end subroutine create_meshes

  !--------------------------------------------------------------------!
  !> @brief
  !> Connect the vertices from the interfaces informations.
  !>
  !> @param[in,out] mesh Primary or secondary mesh
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  subroutine connect_vertices(mesh)
      type(type_mesh_ps), intent(inout) :: mesh

      integer :: i_K, j, CellK
      integer :: NbInterfaces
      logical, dimension(Max_Nb_Nghb)    :: mk
      logical                            :: test

      real(PR) :: SignedArea
      real(PR), dimension(dim_) :: x_i, x_i_p1

#if defined(_IO_ENABLE_VTK_)
      mesh%NbInterfaces = mesh%NbCells
#endif /* _IO_ENABLE_VTK_ */
      do CellK = 1, mesh%NbCells
#if defined(_IO_ENABLE_VTK_)
        mesh%NbInterfaces = mesh%NbInterfaces + mesh%elt(CellK)%NbInterfaces
#endif /* _IO_ENABLE_VTK_ */
        mk(:) = .FALSE.
        mesh%elt(CellK)%vertices_glob(1:2) = mesh%elt(CellK)%e(1:2,1)
        mk(1)=.TRUE.
        do i_K = 2, mesh%elt(CellK)%NbInterfaces-1
          j = 1
          test = .TRUE.
          do while ((j <= mesh%elt(CellK)%NbInterfaces) .and. test)
            if (mesh%elt(CellK)%vertices_glob(i_K)==mesh%elt(CellK)%e(1,j)) then
              if (.not. mk(j)) then
                test  = .FALSE.
                mk(j) = .TRUE.
                mesh%elt(CellK)%vertices_glob(i_K+1)=mesh%elt(CellK)%e(2,j)
              end if
            elseif(mesh%elt(CellK)%vertices_glob(i_K)==mesh%elt(CellK)%e(2,j)) then
              if (.not. mk(j)) then
                test  = .FALSE.
                mk(j) = .TRUE.
                mesh%elt(CellK)%vertices_glob(i_K+1)=mesh%elt(CellK)%e(1,j)
              end if
            end if
            j=j+1
          end do
        end do
        deallocate(mesh%elt(CellK)%e)
      end do

      ! All the vertices of a cell need to turn counterclockwise,
      ! if the signed area of the polygon is < 0, then turn the vertices
      ! using: https://en.wikipedia.org/wiki/Shoelace_formula
      do CellK = 1, mesh%NbCells

        SignedArea = 0._PR
        NbInterfaces = mesh%elt(CellK)%NbInterfaces

        do i_K = 1, NbInterfaces
          x_i(:) = nodes(mesh%elt(CellK)%vertices_glob(i_K))%x(:)
          x_i_p1(:) = nodes(mesh%elt(CellK)%vertices_glob(plus1(i_K,NbInterfaces)))%x(:)
          SignedArea = SignedArea + x_i(1)*x_i_p1(2) - x_i_p1(1)*x_i(2)
        end do

        !SignedArea = 0.5_PR*SignedArea

        if (SignedArea < 0._PR) then
          mesh%elt(CellK)%vertices_glob(2:) = mesh%elt(CellK)%vertices_glob(mesh%elt(CellK)%NbInterfaces:2:-1)
        end if

      end do

  end subroutine connect_vertices

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute some geometrical stuff: normal, area, length...
  !>
  !> @param[in,out] mesh Primary or secondary mesh
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  subroutine compute_geometrical(mesh)
      type(type_mesh_ps), intent(inout) :: mesh

      integer :: CellK, i_K, NbInterfaces
      integer :: CellK_glob, A_glob, B_glob
      real(PR), dimension(dim_) :: xA, xB, xK, xAB, n
      real(PR) :: perimeter, le

      mesh%area = 0._PR
      mesh%MaxNbNghb = 0
      mesh%Delta_x = InitMin

      do CellK = 1, mesh%NbCells
        mesh%elt(CellK)%area = 0._PR
        perimeter = 0._PR

        NbInterfaces = mesh%elt(CellK)%NbInterfaces
        mesh%MaxNbNghb = max(mesh%MaxNbNghb,mesh%elt(CellK)%NbInterfaces)
        CellK_glob = mesh%elt(CellK)%GlobalIdx
        xK(:) = nodes(CellK_glob)%x(:)

        do i_K = 1, NbInterfaces - 1
          A_glob = mesh%elt(CellK)%vertices_glob(i_K)
          B_glob = mesh%elt(CellK)%vertices_glob(i_K+1)
          xA(:) = nodes(A_glob)%x(:)
          xB(:) = nodes(B_glob)%x(:)
          xAB(:) = xB(:) - xA(:)
          if ((CellK_glob /= A_glob) .AND. (CellK_glob /= B_glob)) then
            mesh%elt(CellK)%area = mesh%elt(CellK)%area + TriangleArea(xA,xB,xK)
          !else
            ! In this case the triangle (A,B,CellK) is degenerate, so we
            ! do not need to compute the area
          end if
          le = NORM2(xAB)
          mesh%elt(CellK)%le(i_K) = le
          perimeter = perimeter + le
          mesh%elt(CellK)%tangent(:,i_K) = xAB(:)/le
          n(1) =  xAB(2)
          n(2) = -xAB(1)
          mesh%elt(CellK)%normal(:,i_K)  = n(:)/le
        end do

        i_K = NbInterfaces
        A_glob = mesh%elt(CellK)%vertices_glob(i_K)
        B_glob = mesh%elt(CellK)%vertices_glob(1)
        xA(:) = nodes(A_glob)%x(:)
        xB(:) = nodes(B_glob)%x(:)
        xAB(:) = xB(:) - xA(:)
        if ((CellK_glob /= A_glob) .AND. (CellK_glob /= B_glob)) then
          mesh%elt(CellK)%area = mesh%elt(CellK)%area + TriangleArea(xA,xB,xK)
        !else
          ! In this case the triangle (A,B,CellK) is degenerate, so we
          ! do not need to compute the area
        end if
        le = NORM2(xAB)
        mesh%elt(CellK)%le(i_K) = le
        perimeter = perimeter + le
        mesh%elt(CellK)%tangent(:,i_K) = xAB(:)/le
        n(1) =  xAB(2)
        n(2) = -xAB(1)
        mesh%elt(CellK)%normal(:,i_K)  = n(:)/le

        mesh%area = mesh%area + mesh%elt(CellK)%area
#if defined (_MESH_SAVE_PERIMETER_)
        mesh%elt(CellK)%perimeter = perimeter
#endif /* _MESH_SAVE_PERIMETER_ */
#if defined(_SRC_ENABLE_) || defined(_TP_FLUX_) || defined(_HO_MOOD_)
        mesh%elt(CellK)%Delta_x = mesh%elt(CellK)%area/perimeter
#endif /* _SRC_ENABLE_ || _TP_FLUX_ || _HO_MOOD_ */
        mesh%Delta_x = min(mesh%Delta_x,mesh%elt(CellK)%area/perimeter)
      end do

      write(StdOut,'(A,ES23.16)')'* Total area of the mesh:         ', mesh%area
      write(StdOut,'(A,ES23.16)')'* Space step of the mesh:         ', mesh%Delta_x

  end subroutine compute_geometrical

  !--------------------------------------------------------------------!
  !> @brief
  !> Dispatch the informations from \c infointerface and \c nc into the mesh.
  !>
  !> @param[in] nc, infointerface Informations about the interfaces
  !> @param[in,out] mesh Primary or secondary mesh
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  subroutine dispatch(nc, infointerface, mesh)
      integer, dimension(:), intent(in) :: nc
      integer, dimension(:,:), intent(in) :: infointerface
      type(type_mesh_ps), intent(inout) :: mesh

      integer :: i_K, j, CellK
      integer :: a, b, c, d
      integer :: n1, n2, cpt
      logical :: test

      do CellK = 1, mesh%NbCells
        cpt = 0
        mesh%elt(CellK)%IntInterfaces(:) = 0
        mesh%elt(CellK)%nghb_glob(:)  = 0
        do i_K = 1, mesh%elt(CellK)%NbInterfaces
          test = .FALSE.
          do j = 1, nc(mesh%elt(CellK)%GlobalIdx)-1
            a = dmesh%elt(1,infointerface(j,mesh%elt(CellK)%GlobalIdx))
            b = dmesh%elt(2,infointerface(j,mesh%elt(CellK)%GlobalIdx))
            c = dmesh%elt(3,infointerface(j,mesh%elt(CellK)%GlobalIdx))
            d = dmesh%elt(4,infointerface(j,mesh%elt(CellK)%GlobalIdx))
            n1 = mesh%elt(CellK)%vertices_glob(i_K)
            if (i_K /= mesh%elt(CellK)%NbInterfaces) then
              n2 = mesh%elt(CellK)%vertices_glob(i_K+1)
            else
              n2 = mesh%elt(CellK)%vertices_glob(1)
            end if
            if (a == mesh%elt(CellK)%GlobalIdx) then
              if (((b==n1).OR.(b==n2)) .AND. &
                  ((d==n1).OR.(d==n2))) then
                cpt = cpt + 1
                mesh%elt(CellK)%IntInterfaces(cpt) = infointerface(j,mesh%elt(CellK)%GlobalIdx)
                mesh%elt(CellK)%nghb_glob(cpt)  = c
                mesh%elt(CellK)%nghb(cpt)  = nodes(c)%LocalIdx
                test = .TRUE.
              end if
            elseif (c==mesh%elt(CellK)%GlobalIdx) then
              if (((b==n1).OR.(b==n2)).AND.((d==n1).OR.(d==n2))) then
                cpt = cpt + 1
                mesh%elt(CellK)%IntInterfaces(cpt) = infointerface(j,mesh%elt(CellK)%GlobalIdx)
                mesh%elt(CellK)%nghb_glob(cpt)  = a
                mesh%elt(CellK)%nghb(cpt)  = nodes(a)%LocalIdx
                test                    = .TRUE.
              end if
            elseif (d==mesh%elt(CellK)%GlobalIdx) then
              if (((a==n1).OR.(a==n2)).AND.((c==n1).OR.(c==n2))) then
                cpt = cpt + 1
                mesh%elt(CellK)%IntInterfaces(cpt) = infointerface(j,mesh%elt(CellK)%GlobalIdx)
                mesh%elt(CellK)%nghb_glob(cpt)  = b
                mesh%elt(CellK)%nghb(cpt)  = nodes(b)%LocalIdx
                test                    = .TRUE.
              end if
            elseif (b==mesh%elt(CellK)%GlobalIdx) then
              if (((a==n1).OR.(a==n2)).AND.((c==n1).OR.(c==n2))) then
                cpt = cpt + 1
                mesh%elt(CellK)%IntInterfaces(cpt) = infointerface(j,mesh%elt(CellK)%GlobalIdx)
                mesh%elt(CellK)%nghb_glob(cpt)  = d
                mesh%elt(CellK)%nghb(cpt)  = nodes(d)%LocalIdx
                test                    = .TRUE.
              end if
            end if
          end do
          if (.NOT. test) then
            ! We are on boundary
            cpt = cpt + 1
            mesh%elt(CellK)%IntInterfaces(cpt) = 0
            mesh%elt(CellK)%nghb_glob(cpt)  = 0
            mesh%elt(CellK)%nghb(cpt)  = 0
          end if
        end do

      end do

  end subroutine dispatch

  !--------------------------------------------------------------------!
  !> @brief
  !> Mark the nodes of the diamond mesh in order to create the primary
  !> and the secondary meshes.
  !>
  !> @param[out] NbPrimaryElt, NbSecondaryElt Number of element marked in both meshes
  !>
  !> @authors Rodolphe Turpault
  !> @authors Blachère Florian
  !--------------------------------------------------------------------!
  subroutine mark(NbPrimaryElt, NbSecondaryElt)
      integer, intent(out) :: NbPrimaryElt, NbSecondaryElt

      integer, dimension(:), allocatable :: Queue
      logical, dimension(:), allocatable :: queued
      integer :: LastElt
      integer :: msh_nghb, msh_curr
      integer :: CellK, idx_L, CellL
      logical :: find_nghb_msh


      allocate(queue(dmesh%NbVert))
      Queue(:) = 0
      allocate(queued(dmesh%NbVert))
      queued(:) = .FALSE.

      ! Mark the first node
      CellK = 1

      ! In the primary mesh
      nodes(CellK)%msh = pmesh%id
      NbPrimaryElt = 1
      NbSecondaryElt = 0

      !! Or in the secondary mesh
      !nodes(CellK)%msh = smesh%id
      !NbPrimaryElt = 0
      !NbSecondaryElt = 1

      ! Initialize the queue
      LastElt = nodes(CellK)%NbNghb
      Queue(1:LastElt) = nodes(CellK)%Nghb
      do idx_L=1, nodes(CellK)%NbNghb
        CellL = nodes(CellK)%nghb(idx_L)
        queued(CellL) = .TRUE.
      end do

      do while(LastElt /= 0) ! While the queue is not empty
        CellK = Queue(LastElt) ! Get the element to mark

        Queue(LastElt) = 0
        LastElt = LastElt - 1

        ! Find how we will mark it: if one of its neighbours is in pmesh (smesh)
        ! we will mark it to be in smesh (pmesh).
        ! And in the meantime add all the neighbours that are not already marked and
        ! that are not already in the queue, to the queue.
        find_nghb_msh = .FALSE.

        do idx_L = 1, nodes(CellK)%NbNghb
          CellL = nodes(CellK)%nghb(idx_L)
          msh_nghb = nodes(CellL)%msh

          if (msh_nghb == pmesh%id .and. .not. find_nghb_msh) then
            msh_curr = smesh%id
            find_nghb_msh = .TRUE.
            NbSecondaryElt = NbSecondaryElt + 1

          else if (msh_nghb == smesh%id .and. .not. find_nghb_msh) then
            msh_curr = pmesh%id
            find_nghb_msh = .TRUE.
            NbPrimaryElt = NbPrimaryElt + 1

          else if (msh_nghb == dmesh%id .and. .not. queued(CellL)) then
            LastElt = LastElt + 1
            Queue(LastElt) = CellL
            queued(CellL) = .TRUE.

          end if

        end do

        nodes(CellK)%msh = msh_curr ! Mark the current point

      end do

      deallocate(Queue)
      deallocate(queued)

  end subroutine mark

#if defined(_HO_MOOD_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Get the neighborhood of each cell
  !> (cells that share a geometrical element with the current one)
  !>
  !> This is naïve way to achieve this, maybe there is a clever
  !> (and efficient) way to do it elsewhere
  !>
  !> @param[in,out] mesh Primary or secondary mesh
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  subroutine GetNghbhood(mesh)
      type(type_mesh_ps), intent(inout) :: mesh

      integer :: CellK, i_K
      integer :: CellL
      integer, dimension(:), allocatable :: nb
      integer, dimension(:,:), allocatable :: vertices
      integer, dimension(:,:), allocatable :: interfaces
      integer, dimension(:,:), allocatable :: nghbhood
      integer, dimension(:), allocatable :: NbNghbhood
      integer :: v, vK_p1
      integer :: idx_1, idx_2

      ! Loop over the vertices of each cells
      ! and construct an array which describe the cells connected to the vertices.
      ! Then loop over this new connectivity array to create the neighbours.

      ! Nb of owners of a vertex
      allocate(nb(dmesh%NbVert))
      ! List of the owners of a vertex
      allocate(vertices(Max_Nb_Nghb,dmesh%NbVert))
      ! List of the interface index associated in the owners
      allocate(interfaces(Max_Nb_Nghb,dmesh%NbVert))

      nb(:) = 0
      vertices(:,:) = 0
      interfaces(:,:) = 0

      ! Construct the owners of the vertices
      do CellK = 1, mesh%NbCells
        do i_K = 1, mesh%elt(CellK)%NbInterfaces
          v = mesh%elt(CellK)%vertices_glob(i_K)

          nb(v) = nb(v) + 1
          vertices(nb(v),v) = CellK
          interfaces(nb(v),v) = i_K
        end do
      enddo

      ! Nb of points in the neighborhood of a cell
      allocate(NbNghbhood(mesh%NbCells))
      ! Index of the cells int the neighborhood
      allocate(nghbhood(Max_Nb_Nghb,mesh%NbCells))

      NbNghbhood(:) = 0
      nghbhood(:,:) = -1

      ! Then read the previously constructed array
      loop_vertices: do v = 1, dmesh%NbVert
        test_mesh: if (nodes(v)%msh /= mesh%id) then
          loop_nghb_vert_1: do idx_1 = 1, nb(v)
            CellK = vertices(idx_1,v)
            i_K = interfaces(idx_1,v)
            vK_p1 = mesh%elt(CellK)%vertices_glob(plus1(i_K,mesh%elt(CellK)%NbInterfaces))

            ! Construct the neighborhood arrays that contains all the cells
            ! that share a geometrical elements (vertex or edge in the case of a 2D mesh)
            loop_nghb_vert2_MOOD: do idx_2 = idx_1 + 1, nb(v)
              CellL = vertices(idx_2,v)
              test_not_present_L_in_K: if (.not. IsPresent(CellL,nghbhood(:,CellK))) then
                NbNghbhood(CellK) = NbNghbhood(CellK) + 1
                nghbhood(NbNghbhood(CellK),CellK) = CellL
              end if test_not_present_L_in_K
              test_not_present_K_in_L: if (.not. IsPresent(CellK,nghbhood(:,CellL))) then
                NbNghbhood(CellL) = NbNghbhood(CellL) + 1
                nghbhood(NbNghbhood(CellL),CellL) = CellK
              end if test_not_present_K_in_L
            end do loop_nghb_vert2_MOOD

          end do loop_nghb_vert_1
        end if test_mesh
      end do loop_vertices

      deallocate(nb)
      deallocate(vertices)
      deallocate(interfaces)

      ! Allocate and save the neighborhood arrays
      do CellK = 1, mesh%NbCells
        mesh%elt(CellK)%NbNghbhood = NbNghbhood(CellK)
        allocate(mesh%elt(CellK)%nghbhood(NbNghbhood(CellK)))
        mesh%elt(CellK)%nghbhood(:) = nghbhood(1:NbNghbhood(CellK),CellK)
      end do
      deallocate(NbNghbhood)
      deallocate(nghbhood)

  end subroutine GetNghbhood
#endif /* _HO_MOOD_ */

  !--------------------------------------------------------------------!
  !> @brief
  !> Free all mesh arrays
  !>
  !> @authors Florian Blachère
  !--------------------------------------------------------------------!
  subroutine free_meshes()
      integer :: CellK

      write(StdOut,'(A)') 'Freeing meshes...'

      deallocate(nodes)
      deallocate(BndInterfaces)

      if (allocated(dmesh%elt)) then
        deallocate(dmesh%elt)
        deallocate(dmesh%normals)
        deallocate(dmesh%diags)
        deallocate(dmesh%area)
      end if

      do CellK = 1, pmesh%NbCells
        deallocate(pmesh%elt(CellK)%vertices_glob)
        deallocate(pmesh%elt(CellK)%IntInterfaces)
        deallocate(pmesh%elt(CellK)%nghb_glob)
        deallocate(pmesh%elt(CellK)%nghb)
        deallocate(pmesh%elt(CellK)%normal)
        deallocate(pmesh%elt(CellK)%tangent)
        deallocate(pmesh%elt(CellK)%le)

        deallocate(pmesh%elt(CellK)%idxM)
#if defined(_SRC_ENABLE_)
        deallocate(pmesh%elt(CellK)%delta_K)
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        deallocate(pmesh%elt(CellK)%delta_K_tilde)
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
#endif /* _SRC_ENABLE_ */
        deallocate(pmesh%elt(CellK)%stencil)
        deallocate(pmesh%elt(CellK)%eta)
        deallocate(pmesh%elt(CellK)%BndInterfaces)
#if defined(_HO_MOOD_)
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        deallocate(pmesh%elt(CellK)%T_PI_X)
        deallocate(pmesh%elt(CellK)%stencil_MOOD)
        deallocate(pmesh%elt(CellK)%CoeffPolyMOOD)
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
        deallocate(pmesh%elt(CellK)%nghbhood)
#endif /* _HO_MOOD_ */

      end do
      deallocate(pmesh%elt)

#if defined(_MESH_D2P_ENABLE_SMESH_)
      do CellK=1, smesh%NbCells
        deallocate(smesh%elt(CellK)%vertices_glob)
        deallocate(smesh%elt(CellK)%IntInterfaces)
        deallocate(smesh%elt(CellK)%nghb_glob)
        deallocate(smesh%elt(CellK)%nghb)
        deallocate(smesh%elt(CellK)%normal)
        deallocate(smesh%elt(CellK)%tangent)
        deallocate(smesh%elt(CellK)%le)
      end do
      deallocate(smesh%elt)
#endif /* _MESH_D2P_ENABLE_SMESH_ */

  end subroutine free_meshes

#endif /* _MESH_ */
end module mod_mesh_D2P
