!> Subroutines to read <tt>.msh</tt> files coming from Gmsh (http://geuz.org/gmsh/) \cite Gmsh_2009
!> and create primary and diamond meshes.
!>
!> @authors Florian Blachère
!> @authors Rodolphe Turpault
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.

module mod_mesh_P2D
#if defined(_MESH_P2D_)

  use mod_numerics, only: PR, pi, InitMin, InitMax, eps
  use type_IO, only: StdOut, StdErr
  use type_mesh, only: dim_, Max_Nb_Nghb
  use type_mesh, only: type_mesh_ps
  use type_mesh, only: pmesh, nodes, BndInterfaces
  use type_mesh, only: Neumann_BND, Dirichlet_BND, Wall_BND
  use type_mesh_DLP, only: IntInterfaces
  use mod_tools, only: CrossProduct2D, plus1, minus1
  use mod_tools_array, only:  IsPresent
  use mod_tools_mesh, only: TriangleArea, CDG, Circumcenter, ProjectionPointToLine
#if defined(_MESH_RCM_)
  use mod_tools_time, only: TimeCounters
  use mod_tools_time, only: GetTime
  use mod_RCM, only: RCM, apply_RCM
#endif /* _MESH_RCM_ */
#include "../include/gccversion.h"
#if defined(_FC_PGI_) \
  || (defined(_FC_GNU_) && _GNUC_VERSION_LT_(4,6,0)) \
  || (defined(_FC_GNU_) && _GNUC_VERSION_LT_(6,1,0) && _KIND_ == 16)
  use mod_tools_array, only: NORM2
#endif
  use mod_open_file, only: OpenFile

  implicit none

  private
  public read_msh, create_meshes, free_meshes

  !> Gmsh type for a line
  integer, parameter :: MSH_LINE  = 1
  !> Gmsh type for a triangle
  integer, parameter :: MSH_TRI   = 2
  !> Gmsh type for a quadrangle
  integer, parameter :: MSH_QUAD  = 3
  !> Gmsh type for a point
  integer, parameter :: MSH_POINT = 15

!#define _MESH_P2D_NAIVE_INTER_
#if defined(_HO_MOOD_)
#undef _MESH_P2D_NAIVE_INTER_
#endif /* _HO_MOOD_ */

#define _MESH_CENTER_MASS_
!#define _MESH_CENTER_CIRCLE_

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Read a \c .msh created by Gmsh (http://geuz.org/gmsh/) \cite Gmsh_2009 .
  !>
  !> @param[in] msh_file Path to the \c .msh file
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  subroutine read_msh(msh_file)
      CHARACTER(len=*), intent(in) :: msh_file

      real(PR), dimension(3)    :: coord
      real(PR), dimension(dim_) :: x_min, x_max
      character(len=8)          :: readLine
      integer :: IERR, CellK, d
      integer :: IdxBndInt, NbElements, NbCells
      integer :: idx, ElementType, NbTag, PhysicalTag, GeometricalTag
#if defined(_MESH_PERIODIC_)
      integer :: Period, NbPeriod, NbSidesPeriodic
      integer :: DimensionPeriodicEntity, Slave, Master
      integer :: NbNodes
#endif /* _MESH_PERIODIC_ */
      integer :: FileId

      x_min(:) = InitMin
      x_max(:) = InitMax

      write(StdOut,'(2A)')'Opening the following mesh file: ', TRIM(msh_file)
      call OpenFile(unit=FileId,file=TRIM(msh_file), action='read', status='old')

      ! Read the number of nodes in the mesh
      readLine = ''
      do while (.NOT. readLine(2:6)=='Nodes')
        read (FileId, '(A8)') readLine
        readLine = ADJUSTL(readLine)
      end do
      read (FileId, *) pmesh%NbVert
      write(StdOut,'(A,I9)')'Nb vertices: ',pmesh%NbVert
      if (pmesh%NbVert<2) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " ->  You need at least two nodes in the mesh."
        flush(StdErr)
        STOP
      end if

      allocate(nodes(pmesh%NbVert))
      nodes(:)%onb = .FALSE.

      ! Read the nodes
      do CellK = 1, pmesh%NbVert
        ! Trying to read a node
        read (FileId,*) idx, coord
        nodes(CellK)%x = coord(1:dim_)
        do d = 1,dim_
          x_min(d) = min(coord(d),x_min(d))
          x_max(d) = max(coord(d),x_max(d))
        end do
      end do

      pmesh%x_min(:) = x_min(:)
      pmesh%x_max(:) = x_max(:)
      pmesh%length(:) = pmesh%x_max(:) - pmesh%x_min(:)
      pmesh%x_center(:) = 0.5_PR*(pmesh%x_max(:) + pmesh%x_min(:))

      readLine = ''
      do while (.NOT.(readLine(2:5)=='Elem'))
        read (FileId, '(A8)') readLine
        readLine = ADJUSTL(readLine)
      end do
      read (FileId, *) NbElements

      pmesh%NbCells = 0
      pmesh%NbBndInterfaces   = 0

      ! Then the number of elements (cells and boundary interfaces)
      do CellK = 1, NbElements
        read (FileId, *) idx, ElementType, NbTag, PhysicalTag, GeometricalTag
        if (NbTag /= 2) then
          write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> more than two tags in the mesh file"
          flush(StdErr)
          STOP
        end if

        if (ElementType == MSH_TRI) then ! Triangle
          pmesh%NbCells = pmesh%NbCells + 1
        else if (ElementType == MSH_QUAD) then ! Quadrangle
          pmesh%NbCells = pmesh%NbCells + 1
        elseif (ElementType == MSH_POINT) then ! Single point

        elseif (ElementType == MSH_LINE .and. PhysicalTag >= 100) then ! Interior line

        elseif (ElementType == MSH_LINE .and. PhysicalTag <  100) then ! Boundary line
          pmesh%NbBndInterfaces = pmesh%NbBndInterfaces + 1
        else
          write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, " -> unknown element: ", ElementType
          flush(StdErr)
          STOP
        end if
      end do

      if (pmesh%NbCells<1) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> You nead at least one element in the mesh"
        flush(StdErr)
        STOP
      end if

      write(StdOut,'(A,I8)')'Nb primary cells:    ', pmesh%NbCells

#if defined(_MESH_PERIODIC_)
      readLine = ''
      do while (.NOT.(readLine(2:5)=='Peri'))
        read (FileId, '(A8)') readLine
        readLine = ADJUSTL(readLine)
      end do
      NbSidesPeriodic = 0
      read (FileId, *) NbPeriod

      do Period = 1, NbPeriod
        read (FileId, *) DimensionPeriodicEntity, Slave, Master

        if (DimensionPeriodicEntity == 0) then
          read (FileId, *) NbNodes
          do CellK = 1, NbNodes
            read (FileId, *) Slave, Master
          end do
        else if (DimensionPeriodicEntity == 1) then
          NbSidesPeriodic = NbSidesPeriodic + 1

          read (FileId, *) NbNodes
          do CellK = 1, NbNodes
            read (FileId, *) Slave, Master
          end do
        else
          write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, " -> dimension of periodic entity unknown: ", DimensionPeriodicEntity
          flush(StdErr)
          STOP
        end if
      end do

      pmesh%NbSidesPeriodic = NbSidesPeriodic
      allocate(pmesh%SidesPeriodic(2,pmesh%NbSidesPeriodic))
#endif /* _MESH_PERIODIC_ */

      allocate(pmesh%elt(pmesh%NbCells))
      allocate(BndInterfaces(pmesh%NbBndInterfaces))

      ! Second loop to get the elements, for that we need to restart the reading of the file
      close(FileId)
      readLine = ''
      call OpenFile(unit=FileId, file=TRIM(msh_file), action='read', status='old')
      do while (.NOT.(readLine(2:5)=='Elem'))
        read (FileId, '(A8)') readLine
        readLine = ADJUSTL(readLine)
      end do
      read (FileId, *) NbElements

      IdxBndInt = 0
      NbCells = 0
      do CellK = 1, NbElements
        read (FileId,*) idx, ElementType, NbTag, PhysicalTag, GeometricalTag
        if (ElementType == MSH_TRI) then ! Triangle
          NbCells = NbCells + 1
          pmesh%elt(NbCells)%NbInterfaces = 3
          allocate(pmesh%elt(NbCells)%vertices_glob(3))
          backspace(FileId)  ! Read one line back
          read (FileId, *) idx, ElementType, NbTag, PhysicalTag, GeometricalTag, pmesh%elt(NbCells)%vertices_glob(:)

        else if (ElementType == MSH_QUAD) then ! Quadrangle
          NbCells = NbCells + 1
          pmesh%elt(NbCells)%NbInterfaces = 4
          allocate(pmesh%elt(NbCells)%vertices_glob(4))
          backspace(FileId)  ! Read one line back
          read (FileId, *) idx, ElementType, NbTag, PhysicalTag, GeometricalTag, pmesh%elt(NbCells)%vertices_glob(:)

        elseif (ElementType == MSH_LINE .and. PhysicalTag < 100) then ! Boundary line
          IdxBndInt = IdxBndInt + 1
          backspace(FileId)  ! Read one line back
          allocate(BndInterfaces(IdxBndInt)%BndTag(NbTag))
          read (FileId, *) idx, ElementType, NbTag, BndInterfaces(IdxBndInt)%BndTag(:), BndInterfaces(IdxBndInt)%vertices(:)

          nodes(BndInterfaces(IdxBndInt)%vertices(1))%onb = .TRUE.
          nodes(BndInterfaces(IdxBndInt)%vertices(2))%onb = .TRUE.
        end if

      end do

#if defined(_MESH_PERIODIC_)
      readLine = ''
      do while (.NOT.(readLine(2:5)=='Peri'))
        read (FileId, '(A8)') readLine
        readLine = ADJUSTL(readLine)
      end do
      NbSidesPeriodic = 0
      read (FileId, *) NbPeriod

      ! Then the number of elements (cells and boundary interfaces)
      do Period = 1, NbPeriod
        read (FileId, *) DimensionPeriodicEntity, Slave, Master

        if (DimensionPeriodicEntity == 0) then
          read (FileId, *) NbNodes
          do CellK = 1, NbNodes
            read (FileId, *) Slave, Master
          end do
        else if (DimensionPeriodicEntity == 1) then
          NbSidesPeriodic = NbSidesPeriodic + 1
          pmesh%SidesPeriodic(1,NbSidesPeriodic) = Slave
          pmesh%SidesPeriodic(2,NbSidesPeriodic) = Master

          read (FileId, *) NbNodes
          do CellK = 1, NbNodes
            read (FileId, *) Slave, Master
          end do
        else
          write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, " -> dimension of periodic entity unknown: ", DimensionPeriodicEntity
          flush(StdErr)
          STOP
        end if
      end do
#endif /* _MESH_PERIODIC_ */

      close(FileId)

      write(StdOut,'(A)')'Computational area:'
      write(StdOut,'(A,ES23.16,A,ES23.16,A)') 'x in: [', pmesh%x_min(1),' ;', pmesh%x_max(1),']'
      write(StdOut,'(A,ES23.16,A,ES23.16,A)') 'y in: [', pmesh%x_min(2),' ;', pmesh%x_max(2),']'

      write(StdOut,'(A)')'End of the file reading'
      write(StdOut,'(A)')"-------------------------------------------------------"

  end subroutine read_msh

  !--------------------------------------------------------------------!
  !> @brief
  !> Create the primary and the diamond mesh from the previous reading
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !>
  !--------------------------------------------------------------------!
  subroutine create_meshes()
#if defined (_DEBUG_)
      integer :: CellK, i_K
      real(PR), dimension(dim_) :: sum_ei_ni
#endif /* _DEBUG_ */
      logical :: RCM_applied

      RCM_applied = .FALSE.

      write(StdOut,'(A)')'Computations of the normals, areas...'
      call compute_geometrical()

      write(StdOut,'(A)')'Linking interfaces...'
      call connect_interfaces(RCM_applied)

#if defined(_MESH_RCM_)
      TimeCounters%RCM(1) = GetTime()
      write(StdOut,'(A)')'Find the Reverse Cuthill-McKee permutations...'
      call RCM()
      write(StdOut,'(A)')'Apply these permutations...'
      call apply_RCM()
      RCM_applied = .TRUE.
      write(StdOut,'(A)')'Re-Linking interfaces...'
      call connect_interfaces(RCM_applied)
      TimeCounters%RCM(2) = GetTime() - TimeCounters%RCM(1)
      write(StdOut,'(A)')'RCM done'
#endif /* _MESH_RCM_ */

#if defined(_MESH_PERIODIC_)
      write(StdOut,'(A)') 'Linking for periodic boundary conditions...'
      call connect_bnd_interfaces()
#endif /* _MESH_PERIODIC_ */

#if defined (_DEBUG_)
      do CellK = 1, pmesh%NbCells
        sum_ei_ni(:) = 0._PR
        do i_K = 1, pmesh%elt(CellK)%NbInterfaces
         sum_ei_ni(:) = sum_ei_ni(:) + pmesh%elt(CellK)%le(i_K)*pmesh%elt(CellK)%normal(:,i_K)
        end do
        if (abs(sum_ei_ni(1)) > eps .or. abs(sum_ei_ni(2)) > eps) then
          write(StdOut,'(2A,I4,A)') __FILE__,':', __LINE__, " -> Cell not closed in pmesh:"
          write(StdOut,'(I8,1X,ES23.16,1X,ES23.16)') CellK, sum_ei_ni(1), sum_ei_ni(2)
          read(*,*)
        end if
      end do
#endif /* _DEBUG_ */

      write(StdOut,'(A)')'End of the creation of the primary mesh'

  end subroutine create_meshes

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute some geometrical stuff: normal, area, length...
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  subroutine compute_geometrical()
      integer :: CellK, i_K, NbInterfaces
      integer :: A_glob, B_glob
      real(PR), dimension(dim_) :: xA, xB, xAB, n
      real(PR) :: SignedArea
      real(PR) :: perimeter, le

      pmesh%area = 0._PR
      pmesh%MaxNbNghb = 0
      pmesh%Delta_x = InitMin
#if defined(_IO_ENABLE_VTK_)
      pmesh%NbInterfaces = pmesh%NbCells
#endif /* _IO_ENABLE_VTK_ */

      do CellK = 1, pmesh%NbCells
        NbInterfaces = pmesh%elt(CellK)%NbInterfaces
        pmesh%MaxNbNghb = max(pmesh%MaxNbNghb,pmesh%elt(CellK)%NbInterfaces)
#if defined(_IO_ENABLE_VTK_)
        pmesh%NbInterfaces = pmesh%NbInterfaces + NbInterfaces
#endif /* _IO_ENABLE_VTK_ */

        allocate(pmesh%elt(CellK)%IntInterfaces(NbInterfaces))
        allocate(pmesh%elt(CellK)%nghb(NbInterfaces))
        allocate(pmesh%elt(CellK)%normal(dim_,NbInterfaces))
        allocate(pmesh%elt(CellK)%tangent(dim_,NbInterfaces))
        allocate(pmesh%elt(CellK)%le(NbInterfaces))
        allocate(pmesh%elt(CellK)%BndInterfaces(NbInterfaces))
        pmesh%elt(CellK)%BndInterfaces(:) = 0

        ! All the vertices of a cell need to turn counterclockwise,
        ! if the signed area of the polygon is < 0, then turn the vertices
        ! using: https://en.wikipedia.org/wiki/Shoelace_formula
        SignedArea = 0._PR

        do i_K = 1, NbInterfaces - 1
          A_glob = pmesh%elt(CellK)%vertices_glob(i_K)
          B_glob = pmesh%elt(CellK)%vertices_glob(i_K+1)
          xA(:) = nodes(A_glob)%x(:)
          xB(:) = nodes(B_glob)%x(:)

          SignedArea = SignedArea + xA(1)*xB(2) - xB(1)*xA(2)
        end do

        i_K = NbInterfaces
        A_glob = pmesh%elt(CellK)%vertices_glob(i_K)
        B_glob = pmesh%elt(CellK)%vertices_glob(1)
        xA(:) = nodes(A_glob)%x(:)
        xB(:) = nodes(B_glob)%x(:)

        SignedArea = SignedArea + xA(1)*xB(2) - xB(1)*xA(2)

        if (SignedArea < 0._PR) then
          pmesh%elt(CellK)%vertices_glob(2:) = pmesh%elt(CellK)%vertices_glob(pmesh%elt(CellK)%NbInterfaces:2:-1)
        end if

        pmesh%elt(CellK)%area = 0.5_PR*abs(SignedArea)

        ! Then we can compute the others quantities
        perimeter = 0._PR
        do i_K = 1, NbInterfaces - 1
          A_glob = pmesh%elt(CellK)%vertices_glob(i_K)
          B_glob = pmesh%elt(CellK)%vertices_glob(i_K+1)
          xA(:) = nodes(A_glob)%x(:)
          xB(:) = nodes(B_glob)%x(:)
          xAB(:) = xB(:) - xA(:)
          le = NORM2(xAB)
          pmesh%elt(CellK)%le(i_K) = le
          perimeter = perimeter + le
          pmesh%elt(CellK)%tangent(:,i_K) = xAB(:)/le
          n(1) =  xAB(2)
          n(2) = -xAB(1)
          pmesh%elt(CellK)%normal(:,i_K)  = n(:)/le
        end do

        i_K = NbInterfaces
        A_glob = pmesh%elt(CellK)%vertices_glob(i_K)
        B_glob = pmesh%elt(CellK)%vertices_glob(1)
        xA(:) = nodes(A_glob)%x(:)
        xB(:) = nodes(B_glob)%x(:)
        xAB(:) = xB(:) - xA(:)
        le = NORM2(xAB)
        pmesh%elt(CellK)%le(i_K) = le
        perimeter = perimeter + le
        pmesh%elt(CellK)%tangent(:,i_K) = xAB(:)/le
        n(1) =  xAB(2)
        n(2) = -xAB(1)
        pmesh%elt(CellK)%normal(:,i_K)  = n(:)/le

        pmesh%area = pmesh%area + pmesh%elt(CellK)%area
#if defined (_MESH_SAVE_PERIMETER_)
        mesh%elt(CellK)%perimeter = perimeter
#endif /* _MESH_SAVE_PERIMETER_ */
#if defined(_SRC_ENABLE_) || defined(_TP_FLUX_) || defined(_HO_MOOD_)
        pmesh%elt(CellK)%Delta_x = pmesh%elt(CellK)%area/perimeter
#endif /* _SRC_ENABLE_ || _TP_FLUX_ || _HO_MOOD_ */
        pmesh%Delta_x = min(pmesh%Delta_x,pmesh%elt(CellK)%area/perimeter)

        ! Define the cell center
#if defined(_MESH_CENTER_MASS_)
        ! Use the center of mass
        pmesh%elt(CellK)%x(:) = CDG(CellK)
#elif defined(_MESH_CENTER_CIRCLE_)
        ! or the circumcenter
        pmesh%elt(CellK)%x(:) = Circumcenter(CellK)
#else
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " No method to define the cell center"
        flush(StdErr)
        STOP
#endif /* _MESH_CENTER_ */

      end do

      write(StdOut,'(A,ES23.16)')'* Total area of the mesh:         ', pmesh%area
      write(StdOut,'(A,ES23.16)')'* Space step of the mesh:         ', pmesh%Delta_x

  end subroutine compute_geometrical

  !--------------------------------------------------------------------!
  !> @brief
  !> Connect the elements and create the interiors interfaces
  !>
  !> @param[in] RCM_applied Indicate if the routine is called before of after the RCM
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  subroutine connect_interfaces(RCM_applied)
      logical, intent(in) :: RCM_applied

      integer :: CellK, i_K
      integer :: CellL, i_L
      integer :: NbIntInterfaces
      integer :: idxBndInt, idxIntInt
#if defined(_MESH_P2D_NAIVE_INTER_)
      integer :: NbInterfaces_K, NbInterfaces_L
      integer :: A_K, B_K
      integer :: A_L, B_L
      logical :: nghb_found
#else
      integer, dimension(:), allocatable :: nb
      integer, dimension(:,:), allocatable :: vertices
      integer, dimension(:,:), allocatable ::  interfaces
#if defined(_HO_MOOD_)
      integer, dimension(:,:), allocatable :: nghbhood
      integer, dimension(:), allocatable :: NbNghbhood
#endif /* _HO_MOOD_ */
      integer :: v, vK_p1, vL_m1
      integer :: i_L_m1
      integer :: idx_1, idx_2
      logical :: really_bnd_interface
#endif /* _MESH_P2D_NAIVE_INTER_ */

      NbIntInterfaces = 0

      ! Initialise the nghb array
      do CellK = 1, pmesh%NbCells
        pmesh%elt(CellK)%nghb(:) = -1
      end do


#if defined(_MESH_P2D_NAIVE_INTER_)
      ! Naive method: loop over the interfaces of each cells
      ! and look for another cells that have the same interface.
      loop_elt_K: do CellK = 1, pmesh%NbCells
        pmesh%elt(CellK)%onb = .FALSE.
        NbInterfaces_K = pmesh%elt(CellK)%NbInterfaces
        loop_interfaces_K: do i_K = 1, NbInterfaces_K
          test_done: if (pmesh%elt(CellK)%nghb(i_K) == -1) then
            nghb_found = .FALSE.

            A_K = pmesh%elt(CellK)%vertices_glob(i_K)
            B_K = pmesh%elt(CellK)%vertices_glob(plus1(i_K,NbInterfaces_K))

            loop_elt_L: do CellL = 1, pmesh%NbCells
              NbInterfaces_L = pmesh%elt(CellL)%NbInterfaces
              loop_interfaces_L: do i_L = 1, NbInterfaces_L
                A_L = pmesh%elt(CellL)%vertices_glob(i_L)
                B_L = pmesh%elt(CellL)%vertices_glob(plus1(i_L,NbInterfaces_L))

                if ((A_K == B_L .and. A_L == B_K) .or. &
                    (A_K == B_K .and. A_L == B_L)) then
                  nghb_found = .TRUE.
                  exit loop_elt_L
                end if
              end do loop_interfaces_L
            end do loop_elt_L

            test_found: if (nghb_found) then
              NbIntInterfaces = NbIntInterfaces + 1

#if defined (_MESH_RCM_)
              if (RCM_applied) then
                IntInterfaces(NbIntInterfaces)%nghb(1) = CellK
                IntInterfaces(NbIntInterfaces)%nghb(2) = CellL
              end if
#endif /* _MESH_RCM_ */

              pmesh%elt(CellK)%nghb(i_K) = CellL
              pmesh%elt(CellK)%IntInterfaces(i_K) = NbIntInterfaces

              pmesh%elt(CellL)%nghb(i_L) = CellK
              pmesh%elt(CellL)%IntInterfaces(i_L) = NbIntInterfaces

            else
              ! No nghb found, this is a boundary interface,
              ! we need to look for the boundary interface associated.
              pmesh%elt(CellK)%onb = .TRUE.
              pmesh%elt(CellK)%nghb(i_K) = 0
              pmesh%elt(CellK)%IntInterfaces(i_K) = 0
              loop_BndInterfaces: do idxBndInt = 1, pmesh%NbBndInterfaces
                if ((A_K == BndInterfaces(idxBndInt)%vertices(1) .and. B_K == BndInterfaces(idxBndInt)%vertices(2)) .or. &
                    (A_K == BndInterfaces(idxBndInt)%vertices(2) .and. B_K == BndInterfaces(idxBndInt)%vertices(1))) then
                  BndInterfaces(idxBndInt)%IdxCell = CellK
                  BndInterfaces(idxBndInt)%IdxInterface = i_K
                  pmesh%elt(CellK)%BndInterfaces(i_K) = idxBndInt
                  BndInterfaces(idxBndInt)%x(:) = ProjectionPointToLine(nodes(BndInterfaces(idxBndInt)%vertices(1))%x,nodes(BndInterfaces(idxBndInt)%vertices(2))%x,pmesh%elt(CellK)%x)
                  BndInterfaces(idxBndInt)%dKL = NORM2(pmesh%elt(CellK)%x - BndInterfaces(idxBndInt)%x(:))
                  exit loop_BndInterfaces
                end if
              end do loop_BndInterfaces
            endif test_found
          end if test_done

        end do loop_interfaces_K
      end do loop_elt_K
#else
      ! Second method: loop over the vertices of each cells
      ! and construct an array which describe the cells connected to the vertices.
      ! Then loop over this new connectivity array to create the neighbours.

      ! Nb of owners of a vertex
      allocate(nb(pmesh%NbVert))
      ! List of the owners of a vertex
      allocate(vertices(Max_Nb_Nghb,pmesh%NbVert))
      ! List of the interface index associated in the owners
      allocate(interfaces(Max_Nb_Nghb,pmesh%NbVert))

      nb(:) = 0
      vertices(:,:) = 0
      interfaces(:,:) = 0

      ! Construct the owners of the vertices
      do CellK = 1, pmesh%NbCells
        pmesh%elt(CellK)%onb = .FALSE.
        do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          v = pmesh%elt(CellK)%vertices_glob(i_K)

          nb(v) = nb(v) + 1
          vertices(nb(v),v) = CellK
          interfaces(nb(v),v) = i_K
        end do
      enddo

#if defined(_HO_MOOD_)
      ! Nb of points in the neighborhood of a cell
      allocate(NbNghbhood(pmesh%NbCells))
      ! Index of the cells int the neighborhood
      allocate(nghbhood(Max_Nb_Nghb,pmesh%NbCells))

      NbNghbhood(:) = 0
      nghbhood(:,:) = -1
#endif /* _HO_MOOD_ */

      ! Then read the previously constructed array
      loop_vertices: do v = 1, pmesh%NbVert
        loop_nghb_vert_1: do idx_1 = 1, nb(v)
          CellK = vertices(idx_1,v)
          i_K = interfaces(idx_1,v)
          vK_p1 = pmesh%elt(CellK)%vertices_glob(plus1(i_K,pmesh%elt(CellK)%NbInterfaces))

          test_done: if (pmesh%elt(CellK)%nghb(i_K) == -1) then

            really_bnd_interface = .FALSE.
            test_bnd_interfaces: if (nodes(v)%onb .and. nodes(vK_p1)%onb) then
              ! This test can lead to a false positive (two points can
              ! be on the boundary but without creating a boundary interface).
              ! So, in order in order to check that those two points really
              ! create a boundary interfaces we need a flag (really_bnd_interface).

              loop_BndInterfaces: do idxBndInt = 1, pmesh%NbBndInterfaces
                if ((v == BndInterfaces(idxBndInt)%vertices(1) .and. vK_p1 == BndInterfaces(idxBndInt)%vertices(2)) .or. &
                    (v == BndInterfaces(idxBndInt)%vertices(2) .and. vK_p1 == BndInterfaces(idxBndInt)%vertices(1))) then
                  BndInterfaces(idxBndInt)%IdxCell = CellK
                  BndInterfaces(idxBndInt)%IdxInterface = i_K
                  pmesh%elt(CellK)%BndInterfaces(i_K) = idxBndInt
                  pmesh%elt(CellK)%onb = .TRUE.
                  pmesh%elt(CellK)%nghb(i_K) = 0
                  pmesh%elt(CellK)%IntInterfaces(i_K) = 0
                  BndInterfaces(idxBndInt)%x(:) = ProjectionPointToLine(nodes(BndInterfaces(idxBndInt)%vertices(1))%x,nodes(BndInterfaces(idxBndInt)%vertices(2))%x,pmesh%elt(CellK)%x)
                  BndInterfaces(idxBndInt)%dKL = NORM2(pmesh%elt(CellK)%x - BndInterfaces(idxBndInt)%x(:))
                  really_bnd_interface = .TRUE.
                  exit loop_BndInterfaces
                end if
              end do loop_BndInterfaces
            end if test_bnd_interfaces
            test_not_really_bnd_interface: if (.not. really_bnd_interface) then
              loop_nghb_vert_2: do idx_2 = 1, nb(v)
                test_different: if (idx_1 /= idx_2) then
                  CellL = vertices(idx_2,v)
                  i_L = interfaces(idx_2,v)
                  i_L_m1 = minus1(i_L,pmesh%elt(CellL)%NbInterfaces)
                  vL_m1 = pmesh%elt(CellL)%vertices_glob(i_L_m1)
                  test_nghb: if (vK_p1 == vL_m1) then
                    NbIntInterfaces = NbIntInterfaces + 1

#if defined (_MESH_RCM_)
                    if (RCM_applied) then
                      IntInterfaces(NbIntInterfaces)%nghb(1) = CellK
                      IntInterfaces(NbIntInterfaces)%nghb(2) = CellL
                    end if
#endif /* _MESH_RCM_ */

                    pmesh%elt(CellK)%nghb(i_K) = CellL
                    pmesh%elt(CellK)%IntInterfaces(i_K) = NbIntInterfaces

                    pmesh%elt(CellL)%nghb(i_L_m1) = CellK
                    pmesh%elt(CellL)%IntInterfaces(i_L_m1) = NbIntInterfaces
                    exit loop_nghb_vert_2
                  end if test_nghb
                end if test_different
              end do loop_nghb_vert_2

            end if test_not_really_bnd_interface
          end if test_done

#if defined(_HO_MOOD_)
          ! Construct the neighborhood arrays that contains all the cells
          ! that share a geometrical elements (vertex or edge in the case of a 2D mesh)
          loop_nghb_vert2_MOOD: do idx_2 = idx_1 + 1, nb(v)
            CellL = vertices(idx_2,v)
            test_not_present_L_in_K: if (.not. IsPresent(CellL,nghbhood(:,CellK))) then
              NbNghbhood(CellK) = NbNghbhood(CellK) + 1
              nghbhood(NbNghbhood(CellK),CellK) = CellL
            end if test_not_present_L_in_K
            test_not_present_K_in_L: if (.not. IsPresent(CellK,nghbhood(:,CellL))) then
              NbNghbhood(CellL) = NbNghbhood(CellL) + 1
              nghbhood(NbNghbhood(CellL),CellL) = CellK
            end if test_not_present_K_in_L
          end do loop_nghb_vert2_MOOD
#endif /* _HO_MOOD_ */

        end do loop_nghb_vert_1
      end do loop_vertices

      deallocate(nb)
      deallocate(vertices)
      deallocate(interfaces)

#endif /* _MESH_P2D_NAIVE_INTER_ */

#if defined(_DEBUG_)
      ! Check if all the all the cells are really connected
      do CellK = 1, pmesh%NbCells
        do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          if (pmesh%elt(CellK)%nghb(i_K) == -1) then
            write(StdErr,'(2A,I4,A,I9)')__FILE__,':', __LINE__, " ->  Cell not connected: ", CellK
            write(StdErr,'(A,I9)')'interface: ', i_K
            flush(StdErr)
            STOP
          end if
        end do
      end do
#endif /* _DEBUG_ */

#if defined(_HO_MOOD_)
      ! Allocate and save the neighborhood arrays
      do CellK = 1, pmesh%NbCells
        pmesh%elt(CellK)%NbNghbhood = NbNghbhood(CellK)
#if defined(_MESH_RCM_)
        if (RCM_applied) deallocate(pmesh%elt(CellK)%nghbhood)
#endif /* _MESH_RCM_ */
        allocate(pmesh%elt(CellK)%nghbhood(NbNghbhood(CellK)))
        pmesh%elt(CellK)%nghbhood(:) = nghbhood(1:NbNghbhood(CellK),CellK)
      end do
      deallocate(NbNghbhood)
      deallocate(nghbhood)
#endif /* _HO_MOOD_ */

#if defined(_MESH_RCM_)
      if (RCM_applied) return
#endif /* _MESH_RCM_ */
      ! Allocate and initialise the IntInterfaces array
      pmesh%NbIntInterfaces = NbIntInterfaces
      allocate(IntInterfaces(NbIntInterfaces))
      do idxIntInt = 1, pmesh%NbIntInterfaces
        IntInterfaces(idxIntInt)%nghb(:) = -1
      end do

      do CellK = 1, pmesh%NbCells
        do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          CellL = pmesh%elt(CellK)%nghb(i_K)

          ! Then link the interior interfaces to the elements
          test_int_interfaces: if (CellL /= 0) then
            idxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)
            if (IntInterfaces(idxIntInt)%nghb(1) == -1) then
              IntInterfaces(idxIntInt)%nghb(1) = CellK
              IntInterfaces(idxIntInt)%nghb(2) = CellL
            endif
          endif test_int_interfaces

        end do
      end do

  end subroutine connect_interfaces

#if defined(_MESH_PERIODIC_)
  !--------------------------------------------------------------------!
  !> @brief
  !> When using periodic boundary conditions, connect periodic cells
  !>
  !> @authors Florian Blachère
  !--------------------------------------------------------------------!
  subroutine connect_bnd_interfaces()
    integer :: NbSidesPeriodic
    integer :: side1, side2
    integer :: idxBndInt, idxBndInt_inv

    loop_init: do idxBndInt = 1, pmesh%NbBndInterfaces
      BndInterfaces(IdxBndInt)%IdxCellPeriodic = -1
    end do loop_init

    loop_SidePeriodic: do NbSidesPeriodic = 1, pmesh%NbSidesPeriodic
      side1 = pmesh%SidesPeriodic(1,NbSidesPeriodic)
      side2 = pmesh%SidesPeriodic(2,NbSidesPeriodic)


      loop_BndInterfaces: do idxBndInt = 1, ISHFT(pmesh%NbBndInterfaces,-1)
        loop_BndInterfaces_inv: do idxBndInt_inv = pmesh%NbBndInterfaces, ISHFT(pmesh%NbBndInterfaces,-1)+1, -1
          test_same_side: if (BndInterfaces(idxBndInt)%BndTag(2) == side1 .and. &
              BndInterfaces(idxBndInt_inv)%BndTag(2) == side2) then
            test_already_done: if (BndInterfaces(idxBndInt)%IdxCellPeriodic < 0 .and. &
                BndInterfaces(idxBndInt_inv)%IdxCellPeriodic < 0) then
              BndInterfaces(IdxBndInt)%IdxCellPeriodic = BndInterfaces(idxBndInt_inv)%IdxCell
              BndInterfaces(idxBndInt_inv)%IdxCellPeriodic = BndInterfaces(IdxBndInt)%IdxCell
            end if test_already_done
          end if test_same_side
        end do loop_BndInterfaces_inv
      end do loop_BndInterfaces

    end do loop_SidePeriodic

  end subroutine connect_bnd_interfaces
#endif /* _MESH_PERIODIC_ */

  !--------------------------------------------------------------------!
  !> @brief
  !> Free all mesh arrays
  !>
  !> @authors Florian Blachère
  !--------------------------------------------------------------------!
  subroutine free_meshes()
      integer :: CellK

      write(StdOut,'(A)') 'Freeing meshes...'

      deallocate(nodes)
      deallocate(BndInterfaces)

      do CellK = 1, pmesh%NbCells
        deallocate(pmesh%elt(CellK)%vertices_glob)
        deallocate(pmesh%elt(CellK)%IntInterfaces)
        deallocate(pmesh%elt(CellK)%nghb)
        deallocate(pmesh%elt(CellK)%normal)
        deallocate(pmesh%elt(CellK)%tangent)
        deallocate(pmesh%elt(CellK)%le)

        deallocate(pmesh%elt(CellK)%idxM)
#if defined(_SRC_ENABLE_)
        deallocate(pmesh%elt(CellK)%delta_K)
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        deallocate(pmesh%elt(CellK)%delta_K_tilde)
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
#endif /* _SRC_ENABLE_ */
        deallocate(pmesh%elt(CellK)%stencil)
        deallocate(pmesh%elt(CellK)%eta)
        deallocate(pmesh%elt(CellK)%BndInterfaces)
#if defined(_HO_MOOD_)
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        deallocate(pmesh%elt(CellK)%T_PI_X)
        deallocate(pmesh%elt(CellK)%CoeffPolyMOOD)
        deallocate(pmesh%elt(CellK)%stencil_MOOD)
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
        deallocate(pmesh%elt(CellK)%nghbhood)
#endif /* _HO_MOOD_ */
      end do
      deallocate(pmesh%elt)

  end subroutine free_meshes

#endif /* _MESH_ */
end module mod_mesh_P2D
