!> Definition of the different types used to save the mesh
!>
!> @authors Florian Blachère
!> @authors Rodolphe Turpault
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module type_mesh

  use mod_numerics, only: PR, MaxLenStr

  implicit none

  !> Space dimension (need to be equal to two, it's a 2D code ;) !)
  integer, parameter :: dim_ = 2

  !> Maximal number of neighbours that a cell could have
  integer, parameter :: Max_Nb_Nghb = 30

  !> Tag for the Dirichlet boundary conditions
  integer, parameter :: Dirichlet_BND = 20
  !> Tag for the Neumann boundary conditions
  integer, parameter :: Neumann_BND = 10
  !> Tag for the wall boundary conditions
  integer, parameter :: Wall_BND = 30
#if defined(_MESH_PERIODIC_)
  !> Tag for periodic boundary conditions
  integer, parameter :: Periodic_BND = 40
#endif /* _MESH_PERIODIC */

#if defined(_FC_PAR_OMP_)
  !> Chunksize for the primary mesh
  integer, save :: chunk_NbPrimaryCells
  !> Chunksize for the interior interfaces
  integer, save :: chunk_IntInterfaces
  !> Chunksize for the boundary interfaces
  integer, save :: chunk_BndInterfaces
#endif /* _FC_PAR_OMP_*/

  !> Basename of the mesh file
  character(len=MaxLenStr), save :: basename_mesh_file

#if defined(_MESH_D2P_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Type definition for the diamond mesh
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  type type_mesh_diamond
    !> Number of vertices in the mesh
    integer :: NbVert
    !> Number of elements (=diamonds)
    integer :: NbDiams
    !> Tag of the mesh, should be equal to 0
    integer :: id
    !> Number of boundary interfaces
    integer :: NbBndInterfaces
    !> <tt>elt(4,NbDiams)</tt>, where <tt>elt(:,diam)</tt> are the indexes of the vertices of the diamond <tt>diam</tt>
    integer, dimension(:,:), allocatable :: elt
    !> <tt>diags(2,NbDiams)</tt>, length of the diagonals:
    !> * <tt>diags(1,diam)</tt> = AC
    !> * <tt>diags(2,diam)</tt> = BD
    real(PR), dimension(:,:), allocatable :: diags
    !> <tt>normals(4,NbDiams)</tt>, normals to the diagonals:
    !> * <tt>normals(1:2,diam)</tt>: normal to \f$\vec{BD}\f$ from A to C
    !> * <tt>normals(2:3,diam)</tt>: normal to \f$\vec{AC}\f$ from B to D
    real(PR), dimension(:,:), allocatable :: normals
    !> <tt>area(NbDiams)</tt>, <tt>area(diam)</tt>= area of the diamond <tt>d</tt>
    real(PR), dimension(:), allocatable :: area
    !> Space step of the mesh: \f$ \Delta x = \min\limits_d \frac{|d|}{p_d} \f$
    real(PR) :: Delta_x
  end type type_mesh_diamond
#endif /* _MESH_ */

  !--------------------------------------------------------------------!
  !> @brief
  !> Type definition for a boundary interface
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  type type_bnd_interface
    !> Indexes of the two vertices
    integer, dimension(2) :: vertices
    !> Index of the cell in the primary mesh which the interface belongs to
    integer :: IdxCell
    !> Index of the interface in the primary cell
    integer :: IdxInterface
#if defined(_MESH_D2P_)
    !> Index of the cell in the secondary mesh which the interface belongs to
    integer :: IdxSecCell
    !> Index of the interface in the secondary cell
    integer :: IdxSecInterface
#endif /* _MESH_ */
    !> Information about the type of boundary conditions
    integer, dimension(:), allocatable :: BndTag
    !> Coordinates of the point used to compute the boundary value
    real(PR), dimension(dim_) :: x
    !> Distance between the cell center and the point used to compute the boundary value
    real(PR) :: dKL
#if defined(_MESH_PERIODIC_)
    !> Index of the cell linked with periodic boundary condition
    integer :: IdxCellPeriodic
#endif /* _MESH_PERIODIC */
#if defined(_HO_MOOD_)
    !> <tt>FacePD</tt>: degree of the polynomial reconstruction on the interfaces
    !> equal to CellPD in most case except with periodic boundary conditions
    integer :: FacePD
#endif /* _HO_MOOD_ */
  end type type_bnd_interface

  !--------------------------------------------------------------------!
  !> @brief
  !> Type definition of a point
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  type type_node
    !> Coordinates of the point
    real(PR), dimension(dim_) :: x
    !> Flag for boundary points
    logical :: onb
#if defined(_MESH_D2P_)
    !> Neigbours by the vertices
    integer, dimension(:), allocatable :: nghb
    !> Number of neighbours
    integer :: NbNghb
    !> Tag of the mesh which the point belongs to as a center
    integer :: msh
    !> Local index in the primary or secondary mesh
    integer :: LocalIdx
#endif /* _MESH_ */
  end type type_node

  !--------------------------------------------------------------------!
  !> @brief
  !> Type definition of a cell
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  type type_centroid
#if defined(_MESH_D2P_)
    !> Global index in the array <tt>nodes</tt>
    integer :: GlobalIdx
#endif /* _MESH_ */
    !> Number of interfaces/neighbors
    integer :: NbInterfaces
    !> <tt>x</tt>: coordinates of the cell center
    real(PR), dimension(dim_) :: x
    !> <tt>n(dim_,NbInterfaces)</tt>: normals to the interfaces outgoing from the cell
    real(PR), dimension(:,:), allocatable :: normal
    !> <tt>t(dim_,NbInterfaces)</tt>: tangents to the interfaces, \f$(\vec{n},\vec{t})\f$ is direct
    real(PR), dimension(:,:), allocatable :: tangent
#if defined(_MESH_D2P_)
    !> <tt>nghb_glob(NbInterfaces)</tt>: indexes of the nghb in the type_mesh_d2p::nodes array
    integer, dimension(:), allocatable :: nghb_glob
#endif /* _MESH_ */
    !> <tt>nghb(NbInterfaces)</tt>: indexes of the nghb in the mesh
    integer, dimension(:), allocatable :: nghb
    !> <tt>vertices_glob(NbInterfaces)</tt>: indexes of the vertices in the array type_mesh_d2p::nodes
    integer, dimension(:), allocatable :: vertices_glob
    !> <tt>IntInterfaces(NbInterfaces)</tt>: connectivity to the interior interfaces
    integer, dimension(:), allocatable :: IntInterfaces
    !> <tt>le(NbInterfaces)</tt>: length of the interfaces
    real(PR), dimension(:), allocatable :: le
    !> area of the cell
    real(PR) :: area
#if defined(_MESH_SAVE_PERIMETER_)
    !> perimeter of the cell
    real(PR) :: perimeter
#endif /* _MESH_SAVE_PERIMETER_ */
    !> Number of points in the stencil
    integer :: NbSten
    !> <tt>stencil(NbSten)</tt>: local indexes of the points used in the stencil
    integer, dimension(:), allocatable :: stencil
    !> <tt>eta(dim_,NbSten)</tt>: unitary vector \f$\vec{\eta}_{KJ}\f$
    real(PR),dimension(:,:), allocatable :: eta
    !> <tt>eta(dim_,NbSten)</tt>: unitary vector \f$\vec{\tau}_{KJ}\f$
    real(PR),dimension(:,:), allocatable :: tau
    !> Indexes of the point M used to construct the flux on this interfaces (1 or 2) \cite DroLeP_2011 .
    integer, dimension(:), allocatable :: idxM
#if defined(_MESH_D2P_)
    !> Temporary stuff for the construction
    integer, dimension(:,:), allocatable :: e
#endif /* _MESH_ */
#if defined(_SRC_ENABLE_)
    !> <tt>delta_K(NbEq)</tt>: coefficient \f$\delta_{KJ}\f$
    real(PR), dimension(:), allocatable :: delta_K
    !> <tt>delta_K_tilde(NbEq,nb_points_quad)</tt>: coefficient \f$\tilde{\delta}_{KJ,r}\f$
    real(PR), dimension(:,:), allocatable :: delta_K_tilde
#endif /* _SRC_ENABLE_ */
#if defined(_SRC_ENABLE_) || defined(_TP_FLUX_) || defined(_HO_MOOD_)
    !> Coefficient for \f$\delta_{KK} = \Delta x = \frac{|K|}{p_K}\f$, space step
    real(PR) :: Delta_x
#endif /* _SRC_ENABLE_ || _TP_FLUX_ || _HO_MOOD_ */
    !> Connectivity with the array boundary interfaces
    integer, dimension(:), allocatable :: BndInterfaces
#if defined(_TP_CORREC_)
#endif /* _TP_CORREC_ */
    !> Indicate if a cell has at least one boundary interface
    logical :: onb
#if defined(_HO_MOOD_)
    !> Number of cells that shares a geometrical element with the current cell
    integer :: NbNghbhood
    !> <tt>nghbhood(NbNghbhood)</tt>: indexes of cells that share a geometrical element with the current cell
    integer, dimension(:), allocatable :: nghbhood
    !> Number of points in the stencil of the MOOD reconstruction
    integer :: NbS_MOOD
    !> <tt>T_PI_X(NbS_MOOD,Nd_max)</tt> : transposed of the pseudo-inverse of the matrix used for the polynomial reconstruction of MOOD
    real(PR), dimension(:,:), allocatable :: T_PI_X
    !> <tt>CoeffPolyMOOD(NbEq,Nd_max)</tt>: coefficients of the polynomial reconstruction
    real(PR), dimension(:,:), allocatable :: CoeffPolyMOOD
    !> <tt>stencil_MOOD(NbS_MOOD)</tt>: stencil used for the polynomial reconstruction of MOOD
    integer, dimension(:), allocatable :: stencil_MOOD
    !> Degree of the polynomial reconstruction on the cell
    integer :: CellPD
#endif /* _HO_MOOD_ */
  end type type_centroid

  !--------------------------------------------------------------------!
  !> @brief
  !> Type definition for the primary and the secondary mesh
  !>
  !> @authors Florian Blachère
  !> @authors Rodolphe Turpault
  !--------------------------------------------------------------------!
  type type_mesh_ps
    !> Number of cells
    integer :: NbCells
    !> Number of vertices in the mesh
    integer :: NbVert
    !> Number of interior interfaces
    integer :: NbIntInterfaces
    !> Number of boundary interfaces
    integer :: NbBndInterfaces
#if defined(_IO_ENABLE_VTK_)
    !> Number of interfaces (for the VTK files)
    integer :: NbInterfaces
#endif /* _IO_ENABLE_VTK_ */
#if defined(_MESH_D2P_)
    !> Tag of the mesh, 1 for <tt>pmesh</tt> and 2 for <tt>smesh</tt>
    integer :: id
#endif /* _MESH_ */
    !> Space step of the mesh: \f$ \Delta x = \max\limits_K \frac{|K|}{p_K} \f$
    real(PR) :: Delta_x
    !> Coordinates of the bottom-left vertex
    real(PR), dimension(dim_) :: x_min
    !> Coordinates of the top-right vertex
    real(PR), dimension(dim_) :: x_max
    !> Length of the domain
    real(PR), dimension(dim_) :: length
    !> Coordinates of the center of the domain
    real(PR), dimension(dim_) :: x_center
    !> Total area of the domain
    real(PR) :: area
    !> Describe each cells see type_mesh::type_centroid
    type(type_centroid), dimension(:), allocatable :: elt
#if defined(_MESH_PERIODIC_)
    !> Number of sides with periodic boundary conditions
    integer :: NbSidesPeriodic
    !> <tt>SidesPeriodic(2,NbSidesPeriodic)</tt>: Connectivity for perdiodic sides
    integer, dimension(:,:), allocatable :: SidesPeriodic
#endif /* _MESH_PERIODIC */
    !> Maximal number of neighbours of a cell
    integer :: MaxNbNghb
  end type type_mesh_ps

#if defined(_MESH_D2P_)
  !> Contains the diamond mesh
  type(type_mesh_diamond), save :: dmesh
#endif /* _MESH_D2P_ */
  !> Contains the primary mesh
  type(type_mesh_ps), save :: pmesh
#if defined(_MESH_D2P_)
  !> Contains the secondary mesh
  type(type_mesh_ps), save  :: smesh
#endif /* _MESH_D2P_ */
  !> Contains the boundary interfaces
  type(type_bnd_interface), dimension(:), allocatable, save :: BndInterfaces
  !> Contains all the nodes
  type(type_node), dimension(:), allocatable, save :: nodes

end module type_mesh
