!> Stuff related to PETSc (https://www.mcs.anl.gov/petsc/) \cite petsc-web-page
!>
!> @authors Florian Blachère
!> @authors Sébastien Guisset
!> @authors Maxime Stauffert
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_petsc
#if defined(_ENABLE_PETSC_) && defined(_FLUX_AT_IMEX_)

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: pmesh
  use mod_numerics, only: eps
  use type_mesh_DLP, only: IntInterfaces
  use type_mesh, only: dim_
  use type_mesh, only: BndInterfaces
  use type_mesh, only: Neumann_BND, Dirichlet_BND, Wall_BND
  use mod_bnd_values, only: compute_value_bnd
#if defined(_MESH_PERIODIC_)
  use type_mesh, only: Periodic_BND
#endif /* _MESH_PERIODIC */
  use mod_physics_var, only: NbEq
#if defined(_PHY_EULER_)
  use mod_physics_Euler_AT, only: compute_speed_relax
  use mod_physics_Euler_AT, only: conservative_to_relaxed
  use mod_physics_Euler_AT, only: NbUnkSysIMEX
#endif /* defined(_PHY_EULER_) */
  use mod_tools_array, only: tensor_product, IdentityMatrix2
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells, chunk_IntInterfaces, chunk_BndInterfaces
#endif /* _FC_PAR_OMP_ */

#define _EXPLICIT_RES_
#if defined(_EXPLICIT_RES_)
!#define _DEBUG_IMEX_
#endif /* defined(_EXPLICIT_RES_) */

#include <petscversion.h>
#if PETSC_VERSION_GE(3,8,0)
#include <petsc/finclude/petsc.h>
  use petsc
  implicit none
#else
  implicit none
#include <petsc/finclude/petsc.h90>
#endif

  private
  public petsc_init
  public petsc_create
  public petsc_solve
  public petsc_get_solution
  public petsc_end

  Mat :: matIMEX
  KSP :: kspIMEX
  Vec :: xIMEX
  Vec :: yIMEX
  Vec :: vIMEX

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Initialize arrays relative to PETSc \cite petsc-web-page
  !>
  !> @authors Florian Blachère
  !> @authors Sébastien Guisset
  !> @authors Maxime Stauffert
  !--------------------------------------------------------------------------
  subroutine petsc_init()

    integer :: ierr
    integer :: CellK, CellL, IdxIntInt, IdxBndInt
    integer, dimension(:), allocatable :: NbBlocks
    PC :: pcIMEX

    !================================================================!
    !                   Allocations
    !================================================================!
    call PetscInitialize(PETSC_NULL_CHARACTER,ierr)
    call petsc_check(ierr,__LINE__)
    call MatCreate(PETSC_COMM_WORLD,matIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call MatSetSizes(matIMEX,PETSC_DECIDE,PETSC_DECIDE,NbUnkSysIMEX*pmesh%NbCells,NbUnkSysIMEX*pmesh%NbCells,ierr)
    call petsc_check(ierr,__LINE__)
    call MatSetType(matIMEX,MATSEQBAIJ,ierr)
    call petsc_check(ierr,__LINE__)

    ! Compute and set the number of non zero block in the matrix
    allocate(NbBlocks(pmesh%NbCells))
    NbBlocks(:) = 1 ! One cell always interact with itself
    do IdxIntInt = 1, pmesh%NbIntInterfaces
      CellK = IntInterfaces(IdxIntInt)%nghb(1)
      CellL = IntInterfaces(IdxIntInt)%nghb(2)
      NbBlocks(CellK) = NbBlocks(CellK) + 1
      NbBlocks(CellL) = NbBlocks(CellL) + 1
    end do
#if defined(_MESH_PERIODIC_)
    ! Only add interaction with periodic boundary conditions
    do IdxBndInt = 1, pmesh%NbBndInterfaces
       if (BndInterfaces(IdxBndInt)%BndTag(1) == Periodic_BND) then
          CellK = BndInterfaces(IdxBndInt)%IdxCell
          NbBlocks(CellK) = NbBlocks(CellK) + 1
       end if
    end do
#endif /* _MESH_PERIODIC */
    call MatSeqBAIJSetPreallocation(matIMEX,NbUnkSysIMEX,PETSC_NULL_INTEGER,NbBlocks,ierr)
    call petsc_check(ierr,__LINE__)
    deallocate(NbBlocks)

    ! Initialize the solver
    call KSPCreate(PETSC_COMM_WORLD,kspIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call KSPSetType(kspIMEX,KSPGMRES,ierr)
    call petsc_check(ierr,__LINE__)
    call KSPGetPC(kspIMEX,pcIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call PCSetType(pcIMEX,PCBJACOBI,ierr)
    call petsc_check(ierr,__LINE__)
    call KSPSetTolerances(kspIMEX,eps,PETSC_DEFAULT_REAL,PETSC_DEFAULT_REAL,PETSC_DEFAULT_INTEGER,ierr)
    call petsc_check(ierr,__LINE__)
    call KSPSetOperators(kspIMEX,matIMEX,matIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call KSPSetFromOptions(kspIMEX,ierr)
    call petsc_check(ierr,__LINE__)

    ! Create vectors of unknowns and RHS
    call MatCreateVecs(matIMEX,xIMEX,vIMEX,ierr)
    call petsc_check(ierr,__LINE__)
#if PETSC_VERSION_GE(3,8,0)
    call MatCreateVecs(matIMEX,PETSC_NULL_VEC,yIMEX,ierr)
#else
    call MatCreateVecs(matIMEX,PETSC_NULL_OBJECT,yIMEX,ierr)
#endif
    call petsc_check(ierr,__LINE__)
    !================================================================!
  end subroutine petsc_init

  !--------------------------------------------------------------------!
  !> @brief
  !> Create the matrix and the RHS of the linear system
  !>
  !> @param[in] U_n eulerian conservative variables at time \f$t^n\f$
  !> @param[in] dt Time step
  !> @param[in] t_n Current time
  !>
  !> @authors Sébastien Guisset
  !> @authors Maxime Stauffert
  !--------------------------------------------------------------------!
  subroutine petsc_create(U_n,dt,t_n)
    real(PR), dimension(:,:), intent(in) :: U_n
    real(PR), intent(in) :: dt, t_n

    integer :: CellK
    real(PR), dimension(NbEq) :: W_K
    real(PR), dimension(dim_) :: normal_i
    real(PR), dimension(NbUnkSysIMEX,NbUnkSysIMEX) :: blockD_K, blockD_L
    real(PR), dimension(NbUnkSysIMEX,NbUnkSysIMEX) :: blockE_KL, blockE_LK
    real(PR), dimension(NbEq) :: U_bnd
    real(PR), dimension(dim_,dim_) :: RotationMatrix
    real(PR) :: rho_K, rho_L
    real(PR) :: area_K, area_L
    real(PR) :: le_i, a_i
    integer, dimension(1) :: CellK_m1, CellL_m1
    integer :: ierr
    integer :: i_K
    integer :: CellL
    integer :: IdxIntInt, idxBndInt

    !$OMP SINGLE
    call MatZeroEntries(matIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call VecZeroEntries(yIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    !$OMP END SINGLE

    !$OMP DO &
    !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
    do CellK = 1, pmesh%NbCells
      CellK_m1(1) = CellK-1
      call conservative_to_relaxed(U_n(:,CellK),W_K)
      call VecSetValuesBlocked(vIMEX,1,CellK_m1,W_K(2:4),INSERT_VALUES,ierr)
      call petsc_check(ierr,__LINE__)
    end do
    !$OMP END DO

    blockD_K(:,:)  = 0._PR
    blockD_L(:,:)  = 0._PR
    blockE_KL(:,:) = 0._PR
    blockE_LK(:,:) = 0._PR

    !$OMP DO &
    !$OMP SCHEDULE(STATIC,chunk_IntInterfaces)
    do IdxIntInt = 1, pmesh%NbIntInterfaces
      CellK = IntInterfaces(IdxIntInt)%nghb(1)
      CellK_m1(1) = CellK-1
      CellL = IntInterfaces(IdxIntInt)%nghb(2)
      CellL_m1(1) = CellL-1
      le_i = IntInterfaces(IdxIntInt)%le
      i_K = IntInterfaces(IdxIntInt)%i_Cell(1)
      normal_i(:) = pmesh%elt(CellK)%normal(:,i_K)
      a_i = compute_speed_relax(U_n(:,CellK),U_n(:,CellL))
      area_K = pmesh%elt(CellK)%area
      area_L = pmesh%elt(CellL)%area
      rho_K = U_n(1,CellK)
      rho_L = U_n(1,CellL)

      blockD_K(1:2,1:2) = 0.5_PR*a_i*tensor_product(normal_i,normal_i)
      blockD_L(1:2,1:2) = 0.5_PR*a_i*tensor_product(normal_i,normal_i)
      !blockD_K(1:2,3)   = 0.0_PR
      !blockD_L(1:2,3)   = 0.0_PR
      blockD_K(3,1:2)   = 0.5_PR*a_i**2*normal_i
      blockD_L(3,1:2)   = 0.5_PR*a_i**2*(-normal_i)
      blockD_K(3,3)     = 0.5_PR*a_i
      blockD_L(3,3)     = 0.5_PR*a_i

      blockD_K(:,:) = le_i/(area_K*rho_K)*blockD_K(:,:)
      blockD_L(:,:) = le_i/(area_L*rho_L)*blockD_L(:,:)

      blockE_KL(1:2,1:2) = 0.5_PR*a_i*tensor_product(normal_i,normal_i)
      blockE_LK(1:2,1:2) = 0.5_PR*a_i*tensor_product(normal_i,normal_i)
      blockE_KL(1:2,3)   = -0.5_PR*normal_i
      blockE_LK(1:2,3)   = -0.5_PR*(-normal_i)
      blockE_KL(3,1:2)   = -0.5_PR*a_i**2*normal_i
      blockE_LK(3,1:2)   = -0.5_PR*a_i**2*(-normal_i)
      blockE_KL(3,3)     = 0.5_PR*a_i
      blockE_LK(3,3)     = 0.5_PR*a_i

      blockE_KL(:,:) = le_i/(area_K*rho_K)*blockE_KL(:,:)
      blockE_LK(:,:) = le_i/(area_L*rho_L)*blockE_LK(:,:)

      ! The "transpose" is needed as Fortran is "row-major" and PETSc/C are "column-major"
      call MatSetValuesBlocked(matIMEX,1,CellK_m1,1,CellK_m1,transpose(blockD_K),ADD_VALUES,ierr)
      call petsc_check(ierr,__LINE__)
      call MatSetValuesBlocked(matIMEX,1,CellL_m1,1,CellL_m1,transpose(blockD_L),ADD_VALUES,ierr)
      call petsc_check(ierr,__LINE__)

      call MatSetValuesBlocked(matIMEX,1,CellK_m1,1,CellL_m1,transpose(-blockE_KL),ADD_VALUES,ierr)
      call petsc_check(ierr,__LINE__)
      call MatSetValuesBlocked(matIMEX,1,CellL_m1,1,CellK_m1,transpose(-blockE_LK),ADD_VALUES,ierr)
      call petsc_check(ierr,__LINE__)
    end do
    !$OMP END DO

    !$OMP DO &
    !$OMP SCHEDULE(STATIC,chunk_BndInterfaces)
    do IdxBndInt = 1, pmesh%NbBndInterfaces
      CellK = BndInterfaces(IdxBndInt)%IdxCell
      CellK_m1(1) = CellK-1
      i_K = BndInterfaces(IdxBndInt)%IdxInterface
      normal_i(:) = pmesh%elt(CellK)%normal(:,i_K)
      le_i = pmesh%elt(CellK)%le(i_K)
      area_K = pmesh%elt(CellK)%area
      rho_K = U_n(1,CellK)

      U_bnd(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
      a_i = compute_speed_relax(U_n(:,CellK),U_bnd)

      select case (BndInterfaces(IdxBndInt)%BndTag(1))

      case (Neumann_BND)
        blockD_K(1:2,1:2) = 0._PR
        blockD_K(1:2,3)   = 0.5_PR*normal_i
        blockD_K(3,1:2)   = a_i**2*normal_i
        blockD_K(3,3)     = 0._PR

      case (Dirichlet_BND)
        blockD_K(1:2,1:2) = 0.5_PR*a_i*tensor_product(normal_i,normal_i)
        blockD_K(1:2,3)   = 0.0_PR
        blockD_K(3,1:2)   = 0.5_PR*a_i**2*normal_i
        blockD_K(3,3)     = 0.5_PR*a_i

        blockE_KL(1:2,1:2) = 0.5_PR*a_i*tensor_product(normal_i,normal_i)
        blockE_KL(1:2,3)   = -0.5_PR*normal_i
        blockE_KL(3,1:2)   = -0.5_PR*a_i**2*normal_i
        blockE_KL(3,3)     = 0.5_PR*a_i
        blockE_KL(:,:) = le_i/(area_K*rho_K)*blockE_KL(:,:)

        call conservative_to_relaxed(U_bnd,W_K)

        call VecSetValuesBlocked(yIMEX,1,CellK_m1,dt*matmul(blockE_KL,W_K(2:4)),ADD_VALUES,ierr)
        call petsc_check(ierr,__LINE__)

#if defined(_PHY_M1_) || defined(_PHY_EULER_) || defined(_PHY_SW_) || defined(_PHY_HYDRORAD_) || defined(_PHY_M1MG_)
      case (Wall_BND)
        RotationMatrix(1,1) = -normal_i(1)**2 + normal_i(2)**2
        RotationMatrix(2,1) = -2._PR*normal_i(1)*normal_i(2)
        RotationMatrix(1,2) = -2._PR*normal_i(1)*normal_i(2)
        RotationMatrix(2,2) = -(-normal_i(1)**2 + normal_i(2)**2)
        blockD_K(1:2,1:2) = 0.5_PR*a_i*matmul(tensor_product(normal_i,normal_i),IdentityMatrix2-RotationMatrix)
        blockD_K(1:2,3)   = 0.5_PR*normal_i
        blockD_K(3,1:2)   = 0.5_PR*a_i**2*matmul(normal_i,IdentityMatrix2+RotationMatrix)
        blockD_K(3,3)     = 0._PR
#endif /* _PHY_ */

#if defined(_MESH_PERIODIC_)
      case (Periodic_BND)
        CellL = BndInterfaces(IdxBndInt)%IdxCellPeriodic
        CellL_m1(1) = CellL-1
        blockD_K(1:2,1:2) = 0.5_PR*a_i*tensor_product(normal_i,normal_i)
        blockD_K(1:2,3)   = 0.0_PR
        blockD_K(3,1:2)   = 0.5_PR*a_i**2*normal_i
        blockD_K(3,3)     = 0.5_PR*a_i

        blockE_KL(1:2,1:2) = 0.5_PR*a_i*tensor_product(normal_i,normal_i)
        blockE_KL(1:2,3)   = -0.5_PR*normal_i
        blockE_KL(3,1:2)   = -0.5_PR*a_i**2*normal_i
        blockE_KL(3,3)     = 0.5_PR*a_i
        blockE_KL(:,:) = le_i/(area_K*rho_K)*blockE_KL(:,:)

        call MatSetValuesBlocked(matIMEX,1,CellK_m1,1,CellL_m1,transpose(-blockE_KL),ADD_VALUES,ierr)
        call petsc_check(ierr,__LINE__)
#endif /* _MESH_PERIODIC */

      case default
        write(StdErr,'(2A,I4,A,I9)')__FILE__,':', __LINE__, " -> Unknown boundary conditions: ", BndInterfaces(IdxBndInt)%BndTag(1)
        flush(StdErr)
        STOP
      end select

      blockD_K(:,:) = le_i/(area_K*rho_K)*blockD_K(:,:)
      call MatSetValuesBlocked(matIMEX,1,CellK_m1,1,CellK_m1,transpose(blockD_K),ADD_VALUES,ierr)
      call petsc_check(ierr,__LINE__)
    end do
    !$OMP END DO

    !$OMP SINGLE
    call MatAssemblyBegin(matIMEX,MAT_FINAL_ASSEMBLY,ierr)
    call petsc_check(ierr,__LINE__)
    call MatAssemblyEnd(matIMEX,MAT_FINAL_ASSEMBLY,ierr)
    call petsc_check(ierr,__LINE__)
    call VecAssemblyBegin(yIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call VecAssemblyEnd(yIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call VecAssemblyBegin(vIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call VecAssemblyEnd(vIMEX,ierr)
    call petsc_check(ierr,__LINE__)

#if defined(_DEBUG_)
   !call MatView(matIMEX,PETSC_VIEWER_STDOUT_WORLD,ierr)
   !call petsc_check(ierr,__LINE__)

   !call VecView(vIMEX,PETSC_VIEWER_STDOUT_WORLD,ierr)
   !call petsc_check(ierr,__LINE__)

   !call VecView(yIMEX,PETSC_VIEWER_STDOUT_WORLD,ierr)
   !call petsc_check(ierr,__LINE__)
#endif /* defined(_DEBUG_) */

   ! A = dt*A
   call MatScale(matIMEX,dt,ierr)
   call petsc_check(ierr,__LINE__)

#if defined(_EXPLICIT_RES_)
   ! y = y - A*v
   call VecScale(yIMEX,-1._PR,ierr)
   call petsc_check(ierr,__LINE__)
   call MatMultAdd(matIMEX,vIMEX,yIMEX,yIMEX,ierr)
   call petsc_check(ierr,__LINE__)
   call VecScale(yIMEX,-1._PR,ierr)
   call petsc_check(ierr,__LINE__)
#else
   ! y = y + 1*v
   call VecAXPY(yIMEX,1.0_PR,vIMEX,ierr)
   call petsc_check(ierr,__LINE__)
#endif

#if !defined(_DEBUG_IMEX_)
   ! A = A + 1*Id
   call MatShift(matIMEX,1.0_PR,ierr);
   call petsc_check(ierr,__LINE__)
#endif /* !defined(_DEBUG_IMEX_) */
   !$OMP END SINGLE

  end subroutine petsc_create

  !--------------------------------------------------------------------!
  !> @brief
  !> Solve the linear system
  !>
  !> @authors Sébastien Guisset
  !> @authors Maxime Stauffert
  !--------------------------------------------------------------------!
  subroutine petsc_solve()

    integer :: ierr

#if !defined(_DEBUG_IMEX_)
    call KSPSetUp(kspIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call KSPSolve(kspIMEX,yIMEX,xIMEX,ierr)
    call petsc_check(ierr,__LINE__)
#endif

  end subroutine petsc_solve

  !--------------------------------------------------------------------!
  !> @brief
  !> Get the solution from the resolution
  !>
  !> @param[in] U_n Solution at current time
  !> @param[out] W_alpha Primitive relaxed variables \f$(\rho, u, v, \pi)\f$
  !>
  !> @authors Sébastien Guisset
  !> @authors Maxime Stauffert
  !--------------------------------------------------------------------!
  subroutine petsc_get_solution(U_n,W_alpha)
    real(PR), dimension(:,:), intent(in) :: U_n
    real(PR), dimension(:,:), intent(out) :: W_alpha

    integer :: ierr, CellK
    integer, dimension(NbUnkSysIMEX) :: indexes

    !$OMP SINGLE
#if defined(_EXPLICIT_RES_)
#if defined(_DEBUG_IMEX_)
    ! y = y + 1*v
    call VecAXPY(yIMEX,1.0_PR,vIMEX,ierr)
    call petsc_check(ierr,__LINE__)
#else
    ! x = x + 1*v
    call VecAXPY(xIMEX,1.0_PR,vIMEX,ierr)
    call petsc_check(ierr,__LINE__)
#endif
#endif
    !$OMP END SINGLE

    !$OMP DO &
    !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
    do CellK = 1, pmesh%NbCells
      W_alpha(1,CellK) = U_n(1,CellK)
      indexes(1) = NbUnkSysIMEX*(CellK-1)
      indexes(2) = NbUnkSysIMEX*(CellK-1)+1
      indexes(3) = NbUnkSysIMEX*(CellK-1)+2
#if defined(_DEBUG_IMEX_)
      call VecGetValues(yIMEX,NbUnkSysIMEX,indexes,W_alpha(2:4,CellK),ierr)
#else
      call VecGetValues(xIMEX,NbUnkSysIMEX,indexes,W_alpha(2:4,CellK),ierr)
#endif
      call petsc_check(ierr,__LINE__)
    end do
    !$OMP END DO

  end subroutine petsc_get_solution

  !--------------------------------------------------------------------------
  !> @brief
  !> Check the error code of PETSc \cite petsc-web-page
  !>
  !> In order to have a more precise description of the error compared to
  !> CHKERRQ(ierr)
  !>
  !> @param[in] ierr Error code
  !> @param[in] line Line where the error occured
  !>
  !> @authors Florian Blachère
  !--------------------------------------------------------------------------
  subroutine petsc_check(ierr,line)
    integer, intent(in) :: ierr, line

    if (ierr .ne. 0) then
      write(StdErr,'(2A,I4,A,I9)')__FILE__,':', line, " -> PETSc error: ", ierr
      flush(StdErr)
      STOP
    endif

  end subroutine petsc_check

  !--------------------------------------------------------------------------
  !> @brief
  !> Finalize arrays relative to PETSc \cite petsc-web-page
  !>
  !> @authors Florian Blachère
  !> @authors Sébastien Guisset
  !> @authors Maxime Stauffert
  !--------------------------------------------------------------------------
  subroutine petsc_end()

    integer :: ierr

    call MatDestroy(matIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call KSPDestroy(kspIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call VecDestroy(xIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call VecDestroy(yIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call VecDestroy(vIMEX,ierr)
    call petsc_check(ierr,__LINE__)
    call PetscFinalize(ierr)
    call petsc_check(ierr,__LINE__)
  end subroutine petsc_end

#endif /* defined(_ENABLE_PETSC_) && defined(_FLUX_AT_IMEX_) */
end module mod_petsc
