!> Wrapper used to choose the physics of the problem
!>
!> To choose between physics compile with:
!> * \c -D_PHY_ADV_ for linear advection: physics_Advection.f90
!> * \c -D_PHY_BURGERS_ for the inviscid Burgers equation: physics_Burgers.f90
!> * \c -D_PHY_EULER_ for the isentropic Euler model: physics_Euler.f90
!> * \c -D_PHY_HYDRORAD_ for the radiation hydrodynmics: physics_HydroRad.f90
!> * \c -D_PHY_HYPHEAT_ for the hyperbolic heat equation: physics_HypHeat.f90
!> * \c -D_PHY_M1_ for the M1 model: physics_M1.f90
!> * \c -D_PHY_P1_ for the P1 model: physics_P1.f90
!> * \c -D_PHY_SW_ for the Shallow Water model: physics_SW.f90
!> * \c -D_PHY_TELE_ for the Telegraph equations: physics_Telegraph.f90
!>
!> Each problem could be rewritten as conseravtion laws:
!> \f[\partial_t \textbf{U} + \text{div}(\textbf{F}(\textbf{U}))=
!>      \gamma(\textbf{U})(\textbf{R}(\textbf{U})-\textbf{U})\f]
!> where \f$\textbf{U} \in \mathbb{R}^N\f$, \f$\textbf{F}\f$ is a smooth function
!> whose Jacobian has real eigenvalues, \f$\gamma \geq 0\f$ and \f$\textbf{R}\f$
!> is a smooth function which fulfills the compatibility conditions
!> stated in \cite BerLeFTur_2013 .
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.

module mod_physics

#if defined(_PHY_ADV_)
  use mod_physics_Advection, only: physic_flux, u_dirichlet
  use mod_physics_Advection, only: compute_wave_speed, compute_greatest_wave_speed
  use mod_physics_Advection, only: init_U
  use mod_physics_Advection, only: primitive_to_conservative, conservative_to_primitive
  use mod_physics_Advection, only: IsPhysicallyAdmissible
#elif defined(_PHY_BURGERS_)
  use mod_physics_Burgers, only: physic_flux, u_dirichlet
  use mod_physics_Burgers, only: compute_wave_speed, compute_greatest_wave_speed
  use mod_physics_Burgers, only: init_U
  use mod_physics_Burgers, only: primitive_to_conservative, conservative_to_primitive
  use mod_physics_Burgers, only: IsPhysicallyAdmissible
#elif defined(_PHY_EULER_)
  use mod_physics_Euler, only: physic_flux, u_dirichlet
  use mod_physics_Euler, only: compute_wave_speed, compute_greatest_wave_speed
  use mod_physics_Euler, only: init_U
  use mod_physics_Euler, only: primitive_to_conservative, conservative_to_primitive
  use mod_physics_Euler, only: IsPhysicallyAdmissible
#elif defined(_PHY_HYDRORAD_)
  use mod_physics_HydroRad, only: physic_flux, u_dirichlet
  use mod_physics_HydroRad, only: compute_wave_speed, compute_greatest_wave_speed
  use mod_physics_HydroRad, only: init_U
  use mod_physics_HydroRad, only: primitive_to_conservative, conservative_to_primitive
  use mod_physics_HydroRad, only: IsPhysicallyAdmissible
#elif defined(_PHY_HYPHEAT_)
  use mod_physics_HypHeat, only: physic_flux, u_dirichlet
  use mod_physics_HypHeat, only: compute_wave_speed, compute_greatest_wave_speed
  use mod_physics_HypHeat, only: init_U
  use mod_physics_HypHeat, only: primitive_to_conservative, conservative_to_primitive
  use mod_physics_HypHeat, only: IsPhysicallyAdmissible
#elif defined(_PHY_M1_)
  use mod_physics_M1, only: physic_flux, u_dirichlet
  use mod_physics_M1, only: compute_wave_speed, compute_greatest_wave_speed
  use mod_physics_M1, only: init_U
  use mod_physics_M1, only: primitive_to_conservative, conservative_to_primitive
  use mod_physics_M1, only: IsPhysicallyAdmissible
#elif defined(_PHY_M1MG_)
  use mod_physics_M1MG, only: physic_flux, u_dirichlet
  use mod_physics_M1MG, only: compute_wave_speed, compute_greatest_wave_speed
  use mod_physics_M1MG, only: init_U
  use mod_physics_M1MG, only: primitive_to_conservative, conservative_to_primitive
  use mod_physics_M1MG, only: IsPhysicallyAdmissible
#elif defined(_PHY_P1_)
  use mod_physics_P1, only: physic_flux, u_dirichlet
  use mod_physics_P1, only: compute_wave_speed, compute_greatest_wave_speed
  use mod_physics_P1, only: init_U
  use mod_physics_P1, only: primitive_to_conservative, conservative_to_primitive
  use mod_physics_P1, only: IsPhysicallyAdmissible
#elif defined(_PHY_SW_)
  use mod_physics_SW, only: physic_flux, u_dirichlet
  use mod_physics_SW, only: compute_wave_speed, compute_greatest_wave_speed
  use mod_physics_SW, only: init_U
  use mod_physics_SW, only: primitive_to_conservative, conservative_to_primitive
  use mod_physics_SW, only: IsPhysicallyAdmissible
#elif defined(_PHY_TELE_)
  use mod_physics_Telegraph, only: physic_flux, u_dirichlet
  use mod_physics_Telegraph, only: compute_wave_speed, compute_greatest_wave_speed
  use mod_physics_Telegraph, only: init_U
  use mod_physics_Telegraph, only: primitive_to_conservative, conservative_to_primitive
  use mod_physics_Telegraph, only: IsPhysicallyAdmissible
#endif /* _PHY_ */

  use mod_numerics, only: PR, MaxLenStr

  implicit none

  private
  public physic_flux, u_dirichlet
  public compute_wave_speed, compute_greatest_wave_speed
  public init_U
  public primitive_to_conservative, conservative_to_primitive
  public IsPhysicallyAdmissible

end module mod_physics
