!> Physics module for the advection equation.
!>
!> \f[ \partial_t u + \text{div}(\vec{a} u) = 0\f]
!> where \f$\vec{a}\f$ is the speed of the advection.
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_Advection
#if defined (_PHY_ADV_)

  use mod_numerics, only: PR, Pi, InitMin, InitMax
  use type_IO, only: StdErr
  use type_mesh, only: pmesh, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */
  use mod_read, only: read_solution
  use mod_read, only: Restarting

  use mod_physics_Advection_var, only: T_init
  use mod_physics_Advection_var, only: NbEq, test_case
  use mod_physics_Advection_var, only: speed_adv
  use mod_physics_Advection_var, only: ExactSolutionKnown

#if defined(_HO_MOOD_)
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_int, only: compute_mean_f
#endif /* _HO_MOOD_ */

#include "../include/gccversion.h"
#if defined(_FC_PGI_) \
  || (defined(_FC_GNU_) && _GNUC_VERSION_LT_(4,6,0)) \
  || (defined(_FC_GNU_) && _GNUC_VERSION_LT_(6,1,0) && _KIND_ == 16)
  use mod_tools_array, only: NORM2
#endif

  implicit none

  private
  public physic_flux, U_Dirichlet
  public compute_wave_speed, compute_greatest_wave_speed
  public init_U
  public ExactSolution
  public primitive_to_conservative, conservative_to_primitive
  public IsPhysicallyAdmissible

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Give the initial condition according to the test case
  !>
  !> @param[out] U0 Initial condition
  !> @param[out] NbIter Number of iterations made
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine init_U(U0,NbIter)
      real(PR), dimension(:,:), intent(out) :: U0
      integer, intent(out) :: NbIter

      real(PR), dimension(dim_) :: xK
      integer :: CellK

      !$OMP SINGLE
      ExactSolutionKnown = .TRUE.
      select case(test_case)
      case ('Gauss_x')
        speed_adv(:) = [0.5_PR, 0.0_PR]
      case ('Gauss_y')
        speed_adv(:) = [0._PR, 0.5_PR]
      case ('Gauss_xy')
        speed_adv(:) = 0.25_PR*sqrt(2._PR)
      case ('cst', 'Sin_xy', 'Square_xy')
        speed_adv(:) = [1._PR, 1._PR]
      case ('Sin_x','Square_x','x','x2','x3','x4','x5')
        speed_adv(:) = [1._PR, 0._PR]
      case ('Square_y', 'y','y2','y3','y4','y5')
        speed_adv(:) = [0._PR, 1._PR]
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        flush(StdErr)
        STOP
      end select
      !$OMP END SINGLE

      test_restart: if (Restarting) then
        !$OMP SINGLE
        call read_solution(U0,NbIter,T_init)
        !$OMP END SINGLE COPYPRIVATE(NbIter)

      else
        NbIter = 0
        !$OMP SINGLE
        T_init = 0._PR
        !$OMP END SINGLE

        select case(test_case)
        case ('cst')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells) &
          !$OMP PRIVATE(CellK,xK)
          do CellK = 1, pmesh%NbCells
            U0(1,CellK) = 1._PR
          end do
          !$OMP END DO

        case ('Gauss_x', 'Gauss_y', 'Gauss_xy',&
              'Sin_x', 'Sin_xy',&
              'x', 'y',&
              'x2','y2',&
              'x3','y3',&
              'x4','y4',&
              'x5','y5')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells) &
          !$OMP PRIVATE(CellK,xK)
          do CellK = 1, pmesh%NbCells
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
            U0(:,CellK) = compute_mean_f(CellK,T_init,MaxDegree,InitialCondition,NbEq)
#else
            xK(:) = pmesh%elt(CellK)%x(:)
            U0(:,CellK) = InitialCondition(xK,T_init,NbEq)
#endif
          end do
          !$OMP END DO

        case ('Square_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells) &
          !$OMP PRIVATE(CellK,xK)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) > 0.3_PR .and. xK(1) < 0.4_PR .and.  &
                xK(2) > 0.4_PR .and. xK(2) < 0.6_PR) then
              U0(1,CellK) = 5._PR
            else
              U0(1,CellK) = 0._PR
            end if

          end do
          !$OMP END DO

        case ('Square_y')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells) &
          !$OMP PRIVATE(CellK,xK)
          do CellK = 1, pmesh%NbCells
            xK(:) =pmesh%elt(CellK)%x(:)

            if (xK(2) > 0.1_PR .and. xK(2) < 0.2_PR .and. &
                xK(1) > 0.4_PR .and. xK(1) < 0.6_PR) then
              U0(1,CellK) = 5._PR
            else
              U0(1,CellK) = 0._PR
            end if

          end do
          !$OMP END DO

        case ('Square_xy')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells) &
          !$OMP PRIVATE(CellK,xK)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(2) > 0.1_PR .and. xK(2) < 0.2_PR .and. &
                xK(1) > 0.1_PR .and. xK(1) < 0.2_PR) then
              U0(1,CellK) = 5._PR
            else
              U0(1,CellK) = 0._PR
            end if

          end do
          !$OMP END DO
        end select

      end if test_restart

  end subroutine init_U

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the initial condition
  !>
  !> @param[in] xK Coordinates
  !> @param[in] t_n Current time
  !> @param[in] NbEq Number of equations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function InitialCondition(xK,t_n,NbEq) result(U0)
      real(PR), dimension(:), intent(in) :: xK
      real(PR), intent(in) :: t_n
      integer, intent(in) :: NbEq
      real(PR), dimension(NbEq) :: U0

      U0(:) = ExactSolution(xK,t_n,NbEq)

  end function InitialCondition

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the exact solution
  !>
  !> @param[in] xK Coordinates
  !> @param[in] t_n Current time
  !> @param[in] NbEq Number of equations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function ExactSolution(xK,t_n,NbEq) result(U_ex)
      real(PR), dimension(:), intent(in) :: xK
      real(PR), intent(in) :: t_n
      integer, intent(in) :: NbEq
      real(PR), dimension(NbEq) :: U_ex

      real(PR), dimension(dim_) :: x_t

      x_t(:) = xK(:) - speed_adv(:)*t_n

      select case(test_case)
      case ('cst')
        U_ex(1) = 1._PR
      case ('Gauss_x')
        U_ex(1) = exp(- ((x_t(1) - 0.2_PR)**2/(2._PR*(0.05_PR)**2)) )
      case ('Gauss_y')
        U_ex(1) = exp(- ((x_t(2) - 0.2_PR)**2/(2._PR*(0.05_PR)**2)) )
      case ('Gauss_xy')
        U_ex(1) = exp(- ((x_t(1) - 0.2_PR)**2/(2._PR*(0.05_PR)**2) &
                       + (x_t(2) - 0.2_PR)**2/(2._PR*(0.05_PR)**2)) )
      case ('Sin_x')
        U_ex(1) = sin(2._PR*Pi*x_t(1))
      case ('Sin_xy')
        U_ex(1) = sin(2._PR*Pi*x_t(1))*sin(2._PR*Pi*x_t(2))
      case ('x')
        U_ex(1) = x_t(1)
      case ('y')
        U_ex(1) = x_t(2)
      case ('x2')
        U_ex(1) = x_t(1)**2
      case ('y2')
        U_ex(1) = x_t(2)**2
      case ('x3')
        U_ex(1) = x_t(1)**3
      case ('y3')
        U_ex(1) = x_t(2)**3
      case ('x4')
        U_ex(1) = x_t(1)**4
      case ('y4')
        U_ex(1) = x_t(2)**4
      case ('x5')
        U_ex(1) = x_t(1)**5
      case ('y5')
        U_ex(1) = x_t(2)**5
      case ('Square_x')
        if (x_t(1) > - 0.6_PR .and. x_t(1) < - 0.4_PR .and.  &
            x_t(2) > - 0.1_PR .and. x_t(2) <   0.1_PR) then
          U_ex(1) = 5._PR
        else
          U_ex(1) = 1._PR
        end if
      case ('Square_y')
        if (x_t(2) > - 0.6_PR .and. x_t(2) < - 0.4_PR .and. &
            x_t(1) > - 0.1_PR .and. x_t(1) <   0.1_PR) then
          U_ex(1) = 5._PR
        else
          U_ex(1) = 1._PR
        end if
      case ('Square_xy')
        if (x_t(2) > 0.1_PR .and. x_t(2) < 0.2_PR .and. &
            x_t(1) > 0.1_PR .and. x_t(1) < 0.2_PR) then
          U_ex(1) = 5._PR
        else
          U_ex(1) = 0._PR
        end if
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        flush(StdErr)
        STOP
      end select

  end function ExactSolution

  !--------------------------------------------------------------------!
  !> @brief
  !> Give the limit boundary value
  !>
  !> @param[in] x Coordinates of the point on the boundary
  !> @param[in] t_n Current time
  !>
  !> @return Value of the solution on the boundary at point
  !> \f$x\f$ and time \f$t^n\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function U_Dirichlet(x,t_n)
      real(PR), dimension(:), intent(in) :: x
      real(PR), intent(in) :: t_n
      real(PR), dimension(NbEq) :: U_Dirichlet

      U_Dirichlet(:) = ExactSolution(x,t_n,NbEq)

  end function U_Dirichlet

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the wave speeds waves on a interface
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @param[out] a_l, a_r  Wave speeds on the left and right
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent
      real(PR), intent(out) :: a_l, a_r

      a_l = - NORM2(speed_adv)
      a_r =   NORM2(speed_adv)

  end subroutine compute_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the largest wave speed
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @return Largest wave speed
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_greatest_wave_speed(Ul,Ur,normal,tangent)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent

      real(PR) :: compute_greatest_wave_speed

      compute_greatest_wave_speed = NORM2(speed_adv)

  end function compute_greatest_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the physic flux
  !>
  !> @remark
  !> For the ADV model: \f${\bf F}({\bf U}) = U \vec{a}\f$
  !>
  !> @param[in] U_K solution on cell CellK
  !> @return the physic flux on the cell CellK
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function physic_flux(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(NbEq,dim_)     :: physic_flux

      physic_flux(1,:) = speed_adv(:)*U_K(1)

  end function physic_flux

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the primitive to conservative variables
  !>
  !> @param[in] V_K primitive variables
  !> @param[out] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine primitive_to_conservative(V_K,U_K)
      real(PR), dimension(:), intent(in)     :: V_K
      real(PR), dimension(:), intent(out) :: U_K

      U_K(:) = V_K(:)

  end subroutine primitive_to_conservative

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the conservative to primitive variables
  !>
  !> @param[in] U_K conservative variables
  !> @param[out] V_K primitive variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine conservative_to_primitive(U_K,V_K)
      real(PR), dimension(:), intent(in)     :: U_K
      real(PR), dimension(:), intent(out) :: V_K

      V_K(:) = U_K(:)

  end subroutine conservative_to_primitive

  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the value of the solution on a cell is physically admissible.
  !>
  !> @param[in] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function IsPhysicallyAdmissible(U_K)
      real(PR), dimension(:), intent(in) :: U_K

      logical :: IsPhysicallyAdmissible

      IsPhysicallyAdmissible = .TRUE.

  end function IsPhysicallyAdmissible

#endif /* _PHY_ADV_ */
end module mod_physics_Advection
