!> Physics of the Euler model with with acoustic-transport splitting
!>
!> \cite Gir_2014 , \cite ChaGirKok_2016 .
!>
!> @authors Florian Blachère
!> @authors Sébastien Guisset
!> @authors Maxime Stauffert
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_Euler_AT
#if defined(_PHY_EULER_) && defined(_FLUX_AT_)

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: pmesh
  use mod_physics_Euler_var, only: NbEq
  use mod_physics_Euler, only: sound_speed, compute_pressure

  implicit none

  !> Number of unknowns in the linear system \f$(u, v, \pi)\f$
  integer, parameter :: NbUnkSysIMEX = 3

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute relaxation speed
  !>
  !> @param[in] U_K, U_L Conservative variables
  !>
  !> @authors Florian Blachère
  !> @authors Maxime Stauffert
  !--------------------------------------------------------------------!
  function compute_speed_relax(U_K,U_L)
    real(PR), dimension(:), intent(in) :: U_K, U_L
    real(PR) :: compute_speed_relax

    ! Density of the fluid
    real(PR) :: rho_K, rho_L
    ! Speed of sound in the fluid
    real(PR) :: c_K, c_L
    ! Coefficient strictly greater than one
    real(PR) :: coeff_greater = 1._PR + 0.1_PR

    rho_K = U_K(1)
    rho_L = U_L(1)

    c_K = sound_speed(U_K)
    c_L = sound_speed(U_L)

    compute_speed_relax = coeff_greater*max(rho_K*c_K, rho_L*c_L)

  end function compute_speed_relax

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$u^\star\f$
  !>
  !> @param[in] speed_acoustic vitesse de relaxation locale
  !> @param[in] W_K, W_L Lagrangian variable
  !> @param[in] normal normal locale
  !>
  !> @authors Florian Blachère
  !> @authors Sébastien Guisset
  !--------------------------------------------------------------------!
  function compute_u_star(speed_acoustic,W_K,W_L,normal)
    real(PR), dimension(:), intent(in) :: W_K, W_L
    real(PR), dimension(:), intent(in) :: normal
    real(PR), intent(in) :: speed_acoustic
    real(PR) :: compute_u_star

    compute_u_star = 0.5_PR*(dot_product(W_K(2:3) + W_L(2:3),normal(1:2)) - (W_L(4) - W_K(4))/speed_acoustic)

  end function compute_u_star

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$\pi^\star\f$
  !>
  !> @param[in] speed_acoustic vitesse de relaxation locale
  !> @param[in] W_K, W_L Lagrangian variable
  !> @param[in] normal normal locale
  !>
  !> @authors Florian Blachère
  !> @authors Sébastien Guisset
  !--------------------------------------------------------------------!
  function compute_pi_star(speed_acoustic,W_K,W_L,normal)
    real(PR), dimension(:), intent(in) :: W_K, W_L
    real(PR), dimension(:), intent(in) :: normal
    real(PR), intent(in) :: speed_acoustic
    real(PR) :: compute_pi_star

    compute_pi_star = 0.5_PR*((W_K(4) + W_L(4)) - speed_acoustic*dot_product(W_L(2:3) - W_K(2:3),normal))

  end function compute_pi_star

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the lagrangian coordinates \f$(\tau, u, v, \tau E)\f$
  !> in eulerian coordinates \f$(\rho, \rho u, \rho v, E)\f$
  !>
  !> @param[in] V_K in lagrangian coordinates
  !> @param[out] U_K in eulerian conservative coordinates
  !>
  !> @authors Florian Blachère
  !> @authors Maxime Stauffert
  !--------------------------------------------------------------------!
  subroutine lagrangian_to_conservative(V_K,U_K)
      ! [tau, u, v, tau*E] => [rho, rho*u, rho*v, E]
      real(PR), dimension(:), intent(in) :: V_K
      real(PR), dimension(:), intent(out) :: U_K

      ! rho = 1/tau
      U_K(1) = 1.0_PR/V_K(1)
      ! rho*u
      U_K(2) = U_K(1)*V_K(2)
      ! rho*v
      U_K(3) = U_K(1)*V_K(3)
      ! E = rho*(tau*E)
      U_K(4) = U_K(1)*V_K(4)

  end subroutine lagrangian_to_conservative

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the eulerian coordinates  \f$(\rho, \rho u, \rho v, E)\f$
  !> in lagrangian coordinates \f$(\tau, u, v, \tau E)\f$
  !>
  !> @param[in] U_K in eulerian conservative coordinates
  !> @param[out] V_K in lagrangian coordinates
  !>
  !> @authors Florian Blachère
  !> @authors Maxime Stauffert
  !--------------------------------------------------------------------!
  subroutine conservative_to_lagrangian(U_K,V_K)
      ! [rho, rho*u, rho*v, E] => [tau, u, v, tau*E]
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(:), intent(out) :: V_K

      ! tau = 1/rho
      V_K(1) = 1.0_PR/U_K(1)
      ! u
      V_K(2) = V_K(1)*U_K(2)
      ! v
      V_K(3) = V_K(1)*U_K(3)
      ! tau*E
      V_K(4) = V_K(1)*U_K(4)

  end subroutine conservative_to_lagrangian

  !--------------------------------------------------------------------!
  !> @brief
  !> Calculus of the primitive relaxed variables
  !> \f$(\rho,u,v,\pi)\f$ from the eulerian conservative variables
  !> \f$(\rho,\rho u, \rho v, E)\f$
  !>
  !> @param[in] U_K eulerian conservative variableso
  !> @param[in] W_K primitives relaxed variables
  !>
  !> @authors Florian Blachère
  !> @authors Sébastien Guisset
  !--------------------------------------------------------------------!
  subroutine conservative_to_relaxed(U_K,W_K)
      ! [rho, rho*u, rho*v, E] =>  [rho, u, v, Pi]
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(:), intent(out) :: W_K

      ! rho
      W_K(1) = U_K(1)
      ! u
      W_K(2) = U_K(2)/U_K(1)
      ! v
      W_K(3) = U_K(3)/U_K(1)
      ! p
      W_K(4) = compute_pressure(U_K)
    end subroutine conservative_to_relaxed

#endif /* defined(_PHY_EULER_) && defined(_FLUX_AT_) */
end module mod_physics_Euler_AT
