!> Physics of the radiative hydrodynamic model \cite Pom_1973 \cite Tur_2003 .
!>
!> \f[
!>   \left\{
!>     \begin{array}{rcl}
!>       \partial_t \rho         + \text{div}(\rho {\bf u})                                      & = &  0 \\
!>       \partial_t \rho {\bf u} + \text{div}(\rho {\bf u}\otimes {\bf u}+ {\bf P}_R) + \nabla p & = &  \frac{1}{c} \sigma^f {\bf F}_R \\
!>       \partial_t E            + \text{div}((E+p){\bf u})                                      & = &  c \left(\sigma^a E_R - \sigma^e a T^4 \right)\\
!>       \partial_t E_R          + \text{div}({\bf F}_R)                                         & = &  c \left(\sigma^e a T^4 - \sigma^a E_R \right)\\
!>       \partial_t {\bf F}_R    + c^2 \text{div}(({\bf P}_R)                                    & = & -c \sigma^f {\bf F}_R
!>     \end{array}
!>   \right..
!> \f]
!>
!> with:
!> * \f$ \rho \f$: density of the fluid,
!> * \f$ {\bf u} \f$: speed of the fluid,
!> * \f$ E = \rho(\frac{u^2}{2} + e) \f$: total energy per unit volume, \f$e\f$ specific internal energy,
!> * \f$ \kappa > 0 \f$,
!> * the equation of state (EOS): \f$e=\frac{p}{\rho(\gamma-1)}\f$, \f$p'(\rho) > 0\f$,
!> * the speed of the sound in the fluid is defined as: \f$c = \sqrt{\frac{\gamma p}{\rho}}\f$,
!> * \f$ E_R \f$ the radiative energy,
!> * \f$ {\bf F}_R \f$ the radiative flux,
!> * \f$ T \f$ the temperature,
!> * \f$ {\bf P}_R = \frac{E_R}{2}\left( (1-\chi)  + (3\chi-1)\frac{{\bf F}_R \otimes {\bf F}}{\Vert {\bf F} \Vert_2^2}\right) \f$ the radiative pressure,
!> * \f$\chi = \frac{3 + 4f^2}{5+ \sqrt{4-3f^2}} \f$ the Eddington factor, in order that \f$\chi\f$ is well defined:
!>     \f$ f=\frac{\Vert {\bf F} \Vert}{c E} \leq 1 \f$,
!> * \f$ c \f$ the speed of light.
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_HydroRad
#if defined(_PHY_HYDRORAD_)

#define _PHY_HYDRORAD_HYDRO_SPEED_DAVIS_
!#define _PHY_HYDRORAD_HYDRO_SPEED_ROE_
!#define _PHY_HYDRORAD_HYDRO_SPEED_EINFELDT_
!#define _PHY_HYDRORAD_HYDRO_SPEED_PRESSURE_BASED_

!#define _PHY_HYDRORAD_COMPUTE_WAVE_SPEED_RAD_
!#define _PHY_HYDRORAD_COMPUTE_WAVE_SPEED_
!#define _PHY_HYDRORAD_ONLY_HYDRO_

  use mod_numerics, only: PR, pi, eps
  use mod_numerics, only: zero, third, one
  use mod_numerics, only: sqrt_2, sqrt_3
  use type_IO, only: StdOut, StdErr
  use mod_tools, only: min_NaN, max_NaN
  use mod_tools_array, only: tensor_product
  use type_mesh, only: dim_
  use type_mesh, only: pmesh, nodes
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* defined(_FC_PAR_OMP_) */
  use mod_read, only: read_solution
  use mod_read, only: Restarting

  use mod_physics_HydroRad_var, only: NbEq, test_case
  use mod_physics_HydroRad_var, only: T_init, T_end
  use mod_physics_HydroRad_var, only: adiabatic_index, adiabatic_index_m1
#if defined(_PHY_HYDRORAD_HYDRO_SPEED_PRESSURE_BASED_)
  use mod_physics_HydroRad_var, only: adiabatic_index_p1
#endif /* defined(_PHY_HYDRORAD_HYDRO_SPEED_PRESSURE_BASED_) */
  use mod_physics_HydroRad_var, only: c, a
  use mod_physics_HydroRad_var, only: molar_mass, ideal_gas_constant

#if defined(_HO_MOOD_)
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_int, only: compute_mean_f
#endif /* defined(_HO_MOOD_) */

#include "../include/gccversion.h"
#if defined(_FC_PGI_) \
  || (defined(_FC_GNU_) && _GNUC_VERSION_LT_(4,6,0)) \
  || (defined(_FC_GNU_) && _GNUC_VERSION_LT_(6,1,0) && _KIND_ == 16)
  use mod_tools_array, only: NORM2
#endif

  implicit none

  private
  public physic_flux, U_Dirichlet
  public compute_wave_speed_hydro, compute_wave_speed_rad
  public compute_greatest_wave_speed_hydro, compute_greatest_wave_speed_rad
  public compute_wave_speed, compute_greatest_wave_speed
  public init_U
  public primitive_to_conservative, conservative_to_primitive
  public IsPhysicallyAdmissible

  public compute_temperature, compute_pressure

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Give the initial condition according to the test case
  !>
  !> @param[out] U0 array containing the initial condition
  !> @param[out] NbIter Number of iterations made
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_primitive_to_conservative
  subroutine init_U(U0,NbIter)
    real(PR), dimension(:,:), intent(out) :: U0
    integer, intent(out) :: NbIter

    real(PR), dimension(NbEq) :: U_K, V_K
    real(PR), dimension(dim_) :: xK
    integer :: CellK

    real(PR) :: temp

    !$OMP SINGLE
    select case(test_case)
    case ('cst', 'cst_src', 'dis_rad_x')
      T_end = 0.5_PR/c
#if defined (_PHY_HYDRORAD_ONLY_HYDRO_)
    case ('dis_hydro_x')
      T_end = 0.15_PR
#endif /* defined (_PHY_HYDRORAD_ONLY_HYDRO_) */
    case ('sonde_80km')
      T_end = 0.1_PR
    case ('ball')
      T_end = 1.E-6_PR
    case default
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
      flush(StdErr)
      STOP
    end select
    !$OMP END SINGLE NOWAIT

    test_restart: if (Restarting) then
      !$OMP SINGLE
      call read_solution(U0,NbIter,T_init)
      !$OMP END SINGLE COPYPRIVATE(NbIter)

    else
      NbIter = 0
      !$OMP SINGLE
      T_init = 0._PR
      !$OMP END SINGLE

      select case(test_case)
      case ('cst')
        !$OMP SINGLE
        V_K(1:5) = 1._PR
        V_K(6:7) = 0.5_PR*c
        call primitive_to_conservative(V_K,U_K)
        !$OMP END SINGLE COPYPRIVATE(U_K)

        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          U0(:,CellK) = U_K(:)
        end do
        !$OMP END DO

      case ('cst_src')
        !$OMP SINGLE
        V_K(1:4) = 1._PR
        V_K(5) = a*((V_K(3)*molar_mass)/(V_K(1)*ideal_gas_constant))**4
        V_K(6:7) = 0._PR
        call primitive_to_conservative(V_K,U_K)
        !$OMP END SINGLE COPYPRIVATE(U_K)

        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          U0(:,CellK) = U_K(:)
        end do
        !$OMP END DO

#if defined (_PHY_HYDRORAD_ONLY_HYDRO_)
      case ('dis_hydro_x')
        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          xK(:) = pmesh%elt(CellK)%x(:)

          if (xK(1) < 0.5_PR*pmesh%length(1)) then
            V_K(1) = 1._PR
            V_K(2) = 0._PR
            V_K(3) = 0._PR
            V_K(4) = 1._PR
            V_K(5) = 1._PR
            V_K(6:7) = 0._PR*c
          else
            V_K(1) = 0.125_PR
            V_K(2) = 0._PR
            V_K(3) = 0._PR
            V_K(4) = 0.1_PR
            V_K(5) = 1._PR
            V_K(6:7) = 0._PR*c
          end if

          call primitive_to_conservative(V_K,U_K)
          U0(:,CellK) = U_K(:)

        end do
        !$OMP END DO
#endif /* defined (_PHY_HYDRORAD_ONLY_HYDRO_) */

      case ('dis_rad_x')
        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          xK(:) = pmesh%elt(CellK)%x(:)

          if (xK(1) < 0.5_PR*pmesh%length(1)) then
            V_K(1) = 1._PR
            V_K(2:3) = 1._PR
            V_K(4) = 1._PR
            V_K(5) = 1._PR
            V_K(6) = 0.1_PR*c
            V_K(7) = 0._PR*c
          else
            V_K(1) = 1._PR
            V_K(2:3) = 1._PR
            V_K(4) = 1._PR
            V_K(5) = 10._PR
            V_K(6) = -1._PR*c
            V_K(7) = 0._PR*c
          end if

          call primitive_to_conservative(V_K,U_K)
          U0(:,CellK) = U_K(:)

        end do
        !$OMP END DO

      case ('sonde_80km')
        !$OMP SINGLE
        V_K(1) = 1.57005E-5_PR
        V_K(2) = 1.E4_PR
        V_K(3) = 0._PR
        V_K(4) = 8.86280E-1_PR
        V_K(5) = a*((V_K(4)*molar_mass)/(V_K(1)*ideal_gas_constant))**4
        V_K(6) = 0._PR
        V_K(7) = 0._PR
        call primitive_to_conservative(V_K,U_K)
        !$OMP END SINGLE COPYPRIVATE(U_K)
        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          U0(:,CellK) = U_K(:)
        end do
        !$OMP END DO

      case ('ball')
        !$OMP SINGLE
        V_K(1) = 0.00042_PR
        V_K(2) = 16.E3_PR
        V_K(3) = 0._PR
        V_K(4) = 30.6305_PR
        V_K(5) = a*((V_K(4)*molar_mass)/(V_K(1)*ideal_gas_constant))**4
        V_K(6) = 0._PR
        V_K(7) = 0._PR
        call primitive_to_conservative(V_K,U_K)
        !$OMP END SINGLE COPYPRIVATE(U_K)
        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          U0(:,CellK) = U_K(:)
        end do
        !$OMP END DO

      end select

    end if test_restart

  end subroutine init_U

  !--------------------------------------------------------------------!
  !> @brief
  !> Give the limit boundary value
  !> @param[in] x Coordinates of the point on the boundary
  !> @param[in] t_n Current time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_primitive_to_conservative
  function U_Dirichlet(x,t_n)
    real(PR), dimension(:), intent(in) :: x
    real(PR), intent(in) :: t_n
    real(PR), dimension(NbEq) :: U_Dirichlet

    real(PR), dimension(NbEq) :: U_K, V_K

    select case(test_case)
    case ('cst')
      V_K(:) = 1._PR
    case ('sonde_80km')
      V_K(1) = 1.57005E-5_PR
      V_K(2) = 1.E4_PR
      V_K(3) = 0._PR
      V_K(4) = 8.86280E-1_PR
      V_K(5) = a*((V_K(4)*molar_mass)/(V_K(1)*ideal_gas_constant))**4
      V_K(6) = 0._PR
      V_K(7) = 0._PR
    case ('ball')
      V_K(1) = 0.00042_PR
      V_K(2) = 16.E3_PR
      V_K(3) = 0._PR
      V_K(4) = 30.6305_PR
      V_K(5) = a*((V_K(4)*molar_mass)/(V_K(1)*ideal_gas_constant))**4
      V_K(6) = 0._PR
      V_K(7) = 0._PR
    case default
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
      flush(StdErr)
      STOP
    end select

    call primitive_to_conservative(V_K,U_K)
    U_Dirichlet(:) = U_K(:)

  end function U_Dirichlet

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the hydrodynamic wave speeds on a interface
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent  Normal and tangent to the interface
  !> @param[out] a_l, a_r  Wave speeds on the left and right
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_sound_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_pressure
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_min_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  subroutine compute_wave_speed_hydro(Ul,Ur,normal,tangent,a_l,a_r)
    real(PR), dimension(:), intent(in) :: Ul, Ur
    real(PR), dimension(:), intent(in) :: normal, tangent
    real(PR), intent(out) :: a_l, a_r

    ! Speed of the fluid
    real(PR), dimension(dim_) :: u_l, u_r
    ! Normal speed of the fluid
    real(PR) :: u_l_n, u_r_n
    ! Density of the fluid
    real(PR) :: rho_l, rho_r
#if defined(_PHY_HYDRORAD_HYDRO_SPEED_DAVIS_)
    ! Speed of sound in the fluid
    real(PR) :: c_l, c_r
#elif defined(_PHY_HYDRORAD_HYDRO_SPEED_ROE_)
    real(PR) :: u_tilde, a_tilde
    real(PR) :: sqrt_rho_l, sqrt_rho_r
    ! Pressure in the fluid
    real(PR) :: p_l, p_r
    ! Enthalpy
    real(PR) :: H_l, H_r, H_tilde
#elif defined(_PHY_HYDRORAD_HYDRO_SPEED_EINFELDT_)
    real(PR) :: sqrt_rho_l, sqrt_rho_r
    real(PR) :: u_bar, d_bar, eta_2
    ! Speed of sound in the fluid
    real(PR) :: c_l, c_r
#elif defined(_PHY_HYDRORAD_HYDRO_SPEED_PRESSURE_BASED_)
    real(PR) :: rho_bar, c_bar, c_l, c_r, p_pvrs, p_star, p_l, p_r, q_l, q_r
#endif /* _PHY_EULER_WAVE_SPEED_ */

    rho_l = Ul(1)
    u_l(:) = Ul(2:3)/rho_l
    u_l_n = dot_product(u_l,normal)

    rho_r = Ur(1)
    u_r(:) = Ur(2:3)/rho_r
    u_r_n = dot_product(u_r,normal)

#if defined(_PHY_HYDRORAD_HYDRO_SPEED_DAVIS_)
    c_l = sound_speed(Ul)
    c_r = sound_speed(Ur)

    !a_l = min_NaN(u_l_n - c_l, u_r_n - c_r)
    !a_r = max_NaN(u_l_n + c_l, u_r_n + c_r)

    !a_l = min_NaN(u_l_n - c_l, min_NaN(u_r_n - c_r, 0._PR))
    !a_r = max_NaN(u_l_n + c_l, max_NaN(u_r_n + c_r, 0._PR))

    a_l = min_NaN(- abs(u_l_n) - c_l, - abs(u_r_n) - c_r)
    a_r = max_NaN(  abs(u_l_n) + c_l,   abs(u_r_n) + c_r)

#elif defined(_PHY_HYDRORAD_HYDRO_SPEED_PRESSURE_BASED_)
    c_l = sound_speed(Ul)
    c_r = sound_speed(Ur)

    p_l = compute_pressure(Ul)
    p_r = compute_pressure(Ur)

    rho_bar = 0.5_PR*(rho_l + rho_r)
    c_bar = 0.5_PR*(c_l + c_r)

    p_pvrs = 0.5_PR*((p_l + p_r) - (u_r_n - u_l_n)*rho_bar*c_bar)
    p_star = max(0._PR,p_pvrs)

    if (p_star <= p_l) then
      q_l = 1._PR
    else
      q_l = sqrt(1._PR + (adiabatic_index_p1)/(2._PR*adiabatic_index)*(p_star/p_l - 1._PR))
    end if

    if (p_star <= p_r) then
      q_r = 1._PR
    else
      q_r = sqrt(1._PR + (adiabatic_index_p1)/(2._PR*adiabatic_index)*(p_star/p_r - 1._PR))
    end if

    a_l = u_l_n - c_l*q_l
    a_r = u_r_n + c_r*q_r

    !a_l = - abs(u_l_n) - c_l*q_l
    !a_r =   abs(u_r_n) + c_r*q_r

#elif defined(_PHY_HYDRORAD_HYDRO_SPEED_ROE_)
    p_l = compute_pressure(Ul)
    p_r = compute_pressure(Ur)
    ! Enthalpy
    H_l = (Ul(4) + p_l)/rho_l
    H_r = (Ur(4) + p_r)/rho_r

    sqrt_rho_l = sqrt(rho_l)
    sqrt_rho_r = sqrt(rho_r)

    H_tilde = (sqrt_rho_l*H_l + sqrt_rho_l*H_r)/(sqrt_rho_l + sqrt_rho_l)
    u_tilde = (sqrt_rho_l*u_l_n + sqrt_rho_l*u_r_n)/(sqrt_rho_l + sqrt_rho_l)
    a_tilde = sqrt((adiabatic_index_m1)*(H_tilde - 0.5_PR*u_tilde**2))

    a_l = u_tilde - a_tilde
    a_r = u_tilde + a_tilde

    !a_l = - abs(u_tilde) - a_tilde
    !a_r =   abs(u_tilde) + a_tilde

#elif defined(_PHY_HYDRORAD_HYDRO_SPEED_EINFELDT_)
    c_l = sound_speed(Ul)
    c_r = sound_speed(Ur)

    sqrt_rho_l = sqrt(rho_l)
    sqrt_rho_r = sqrt(rho_r)

    eta_2 = 0.5_PR*sqrt_rho_l*sqrt_rho_r/(sqrt_rho_l+sqrt_rho_r)**2

    d_bar = sqrt((sqrt_rho_l*c_l**2 + sqrt_rho_r*c_r**2)/(sqrt_rho_l + sqrt_rho_r) &
        + eta_2*(u_r_n - u_l_n)**2)

    u_bar = (sqrt_rho_l*u_l_n + sqrt_rho_l*u_r_n)/(sqrt_rho_l + sqrt_rho_l)

    a_l = u_bar - d_bar
    a_r = u_bar + d_bar

    !a_l = - abs(u_bar) - d_bar
    !a_r =   abs(u_bar) + d_bar
#else
    write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> No method defined to compute wave speed."
    flush(StdErr)
    STOP
#endif /* _PHY_EULER_WAVE_SPEED_ */
  end subroutine compute_wave_speed_hydro

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the largest hydrodynamic wave speed
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @return Largest wave speed
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed_hydro
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  function compute_greatest_wave_speed_hydro(Ul,Ur,normal,tangent)
    real(PR), dimension(:), intent(in) :: Ul, Ur
    real(PR), dimension(:), intent(in) :: normal, tangent
    real(PR) :: compute_greatest_wave_speed_hydro

    real(PR) :: a_l, a_r

    call compute_wave_speed_hydro(Ul,Ur,normal,tangent,a_l,a_r)
    compute_greatest_wave_speed_hydro = max_NaN(abs(a_l),abs(a_r))

  end function compute_greatest_wave_speed_hydro

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the radiative wave speeds waves on a interface
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @param[out] a_l, a_r  Wave speeds on the left and right
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_min_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_chi
  subroutine compute_wave_speed_rad(Ul,Ur,normal,tangent,a_l,a_r)
    real(PR), dimension(:), intent(in) :: Ul, Ur
    real(PR), dimension(:), intent(in) :: normal, tangent
    real(PR), intent(out) :: a_l, a_r

#if defined(_PHY_HYDRORAD_COMPUTE_WAVE_SPEED_RAD_)
      real(PR), dimension(dim_) :: F_l, ff_l
      real(PR), dimension(dim_) :: F_r, ff_r
      real(PR) :: E_l, xi_l, chi_l, ff_l_n, ff_l_t, norm_ff_l
      real(PR) :: E_r, xi_r, chi_r, ff_r_n, ff_r_t, norm_ff_r
      real(PR) :: lambda_m_l, lambda_p_l
      real(PR) :: lambda_m_r, lambda_p_r

      E_l    = Ul(5)
      F_l(:) = Ul(6:7)
      chi_l  = compute_chi(Ul)
      ff_l(:)= F_l(:)/(c*E_l)
      ff_l_n = dot_product(ff_l,normal)
      ff_l_t = dot_product(ff_l,tangent)
      norm_ff_l = NORM2(ff_l)
      xi_l   = sqrt(4._PR - 3._PR*norm_ff_l**2)

      lambda_p_l = c*(ff_l_n/xi_l + sqrt_2*sqrt((xi_l - 1._PR)*(xi_l + 2._PR)*&
         (2._PR*(xi_l - 1._PR)*(xi_l + 2._PR) + 3._PR*ff_l_t**2))/(sqrt_3*xi_l*(xi_l + 2._PR)))

      lambda_m_l = c*(ff_l_n/xi_l - sqrt_2*sqrt((xi_l - 1._PR)*(xi_l + 2._PR)*&
         (2._PR*(xi_l - 1._PR)*(xi_l + 2._PR) + 3._PR*ff_l_t**2))/(sqrt_3*xi_l*(xi_l + 2._PR)))

      E_r    = Ur(5)
      F_r(:) = Ur(6:7)
      chi_r  = compute_chi(Ur)
      ff_r(:)= F_r(:)/(c*E_r)
      ff_r_n = dot_product(ff_r,normal)
      ff_r_t = dot_product(ff_r,tangent)
      norm_ff_r = NORM2(ff_r)
      xi_r   = sqrt(4._PR - 3._PR*norm_ff_r**2)

      lambda_p_r = c*(ff_r_n/xi_r + sqrt_2*sqrt((xi_r - 1._PR)*(xi_r + 2._PR)*&
         (2._PR*(xi_r - 1._PR)*(xi_r + 2._PR) + 3._PR*ff_r_t**2))/(sqrt_3*xi_r*(xi_r + 2._PR)))

      lambda_m_r = c*(ff_r_n/xi_r - sqrt_2*sqrt((xi_r - 1._PR)*(xi_r + 2._PR)*&
         (2._PR*(xi_r - 1._PR)*(xi_r + 2._PR) + 3._PR*ff_r_t**2))/(sqrt_3*xi_r*(xi_r + 2._PR)))

      a_l = min_NaN(lambda_m_l, lambda_m_r)
      a_r = max_NaN(lambda_p_l, lambda_p_r)

#if defined(_SRC_ENABLE_)
      ! Needed as we made assumptions on the sign of the wave speeds, in flux_1D_source
      a_l = min_NaN(a_l, 0._PR)
      a_r = max_NaN(a_r, 0._PR)
#endif /* defined(_SRC_ENABLE_) */

      !a_l = min_NaN(c*norm_ff_l, c*norm_ff_r, &
          !(norm_ff_l - chi_l)/(1._PR - norm_ff_l), (norm_ff_r - chi_r)/(1._PR - norm_ff_r), &
          !(norm_ff_l + chi_l)/(1._PR + norm_ff_l), (norm_ff_r + chi_r)/(1._PR + norm_ff_r), &
          !lambda_m_l, lambda_m_r, 0._PR)
      !a_r = max_NaN(c*norm_ff_l, c*norm_ff_r, &
          !(norm_ff_l - chi_l)/(1._PR - norm_ff_l), (norm_ff_r - chi_r)/(1._PR - norm_ff_r), &
          !(norm_ff_l + chi_l)/(1._PR + norm_ff_l), (norm_ff_r + chi_r)/(1._PR + norm_ff_r), &
          !lambda_p_l, lambda_p_r, 0._PR)

#else
      a_l = -c
      a_r = c
#endif /* defined(_PHY_HYDRORAD_COMPUTE_WAVE_SPEED_RAD_) */

  end subroutine compute_wave_speed_rad

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the largest radiative wave speed
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @return Largest wave speed
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed_rad
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  function compute_greatest_wave_speed_rad(Ul,Ur,normal,tangent)
    real(PR), dimension(:), intent(in) :: Ul, Ur
    real(PR), dimension(:), intent(in) :: normal, tangent
    real(PR) :: compute_greatest_wave_speed_rad

#if defined(_PHY_HYDRORAD_COMPUTE_WAVE_SPEED_RAD_)
    real(PR) :: a_l, a_r

    call compute_wave_speed_rad(Ul,Ur,normal,tangent,a_l,a_r)
    compute_greatest_wave_speed_rad = max_NaN(abs(a_l),abs(a_r))
#else
    compute_greatest_wave_speed_rad = c
#endif /* defined(_PHY_HYDRORAD_COMPUTE_WAVE_SPEED_RAD_) */

  end function compute_greatest_wave_speed_rad

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the wave speeds waves on a interface
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @param[out] a_l, a_r  Wave speeds on the left and right
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_min_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed_hydro
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed_rad
  subroutine compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)
    real(PR), dimension(:), intent(in) :: Ul, Ur
    real(PR), dimension(:), intent(in) :: normal, tangent
    real(PR), intent(out) :: a_l, a_r

#if defined(_PHY_HYDRORAD_COMPUTE_WAVE_SPEED_)
#if defined (_PHY_HYDRORAD_ONLY_HYDRO_)
    call compute_wave_speed_hydro(Ul,Ur,normal,tangent,a_l,a_r)
#else
    real(PR) :: a_l_hydro, a_r_hydro
    real(PR) :: a_l_rad, a_r_rad

    call compute_wave_speed_hydro(Ul,Ur,normal,tangent,a_l_hydro,a_r_hydro)
    call compute_wave_speed_rad(Ul,Ur,normal,tangent,a_l_rad,a_r_rad)
    a_l = min_NaN(a_l_hydro,a_l_rad)
    a_r = max_NaN(a_r_hydro,a_r_rad)
#endif /* defined (_PHY_HYDRORAD_ONLY_HYDRO_) */

#else
    a_l = - c
    a_r = c
#endif /* defined(_PHY_HYDRORAD_COMPUTE_WAVE_SPEED_) */

  end subroutine compute_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the largest wave speed
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @return Largest wave speed
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed_hydro
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed_rad
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  function compute_greatest_wave_speed(Ul,Ur,normal,tangent)
    real(PR), dimension(:), intent(in) :: Ul, Ur
    real(PR), dimension(:), intent(in) :: normal, tangent

    real(PR) :: compute_greatest_wave_speed

#if defined(_PHY_HYDRORAD_COMPUTE_WAVE_SPEED_)
#if defined (_PHY_HYDRORAD_ONLY_HYDRO_)
    compute_greatest_wave_speed = compute_greatest_wave_speed_hydro(Ul,Ur,normal,tangent)
#else
    compute_greatest_wave_speed = max_NaN(compute_greatest_wave_speed_hydro(Ul,Ur,normal,tangent),compute_greatest_wave_speed_rad(Ul,Ur,normal,tangent))
#endif /* defined (_PHY_HYDRORAD_ONLY_HYDRO_) */

#else
    compute_greatest_wave_speed = c
#endif /* defined(_PHY_HYDRORAD_COMPUTE_WAVE_SPEED_) */

  end function compute_greatest_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the sound speed in the fluid
  !>
  !> @param[in] U Conservative variables
  !> @return the sound speed:
  !> \f$c_s  = \sqrt{\frac{\gamma p}{\rho}}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_pressure
  function sound_speed(U)
    real(PR),  dimension(:), intent(in) :: U
    real(PR) :: sound_speed

    sound_speed = sqrt(adiabatic_index*compute_pressure(U))/sqrt(U(1))

  end function sound_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the physic flux
  !>
  !> @remark
  !> \f${\bf F}({\bf U}) = \begin{pmatrix}{c}
  !>   \rho {\bf u} \\ \rho u^2 + P_R + p \\ (E+p) u \\
  !>   F_R \\ c^2 P_R
  !> \end{pmatrix}\f$
  !>
  !> @param[in] U_K solution on cell CellK
  !> @return the physic flux on the cell CellK
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_pressure
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_radiative_pressure
  function physic_flux(U_K)
    real(PR), dimension(:), intent(in) :: U_K
    real(PR), dimension(NbEq,dim_) :: physic_flux

    real(PR) :: rho, p, E, E_R
    real(PR), dimension(dim_) :: rho_u, u, F_R
    real(PR), dimension(dim_,dim_) :: P_R

    rho      = U_K(1)
    rho_u(:) = U_K(2:3)
    u(:)     = rho_u(:)/rho
    E        = U_K(4)
    p        = compute_pressure(U_K)
    E_R      = U_K(5)
    F_R(:)   = U_K(6:7)
    P_R(:,:) = compute_radiative_pressure(U_K)

    physic_flux(1,1)   = rho_u(1)
    physic_flux(2,1)   = rho*u(1)**2 + p ! + P_R(1,1)
    physic_flux(3,1)   = rho*u(1)*u(2) ! + P_R(2,1)
    physic_flux(4,1)   = (E+p)*u(1)
    physic_flux(5,1)   = F_R(1)
    physic_flux(6:7,1) = c**2*P_R(1:2,1)

    physic_flux(1,2)   = rho_u(2)
    physic_flux(2,2)   = rho*u(1)*u(2) ! + P_R(1,2)
    physic_flux(3,2)   = rho*u(2)**2 + p ! + P_R(2,2)
    physic_flux(4,2)   = (E+p)*u(2)
    physic_flux(5,2)   = F_R(2)
    physic_flux(6:7,2) = c**2*P_R(1:2,2)

  end function physic_flux

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the pressure
  !>
  !> @param[in] U conservative variables
  !> @return the pressure: \f$p  = (\gamma - 1) (E - \frac{\rho (u^2 + v^2)}{2}) \f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_pressure(U)
    real(PR), dimension(:), intent(in) :: U
    real(PR) :: compute_pressure

    real(PR) :: rho_e

    rho_e = U(4) - 0.5_PR*dot_product(U(2:3),U(2:3))/U(1)
    compute_pressure = adiabatic_index_m1*rho_e

  end function compute_pressure

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the temperature
  !>
  !> @param[in] U conservative variables
  !> @return the temperature: \f$ T = \frac{p M}{\rho R} \f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_pressure
  function compute_temperature(U)
    real(PR), dimension(:), intent(in) :: U
    real(PR) :: compute_temperature

    compute_temperature = compute_pressure(U)*molar_mass/(U(1)*ideal_gas_constant)

  end function compute_temperature

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the radiative pressure
  !>
  !> @param[in] U conservative variables
  !> @return the radiative pressure:
  !> \f$P = \frac{E}{2}\left( (1-\chi)Id  + (3\chi-1)\frac{F \otimes F}{\Vert F \Vert_2^2}\right) \f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_chi
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_array_mp_tensor_product
  function compute_radiative_pressure(U) result(P_R)
    real(PR), dimension(:), intent(in) :: U
    real(PR), dimension(dim_,dim_) :: P_R

    real(PR), dimension(dim_) :: F
    real(PR) :: chi, NORM2_F
    integer  :: i

    F(:) = U(6:7)
    NORM2_F = NORM2(F)

    if (NORM2_F < eps) then
      P_R(:,:) = zero
      do i = 1, dim_
        P_R(i,i) = third
      end do

    else
      chi = compute_chi(U)

      P_R(:,:) = 0.5_PR*(3._PR*chi - 1._PR)*tensor_product(F,F)/NORM2_F**2

      do i = 1, dim_
        P_R(i,i) = P_R(i,i) + 0.5_PR*(1._PR - chi)
      end do

    end if

    P_R(:,:) = U(5)*P_R(:,:)

  end function compute_radiative_pressure

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the Eddington factor
  !>
  !> @remark
  !> In order that \f$\chi\f$ is well defined the anistropy factor
  !> need to satisfy \f$f=\frac{\Vert F \Vert}{c E} \leq 1\f$
  !>
  !> @param[in] U conservative variables
  !> @return the Eddington factor: \f$\chi = \frac{3 + 4f^2}{5+ \sqrt{4-3f^2}} \f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_chi(U)
    real(PR), dimension(:), intent(in) :: U
    real(PR) :: compute_chi

    real(PR) :: anisotropy_factor_2

    anisotropy_factor_2 = (NORM2(U(6:7))/(c*U(5)))**2

#if defined(_DEBUG_)
    if (anisotropy_factor_2 > one) then
      write(StdOut,'(A,ES23.16,A)') '||F||/(c*E) =', sqrt(anisotropy_factor_2), ' > 1'
    end if
#endif /* _DEBUG_ */

    compute_chi = (3._PR + 4._PR*anisotropy_factor_2)&
        /(5._PR + 2._PR*sqrt(4._PR - 3._PR*anisotropy_factor_2))
  end function compute_chi

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the primitive to conservative variables
  !>
  !> @remark
  !> For the radiative hydrodynamic model the primitive variables are:
  !> * \f$ \rho \f$  the density,
  !> * \f$ u \f$ the speed,
  !> * \f$ p \f$ the pressure,
  !> * \f$ E_R \f$ the radiative energy
  !> * \f$ {\bf F}_R \f$ the radiative flux.
  !>
  !> @remark
  !> For the radiative hydrodynamic model the conservative variables are:
  !> * \f$ \rho \f$  the density,
  !> * \f$ \rho u \f$ the momentum,
  !> * \f$ E \f$ the energy,
  !> * \f$ E_R \f$ the radiative energy,
  !> * \f$ {\bf F}_R \f$ the radiative flux.
  !>
  !> @param[in] V_K primitive variables
  !> @param[out] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine primitive_to_conservative(V_K,U_K)
    ! [rho, u, p, E_R, F_R] => [rho, rho*u, E, E_R, F_R]
    real(PR), dimension(:), intent(in) :: V_K
    real(PR), dimension(:), intent(out) :: U_K

    ! rho
    U_K(1) = V_K(1)
    ! rho*u
    U_K(2:3) = V_K(2:3)*V_K(1)
    ! E = rho*0.5*(u**2 + v**2) + p/(gamma-1)
    U_K(4) = V_K(1)*0.5_PR*dot_product(V_K(2:3),V_K(2:3)) + V_K(4)/adiabatic_index_m1
    ! E_R
    U_K(5) = V_K(5)
    ! F_R
    U_K(6:7) = V_K(6:7)

  end subroutine primitive_to_conservative

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the conservative to primitive variables
  !>
  !> @remark
  !> For the radiative hydrodynamic model the primitive variables are:
  !> * \f$ \rho \f$  the density,
  !> * \f$ u \f$ the speed,
  !> * \f$ p \f$ the pressure,
  !> * \f$ E_R \f$ the radiative energy
  !> * \f$ {\bf F}_R \f$ the radiative flux.
  !>
  !> @remark
  !> For the radiative hydrodynamic model the conservative variables are:
  !> * \f$ \rho \f$  the density,
  !> * \f$ \rho u \f$ the momentum,
  !> * \f$ E \f$ the energy,
  !> * \f$ E_R \f$ the radiative energy,
  !> * \f$ {\bf F}_R \f$ the radiative flux.
  !>
  !> @param[in] U_K conservative variables
  !> @param[out] V_K primitive variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine conservative_to_primitive(U_K,V_K)
    ! [rho, rho*u, E, E_R, F_R] =>  [rho, u, p, E_R, F_R]
    real(PR), dimension(:), intent(in) :: U_K
    real(PR), dimension(:), intent(out) :: V_K

    ! rho
    V_K(1) = U_K(1)
    ! u
    V_K(2:3) = U_K(2:3)/U_K(1)
    ! p
    V_K(4) = compute_pressure(U_K)
    ! E_R
    V_K(5) = U_K(5)
    ! F_R
    V_K(6:7) = U_K(6:7)

  end subroutine conservative_to_primitive

  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the value of the solution on a cell is physically admissible.
  !>
  !> @param[in] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_temperature
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_pressure
  function IsPhysicallyAdmissible(U_K)
    real(PR), dimension(:), intent(in) :: U_K

    logical :: IsPhysicallyAdmissible

    if (U_K(1) < zero) then
      ! \rho < 0
      IsPhysicallyAdmissible = .FALSE.
    else if (compute_pressure(U_K) < 0._PR) then
      ! p < 0
      IsPhysicallyAdmissible = .FALSE.
    else if (U_K(4) < zero) then
      ! E < 0
      IsPhysicallyAdmissible = .FALSE.
    else if (U_K(5) < zero) then
      ! E_R < 0
      IsPhysicallyAdmissible = .FALSE.
    else if (NORM2(U_K(6:7)) > c*U_K(5)) then
      ! ||F_R|| > (c*E_R)
      IsPhysicallyAdmissible = .FALSE.
    else if (compute_temperature(U_K) < zero) then
      ! T < 0
      IsPhysicallyAdmissible = .FALSE.
    else
      IsPhysicallyAdmissible = .TRUE.
    end if
  end function IsPhysicallyAdmissible

#endif /* _PHY_HYDRORAD_ */
end module mod_physics_HydroRad
