!> Physics of the hyperbolic heat equation
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t p   + \partial_x u_x  + \partial_y u_y  & = & 0 \\
!>   \partial_t u_x + \partial_x p                      & = & -\sigma u_x \\
!>   \partial_t u_y                   + \partial_y p    & = & -\sigma u_y
!> \end{array}\right.\f]
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_HypHeat
#if defined(_PHY_HYPHEAT_)

  use mod_numerics, only: PR, Pi, eps, sqrt_2
  use type_IO, only: StdErr
  use type_mesh, only: pmesh, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */
  use mod_read, only: read_solution
  use mod_read, only: Restarting

  use mod_physics_HypHeat_var, only: T_init
  use mod_physics_HypHeat_var, only: NbEq, test_case
#if defined(_SRC_ENABLE_)
  use mod_physics_HypHeat_var, only: sigma
#endif /* defined(_SRC_ENABLE_) */
  use mod_physics_HypHeat_var, only: ExactSolutionKnown

#if defined(_HO_MOOD_)
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_int, only: compute_mean_f
#endif /* _HO_MOOD_ */

  implicit none

  private
  public physic_flux, U_Dirichlet
  public compute_wave_speed, compute_greatest_wave_speed
  public init_U
  public ExactSolution
  public primitive_to_conservative, conservative_to_primitive
  public IsPhysicallyAdmissible


contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Give the initial condition according to the test case
  !>
  !> @param[out] U0 array containing the initial condition
  !> @param[out] NbIter Number of iterations made
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine init_U(U0,NbIter)
      real(PR), dimension(:,:), intent(out) :: U0
      integer, intent(out) :: NbIter

      real(PR), dimension(dim_) :: xK
      integer :: CellK

      !$OMP SINGLE
      ExactSolutionKnown = .FALSE.
      select case(test_case)
      case ('cst')
#if defined(_SRC_ENABLE_)
      case ('cont', &
            'heat_fonda', &
            'heat_exp', 'heat_trig', &
            'heat_exp2', 'heat_trig2')
        ExactSolutionKnown = .TRUE.
#endif /* _SRC_ENABLE_ */
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        flush(StdErr)
        STOP
      end select
      !$OMP END SINGLE NOWAIT

      test_restart: if (Restarting) then
        !$OMP SINGLE
        call read_solution(U0,NbIter,T_init)
        !$OMP END SINGLE COPYPRIVATE(NbIter)

      else
        NbIter = 0
        !$OMP SINGLE
        T_init = 0._PR
        !$OMP END SINGLE

        select case(test_case)
        case ('cst')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(1,CellK) = 1._PR
            U0(2,CellK) = 0._PR
            U0(3,CellK) = 0._PR
          end do
          !$OMP END DO

#if defined(_SRC_ENABLE_)
        case ('cont', &
            'heat_fonda', &
            'heat_exp', 'heat_trig', &
            'heat_exp2', 'heat_trig2')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
#if (_HO_MOOD_MAX_DEGREE_ >= 1) || (_HO_MUSCL_MAX_DEGREE_ == 1)
            U0(:,CellK) = compute_mean_f(CellK,T_init,MaxDegree,InitialCondition,NbEq)
#else
            xK(:) = pmesh%elt(CellK)%x(:)
            U0(:,CellK) = InitialCondition(xK,T_init,NbEq)
#endif
          end do
          !$OMP END DO
#endif /* _SRC_ENABLE_ */
        end select

      end if test_restart

  end subroutine init_U

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the initial condition
  !>
  !> @param[in] xK Coordinates
  !> @param[in] t_n Current time
  !> @param[in] NbEq Number of equations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function InitialCondition(xK,t_n,NbEq) result(U0)
      real(PR), dimension(:), intent(in) :: xK
      real(PR), intent(in) :: t_n
      integer, intent(in) :: NbEq
      real(PR), dimension(NbEq) :: U0

      U0(:) = ExactSolution(xK,t_n,NbEq)

  end function InitialCondition

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the exact solution
  !>
  !> @param[in] xK Coordinates
  !> @param[in] t_n Current time
  !> @param[in] NbEq Number of equations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function ExactSolution(xK,t_n,NbEq) result(U_ex)
      real(PR), dimension(:), intent(in) :: xK
      real(PR), intent(in) :: t_n
      integer, intent(in) :: NbEq
      real(PR), dimension(NbEq) :: U_ex

#if defined(_SRC_ENABLE_)
      real(PR) :: sigma_tmp, t_offset
#endif /* defined(_SRC_ENABLE_) */

      select case(test_case)
      case ('cst')
        U_ex(:) = 1._PR
#if defined(_SRC_ENABLE_)
      case ('cont')
        ! See Fra, BueDesFra et BuesDesFraLer
        ! Trick to avoid a compilation error with negative square root
        sigma_tmp = sigma
        if (sigma > sqrt(32._PR)*Pi) then
          U_ex(1) = 0.5_PR/sigma*(&
              + sigma*exp( 0.5_PR*(-sigma + sqrt(sigma_tmp**2 - 32._PR*Pi**2))*t_n) &
              + sigma*exp(-0.5_PR*( sigma + sqrt(sigma_tmp**2 - 32._PR*Pi**2))*t_n) &
              + sqrt(sigma_tmp**2 - 32._PR*Pi**2)*exp( 0.5_PR*(-sigma + sqrt(sigma_tmp**2 - 32._PR*Pi**2))*t_n) &
              - sqrt(sigma_tmp**2 - 32._PR*Pi**2)*exp(-0.5_PR*( sigma + sqrt(sigma_tmp**2 - 32._PR*Pi**2))*t_n) &
              )
          U_ex(2:3) = 2._PR*Pi/sigma*(&
              exp( 0.5_PR*(-sigma + sqrt(sigma_tmp**2 - 32._PR*Pi**2))*t_n) +&
              exp(-0.5_PR*( sigma + sqrt(sigma_tmp**2 - 32._PR*Pi**2))*t_n) &
              )
        else if (sigma < sqrt(32._PR)*Pi) then
          U_ex(1) = - 0.5_PR*exp(-0.5_PR*sigma*t_n)/sigma*(&
              - sigma*sin(0.5_PR*sqrt(32._PR*Pi**2 - sigma_tmp**2)*t_n) &
              - sigma*cos(0.5_PR*sqrt(32._PR*Pi**2 - sigma_tmp**2)*t_n) &
              - sqrt(32._PR*Pi**2 - sigma_tmp**2)*cos(0.5_PR*sqrt(32._PR*Pi**2 - sigma_tmp**2)*t_n) &
              + sqrt(32._PR*Pi**2 - sigma_tmp**2)*sin(0.5_PR*sqrt(32._PR*Pi**2 - sigma_tmp**2)*t_n) &
              )
          U_ex(2:3) = 2._PR*Pi*exp(-0.5_PR*sigma*t_n)/sigma*(&
              sin(0.5_PR*sqrt(32._PR*Pi**2 - sigma_tmp**2)*t_n) +&
              cos(0.5_PR*sqrt(32._PR*Pi**2 - sigma_tmp**2)*t_n)&
              )
        else
          U_ex(1) = 1._PR/(8._PR*Pi)*exp(-2._PR*sqrt_2*t_n)*(4._PR*Pi + 4._PR*Pi*t_n + sqrt_2)
          U_ex(2:3) = 0.25_PR*sqrt_2*exp(-2._PR*sqrt_2*t_n)*(1._PR + t_n)
        end if
        U_ex(1) = U_ex(1)*cos(2._PR*Pi*xK(1))*cos(2._PR*Pi*xK(2))
        U_ex(2) = U_ex(2)*sin(2._PR*Pi*xK(1))*cos(2._PR*Pi*xK(2))
        U_ex(3) = U_ex(3)*cos(2._PR*Pi*xK(1))*sin(2._PR*Pi*xK(2))
      case ('heat_fonda')
        ! See https://en.wikipedia.org/wiki/Heat_equation#Fundamental_solutions
        t_offset = t_n + sigma/1000._PR
        U_ex(1) = sigma/(4._PR*Pi*t_offset)*exp(- sigma*sum((xK(:) - 0.5_PR*pmesh%length(:))**2)/(4._PR*t_offset))
        U_ex(2:3) = 0._PR
      case ('heat_exp')
        U_ex(1) = (exp(xK(1)) + exp(-xK(1)))*(exp(xK(2)) + exp(-xK(2)))*exp(2._PR*t_n/sigma)
        U_ex(2:3) = 0._PR
      case ('heat_exp2')
        U_ex(1) = (exp(sqrt(sigma)*xK(1)) + exp(-sqrt(sigma)*xK(1)))*(exp(sqrt(sigma)*xK(1)) + exp(-sqrt(sigma)*xK(1)))*exp(2._PR*t_n)
        U_ex(2:3) = 0._PR
      case ('heat_trig')
        U_ex(1) = (cos(xK(1)) + sin(xK(1)))*(cos(xK(2)) + sin(xK(2)))*exp(2._PR*t_n/sigma)
        U_ex(2:3) = 0._PR
      case ('heat_trig2')
        U_ex(1) = (cos(sqrt(sigma)*xK(1)) + sin(sqrt(sigma)*xK(1)))*(cos(sqrt(sigma)*xK(2)) + sin(sqrt(sigma)*xK(2)))*exp(-2._PR*t_n)
        U_ex(2:3) = 0._PR
#endif /* _SRC_ENABLE_ */
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        flush(StdErr)
        STOP
      end select

  end function ExactSolution

  !--------------------------------------------------------------------!
  !> @brief
  !> Give the limit boundary value
  !>
  !> @param[in] x Coordinates of the point on the boundary
  !> @param[in] t_n Current time
  !>
  !> @return Value of the solution on the boundary at point
  !> \f$x\f$ and time \f$t^n\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function U_Dirichlet(x,t_n)
      real(PR), dimension(:), intent(in) :: x
      real(PR), intent(in) :: t_n
      real(PR), dimension(NbEq) :: U_Dirichlet

      U_Dirichlet(:) = ExactSolution(x,t_n,NbEq)

  end function U_Dirichlet

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the wave speeds waves on a interface
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @param[out] a_l, a_r  Wave speeds on the left and right
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent
      real(PR), intent(out) :: a_l, a_r

      a_l = -1._PR
      a_r = 1._PR
  end subroutine compute_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the largest wave speed
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @return Largest wave speed
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_wave_speed
  function compute_greatest_wave_speed(Ul,Ur,normal,tangent)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent

      real(PR) :: compute_greatest_wave_speed

      compute_greatest_wave_speed = 1._PR

  end function compute_greatest_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the physic flux
  !>
  !> @remark
  !> For the hyperbolic heat model:
  !> \f${\bf F}({\bf U}) = \left(\begin{array}{c}u^t \\ p I \end{array}\right)\f$
  !>
  !> @param[in] U_K solution on cell CellK
  !> @return the physic flux on the cell CellK
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function physic_flux(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(NbEq,dim_)     :: physic_flux

      physic_flux(1,1) = U_K(2)
      physic_flux(2,1) = U_K(1)
      physic_flux(3,1) = 0._PR

      physic_flux(1,2) = U_K(3)
      physic_flux(2,2) = 0._PR
      physic_flux(3,2) = U_K(1)

  end function physic_flux

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the primitive to conservative variables
  !>
  !> @remark
  !> For the hyperbolic heat model the conservative and primitive variables are the same.
  !>
  !> @param[in] V_K primitive variables
  !> @param[out] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine primitive_to_conservative(V_K,U_K)
      real(PR), dimension(:), intent(in)     :: V_K
      real(PR), dimension(:), intent(out) :: U_K

      U_K(:) = V_K(:)

  end subroutine primitive_to_conservative

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the conservative to primitive variables
  !>
  !> @remark
  !> For the hyperbolic heat model the conservative and primitive variables are the same.
  !>
  !> @param[in] U_K conservative variables
  !> @param[out] V_K primitive variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine conservative_to_primitive(U_K,V_K)
      real(PR), dimension(:), intent(in)     :: U_K
      real(PR), dimension(:), intent(out) :: V_K

      V_K(:) = U_K(:)

  end subroutine conservative_to_primitive

  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the value of the solution on a cell is physically admissible.
  !>
  !> @param[in] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function IsPhysicallyAdmissible(U_K)
      real(PR), dimension(:), intent(in) :: U_K

      logical :: IsPhysicallyAdmissible

      IsPhysicallyAdmissible = .TRUE.

  end function IsPhysicallyAdmissible

#endif /* _PHY_HYPHEAT_ */
end module mod_physics_HypHeat
