!> Physics of the multigroup M1 model for radiative transfer.
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t E_q       +     \text{div} ({\bf F}_q) & = & c (\sigma^e_q a \theta_q^4(T) - \sigma^a_q E_q) \\
!>   \partial_t {\bf F}_q + c^2 \text{div} ({\bf P_q}) & = & -c \sigma^f_q {\bf F}_q \\
!>   \rho C_v \partial_t T                             & = & -c \sum_{q=1}^Q (\sigma^e_q a \theta_q^4(T) - \sigma^a_q E_q)
!> \end{array}\right.\f]
!>
!> where the index \f$q\f$ defines each group.
!>
!> with :
!> * \f$ E_q \f$ the radiative energy,
!> * \f$ {\bf F}_q \f$ the radiative flux,
!> * \f$ T \f$ the temperature,
!> * \f$ {\bf P}_q = \frac{E_q}{2}\left( (1-\chi_q) {\bf I}  + (3\chi_q-1)\frac{{\bf F}_q\otimes {\bf F}_q}{\Vert {\bf F}_q \Vert_2^2}\right) \f$ the radiative pressure,
!> * \f$\chi_q\f$ is the Eddington factor and it is no more explicit as in the grey M1 model, but as a fist approxamtion we use it:
!>     \f$\chi_q = \frac{3 + 4f_q^2}{5+ \sqrt{4-3f_q^2}}\f$
!> * \f$f_q\f$ is the anistropy factor: \f$f_q=\frac{\Vert {\bf F}_q \Vert}{c E_q} \leq 1 \f$,
!> * \f$ c \f$ the speed of light.
!>
!> @remark See \cite Tur_2002 \cite Tur_2003 \cite Sar_2013 for the definition.
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_M1MG_var
#if defined(_PHY_M1MG_)

  use type_mesh, only: dim_
  use mod_numerics, only: PR, pi, MaxLenStr, eps
  use mod_tools, only: LenInt, Int2Str
  use type_IO, only: StdErr
  use mod_open_file, only: OpenFile

  implicit none

  private
  public NbEq, NbEq_str, T_init, T_end
  public len_model, len_primitive, len_conservative
  public model_name, primitive_name, conservative_name, test_case

  public c, a, h, k_B
#if defined(_SRC_ENABLE_)
  public rho, C_v
#endif /* _SRC_ENABLE_ */

  public NbGroup, frequencies
  public Init_M1MG, Finalize_M1MG

  public tab_T
  public tab_f
  public tab_T_R
  public tab_sigma_a, tab_sigma_e, tab_sigma_f, tab_sigma_R
  public tab_chi

  integer, parameter                  :: len_model = 4
  !> Name of the model
  character(len=len_model), parameter :: model_name = 'M1MG'
  !> Number of groups in the model
  integer, parameter                  :: NbGroup = _PHY_M1MG_NB_GROUP_
  !> Number of equations (unknowns) in the model
  integer, parameter                  :: NbEq = (1+dim_)*NbGroup + 1
  !> Number of equations (unknowns) in the model, in a string
  character(len=:), allocatable, save :: NbEq_str

  !> Speed of light (\f$c \approx 299\ 792\ 458\ m.s^{-1}\f$)
  real(PR), parameter :: c = 299792458._PR
  !real(PR), parameter :: c = 1._PR

  !> Gas constant (\f$R \approx 3.314\ 462\ 1 \ J.mol^{-1}.K^{-1}\f$)
  real(PR), parameter :: ideal_gas_constant = 8.3144621_PR

  !> Avogadro constant (\f$N_A \approx 6.022\ 141\ 29\ \times 10^{23}\ mol^{-1}\f$)
  real(PR), parameter :: N_A = 6.02214129E+23_PR

  !> Boltzmann constant (\f$k_B = \frac{R}{N_A} \approx 1.380\ 648\ 8 \times 10^{-23}\ J.K^{-1}\f$)
  real(PR), parameter :: k_B = ideal_gas_constant/N_A

  !> Planck constant (\f$h \approx 6.626\ 069\ 57 \times10^{-34}\ J.s^{-1}\f$)
  real(PR), parameter :: h = 6.62606957E-34_PR

  !> Reduced Planck constant (\f$\bar{h}=\frac{h}{2\pi}\ J.s^{-1}\f$)
  real(PR), parameter :: h_bar = h/(2._PR*pi)

  !> Stefan-Boltzmann constant (\f$\sigma =\frac{2\pi^5R^4}{15h^3c^2N_A^4} \approx 5.670\ 373\ 21\times10^{-8}\ W.m^{-2}.K^{-4}\f$)
  real(PR), parameter :: sigma = (2._PR*pi**5*ideal_gas_constant**4)/(15._PR*h**3*c**2*N_A**4)

  !> Radiation constant (or radiation density constant) \f$a = \frac{4\sigma}{c} \approx 7.5657 \times10^{-16}\ J.m^{-3}.K^{-4}\f$)
  real(PR), parameter :: a = 4._PR*sigma/c

#if defined(_SRC_ENABLE_)
  !> Density of the fluid (\f$kg.m^{-3}\f$)
  real(PR), dimension(:), allocatable, save :: rho
  !> Volumic heat capacity (\f$J.K^{-1}\f$)
  real(PR), dimension(:), allocatable, save :: C_v
#endif /* _SRC_ENABLE_ */

  !> Intial time of the simulation
  real(PR), save :: T_init
  !> Final time of the simulation
  real(PR), save :: T_end

  integer, save :: len_primitive
  integer, save :: len_conservative

  !> Primitive variables used in the model
  character(len=:), dimension(:), allocatable, save :: primitive_name
  !> Conservative variables used in the model
  character(len=:), dimension(:), allocatable, save :: conservative_name

  !> Name of the test case
  character(len=MaxLenStr), save :: test_case

  !> Boundaries of each group
  real(PR), dimension(:), allocatable, save :: frequencies


  !> Number of samples for the anistropy factor
  integer, parameter :: nb_sampling_f = 30
  !> Anisotropy factor samples
  real(PR), dimension(:), allocatable, save :: tab_f

  !> Number of samples for the anistropy factor
  integer, parameter :: nb_sampling_T_R = 30
  !> Minimum of radiative temperature
  real(PR), parameter :: min_sampling_T_R = 11604.505_PR
  !> Maxmum of radiative temperature
  real(PR), parameter :: max_sampling_T_R = 232090.10_PR
  !> Radiative temperature samples
  real(PR), dimension(:), allocatable, save :: tab_T_R

  !> Number of samples of temperature
  integer, parameter :: nb_sampling_T = 27
  !> Temperature samples
  real(PR), dimension(:), allocatable, save :: tab_T
  !> Temperature samples (string version)
  character(len=6), dimension(:), allocatable, save :: tab_T_str

  !> Mean opacity of absorption
  real(PR), dimension(:,:,:,:), allocatable, save :: tab_sigma_a
  !> Mean opacity of emission
  real(PR), dimension(:,:), allocatable, save :: tab_sigma_e
  !> Mean opacity of absorption of flux
  real(PR), dimension(:,:,:,:), allocatable, save :: tab_sigma_f
  !> Mean opacity of Rosseland (for the limit)
  real(PR), dimension(:), allocatable, save :: tab_sigma_R

  !> Eddington factor \f$\chi\f$
  real(PR), dimension(:,:,:,:), allocatable, save :: tab_chi

contains

  !--------------------------------------------------------------------------
  !> @brief
  !> Initialize some parameters for the multigroup M1 model
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine Init_M1MG()

    integer :: len_NbGroup, len_NbEq
    integer :: group
    integer :: idx_file, FileUnit
    integer :: idx_f, idx_T_R
    integer :: tmp1, tmp2
    real(PR) :: tmp3, tmp4
    real(PR) :: temp

    !================================================================!
    ! Definition of the name of the variables
    !================================================================!
    len_NbEq = LenInt(NbEq)
    allocate(character(len=len_NbEq) :: NbEq_str)
    NbEq_str = Int2Str(NbEq)

    len_NbGroup = LenInt(NbGroup)
    len_primitive    = len_NbGroup + len('F_Rx_')
    len_conservative = len_NbGroup + len('F_Rx_')

    allocate(character(len=len_primitive) :: primitive_name(NbEq))
    allocate(character(len=len_conservative) :: conservative_name(NbEq))

    do group = 1, NbGroup
      primitive_name((1+dim_)*(group-1) + 1) = 'E_R_'//Int2Str(group,len_NbGroup)
      primitive_name((1+dim_)*(group-1) + 2) = 'F_Rx_'//Int2Str(group,len_NbGroup)
      primitive_name((1+dim_)*(group-1) + 3) = 'F_Ry_'//Int2Str(group,len_NbGroup)

      conservative_name((1+dim_)*(group-1) + 1) = 'E_R_'//Int2Str(group,len_NbGroup)
      conservative_name((1+dim_)*(group-1) + 2) = 'F_Rx_'//Int2Str(group,len_NbGroup)
      conservative_name((1+dim_)*(group-1) + 3) = 'F_Ry_'//Int2Str(group,len_NbGroup)
    end do

    primitive_name(NbEq) = 'T'
    conservative_name(NbEq) = 'T'
    !================================================================!

    !================================================================!
    ! Definition of the groups of frequencies
    !================================================================!
    allocate(frequencies(NbGroup+1))
    if (NbGroup == 6) then
      frequencies(1) = 1.0E13_PR
      frequencies(2) = 3.5E14_PR
      frequencies(3) = 3.0E15_PR
      frequencies(4) = 6.0E15_PR
      frequencies(5) = 9.0E15_PR
      frequencies(6) = 4.6E16_PR
      frequencies(7) = 5.0E17_PR
    else if (NbGroup == 3) then
      frequencies(1) = 1.0E13_PR
      frequencies(2) = 3.0E15_PR
      frequencies(3) = 9.0E15_PR
      frequencies(4) = 5.0E17_PR
    else if (NbGroup == 2) then
      frequencies(1) = 1.0E13_PR
      frequencies(2) = 6.0E15_PR
      frequencies(3) = 5.0E17_PR
    else if (NbGroup == 1) then
      frequencies(1) = 1.0E13_PR
      frequencies(2) = 5.0E17_PR
    else
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Groups not properly defined"
      flush(StdErr)
      STOP
    end if

    !================================================================!
    ! Reading of the precomputed mean opacities
    !================================================================!
    allocate(tab_T(nb_sampling_T))
    allocate(tab_T_str(nb_sampling_T))

    allocate(tab_f(nb_sampling_f))
    do idx_f = 1, nb_sampling_f
      tab_f(idx_f) = 0._PR + (1._PR - 0._PR)*real(idx_f-1,PR)/real(nb_sampling_f - 1,PR)
    end do

    allocate(tab_T_R(nb_sampling_T_R))
    do idx_T_R = 1, nb_sampling_T_R
      tab_T_R(idx_T_R) = min_sampling_T_R &
          + (max_sampling_T_R - min_sampling_T_R)*real(idx_T_R-1,PR)/real(nb_sampling_T_R - 1,PR)
    end do

#if defined(_SRC_ENABLE_)
    allocate(tab_sigma_a(nb_sampling_f,nb_sampling_T_R,nb_sampling_T,NbGroup))
    allocate(tab_sigma_e(nb_sampling_T,NbGroup))
    allocate(tab_sigma_f(nb_sampling_f,nb_sampling_T_R,nb_sampling_T,NbGroup))
    allocate(tab_sigma_R(nb_sampling_T))
#endif /* defined(_SRC_ENABLE_) */
    allocate(tab_chi(nb_sampling_f,nb_sampling_T_R,nb_sampling_T,NbGroup))

    tab_T_str = ['011877', '013342', '014988',&
        '016836', '018912', '021245',&
        '023865', '026808', '030114',&
        '033828', '038000', '042686',&
        '047950', '053863', '060506',&
        '067968', '076350', '085766',&
        '096343', '108224', '121571',&
        '136563', '153405', '172323',&
        '193575', '217447', '244264']

    do idx_file = 1, nb_sampling_T
#if defined(_SRC_ENABLE_)
      !================================================================!
      ! Reading of sigma^a_q(f,T_R,T)
      !================================================================!
      call OpenFile(unit=FileUnit,file='data_M1mg/sigma_a_Kr_'//Int2Str(NbGroup,2)//'g/sigma_a_Kr_'//Int2Str(NbGroup,2)//'g_T'//tab_T_str(idx_file)//'.data',&
          action='read',status='old')
      read(FileUnit,*) tmp1, tmp2, tmp3, tmp4
      if (tmp1 /= nb_sampling_T_R) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same sampling on radiative temperature"
        flush(StdErr)
        STOP
      end if
      if (tmp2 /= nb_sampling_f) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same sampling on anistropy factor"
        flush(StdErr)
        STOP
      end if
      if (abs(tmp3 - min_sampling_T_R) > eps) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same minimum for radiative temperature"
        flush(StdErr)
        STOP
      end if
      if (abs(tmp4 - max_sampling_T_R) > eps) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same maximum for radiative temperature"
        flush(StdErr)
        STOP
      end if

      do group = 1, NbGroup
        do idx_T_R = 1, nb_sampling_T_R
          do idx_f = 1, nb_sampling_f
            read(FileUnit,*) tab_sigma_a(idx_f,idx_T_R,idx_file,group)
          end do
        end do
      end do

      close(FileUnit)
      !================================================================!

      !================================================================!
      ! Reading of sigma^e_q(T)
      !================================================================!
      call OpenFile(unit=FileUnit,file='data_M1mg/sigma_e_Kr_'//Int2Str(NbGroup,2)//'g/sigma_e_Kr_'//Int2Str(NbGroup,2)//'g_T'//tab_T_str(idx_file)//'.data',&
          action='read',status='old')
      read(FileUnit,*) tab_T(idx_file)
      do group = 1, NbGroup
        read(FileUnit,*) tab_sigma_e(idx_file,group)
      end do
      close(FileUnit)
      !================================================================!

      !================================================================!
      ! Reading of sigma^f_q(f,T_R,T)
      !================================================================!
      call OpenFile(unit=FileUnit,file='data_M1mg/sigma_f_Kr_'//Int2Str(NbGroup,2)//'g/sigma_f_Kr_'//Int2Str(NbGroup,2)//'g_T'//tab_T_str(idx_file)//'.data',&
          action='read',status='old')
      read(FileUnit,*) tmp1, tmp2, tmp3, tmp4
      if (tmp1 /= nb_sampling_T_R) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same sampling on radiative temperature"
        flush(StdErr)
        STOP
      end if
      if (tmp2 /= nb_sampling_f) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same sampling on anistropy factor"
        flush(StdErr)
        STOP
      end if
      if (abs(tmp3 - min_sampling_T_R) > eps) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same minimum for radiative temperature"
        flush(StdErr)
        STOP
      end if
      if (abs(tmp4 - max_sampling_T_R) > eps) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same maximum for radiative temperature"
        flush(StdErr)
        STOP
      end if

      do group = 1, NbGroup
        do idx_T_R = 1, nb_sampling_T_R
          do idx_f = 1, nb_sampling_f
            read(FileUnit,*) tab_sigma_f(idx_f,idx_T_R,idx_file,group)
          end do
        end do
      end do

      close(FileUnit)
      !================================================================!

      !================================================================!
      ! Reading of sigma^R(T)
      ! This is for the limit equation, so we use the value integrated
      ! over the whole set of frequencies
      !================================================================!
      call OpenFile(unit=FileUnit,file='data_M1mg/sigma_R_Kr_01g/sigma_R_Kr_01g_T'//tab_T_str(idx_file)//'.data',&
          action='read',status='old')
      read(FileUnit,*) temp
      if (abs(temp - tab_T(idx_file)) > eps) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same sampling on temperature"
        flush(StdErr)
        STOP
      end if
      read(FileUnit,*) tab_sigma_R(idx_file)
      close(FileUnit)
      !================================================================!
#endif /* defined(_SRC_ENABLE_) */


      !================================================================!
      ! Reading of the precomputed factor of Eddington (\chi)
      !================================================================!
      call OpenFile(unit=FileUnit,file='data_M1mg/chi_Kr_'//Int2Str(NbGroup,2)//'g/chi_Kr_'//Int2Str(NbGroup,2)//'g_T'//tab_T_str(idx_file)//'.data',&
          action='read',status='old')
      read(FileUnit,*) tmp1, tmp2, tmp3, tmp4
      if (tmp1 /= nb_sampling_T_R) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same sampling on radiative temperature"
        flush(StdErr)
        STOP
      end if
      if (tmp2 /= nb_sampling_f) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same sampling on anistropy factor"
        flush(StdErr)
        STOP
      end if
      if (abs(tmp3 - min_sampling_T_R) > eps) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same minimum for radiative temperature"
        flush(StdErr)
        STOP
      end if
      if (abs(tmp4 - max_sampling_T_R) > eps) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Not the same maximum for radiative temperature"
        flush(StdErr)
        STOP
      end if

      do group = 1, NbGroup
        do idx_T_R = 1, nb_sampling_T_R
          do idx_f = 1, nb_sampling_f
            read(FileUnit,*) tab_chi(idx_f,idx_T_R,idx_file,group)
          end do
        end do
      end do

      close(FileUnit)
      !================================================================!
    end do
    !================================================================!

  end subroutine Init_M1MG

  !--------------------------------------------------------------------------
  !> @brief
  !> Dealloacte memory used by the parameters of the multigroup M1 model
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine Finalize_M1MG()

    deallocate(NbEq_str)
    deallocate(primitive_name)
    deallocate(conservative_name)

    deallocate(frequencies)

    deallocate(tab_T)
    deallocate(tab_T_str)

    deallocate(tab_f)
    deallocate(tab_T_R)

#if defined(_SRC_ENABLE_)
    deallocate(tab_sigma_a)
    deallocate(tab_sigma_e)
    deallocate(tab_sigma_f)
    deallocate(tab_sigma_R)
#endif /* defined(_SRC_ENABLE_) */

    deallocate(tab_chi)

  end subroutine Finalize_M1MG

#endif /* _PHY_M1MG_ */
end module mod_physics_M1MG_var
