!> Physics of the P1 model for radiative transfer.
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t E   +               \partial_x F_x  +               \partial_y F_y  & = & 0 \\
!>   \partial_t F_x + \frac{c^2}{3} \partial_x E                                    & = & -c \sigma^f F_x \\
!>   \partial_t F_y +                                  \frac{c^2}{3}\partial_y E    & = & -c \sigma^f F_y
!> \end{array}\right.\f]
!>
!> with:
!> * \f$ E \f$ the radiative energy,
!> * \f$ F \f$ the radiative flux,
!> * \f$ c \f$ the speed of light,
!> * \f$ \sigma_f \f$ the mean opacity of absorption of flux.
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_P1
#if defined(_PHY_P1_)

  use mod_numerics, only: PR, pi, eps
  use mod_numerics, only: zero
  use type_IO, only: StdErr
  use type_mesh, only: pmesh, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */
  use mod_read, only: read_solution
  use mod_read, only: Restarting

  use mod_physics_P1_var, only: NbEq, test_case
  use mod_physics_P1_var, only: T_init, T_end
  use mod_physics_P1_var, only: c

#if defined(_HO_MOOD_)
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_int, only: compute_mean_f
#endif /* _HO_MOOD_ */

  implicit none

  private
  public physic_flux, U_Dirichlet
  public compute_wave_speed, compute_greatest_wave_speed
  public init_U
  public primitive_to_conservative, conservative_to_primitive
  public IsPhysicallyAdmissible


contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Give the initial condition according to the test case
  !>
  !> @param[out] U0 array containing the initial condition
  !> @param[out] NbIter Number of iterations made
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine init_U(U0,NbIter)
      real(PR), dimension(:,:), intent(out) :: U0
      integer, intent(out) :: NbIter

      real(PR), dimension(dim_) :: xK, x_c
      integer :: CellK

      !$OMP SINGLE
      select case(test_case)
      case ('cst')
        T_end = 1._PR/c
      case ('x_1')
        T_end = 1.0_PR/(3._PR*c)
      case ('x_2')
        T_end = 1.0_PR/(3._PR*c)
      case ('Gauss_x')
        T_end = 1.E-7_PR
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        flush(StdErr)
        STOP
      end select
      !$OMP END SINGLE NOWAIT

      test_restart: if (Restarting) then
        !$OMP SINGLE
        call read_solution(U0,NbIter,T_init)
        !$OMP END SINGLE COPYPRIVATE(NbIter)

      else
        NbIter = 0
        !$OMP SINGLE
        T_init = 0._PR
        !$OMP END SINGLE

        select case(test_case)
        case ('cst')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(1,CellK) = 1._PR
            U0(2:3,CellK) = c*U0(1,CellK)
          end do
          !$OMP END DO

        case ('x_1')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) < 0.5_PR) then
              U0(1,CellK) = 1._PR
              U0(2,CellK) = 0.1_PR
              U0(3,CellK) = 0._PR
            else
              U0(1,CellK) = 10._PR
              U0(2,CellK) = -1._PR
              U0(3,CellK) = 0._PR
            end if
          end do
          !$OMP END DO

        case ('x_2')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) < 0.5_PR) then
              U0(1,CellK) = 2._PR
              U0(2,CellK) = -0.5_PR
              U0(3,CellK) = 0.0_PR
            else
              U0(1,CellK) = 2._PR
              U0(2,CellK) = 1._PR
              U0(3,CellK) = 0._PR
            end if
          end do
          !$OMP END DO

        case ('Gauss_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
            U0(:,CellK) = compute_mean_f(CellK,T_init,MaxDegree,InitialCondition,NbEq)
#else
            xK(:) = pmesh%elt(CellK)%x(:)
            U0(:,CellK) = InitialCondition(xK,T_init,NbEq)
#endif
          end do
          !$OMP END DO
        end select

      end if test_restart

  end subroutine init_U

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the initial condition
  !>
  !> @param[in] xK Coordinates
  !> @param[in] t_n Current time
  !> @param[in] NbEq Number of equations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function InitialCondition(xK,t_n,NbEq) result(U0)
      real(PR), dimension(:), intent(in) :: xK
      real(PR), intent(in) :: t_n
      integer, intent(in) :: NbEq
      real(PR), dimension(NbEq) :: U0

      select case(test_case)
      case ('Gauss_x')
        U0(1) = 300._PR*exp(- 0.5_PR*(xK(1) - 0.5_PR)**2/(0.05_PR)**2) + 300._PR
        U0(2:3) = 0._PR
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        flush(StdErr)
        STOP
      end select

  end function InitialCondition

  !--------------------------------------------------------------------!
  !> @brief
  !> Give the limit boundary value
  !>
  !> @param[in] x Coordinates of the point on the boundary
  !> @param[in] t_n Current time
  !>
  !> @return Value of the solution on the boundary at point
  !> \f$x\f$ and time \f$t^n\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function U_Dirichlet(x,t_n)
      real(PR), dimension(:), intent(in) :: x
      real(PR), intent(in) :: t_n
      real(PR), dimension(NbEq) :: U_Dirichlet

      select case(test_case)
      case ('cst')
        U_Dirichlet(1)   = 1._PR
        U_Dirichlet(2:3) = 0._PR
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        flush(StdErr)
        STOP
      end select

  end function U_Dirichlet

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the wave speeds waves on a interface
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @param[out] a_l, a_r  Wave speeds on the left and right
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent
      real(PR), intent(out) :: a_l, a_r

      a_l = -c
      a_r = c
  end subroutine compute_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the largest wave speed
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @return Largest wave speed
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_wave_speed
  function compute_greatest_wave_speed(Ul,Ur,normal,tangent)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent

      real(PR) :: compute_greatest_wave_speed

      compute_greatest_wave_speed = c

  end function compute_greatest_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the physic flux
  !>
  !> @remark
  !> For the P1 model:
  !> \f${\bf F}({\bf U}) = \left(\begin{array}{c}F^t \\ c^2 E \end{array}\right)\f$
  !>
  !> @param[in] U_K solution on cell CellK
  !> @return the physic flux on the cell CellK
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_radiative_pressure
  function physic_flux(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(NbEq,dim_)     :: physic_flux

      real(PR) :: c2_E_3
      real(PR), dimension(dim_) :: F

      c2_E_3 = c**2*U_K(1)/3._PR
      F(:) = U_K(2:3)

      physic_flux(1,1) = F(1)
      physic_flux(2,1) = c2_E_3
      physic_flux(3,1) = 0._PR

      physic_flux(1,2) = F(2)
      physic_flux(2,2) = 0._PR
      physic_flux(3,2) = c2_E_3

  end function physic_flux

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the primitive to conservative variables
  !>
  !> @remark
  !> For the P1 model the conservative and primitive variables are the same:
  !> * \f$ E \f$ the radiative energy
  !> * \f$ F \f$ the radiative flux
  !>
  !> @param[in] V_K primitive variables
  !> @param[out] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine primitive_to_conservative(V_K,U_K)
      real(PR), dimension(:), intent(in)     :: V_K
      real(PR), dimension(:), intent(out) :: U_K

      U_K(:) = V_K(:)

  end subroutine primitive_to_conservative

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the conservative to primitive variables
  !>
  !> @remark
  !> For the P1 model the conservative and primitive variables are the same:
  !> * \f$ E \f$ the radiative energy
  !> * \f$ F \f$ the radiative flux
  !>
  !> @param[in] U_K conservative variables
  !> @param[out] V_K primitive variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine conservative_to_primitive(U_K,V_K)
      real(PR), dimension(:), intent(in)     :: U_K
      real(PR), dimension(:), intent(out) :: V_K

      V_K(:) = U_K(:)

  end subroutine conservative_to_primitive

  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the value of the solution on a cell is physically admissible.
  !>
  !> @param[in] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function IsPhysicallyAdmissible(U_K)
      real(PR), dimension(:), intent(in) :: U_K

      logical :: IsPhysicallyAdmissible

      if (U_K(1) < zero) then
        IsPhysicallyAdmissible = .FALSE.
      else
        IsPhysicallyAdmissible = .TRUE.
      end if

  end function IsPhysicallyAdmissible

#endif /* _PHY_P1_ */
end module mod_physics_P1
