!> Physics of the P1 model for radiative transfer.
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t E   +               \partial_x F_x  +               \partial_y F_y  & = & 0 \\
!>   \partial_t F_x + \frac{c^2}{3} \partial_x E    +  \frac{c^2}{3}\partial_y E    & = & -c \sigma^f F_x \\
!>   \partial_t F_y + \frac{c^2}{3} \partial_x E    +  \frac{c^2}{3}\partial_y E    & = & -c \sigma^f F_y
!> \end{array}\right.\f]
!>
!> with:
!> * \f$ E \f$ the radiative energy,
!> * \f$ F \f$ the radiative flux,
!> * \f$ c \f$ the speed of light,
!> * \f$ \sigma_f \f$ the mean opacity of absorption of flux.
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_P1_var
#if defined(_PHY_P1_)

  use mod_numerics, only: PR,  MaxLenStr

  implicit none

  private
  public NbEq, NbEq_str, T_init, T_end
  public len_model, len_primitive, len_conservative
  public model_name, primitive_name, conservative_name, test_case

  public c, sigma_f

  integer, parameter                  :: len_model = 2
  !> Name of the model
  character(len=len_model), parameter :: model_name = 'P1'
  !> Number of equations (unknowns) in the model
  integer, parameter                  :: NbEq = 3
  !> Number of equations (unknowns) in the model, in a string
  character(len=1), parameter :: NbEq_str = '3'

  !> Speed of light (\f$c \approx 299\ 792\ 458\ m.s^{-1}\f$)
  real(PR), parameter :: c = 299792458._PR

  !> Mean opacity of absorption of flux
  real(PR), parameter :: sigma_f = 1500._PR

  !> Intial time of the simulation
  real(PR), save :: T_init
  !> Final time of the simulation
  real(PR), save :: T_end

  integer, parameter :: len_primitive    = 2
  integer, parameter :: len_conservative = 2

  !> Primitive variables used in the model
  character(len=len_primitive),    dimension(NbEq), parameter :: primitive_name    = ['E ', 'Fx', 'Fy']
  !> Conservative variables used in the model
  character(len=len_conservative), dimension(NbEq), parameter :: conservative_name = ['E ', 'Fx', 'Fy']

  !> Name of the test case
  character(len=MaxLenStr), save :: test_case

#endif /* _PHY_P1_ */
end module mod_physics_P1_var
