!> Physics of the Shallow Water equations \cite StV_1871 .
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t h   + \partial_x h u            + \partial_y h v            & = & 0 \\
!>   \partial_t h u + \partial_x (h u^2 + p(h)) + \partial_y h uv           & = & - \frac{\kappa}{h^\eta} h u \\
!>   \partial_t h v + \partial_x  h u v         + \partial_y (h v^2 + p(h)) & = & - \frac{\kappa}{h^\eta} h u \\
!> \end{array} \right.  \f]
!> with:
!> * \f$ h \f$: height,
!> * \f$ \vec{V} =(u,v) \f$: speed of the fluid,
!> * \f$ g \f$: gravity constannt,
!> * \f$ \kappa \f$: friction,
!> * \f$ p \f$: pressure,
!> * and the speed of the sound in the fluid is defined as: \f$c = \sqrt{gh}\f$.
!>
!> \cite StV_1871
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_SW
#if defined(_PHY_SW_)

  use mod_numerics, only: PR, eps
  use type_IO, only: StdErr
  use mod_tools, only: min_NaN, max_NaN
  use type_mesh, only: pmesh, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */
  use mod_read, only: read_solution
  use mod_read, only: Restarting

  use mod_physics_SW_var, only: NbEq, test_case
  use mod_physics_SW_var, only: T_init, T_end
  use mod_physics_SW_var, only: gravity

  implicit none

  private
  public physic_flux, U_Dirichlet
  public compute_wave_speed, compute_greatest_wave_speed
  public init_U
  public primitive_to_conservative, conservative_to_primitive
  public IsPhysicallyAdmissible

#if defined(_SRC_ENABLE_)
  public compute_pressure, p_prime_h
#endif /* _SRC_ENABLE_ */

!#define _PHY_SW_DRY_WET_

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Give the initial condition according to the test case
  !>
  !> @param[out] U0 array containing the initial condition
  !> @param[out] NbIter Number of iterations made
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_primitive_to_conservative
  subroutine init_U(U0,NbIter)
      real(PR), dimension(:,:), intent(out) :: U0
      integer, intent(out) :: NbIter

      real(PR), dimension(NbEq) :: U_K, V_K
      real(PR), dimension(dim_) :: xK
      integer :: CellK

      !$OMP SINGLE
      select case(test_case)
      case ('cst')
        T_end = 1._PR
      case ('DB_x')
        T_end = 0.03_PR
      case ('river')
        T_end = 2._PR
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        flush(StdErr)
        STOP
      end select
      !$OMP END SINGLE NOWAIT

      test_restart: if (Restarting) then
        !$OMP SINGLE
        call read_solution(U0,NbIter,T_init)
        !$OMP END SINGLE COPYPRIVATE(NbIter)

      else
        NbIter = 0
        !$OMP SINGLE
        T_init = 0._PR
        !$OMP END SINGLE

        select case(test_case)
        case ('cst')
          !$OMP SINGLE
          V_K(1) = 1._PR
          V_K(2) = 1.0_PR
          V_K(3) = 1.0_PR

          call primitive_to_conservative(V_K,U_K)
          !$OMP END SINGLE COPYPRIVATE(U_K)

          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('DB_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) < 0.5_PR) then
              V_K(1) = 6._PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
            else
#if defined (_PHY_SW_DRY_WET_)
              V_K(1) = 0._PR
#else
              V_K(1) = 0.5_PR
#endif /* _PHY_SW_DRY_WET_ */
              V_K(2) = 0._PR
              V_K(3) = 0._PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('river')
          !$OMP SINGLE
#if defined (_PHY_SW_DRY_WET_)
          V_K(1) = 0._PR
#else
          V_K(1) = 1._PR
#endif /* _PHY_SW_DRY_WET_ */
          V_K(2) = 0._PR
          V_K(3) = 0._PR
          call primitive_to_conservative(V_K,U_K)
          !$OMP END SINGLE COPYPRIVATE(U_K)

          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO
        end select

      end if test_restart

  end subroutine init_U

  !--------------------------------------------------------------------!
  !> @brief
  !> Give the limit boundary value
  !>
  !> @param[in] x Coordinates of the point on the boundary
  !> @param[in] t_n Current time
  !>
  !> @return Value of the solution on the boundary at point
  !> \f$x\f$ and time \f$t^n\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_primitive_to_conservative
  function U_Dirichlet(x,t_n)
      real(PR), dimension(:), intent(in) :: x
      real(PR), intent(in) :: t_n
      real(PR), dimension(NbEq) :: U_Dirichlet

      real(PR), dimension(NbEq) :: U_K, V_K

      select case(test_case)
      case ('river')
        V_K(1) = 2._PR
        V_K(2) = 2._PR
        V_K(3) = 0._PR

      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        flush(StdErr)
        STOP
      end select

      call primitive_to_conservative(V_K,U_K)
      U_Dirichlet(:) = U_K(:)

  end function U_Dirichlet

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the wave speeds waves on a interface
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent  Normal and tangent to the interface
  !> @param[out] a_l, a_r  Wave speeds on the left and right
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_min_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  subroutine compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent
      real(PR), intent(out) :: a_l, a_r

      ! speed of the fluid
      real(PR), dimension(dim_) :: u_l, u_r
      ! normal speed of the fluid
      real(PR) :: u_l_n, u_r_n
      ! speed of sound in the fluid
      real(PR) :: c_l, c_r

#if defined(_PHY_SW_DRY_WET_)
      test_IsNotZero_l: if (abs(Ul(1)) > eps) then
#endif /* _PHY_SW_DRY_WET_ */
        u_l(1) = Ul(2)/Ul(1)
        u_l(2) = Ul(3)/Ul(1)

#if defined (_PHY_SW_DRY_WET_)
      else
        u_l(:) = - eps
      endif test_IsNotZero_l
#endif /* _PHY_SW_DRY_WET_ */

#if defined(_PHY_SW_DRY_WET_)
      test_IsNotZero_r: if (abs(Ur(1)) > eps) then
#endif /* _PHY_SW_DRY_WET_ */
        u_r(1) = Ur(2)/Ur(1)
        u_r(2) = Ur(3)/Ur(1)
#if defined (_PHY_SW_DRY_WET_)
      else
        u_r(:) = eps
      endif test_IsNotZero_r
#endif /* _PHY_SW_DRY_WET_ */

      u_l_n = dot_product(u_l,normal)
      u_r_n = dot_product(u_r,normal)

      c_l = sound_speed(Ul)
      c_r = sound_speed(Ur)

      a_l = min_NaN(u_l_n - c_l, u_r_n - c_r)
      a_r = max_NaN(u_l_n + c_l, u_r_n + c_r)

  end subroutine compute_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the largest wave speed
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @return Largest wave speed
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  function compute_greatest_wave_speed(Ul,Ur,normal,tangent)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent

      real(PR) :: compute_greatest_wave_speed

      real(PR) :: a_l, a_r

      call compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)
      compute_greatest_wave_speed = max_NaN(abs(a_l),abs(a_r))

  end function compute_greatest_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the sound speed in the fluid
  !>
  !> @param[in] U_K Conservative variables
  !> @return the sound speed: \f$c = \sqrt{gh}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function sound_speed(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR) :: sound_speed

      sound_speed = sqrt(gravity*U_K(1))

  end function sound_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the physic flux
  !>
  !> @remark
  !> For the SW model:
  !> \f${\bf F}({\bf U}) = \left(\begin{array}{c}(h u)^t \\ h (\frac{1}{2} Id + u \otimes u) \end{array}\right)\f$
  !>
  !> @param[in] U_K solution on cell CellK
  !> @return the physic flux on the cell CellK
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function physic_flux(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(NbEq,dim_) :: physic_flux

      real(PR) :: h, h_u, h_v, p
#if defined (_PHY_SW_DRY_WET_)
      real(PR) :: u, v
#endif /* _PHY_SW_DRY_WET_ */

      h   = U_K(1)
      h_u = U_K(2)
      h_v = U_K(3)

#if defined (_PHY_SW_DRY_WET_)
      test_IsNotZero: if (abs(h) > eps) then
#endif /* _PHY_SW_DRY_WET_ */
        p = compute_pressure(U_K)

        physic_flux(1,1) = h_u
        physic_flux(2,1) = h_u**2/h + p
        physic_flux(3,1) = h_u*h_v/h

        physic_flux(1,2) = h_v
        physic_flux(2,2) = h_u*h_v/h
        physic_flux(3,2) = h_v**2/h + p

#if defined (_PHY_SW_DRY_WET_)
      else
        physic_flux(:,:) = 0._PR
      endif test_IsNotZero
#endif /* _PHY_SW_DRY_WET_ */

  end function physic_flux

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the pressure
  !>
  !> @param[in] U conservative variables
  !> @return the pressure : \f$p = g\frac{h^2}{2} \f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_pressure(U)
      real(PR),  dimension(:), intent(in) :: U
      real(PR) :: compute_pressure

      compute_pressure = 0.5_PR*gravity*U(1)**2

  end function compute_pressure

#if defined(_SRC_ENABLE_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the derivative of the pressure regarding to the density: \f$p'(h)\f$
  !>
  !> @param[in] U conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function p_prime_h(U)
      real(PR),  dimension(:), intent(in) :: U
      real(PR) :: p_prime_h

      p_prime_h = gravity*U(1)

  end function p_prime_h
#endif /* _SRC_ENABLE_ */

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the primitive to conservative variables
  !>
  !> @remark
  !> For the SW model the primitive variables are:
  !> * \f$ h \f$ the density
  !> * \f$ (u,v) \f$ the speed
  !>
  !> @remark
  !> For the SW model the conservative variables are:
  !> * \f$ h \f$  the density
  !> * \f$ (h u, h v) \f$ the discharge
  !>
  !> @param[in] V_K primitive variables
  !> @param[out] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine primitive_to_conservative(V_K,U_K)
      ! [h, u, v] => [h, h*u, h*v]
      real(PR), dimension(:), intent(in) :: V_K
      real(PR), dimension(:), intent(out) :: U_K

      ! h
      U_K(1) = V_K(1)
      ! h*u
      U_K(2) = V_K(2)*V_K(1)
      ! h*v
      U_K(3) = V_K(3)*V_K(1)

  end subroutine primitive_to_conservative

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the conservative to primitive variables
  !>
  !> @remark
  !> For the SW model the primitive variables are:
  !> * \f$ h \f$ the density
  !> * \f$ (u,v) \f$ the speed
  !>
  !> @remark
  !> For the SW model the conservative variables are:
  !> * \f$ h \f$  the density
  !> * \f$ (h u, h v) \f$ the discharge
  !>
  !> @param[in] U_K conservative variables
  !> @param[out] V_K primitive variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine conservative_to_primitive(U_K,V_K)
      ! [h, h*u, h*v] =>  [h, u, v]
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(:), intent(out) :: V_K


#if defined (_PHY_SW_DRY_WET_)
      test_IsNotZero: if (abs(U_K(1)) > eps) then
#endif /* _PHY_SW_DRY_WET_ */
        ! h
        V_K(1) = U_K(1)
        ! u
        V_K(2) = U_K(2)/U_K(1)
        ! v
        V_K(3) = U_K(3)/U_K(1)

#if defined (_PHY_SW_DRY_WET_)
      else
        V_K(:) = 0._PR
      endif test_IsNotZero
#endif /* _PHY_SW_DRY_WET_ */

  end subroutine conservative_to_primitive

  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the value of the solution on a cell is physically admissible.
  !>
  !> @param[in] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function IsPhysicallyAdmissible(U_K)
      real(PR), dimension(:), intent(in) :: U_K

      logical :: IsPhysicallyAdmissible

      if (U_K(1) < 0._PR) then
        IsPhysicallyAdmissible = .FALSE.
      else
        IsPhysicallyAdmissible = .TRUE.
      end if

  end function IsPhysicallyAdmissible

#endif /* _PHY_SW_ */
end module mod_physics_SW
