!> Routines for the entropy for shallow-water model see \cite TadZho_2008
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.

module mod_physics_SW_entropy
#if defined(_PHY_SW_) && defined(_HO_MOOD_ENTROPIC_)

  use mod_numerics, only: PR, eps
  use type_mesh, only: dim_, pmesh
  use mod_physics_SW_var, only: gravity, NbEq

  implicit none

  private
  public compute_entropy_math
  public entropy_flux

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the mathematical entropy: \f$\eta = \frac{1}{2}(gh^2 + u^2 h + v^2 h)\f$
  !>
  !> @param[in] U_K Mean values
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_entropy_math(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR) :: compute_entropy_math

      compute_entropy_math = 0.5_PR*(gravity*U_K(1) &
          + U_K(2)**2/U_K(1) + U_K(3)**2/U_K(1))

  end function compute_entropy_math

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the entropy flux: \f$G(U) = \left(g u h^2 + \frac{u^3h + u v^2 h}{2},
  !> g v h^2 + \frac{u^2 v h + v^3h}{2}\right)\f$
  !>
  !> @param[in] U_K Mean values
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function entropy_flux(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(dim_) :: entropy_flux

      entropy_flux(1) = gravity*U_K(2)*U_K(1) + 0.5_PR*(U_K(2)**3 + U_K(2)*U_K(3)**2)/U_K(1)**2
      entropy_flux(2) = gravity*U_K(3)*U_K(1) + 0.5_PR*(U_K(3)**3 + U_K(3)*U_K(2)**2)/U_K(1)**2

  end function entropy_flux

#endif /* _PHY_SW_ && _HO_MOOD_ENTROPIC_ */
end module mod_physics_SW_entropy
