!> Physics module for the Telegraph (or Goldstein-Taylor) equations
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t u + \text{div}(\vec{a} u) & = & \sigma (v - u) \\
!>   \partial_t v - \text{div}(\vec{a} v) & = & \sigma (u - v)
!> \end{array}\right.\f]
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_Telegraph
#if defined (_PHY_TELE_)

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: pmesh, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */
  use mod_read, only: read_solution
  use mod_read, only: Restarting

  use mod_physics_Telegraph_var, only: T_init
  use mod_physics_Telegraph_var, only: NbEq, test_case
  use mod_physics_Telegraph_var, only: speed_tele

  implicit none

  private
  public physic_flux, U_Dirichlet
  public compute_wave_speed, compute_greatest_wave_speed
  public init_U
  public primitive_to_conservative, conservative_to_primitive
  public IsPhysicallyAdmissible

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Give the initial condition according to the test case
  !>
  !> @param[out] U0 Initial condition
  !> @param[out] NbIter Number of iterations made
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine init_U(U0,NbIter)
      real(PR), dimension(:,:), intent(out) :: U0
      integer, intent(out) :: NbIter

      real(PR), dimension(dim_) :: xK
      integer :: CellK

      !$OMP SINGLE
      select case(test_case)
      case ('cst')
        speed_tele = 1.0_PR
      case ('AP')
        speed_tele = 2.0_PR
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        flush(StdErr)
        STOP
      end select
      !$OMP END SINGLE

      test_restart: if (Restarting) then
        !$OMP SINGLE
        call read_solution(U0,NbIter,T_init)
        !$OMP END SINGLE COPYPRIVATE(NbIter)

      else
        NbIter = 0
        !$OMP SINGLE
        T_init = 0._PR
        !$OMP END SINGLE

        select case(test_case)
        case ('cst')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(1,CellK) = 1._PR
            U0(2,CellK) = 1._PR
          end do
          !$OMP END DO
        case ('AP')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) < 0.5_PR) then
              U0(1,CellK) = 1._PR
              U0(2,CellK) = 1._PR
            else
              U0(1,CellK) = 5.E-3_PR
              U0(2,CellK) = 5.E-3_PR
            end if
          end do
          !$OMP END DO
        end select

      end if test_restart

  end subroutine init_U

  !--------------------------------------------------------------------!
  !> @brief
  !> Give the limit boundary value
  !>
  !> @param[in] x Coordinates of the point on the boundary
  !> @param[in] t_n Current time
  !>
  !> @return Value of the solution on the boundary at point
  !> \f$x\f$ and time \f$t^n\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function U_Dirichlet(x,t_n)
      real(PR), dimension(:), intent(in) :: x
      real(PR), intent(in) :: t_n
      real(PR), dimension(NbEq) :: U_Dirichlet

      select case(test_case)
      case ('cst')
        U_Dirichlet(:) = 1._PR
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        flush(StdErr)
        STOP
      end select

  end function U_Dirichlet

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the wave speeds waves on a interface
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal,tangent Normal and tangent to the interface
  !> @param[out] a_l, a_r  Wave speeds on the left and right
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal,tangent
      real(PR), intent(out) :: a_l, a_r

      a_l = - speed_tele
      a_r =   speed_tele

  end subroutine compute_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the largest wave speed
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal,tangent Normal and tangent to the interface
  !> @return Largest wave speed
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_greatest_wave_speed(Ul,Ur,normal,tangent)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal,tangent

      real(PR) :: compute_greatest_wave_speed

      compute_greatest_wave_speed = speed_tele

  end function compute_greatest_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the physic flux
  !>
  !> @remark
  !> For the Telegraph equations:
  !> \f${\bf F}({\bf U}) =  \left(\begin{array}{c} \vec{a}u  \\ -\vec{a}v \end{array}\right)\f$
  !>
  !> @param[in] U_K solution on cell CellK
  !> @return the physic flux on the cell CellK
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function physic_flux(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(NbEq,dim_)     :: physic_flux

      physic_flux(1,:) =  speed_tele*U_K(1)
      physic_flux(2,:) = -speed_tele*U_K(2)

  end function physic_flux

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the primitive to conservative variables
  !>
  !> @param[in] V_K primitive variables
  !> @param[out] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine primitive_to_conservative(V_K,U_K)
      real(PR), dimension(:), intent(in)     :: V_K
      real(PR), dimension(:), intent(out) :: U_K

      U_K(:) = V_K(:)

  end subroutine primitive_to_conservative

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the conservative to primitive variables
  !>
  !> @param[in] U_K conservative variables
  !> @param[out] V_K primitive variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine conservative_to_primitive(U_K,V_K)
      real(PR), dimension(:), intent(in)     :: U_K
      real(PR), dimension(:), intent(out) :: V_K

      V_K(:) = U_K(:)

  end subroutine conservative_to_primitive

  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the value of the solution on a cell is physically admissible.
  !>
  !> @param[in] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function IsPhysicallyAdmissible(U_K)
      real(PR), dimension(:), intent(in) :: U_K

      logical :: IsPhysicallyAdmissible

      IsPhysicallyAdmissible = .TRUE.

  end function IsPhysicallyAdmissible

#endif /* _PHY_TELE_ */
end module mod_physics_Telegraph
