!> Main scheme file
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_scheme

  use mod_numerics, only: PR
  use type_IO, only: StdOut
  use mod_tools_time, only: InitTimeCounters, GetTime, HMS
  use mod_tools_time, only: TimeCounters
#if defined(_FC_PAR_OMP_)
  use mod_tools_omp, only: compute_chunk
#endif /* _FC_PAR_OMP_ */
  use mod_mesh, only: read_msh, create_meshes, free_meshes
  use mod_commons_DLP , only: create_nodes_DLP
  use mod_commons_DLP , only: create_stencil_DLP, free_IntInterfaces
#if defined(_MESH_D2P_) && defined(_IO_ENABLE_VTK_)
  use mod_write_VTK, only: write_VTK_diamond, write_VTK_secondary
#endif /* _MESH_ && _IO_ENABLE_VTK_ */
  use mod_physics_var, only: test_case, model_name, T_end
#if defined(_HO_MOOD_)
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_quad, only: InitQuadratureRules
  use mod_HO_MOOD_quad, only: EndQuadratureRules
  use mod_HO_MOOD, only: InitMood
#endif /* _HO_MOOD_ */
#if defined(_PHY_M1MG_)
  use mod_physics_M1MG_var, only: Init_M1MG, Finalize_M1MG
#endif /* defined(_PHY_M1MG_) */
  use type_IO, only: out_screen
#if defined(_IO_ENABLE_VTK_)
  use type_IO, only: out_disk
#endif /* _IO_ENABLE_VTK_ */
#if defined(_IO_ENABLE_DUMP_)
  use type_IO, only: dump_disk
#endif /* _IO_ENABLE_DUMP_ */
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
  use type_recompute, only: InitCellsToUpdate, EndCellsToUpdate
#endif /* _TP_CORREC_ || _HO_MOOD_ */
  use mod_iterate, only: iterate


  implicit none

  private
  public scheme

contains

  !--------------------------------------------------------------------!
  !> @brief
  !> Call all the necessary routines to have the solution
  !>
  !> @param[in] mesh_file Full path to the mesh file
  !> @param[in] info      Informations about the run
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine scheme(mesh_file,info)
    character(len=*), intent(in) :: mesh_file, info

    real(PR) :: seconds
    integer  :: hours, minutes

    ! Read .msh file and then create the right meshes
    !====================================================================!
    TimeCounters%Mesh(1) = GetTime()
    call read_msh(trim(mesh_file))
    call create_meshes()
    TimeCounters%Mesh(2) = GetTime() - TimeCounters%Mesh(1)

    call HMS(TimeCounters%Mesh(2),hours,minutes,seconds)
    write(StdOut,'(A)')'--------------------------'
    write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')"Meshes creation          : ",&
       & hours, ' h ', minutes, ' min ', seconds, ' s'
   write(StdOut,'(A)')"-------------------------------------------------------"
#if defined(_MESH_D2P_) && defined(_IO_ENABLE_VTK_)
    ! Save all the additional meshes (diamond and secondary) in VTK files
    call write_VTK_diamond  (info)
    call write_VTK_secondary(info)
#endif /* _MESH_ && _IO_ENABLE_VTK_ */
    !====================================================================!

    ! Initialisation of time counters
    call InitTimeCounters()
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
    ! Initialisation of arrays for the detection
    call InitCellsToUpdate()
#endif /* _TP_CORREC_ || _HO_MOOD_ */

#if defined(_PHY_M1MG_)
    ! Initialize some coefficients for the multigroup M1 model for radiative transfer
    call Init_M1MG()
#endif /* defined(_PHY_M1MG_) */

    !$OMP PARALLEL

#if defined(_FC_PAR_OMP_)
    call compute_chunk()
#endif /* _FC_PAR_OMP_ */

#if defined(_HO_MOOD_)
    ! MOOD initialisation
    !====================================================================!
    !$OMP SINGLE
    TimeCounters%MOOD(1) = GetTime()
    call InitQuadratureRules()
    !$OMP END SINGLE

    call InitMOOD()

    !$OMP SINGLE
    TimeCounters%MOOD(2) = TimeCounters%MOOD(2) + GetTime() - TimeCounters%MOOD(1)
    call HMS(TimeCounters%MOOD(2),hours,minutes,seconds)
    write(StdOut,'(A)')'--------------------------'
    write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')"MOOD initialization      : ",&
       & hours, ' h ', minutes, ' min ', seconds, ' s'
    write(StdOut,'(A)') "-------------------------------------------------------"
    !$OMP END SINGLE
    !====================================================================!
#endif /* _HO_MOOD_ */

    ! Create the additional points (M) for the DLP scheme
    ! and the connectivity associated
    !====================================================================!
    !$OMP SINGLE
    TimeCounters%NodesM(1) = GetTime()
    !$OMP END SINGLE NOWAIT

    call create_nodes_DLP()
    call create_stencil_DLP()

    !$OMP SINGLE
    TimeCounters%NodesM(2) = GetTime() - TimeCounters%NodesM(1)
    call HMS(TimeCounters%NodesM(2),hours,minutes,seconds)
    write(StdOut,'(A)')'--------------------------'
    write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')"NodesM creation          : ",&
       & hours, ' h ', minutes, ' min ', seconds, ' s'
    write(StdOut,'(A)') "-------------------------------------------------------"
    !$OMP END SINGLE
    !====================================================================!


    !$OMP SINGLE
    write(StdOut,'(A)')  "======================================================="
    write(StdOut,'(A)')  '               Computing the solution...'
    write(StdOut,'(2A)') 'Physic    : ', model_name
    write(StdOut,'(2A)') 'Test case : ', trim(test_case)
#if defined(_HO_MOOD_)
    write(StdOut,'(A,I1)') 'Using MOOD with a polynomial reconstruction of degree: ', MaxDegree
#else
    write(StdOut,'(A)') 'Using a first order scheme'
#endif /* _HO_MOOD_ */
    write(StdOut,'(A,I2,A)') 'Log file will be updated every      : ', out_screen, '%'
#if defined(_IO_ENABLE_VTK_)
    write(StdOut,'(A,I2,A)') 'VTK files will be written every     : ', out_disk, '%'
#endif /* _IO_ENABLE_VTK_ */
#if defined(_IO_ENABLE_DUMP_)
    write(StdOut,'(A,I2,A)') 'Dump (restart) will be written every: ', dump_disk, '%'
#endif /* _IO_ENABLE_DUMP_ */
    write(StdOut,'(A)') "-------------------------------------------------------"

    TimeCounters%Calc(1) = GetTime()
    !$OMP END SINGLE

    call iterate(info)

    !$OMP SINGLE
    TimeCounters%Calc(2) = GetTime() - TimeCounters%Calc(1)
    !$OMP END SINGLE NOWAIT

    !$OMP END PARALLEL

    write(StdOut,'(A)') "======================================================="
    write(StdOut,'(A)') '                   Finalizing...'
    call free_meshes()
    call free_IntInterfaces()
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
    call EndCellsToUpdate()
#endif /* _TP_CORREC_ || _HO_MOOD_ */
#if defined(_HO_MOOD_)
    call EndQuadratureRules()
#endif /* _HO_MOOD_ */
#if defined(_PHY_M1MG_)
    call Finalize_M1MG()
#endif /* defined(_PHY_M1MG_) */

  end subroutine scheme

end module mod_scheme
