!> Tools to interact with the system
!>
!> See http://fortranwiki.org/fortran/show/Command-line+arguments
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_cli

  use mod_numerics, only: MaxLenStr
  use type_IO, only: StdOut, StdErr
  use type_IO, only: separator
  use type_IO, only: FS
  use type_IO, only: Output_Directory
  use cla, only: cla_char, cla_int, cla_end
  use cla, only: cla_init, cla_register, cla_help, cla_show, cla_validate, cla_key_present, cla_get
#if defined(_FC_INTEL_)
  use IFPORT, only: access
#endif /* defined(_FC_INTEL_) */
  !use mod_directory, only: CreateOutputDirectory

  implicit none

  private
  public GetArguments

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Get and check the arguments with CLA-F90 \cite claf90
  !>
  !> @param[in] Binary Name of the binary
  !> @param[out] mesh_file Path to the mesh file
  !> @param[out] test_case Name of the test case
  !> @param[out] RestartFile Path to the restart file
  !> @param[out] Restarting Indicate if we are doing a restart
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine GetArguments(Binary,mesh_file,test_case,RestartFile,Restarting)
    character(len=*), intent(in) :: Binary
    character(len=*), intent(out) ::  mesh_file, test_case, RestartFile
    logical, intent(out) :: Restarting

    write(StdOut,'(A)') 'CLI arguments:'
    write(StdOut,'(A)') separator(4)

    ! Initialize cla
    call cla_init()

    ! Register the various arguments

    ! Input
    !--------------------------------------------------------------------!
    call cla_register('-m','--mesh','path to a mesh file (.msh)',cla_char,'mesh'//FS//'tri_01.msh')
    call cla_register('-r','--restart', 'optional path to a restart file (.dat, .nc, .h5)',cla_char,'')
    call cla_register('-t','--test_case', 'name of the test case to run (see the physics module)',cla_char,'cst')
    !--------------------------------------------------------------------!

    ! Validate data and check for help options
    call cla_validate(Binary)
#if defined(_DEBUG_)
    call cla_show()
#endif /* defined(_DEBUG_) */


    ! Input
    !--------------------------------------------------------------------!
    write(StdOut,'(A)') ' * Input files:'
    call cla_get('--mesh',mesh_file)
    if (cla_key_present('--mesh')) then
      write(StdOut,'(2A)') 'Mesh file given: ', trim(mesh_file)
    else
      write(StdOut,'(2A)') 'Using default mesh file: ', trim(mesh_file)
    endif
    call CheckFile(trim(mesh_file))
    if (mesh_file(len_trim(mesh_file)-3:len_trim(mesh_file)) /=  '.msh') then
      write(StdErr,'(A,A)') "The mesh file need to .msh generated by Gmsh"
      write(StdErr,'(A)') separator(1)
      flush(StdErr)
    end if

    call cla_get('--restart',RestartFile)
    if (cla_key_present('--restart')) then
      Restarting = .TRUE.
      write(StdOut,'(A,A)') 'File given for the restart: ', trim(RestartFile)
      call CheckFile(trim(RestartFile))
      if (RestartFile(len_trim(RestartFile)-3:len_trim(RestartFile)) /=  '.dat' &
          .or. RestartFile(len_trim(RestartFile)-2:len_trim(RestartFile)) /=  '.h5' &
          .or. RestartFile(len_trim(RestartFile)-2:len_trim(RestartFile)) /=  '.nc' &
          ) then
        write(StdErr,'(A,A)') "The restart file need to be a .dat, .h5 or .nc file"
        write(StdErr,'(A)') separator(1)
        flush(StdErr)
      end if
    else
      Restarting = .FALSE.
      write(StdOut,'(A)') 'No restart file given'
    endif

    call cla_get('--test_case',test_case)
    if (cla_key_present('--test_case') .and. len(trim(test_case)) > 0) then
      write(StdOut,'(2A)') 'Test case to run: ', trim(test_case)
    else
      write(StdOut,'(2A)') 'Using default test case: ', trim(test_case)
    endif
    write(StdOut,'(A)') separator(4)
    !--------------------------------------------------------------------!

    call cla_end()

  end subroutine GetArguments

  !--------------------------------------------------------------------!
  !> @brief
  !> Check if a file exists and is readable
  !>
  !> @param[in] FileName File to test
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine CheckFile(FileName)
    character(len=*), intent(in) :: FileName
#if defined(_FC_PGI_)
    integer :: access
#endif

    if( access(FileName, 'r') /= 0 ) then
      write(StdErr,'(3A)') "The file ", "'"//trim(FileName)//"'", " does not exist or is not readable"
      write(StdErr,'(A)') separator(1)
      flush(StdErr)
      STOP
    endif
  end subroutine CheckFile

end module mod_cli
