!> Execute a command and get the result
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_execute_command

  use mod_numerics, only: PR, MaxLenStr
  use type_IO, only: StdErr, StdOut
  use type_IO, only: Output_Directory
#if defined(_FC_INTEL_)
  use IFPORT, only: SYSTEM, GETPID
#endif /* defined(_FC_INTEL_) */
  use mod_open_file, only: OpenFile

#include "../include/gccversion.h"

  implicit none

  private
  public ExecuteCommand
  !> Execute command with or without output
  interface ExecuteCommand
#if defined(__unix__) || defined(__unix) || defined(unix)
    module procedure ExecuteCommand_wo_output, ExecuteCommand_w_output
#else
    module procedure ExecuteCommand_wo_output
#endif /* defined(__unix__) || defined(__unix) || defined(unix) */
  end interface

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Execute a command
  !>
  !> @remark EXECUTE_COMMAND_LINE appears in GNU>=4.6.0
  !>
  !> @param[in] command Command to execute
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine ExecuteCommand_wo_output(command)
    character(len=*), intent(in) :: command

    integer :: IERR
#if defined(_FC_GNU_)
    integer :: status
    character(len=MaxLenStr) :: MSG
#elif defined(_FC_PGI_)
    integer :: SYSTEM, GETPID
#endif
    character(len=MaxLenStr) :: pid_str
    integer :: pid_int

    pid_int = GETPID()
    write(pid_str,'(I9)') pid_int
    pid_str = trim(adjustl(pid_str))

#if defined(_FC_GNU_) && _GNUC_VERSION_GE_(4,6,0)
    CALL EXECUTE_COMMAND_LINE(trim(command), &
      WAIT=.TRUE., EXITSTAT=IERR, CMDSTAT=status, CMDMSG=msg)
    if (IERR /= 0 .or. status /=0) then
      write(StdErr,'(2A,I4,A,A)')__FILE__,':', __LINE__, " -> Error when executing the command: ", trim(msg)
      flush(StdErr)
      STOP
    end if
#else
    IERR = SYSTEM(trim(command))
    if (IERR /= 0) then
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Error when executing the command"
      flush(StdErr)
      STOP
    end if
#endif /* defined(_FC_GNU_) */

  end subroutine ExecuteCommand_wo_output

#if defined(__unix__) || defined(__unix) || defined(unix)
  !--------------------------------------------------------------------!
  !> @brief
  !> Execute a command and get the result
  !>
  !> @remark EXECUTE_COMMAND_LINE appears in GNU>=4.6.0
  !>
  !> @param[in] command Command to execute
  !> @param[out] output Result of the command
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine ExecuteCommand_w_output(command,output)
    character(len=*), intent(in) :: command
    character(len=*), intent(out) :: output

    integer :: IERR
    integer :: TmpFileId
#if defined(_FC_GNU_)
    integer :: status
    character(len=MaxLenStr) :: MSG
#elif defined(_FC_PGI_)
    integer :: SYSTEM, GETPID
#endif
    character(len=MaxLenStr) :: pid_str
    integer :: pid_int

    pid_int = GETPID()
    write(pid_str,'(I9)') pid_int
    pid_str = trim(adjustl(pid_str))

    ! Execute the command and redirect the stdout to a temporary file
#if defined(_FC_GNU_) && _GNUC_VERSION_GE_(4,6,0)
    CALL EXECUTE_COMMAND_LINE(trim(command)//' > '//trim(Output_Directory)//'/output_'//trim(pid_str)//'.tmp', &
      WAIT=.TRUE., EXITSTAT=IERR, CMDSTAT=status, CMDMSG=msg)
    if (IERR /= 0 .or. status /=0) then
      write(StdErr,'(2A,I4,A,A)')__FILE__,':', __LINE__, " -> Error when executing the command: ", trim(msg)
      flush(StdErr)
      STOP
    end if
#else
    IERR = SYSTEM(trim(command)//' > '//trim(Output_Directory)//'/output_'//trim(pid_str)//'.tmp')
    if (IERR /= 0) then
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Error when executing the command"
      flush(StdErr)
      STOP
    end if
#endif /* defined(_FC_GNU_) */

    ! Read the temporary file
    call OpenFile(unit=TmpFileId,file=trim(Output_Directory)//'/output_'//trim(pid_str)//'.tmp',action='read',status='old')
    read(TmpFileId,'(99A)') output
    close(TmpFileId)

    ! Delete the temporary file
#if defined(_FC_GNU_) && _GNUC_VERSION_GE_(4,6,0)
    CALL EXECUTE_COMMAND_LINE('rm -f '//trim(Output_Directory)//'/output_'//trim(pid_str)//'.tmp', &
      WAIT=.TRUE., EXITSTAT=IERR, CMDSTAT=status, CMDMSG=msg)
    if (IERR /= 0 .or. status /=0) then
      write(StdErr,'(2A,I4,A,A)')__FILE__,':', __LINE__, " -> Error while deleting the temporary file: ", trim(msg)
      flush(StdErr)
      STOP
    end if
#else
    IERR = SYSTEM('rm -f '//trim(Output_Directory)//'/output_'//trim(pid_str)//'.tmp')
    if (IERR /= 0) then
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Error while deleting the temporary file"
      flush(StdErr)
      STOP
    end if
#endif /* defined(_FC_GNU_) */

  end subroutine ExecuteCommand_w_output
#endif /* defined(__unix__) || defined(__unix) || defined(unix) */

end module mod_execute_command
