!> Initialization
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_initialize

  use type_IO, only: StdOut, StdErr
  use type_IO, only: Output_Directory
#include "../include/gccversion.h"
#if defined(_FC_INTEL_)
  use IFPORT, only: GETPID, GETCWD, HOSTNM
#elif defined(_FC_GNU_) && _GNUC_VERSION_GE_(4,6,0)
  use iso_fortran_env, only: compiler_version, compiler_options
#endif /* _FC_ */
  use mod_numerics, only: MaxLenStr, PR
  use mod_physics_var, only: test_case
  use mod_read, only: RestartFile, Restarting
#if defined(_FC_PAR_OMP_)
  use mod_tools_omp, only: DisplayOMP_status
#endif /* _FC_PAR_OMP_ */
  use mod_tools_pbs, only: DisplayPBS_status
  use type_mesh, only: basename_mesh_file
  use mod_directory, only: CreateOutputDirectory
  use mod_cli, only: GetArguments

  implicit none

  private
  public Initialize

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Get and check the arguments
  !>
  !> @param[in] DATE, TIME Date and time of the last compilation
  !> @param[in] GIT_BRANCH, GIT_REVISION Revision and tag of the source code
  !> @param[out] mesh_file Full path to the mesh file
  !> @param[out] info Informations about the run
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine Initialize(DATE,TIME,mesh_file,info,GIT_BRANCH,GIT_REVISION)
      character(len=*), intent(in)  :: DATE, TIME
      character(len=*), intent(in), optional :: GIT_BRANCH, GIT_REVISION
      character(len=*), intent(out) :: mesh_file
      character(len=*), intent(out) :: info

      character(len=MaxLenStr) :: str
      integer               :: pid_int
      integer               :: len_arg
      integer               :: IERR
      integer               :: trim_left, trim_right
      integer, dimension(8) :: date_run

      write(StdOut,'(A)')"======================================================="
      write(StdOut,'(A)')"                        AdmAP2D"
      write(StdOut,'(A)')"-------------------------------------------------------"
      write(StdOut,'(A)')"Copyright © 2017-2018 Université de technologie de Troyes,"
      write(StdOut,'(A)')"Copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,"
      write(StdOut,'(A)')"Copyright © 2013-2018 CNRS,"
      write(StdOut,'(A)')"Copyright © 2013-2016 Université de Nantes."
      write(StdOut,'(A)')" "
      write(StdOut,'(A)')"This program is free software: you can redistribute it and/or modify"
      write(StdOut,'(A)')"it under the terms of the GNU General Public License as published by"
      write(StdOut,'(A)')"the Free Software Foundation, either version 3 of the License, or"
      write(StdOut,'(A)')"(at your option) any later version."
      write(StdOut,'(A)')" "
      write(StdOut,'(A)')"This program is distributed in the hope that it will be useful,"
      write(StdOut,'(A)')"but WITHOUT ANY WARRANTY; without even the implied warranty of"
      write(StdOut,'(A)')"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the"
      write(StdOut,'(A)')"GNU General Public License for more details."
      write(StdOut,'(A)')" "
      write(StdOut,'(A)')"You should have received a copy of the GNU General Public License"
      write(StdOut,'(A)')"along with this program.  If not, see <http://www.gnu.org/licenses/>."
      write(StdOut,'(A)')"-------------------------------------------------------"
#if defined(_BUILD_SYSTEM_)
      write(StdOut,'(2A)')"Built with: ", _BUILD_SYSTEM_
#endif /* defined(_BUILD_SYSTEM_) */
#if defined(_BUILD_HOST_)
      write(StdOut,'(2A)')"Compiled on: ", _BUILD_HOST_
#endif /* defined(_BUILD_HOST_) */
#if defined(_BUILD_ARCH_)
      write(StdOut,'(2A)')"Architecture: ", _BUILD_ARCH_
#endif /* defined(_BUILD_ARCH_) */
#if defined(_BUILD_USER_)
      write(StdOut,'(2A)')"By: ", _BUILD_USER_
#endif /* defined(_BUILD_USER_) */
#if defined(_FC_)
      write(StdOut,'(2A)')"Fortran compiler: ", _FC_
#endif /* defined(_FC_) */
#if defined(_CC_)
      write(StdOut,'(2A)')"C compiler: ", _CC_
#endif /* defined(_CC_) */
#if defined(_FC_INTEL_)
      write(StdOut,'(A,I4,A,I8)')"Compiler version: ", __INTEL_COMPILER, ', build on: ', __INTEL_COMPILER_BUILD_DATE
#elif defined(_FC_GNU_)
#if _GNUC_VERSION_GE_(4,6,0)
      write(StdOut,'(2A)')"Compiler version: ", compiler_version()
      write(StdOut,'(2A)')"Options : ", compiler_options()
#else
      !write(StdOut,'(2A)')"Compiler version: ",__VERSION__
      write(StdOut,'(A,1I1,A,1I1,A,1I1)')"Compiler version: ", __GNUC__, '.', __GNUC_MINOR__, '.', __GNUC_PATCHLEVEL__
#endif
#elif defined(_FC_PGI_)
      write(StdOut,'(A,I2,A,I2,A,I1)')"PGI compiler version: ", __PGIC__, '.', __PGIC_MINOR__, '.', __PGIC_PATCHLEVEL__
#endif /* _FC_ */
#if defined(_FC_PAR_OMP_)
      write(StdOut,'(A,I6)')"OpenMP specification: ", _OPENMP
#endif /* _FC_PAR_OMP_ */
      write(StdOut,'(A,I2)')"Precision for real numbers: ", PR
      if (present(GIT_BRANCH) .and. present(GIT_REVISION)) then
        write(StdOut,'(4A)')"Git branch: ", GIT_BRANCH, ', revision: ', GIT_REVISION
      end if
      write(StdOut,'(4A)')"Date of the latest compilation: ", DATE,' ', TIME
#if defined(_ENABLE_LAPACK_)
      write(StdOut,'(A)')"......................................................."
      write(StdOut,'(2A)')"LAPACK used: ", _ENABLE_LAPACK_
#endif /* _ENABLE_LAPACK_ */
      write(StdOut,'(A)')"......................................................."
#if defined(_FFLAGS_)
      write(StdOut,'(2A)')"FFLAGS  : ", trim(adjustl(_FFLAGS_))
#endif /* defined(_FFLAGS_) */
#if defined(_INCDIR_)
      write(StdOut,'(2A)')"INCLUDE : ", trim(adjustl(_INCDIR_))
#endif /* defined(_INCDIR_) */
#if defined(_LDPATH_)
      write(StdOut,'(2A)')"LDPATH  : ", trim(adjustl(_LDPATH_))
#endif /* defined(_LDPATH_) */
#if defined(_LDLAGS_)
      write(StdOut,'(2A)')"LDFLAGS : ", trim(adjustl(_LDFLAGS_))
#endif /* defined(_LDLAGS_) */
      write(StdOut,'(A)')"......................................................."
#if !defined(_FC_PGI_)
      pid_int = GETPID()
      write(str,'(I9)') pid_int
      write(StdOut,'(2A)')"PID of the run: ", trim(adjustl(str))
      IERR = HOSTNM(str)
      if (IERR == 0) then
        write(StdOut,'(2A)') "HOST          : ", trim(adjustl(str))
      else
        write(StdErr,'(A)')'Trouble to find the hostname'
        write(StdErr,'(A)')"======================================================="
        flush(StdErr)
        STOP
      end if
      IERR = GETCWD(str)
      if (IERR == 0) then
        write(StdOut,'(2A)') "PWD           : ", trim(adjustl(str))
      else
        write(StdErr,'(A)')'Trouble to define the current working directory'
        write(StdErr,'(A)')"======================================================="
        flush(StdErr)
        STOP
      end if
#endif /* !defined(_FC_PGI_) */

      ! Get the name of the program
      call get_command_argument(0, str, len_arg, IERR)
      if (IERR == 0) then
        write(StdOut,'(2A)') "Program run   : ", trim(str)
      endif
      call  date_and_time(VALUES=date_run)
      write(StdOut,'(A,I2.2,A,I2.2,A,I2.2,A,I4.4,A,I2.2,A,I2.2)') &
          "At            : ", date_run(5),':',date_run(6),':',date_run(7), &
          ", on: ", date_run(1),'/',date_run(2),'/',date_run(3)

      write(StdOut,'(A)')"......................................................."
#if defined(_FC_PAR_OMP_)
      call DisplayOMP_status()
#endif /* _FC_PAR_OMP_ */

      call DisplayPBS_status()

      call GetArguments(trim(str),mesh_file,test_case,RestartFile,Restarting)
      trim_left  = index(mesh_file,'/',BACK=.TRUE.)    + 1
      trim_right = index(mesh_file,'.msh',BACK=.TRUE.) - 1
      basename_mesh_file = mesh_file(trim_left:trim_right)

#if defined(_INFO_)
      info = _INFO_//'-'//trim(basename_mesh_file)//'-'//trim(test_case)
      Output_Directory = _IO_OUTDIR_//'/'//trim(info)
#else
      info = 'AdmAP2D-'//trim(basename_mesh_file)//'-'//trim(test_case)
      Output_Directory = './out-'//trim(info)
#endif

      call CreateOutputDirectory(Output_Directory)
      write(StdOut,'(2(A))') "Output directory: ", trim(Output_Directory)

      write(StdOut,'(A)')"======================================================="
      flush(StdOut)

  end subroutine Initialize

end module mod_initialize
