!> Wrapper to open files
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_open_file

  use mod_numerics, only: PR, MaxLenStr
  use type_IO, only: StdErr, StdOut

  implicit none

  private
  public OpenFile

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Open a file
  !>
  !> @remark newunit appears in gfortran>=4.5.0
  !>
  !> @param[in] file Full path to the file to open
  !> @param[in] action How the file is opened (read, write, readwrite)
  !> @param[in] status State of the file (old, new, replace, scratch, unknonwn)
  !> @param[out] unit Unit of the file opened
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine OpenFile(unit,file,action,status)
    character(len=*), intent(in) :: file, action, status
    integer, intent(out) :: unit

#include "../include/gccversion.h"
#if defined(_FC_GNU_) && _GNUC_VERSION_LT_(4,5,0)
    unit = getUnit()
    open(unit=unit,file=trim(file),action=trim(action),status=trim(status))
#else
    open(newunit=unit,file=trim(file),action=trim(action),status=trim(status))
#endif

  end subroutine OpenFile

#if defined(_FC_GNU_) && _GNUC_VERSION_LT_(4,5,0)
  !--------------------------------------------------------------------!
  !> @brief
  !> Find a free unit to read/write a file
  !>
  !> @remark from https://modelingguru.nasa.gov/docs/DOC-2052
  !>
  !> @return a valid unit to write a file
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  integer function getUnit()
    integer :: unit
    logical :: isOpen

    integer, parameter :: MIN_UNIT_NUMBER = 10
    integer, parameter :: MAX_UNIT_NUMBER = 99

    do unit = MIN_UNIT_NUMBER, MAX_UNIT_NUMBER
      inquire(unit = unit, opened = isOpen)
      if (.not. isOpen) then
        getUnit = unit
        return
      end if
    end do
  end function getUnit
#endif

end module mod_open_file
