!> Performs a time iteration with the right time scheme
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_Butcher_tableau

  use mod_numerics, only: PR, eps
  use mod_numerics, only: third, two_third
  use mod_numerics, only: InitMin
  use type_IO, only: StdOut, StdErr

  implicit none

#if !defined(_FC_PGI_)
  private
  public ButcherTableau
  public InitButcherTableau
  public EndButcherTableau
#endif /* !defined(_FC_PGI_) */

  !--------------------------------------------------------------------!
  !> @brief
  !> Type definition for the Butcher tableau \cite SpiRuu_2002 , \cite ShuOsh_1988 ,
  !> \cite KubYeaKet_2014
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  type type_Butcher
    !> \f$\alpha\f$ coefficients
    real(PR), dimension(:,:), allocatable :: alpha
    !> \f$\beta\f$ coefficients
    real(PR), dimension(:,:), allocatable :: beta
    !> \f$c\f$ coefficients
    real(PR), dimension(:,:), allocatable :: c
    !> \f$d\f$ coefficients
    real(PR), dimension(:), allocatable :: d
    !> CFL coefficient such that \cite GotShu_1998 :
    !> \f$ \Delta t^{SSP} \leq C \Delta t^{FE} \f$
    real(PR) :: CFL_coeff
    !> Number of stages
    integer :: NbStages
  end type type_Butcher

  !> Butcher tableau
  type(type_Butcher), save :: ButcherTableau

contains

  !--------------------------------------------------------------------------
  !> @brief
  !> Initialisation of the coefficients of the Butcher tableau with
  !> the time schemes of \cite SpiRuu_2002 , \cite ShuOsh_1988 , \cite KubYeaKet_2014
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine InitButcherTableau()
#if !defined(_TIME_RK3TVD_)

#if defined(_DEBUG_)
      real(PR) :: sum_alpha
#endif /* defined(_DEBUG_) */
      integer :: i, k, l
      !==========================================================================
      !                    Number of stages in the scheme
      !==========================================================================
#if defined (_TIME_FE_) || defined(_TIME_SSPRK11_)
      integer, parameter :: NbStages = 1
#elif defined (_TIME_SSPRK21_) || defined (_TIME_SSPRK22_)
      integer, parameter :: NbStages = 2
#elif defined (_TIME_SSPRK31_) || defined (_TIME_SSPRK32_) || defined (_TIME_SSPRK33_)
      integer, parameter :: NbStages = 3
#elif defined (_TIME_SSPRK42_) || defined (_TIME_SSPRK43_)
      integer, parameter :: NbStages = 4
#elif defined (_TIME_SSPRK52_) || defined (_TIME_SSPRK53_) || defined (_TIME_SSPRK54_)
      integer, parameter :: NbStages = 5
#elif defined (_TIME_SSPRK62_) || defined (_TIME_SSPRK63_) || defined (_TIME_SSPRK64_)
      integer, parameter :: NbStages = 6
#elif defined (_TIME_SSPRK72_) || defined (_TIME_SSPRK73_) || defined (_TIME_SSPRK74_)
      integer, parameter :: NbStages = 7
#elif defined (_TIME_SSPRK82_) || defined (_TIME_SSPRK83_) || defined (_TIME_SSPRK84_)
      integer, parameter :: NbStages = 8
#endif /* _TIME_SCHEME_ */
      !==========================================================================

      write(StdOut,'(A)') 'Initializing Butcher arrays...'

      !==========================================================================
      !                      Allocations and initialisations
      !==========================================================================
      ButcherTableau%NbStages = NbStages
      allocate(ButcherTableau%alpha(0:NbStages-1,1:NbStages))
      ButcherTableau%alpha(:,:) = 0._PR
      allocate(ButcherTableau%beta(0:NbStages-1,1:NbStages))
      ButcherTableau%beta(:,:) = 0._PR
      allocate(ButcherTableau%c(0:NbStages-1,1:NbStages))
      ButcherTableau%c(:,:) = 0._PR
      allocate(ButcherTableau%d(0:NbStages-1))
      ButcherTableau%d(:) = 0._PR
      !==========================================================================

      !==========================================================================
      !                     Definition of alpha and beta
      !==========================================================================
#if defined (_TIME_FE_) || defined(_TIME_SSPRK11_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 1._PR
#elif defined (_TIME_SSPRK21_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.5_PR

      ButcherTableau%alpha(0,2) = 0._PR
      ButcherTableau%alpha(1,2) = 1._PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.5_PR
#elif defined (_TIME_SSPRK31_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = third

      ButcherTableau%alpha(0,2) = 0._PR
      ButcherTableau%alpha(1,2) = 1._PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = third

      ButcherTableau%alpha(0,3) = 0._PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 1._PR
      ButcherTableau%beta(0,3) = 0._PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = third
#elif defined (_TIME_SSPRK22_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 1._PR

      ButcherTableau%alpha(0,2) = 0.5_PR
      ButcherTableau%alpha(1,2) = 0.5_PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.5_PR
#elif defined (_TIME_SSPRK32_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.5_PR

      ButcherTableau%alpha(0,2) = 0._PR
      ButcherTableau%alpha(1,2) = 1._PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.5_PR

      ButcherTableau%alpha(0,3) = third
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = two_third
      ButcherTableau%beta(0,3) = 0._PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = third
#elif defined (_TIME_SSPRK42_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = third

      ButcherTableau%alpha(0,2) = 0._PR
      ButcherTableau%alpha(1,2) = 1._PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = third

      ButcherTableau%alpha(0,3) = 0._PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 1._PR
      ButcherTableau%beta(0,3) = 0._PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = third

      ButcherTableau%alpha(0,4) = 0.25_PR
      ButcherTableau%alpha(1,4) = 0._PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 0.75_PR
      ButcherTableau%beta(0,4) = 0._PR
      ButcherTableau%beta(1,4) = 0._PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.25_PR
#elif defined (_TIME_SSPRK52_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.324840618151514_PR

      ButcherTableau%alpha(0,2) = 0.235593265061659_PR
      ButcherTableau%alpha(1,2) = 0.764406734938341_PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.248310356296551_PR

      ButcherTableau%alpha(0,3) = 0.174017972351526_PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 0.825982027648475_PR
      ButcherTableau%beta(0,3) = 0.108822380501601_PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 0.268312512443371_PR

      ButcherTableau%alpha(0,4) = 0.235264368870758_PR
      ButcherTableau%alpha(1,4) = 0.000058643383967_PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 0.764676987745275_PR
      ButcherTableau%beta(0,4) = 0.054392262422093_PR
      ButcherTableau%beta(1,4) = 0.000019049753098_PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.248398145385413_PR

      ButcherTableau%alpha(0,5) = 0.141720372339803_PR
      ButcherTableau%alpha(1,5) = 0.095374613155521_PR
      ButcherTableau%alpha(2,5) = 0.000311763705780_PR
      ButcherTableau%alpha(3,5) = 0._PR
      ButcherTableau%alpha(4,5) = 0.762593250798895_PR
      ButcherTableau%beta(0,5) = 0.000000180291569_PR
      ButcherTableau%beta(1,5) = 0.030981548293401_PR
      ButcherTableau%beta(2,5) = 0.000101273514903_PR
      ButcherTableau%beta(3,5) = 0._PR
      ButcherTableau%beta(4,5) = 0.247721262987686_PR
#elif defined (_TIME_SSPRK62_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.271370158498047_PR

      ButcherTableau%alpha(0,2) = 0.176902819560407_PR
      ButcherTableau%alpha(1,2) = 0.823097180439593_PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.223364012315188_PR

      ButcherTableau%alpha(0,3) = 0.015893151207488_PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 0.984106848792512_PR
      ButcherTableau%beta(0,3) = 0.174831877653527_PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 0.267057231535837_PR

      ButcherTableau%alpha(0,4) = 0.153504267159468_PR
      ButcherTableau%alpha(1,4) = 0.000003730459625_PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 0.846492002380908_PR
      ButcherTableau%beta(0,4) = 0.024088031667553_PR
      ButcherTableau%beta(1,4) = 0.000001012335420_PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.229712668853436_PR

      ButcherTableau%alpha(0,5) = 0.180356799710441_PR
      ButcherTableau%alpha(1,5) = 0.227796438692973_PR
      ButcherTableau%alpha(2,5) = 0.000004416728347_PR
      ButcherTableau%alpha(3,5) = 0._PR
      ButcherTableau%alpha(4,5) = 0.591842344868240_PR
      ButcherTableau%beta(0,5) = 0.042683362900909_PR
      ButcherTableau%beta(1,5) = 0.061817155673403_PR
      ButcherTableau%beta(2,5) = 0.000001198568272_PR
      ButcherTableau%beta(3,5) = 0._PR
      ButcherTableau%beta(4,5) = 0.160608350932750_PR

      ButcherTableau%alpha(0,6) = 0.098962308653140_PR
      ButcherTableau%alpha(1,6) = 0.000000000411893_PR
      ButcherTableau%alpha(2,6) = 0.151738038514171_PR
      ButcherTableau%alpha(3,6) = 0.019744621964792_PR
      ButcherTableau%alpha(4,6) = 0._PR
      ButcherTableau%alpha(5,6) = 0.729555030456003_PR
      ButcherTableau%beta(0,6) = 0.000000000265385_PR
      ButcherTableau%beta(1,6) = 0.000000000111776_PR
      ButcherTableau%beta(2,6) = 0.041177175561773_PR
      ButcherTableau%beta(3,6) = 0.005358101192070_PR
      ButcherTableau%beta(4,6) = 0._PR
      ButcherTableau%beta(5,6) = 0.197979464247893_PR
#elif defined (_TIME_SSPRK72_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.232788108318814_PR

      ButcherTableau%alpha(0,2) = 0.000058693366670_PR
      ButcherTableau%alpha(1,2) = 0.999941306633330_PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.232774445201016_PR

      ButcherTableau%alpha(0,3) = 0.064464637955751_PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 0.935535362044249_PR
      ButcherTableau%beta(0,3) = 0.165502968997453_PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 0.217781507195638_PR

      ButcherTableau%alpha(0,4) = 0.220086300845449_PR
      ButcherTableau%alpha(1,4) = 0.102225195325932_PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 0.677688503828619_PR
      ButcherTableau%beta(0,4) = 0.103058857428462_PR
      ButcherTableau%beta(1,4) = 0.023796809842445_PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.157757824835672_PR

      ButcherTableau%alpha(0,5) = 0.109581884408778_PR
      ButcherTableau%alpha(1,5) = 0.071060761259314_PR
      ButcherTableau%alpha(2,5) = 0.005348397939225_PR
      ButcherTableau%alpha(3,5) = 0._PR
      ButcherTableau%alpha(4,5) = 0.814008956392683_PR
      ButcherTableau%beta(0,5) = 0.030644252256650_PR
      ButcherTableau%beta(1,5) = 0.016542100189251_PR
      ButcherTableau%beta(2,5) = 0.001245043438808_PR
      ButcherTableau%beta(3,5) = 0._PR
      ButcherTableau%beta(4,5) = 0.189491605113225_PR

      ButcherTableau%alpha(0,6) = 0.155551745312158_PR
      ButcherTableau%alpha(1,6) = 0.125802445334275_PR
      ButcherTableau%alpha(2,6) = 0.078456975942849_PR
      ButcherTableau%alpha(3,6) = 0.018610166400853_PR
      ButcherTableau%alpha(4,6) = 0._PR
      ButcherTableau%alpha(5,6) = 0.621578667009864_PR
      ButcherTableau%beta(0,6) = 0.018654279847214_PR
      ButcherTableau%beta(1,6) = 0.029285313271247_PR
      ButcherTableau%beta(2,6) = 0.018263851014151_PR
      ButcherTableau%beta(3,6) = 0.004332225431953_PR
      ButcherTableau%beta(4,6) = 0._PR
      ButcherTableau%beta(5,6) = 0.144696122064557_PR

      ButcherTableau%alpha(0,7) = 0.068396686509772_PR
      ButcherTableau%alpha(1,7) = 0.054238497539825_PR
      ButcherTableau%alpha(2,7) = 0.022030341614959_PR
      ButcherTableau%alpha(3,7) = 0.004628734371986_PR
      ButcherTableau%alpha(4,7) = 0.103696237861290_PR
      ButcherTableau%alpha(5,7) = 0._PR
      ButcherTableau%alpha(6,7) = 0.747009502102168_PR
      ButcherTableau%beta(0,7) = 0.000000001040139_PR
      ButcherTableau%beta(1,7) = 0.012626077240350_PR
      ButcherTableau%beta(2,7) = 0.005128401550164_PR
      ButcherTableau%beta(3,7) = 0.001077514318365_PR
      ButcherTableau%beta(4,7) = 0.024139251051507_PR
      ButcherTableau%beta(5,7) = 0._PR
      ButcherTableau%beta(6,7) = 0.173894928890543_PR
#elif defined (_TIME_SSPRK82_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.203816348874755_PR

      ButcherTableau%alpha(0,2) = 0.000022706789062_PR
      ButcherTableau%alpha(1,2) = 0.999977293210938_PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.203811720859914_PR

      ButcherTableau%alpha(0,3) = 0.040163352005589_PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 0.959836647994411_PR
      ButcherTableau%beta(0,3) = 0.203690674361979_PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 0.195630401110404_PR

      ButcherTableau%alpha(0,4) = 0.068277481460787_PR
      ButcherTableau%alpha(1,4) = 0.000794112060878_PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 0.930928406478335_PR
      ButcherTableau%beta(0,4) = 0.036121835699129_PR
      ButcherTableau%beta(1,4) = 0.000161853020846_PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.189738428872208_PR

      ButcherTableau%alpha(0,5) = 0.203201797047416_PR
      ButcherTableau%alpha(1,5) = 0.049250734246472_PR
      ButcherTableau%alpha(2,5) = 0.028073821192189_PR
      ButcherTableau%alpha(3,5) = 0._PR
      ButcherTableau%alpha(4,5) = 0.719473647513923_PR
      ButcherTableau%beta(0,5) = 0.084801239646524_PR
      ButcherTableau%beta(1,5) = 0.010038104833517_PR
      ButcherTableau%beta(2,5) = 0.005721903734355_PR
      ButcherTableau%beta(3,5) = 0._PR
      ButcherTableau%beta(4,5) = 0.146640491947890_PR

      ButcherTableau%alpha(0,6) = 0.135232895616102_PR
      ButcherTableau%alpha(1,6) = 0.000233842652217_PR
      ButcherTableau%alpha(2,6) = 0.057212625463321_PR
      ButcherTableau%alpha(3,6) = 0.005935142172269_PR
      ButcherTableau%alpha(4,6) = 0._PR
      ButcherTableau%alpha(5,6) = 0.801385494096090_PR
      ButcherTableau%beta(0,6) = 0._PR
      ButcherTableau%beta(1,6) = 0.000047660955586_PR
      ButcherTableau%beta(2,6) = 0.011660868431473_PR
      ButcherTableau%beta(3,6) = 0.001209679007605_PR
      ButcherTableau%beta(4,6) = 0._PR
      ButcherTableau%beta(5,6) = 0.163335465447857_PR

      ButcherTableau%alpha(0,7) = 0.330663126197853_PR
      ButcherTableau%alpha(1,7) = 0._PR
      ButcherTableau%alpha(2,7) = 0._PR
      ButcherTableau%alpha(3,7) = 0.036269305920717_PR
      ButcherTableau%alpha(4,7) = 0.087351183650948_PR
      ButcherTableau%alpha(5,7) = 0._PR
      ButcherTableau%alpha(6,7) = 0.545716384230481_PR
      ButcherTableau%beta(0,7) = 0.058391312304979_PR
      ButcherTableau%beta(1,7) = 0._PR
      ButcherTableau%beta(2,7) = 0._PR
      ButcherTableau%beta(3,7) = 0.007392277508982_PR
      ButcherTableau%beta(4,7) = 0.017803599321625_PR
      ButcherTableau%beta(5,7) = 0._PR
      ButcherTableau%beta(6,7) = 0.111225920954990_PR

      ButcherTableau%alpha(0,8) = 0.072377654137843_PR
      ButcherTableau%alpha(1,8) = 0.001047652272129_PR
      ButcherTableau%alpha(2,8) = 0._PR
      ButcherTableau%alpha(3,8) = 0.005351824870836_PR
      ButcherTableau%alpha(4,8) = 0.206003319412558_PR
      ButcherTableau%alpha(5,8) = 0.025024389094107_PR
      ButcherTableau%alpha(6,8) = 0._PR
      ButcherTableau%alpha(7,8) = 0.690195160212526_PR
      ButcherTableau%beta(0,8) = 0.004977697629959_PR
      ButcherTableau%beta(1,8) = 0.000213528660996_PR
      ButcherTableau%beta(2,8) = 0._PR
      ButcherTableau%beta(3,8) = 0.001090789404991_PR
      ButcherTableau%beta(4,8) = 0.041986844418748_PR
      ButcherTableau%beta(5,8) = 0.005100379617982_PR
      ButcherTableau%beta(6,8) = 0._PR
      ButcherTableau%beta(7,8) = 0.140673057565544_PR
#elif defined (_TIME_SSPRK33_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 1._PR

      ButcherTableau%alpha(0,2) = 0.75_PR
      ButcherTableau%alpha(1,2) = 0.25_PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.25_PR

      ButcherTableau%alpha(0,3) = third
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = two_third
      ButcherTableau%beta(0,3) = 0._PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = two_third
#elif defined (_TIME_SSPRK43_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.5_PR

      ButcherTableau%alpha(0,2) = 0._PR
      ButcherTableau%alpha(1,2) = 1._PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.5_PR

      ButcherTableau%alpha(0,3) = two_third
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = third
      ButcherTableau%beta(0,3) = 0._PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 1._PR/6._PR

      ButcherTableau%alpha(0,4) = 0._PR
      ButcherTableau%alpha(1,4) = 0._PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 1._PR
      ButcherTableau%beta(0,4) = 0._PR
      ButcherTableau%beta(1,4) = 0._PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.5_PR
#elif defined (_TIME_SSPRK53_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.37726891511710_PR

      ButcherTableau%alpha(0,2) = 0._PR
      ButcherTableau%alpha(1,2) = 1._PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.37726891511710_PR

      ButcherTableau%alpha(0,3) = 0.56656131914033_PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 0.43343868085967_PR
      ButcherTableau%beta(0,3) = 0._PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 0.16352294089771_PR

      ButcherTableau%alpha(0,4) = 0.09299483444413_PR
      ButcherTableau%alpha(1,4) = 0.00002090369620_PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 0.90698426185967_PR
      ButcherTableau%beta(0,4) = 0.00071997378654_PR
      ButcherTableau%beta(1,4) = 0._PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.34217696850008_PR

      ButcherTableau%alpha(0,5) = 0.00736132260920_PR
      ButcherTableau%alpha(1,5) = 0.20127980325145_PR
      ButcherTableau%alpha(2,5) = 0.00182955389682_PR
      ButcherTableau%alpha(3,5) = 0._PR
      ButcherTableau%alpha(4,5) = 0.78952932024253_PR
      ButcherTableau%beta(0,5) = 0.00277719819460_PR
      ButcherTableau%beta(1,5) = 0.00001567934613_PR
      ButcherTableau%beta(2,5) = 0._PR
      ButcherTableau%beta(3,5) = 0._PR
      ButcherTableau%beta(4,5) = 0.29786487010104_PR
#elif defined (_TIME_SSPRK63_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.284220721334261_PR

      ButcherTableau%alpha(0,2) = 0._PR
      ButcherTableau%alpha(1,2) = 1._PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.284220721334261_PR

      ButcherTableau%alpha(0,3) = 0._PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 1._PR
      ButcherTableau%beta(0,3) = 0._PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 0.284220721334261_PR

      ButcherTableau%alpha(0,4) = 0.476769811285196_PR
      ButcherTableau%alpha(1,4) = 0.098511733286064_PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 0.424718455428740_PR
      ButcherTableau%beta(0,4) = 0._PR
      ButcherTableau%beta(1,4) = 0._PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.120713785765930_PR

      ButcherTableau%alpha(0,5) = 0._PR
      ButcherTableau%alpha(1,5) = 0._PR
      ButcherTableau%alpha(2,5) = 0._PR
      ButcherTableau%alpha(3,5) = 0._PR
      ButcherTableau%alpha(4,5) = 1._PR
      ButcherTableau%beta(0,5) = 0._PR
      ButcherTableau%beta(1,5) = 0._PR
      ButcherTableau%beta(2,5) = 0._PR
      ButcherTableau%beta(3,5) = 0._PR
      ButcherTableau%beta(4,5) = 0.284220721334261_PR

      ButcherTableau%alpha(0,6) = 0._PR
      ButcherTableau%alpha(1,6) = 0._PR
      ButcherTableau%alpha(2,6) = 0.155221702560091_PR
      ButcherTableau%alpha(3,6) = 0._PR
      ButcherTableau%alpha(4,6) = 0._PR
      ButcherTableau%alpha(5,6) = 0.844778297439909_PR
      ButcherTableau%beta(0,6) = 0._PR
      ButcherTableau%beta(1,6) = 0._PR
      ButcherTableau%beta(2,6) = 0._PR
      ButcherTableau%beta(3,6) = 0._PR
      ButcherTableau%beta(4,6) = 0._PR
      ButcherTableau%beta(5,6) = 0.240103497065900_PR
#elif defined (_TIME_SSPRK73_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.233213863663009_PR

      ButcherTableau%alpha(0,2) = 0._PR
      ButcherTableau%alpha(1,2) = 1._PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.233213863663009_PR

      ButcherTableau%alpha(0,3) = 0._PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 1._PR
      ButcherTableau%beta(0,3) = 0._PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 0.233213863663009_PR

      ButcherTableau%alpha(0,4) = 0.184962588071072_PR
      ButcherTableau%alpha(1,4) = 0._PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 0.815037411928928_PR
      ButcherTableau%beta(0,4) = 0._PR
      ButcherTableau%beta(1,4) = 0._PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.190078023865845_PR

      ButcherTableau%alpha(0,5) = 0.180718656570380_PR
      ButcherTableau%alpha(1,5) = 0.314831034403793_PR
      ButcherTableau%alpha(2,5) = 0._PR
      ButcherTableau%alpha(3,5) = 0._PR
      ButcherTableau%alpha(4,5) = 0.504450309025826_PR
      ButcherTableau%beta(0,5) = 0._PR
      ButcherTableau%beta(1,5) = 0._PR
      ButcherTableau%beta(2,5) = 0._PR
      ButcherTableau%beta(3,5) = 0._PR
      ButcherTableau%beta(4,5) = 0.117644805593912_PR

      ButcherTableau%alpha(0,6) = 0._PR
      ButcherTableau%alpha(1,6) = 0._PR
      ButcherTableau%alpha(2,6) = 0._PR
      ButcherTableau%alpha(3,6) = 0._PR
      ButcherTableau%alpha(4,6) = 0._PR
      ButcherTableau%alpha(5,6) = 1._PR
      ButcherTableau%beta(0,6) = 0._PR
      ButcherTableau%beta(1,6) = 0._PR
      ButcherTableau%beta(2,6) = 0._PR
      ButcherTableau%beta(3,6) = 0._PR
      ButcherTableau%beta(4,6) = 0._PR
      ButcherTableau%beta(5,6) = 0.233213863663009_PR

      ButcherTableau%alpha(0,7) = 0._PR
      ButcherTableau%alpha(1,7) = 0._PR
      ButcherTableau%alpha(2,7) = 0._PR
      ButcherTableau%alpha(3,7) = 0.120199000000000_PR
      ButcherTableau%alpha(4,7) = 0._PR
      ButcherTableau%alpha(5,7) = 0._PR
      ButcherTableau%alpha(6,7) = 0.879801000000000_PR
      ButcherTableau%beta(0,7) = 0._PR
      ButcherTableau%beta(1,7) = 0._PR
      ButcherTableau%beta(2,7) = 0._PR
      ButcherTableau%beta(3,7) = 0._PR
      ButcherTableau%beta(4,7) = 0._PR
      ButcherTableau%beta(5,7) = 0._PR
      ButcherTableau%beta(6,7) = 0.205181790464579_PR
#elif defined (_TIME_SSPRK83_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.195804015330143_PR

      ButcherTableau%alpha(0,2) = 0._PR
      ButcherTableau%alpha(1,2) = 1._PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.195804015330143_PR

      ButcherTableau%alpha(0,3) = 0._PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 1._PR
      ButcherTableau%beta(0,3) = 0._PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 0.195804015330143_PR

      ButcherTableau%alpha(0,4) = 0._PR
      ButcherTableau%alpha(1,4) = 0._PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 1._PR
      ButcherTableau%beta(0,4) = 0._PR
      ButcherTableau%beta(1,4) = 0._PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.195804015330143_PR

      ButcherTableau%alpha(0,5) = 0.421366967085359_PR
      ButcherTableau%alpha(1,5) = 0.005949401107575_PR
      ButcherTableau%alpha(2,5) = 0._PR
      ButcherTableau%alpha(3,5) = 0._PR
      ButcherTableau%alpha(4,5) = 0.572683631807067_PR
      ButcherTableau%beta(0,5) = 0._PR
      ButcherTableau%beta(1,5) = 0._PR
      ButcherTableau%beta(2,5) = 0._PR
      ButcherTableau%beta(3,5) = 0._PR
      ButcherTableau%beta(4,5) = 0.112133754621673_PR

      ButcherTableau%alpha(0,6) = 0._PR
      ButcherTableau%alpha(1,6) = 0.004254010666365_PR
      ButcherTableau%alpha(2,6) = 0._PR
      ButcherTableau%alpha(3,6) = 0._PR
      ButcherTableau%alpha(4,6) = 0._PR
      ButcherTableau%alpha(5,6) = 0.995745989333635_PR
      ButcherTableau%beta(0,6) = 0._PR
      ButcherTableau%beta(1,6) = 0._PR
      ButcherTableau%beta(2,6) = 0._PR
      ButcherTableau%beta(3,6) = 0._PR
      ButcherTableau%beta(4,6) = 0._PR
      ButcherTableau%beta(5,6) = 0.194971062960412_PR

      ButcherTableau%alpha(0,7) = 0._PR
      ButcherTableau%alpha(1,7) = 0._PR
      ButcherTableau%alpha(2,7) = 0.104380143093325_PR
      ButcherTableau%alpha(3,7) = 0.243265240906726_PR
      ButcherTableau%alpha(4,7) = 0._PR
      ButcherTableau%alpha(5,7) = 0._PR
      ButcherTableau%alpha(6,7) = 0.652354615999950_PR
      ButcherTableau%beta(0,7) = 0._PR
      ButcherTableau%beta(1,7) = 0._PR
      ButcherTableau%beta(2,7) = 0._PR
      ButcherTableau%beta(3,7) = 0._PR
      ButcherTableau%beta(4,7) = 0._PR
      ButcherTableau%beta(5,7) = 0._PR
      ButcherTableau%beta(6,7) = 0.127733653231944_PR

      ButcherTableau%alpha(0,8) = 0._PR
      ButcherTableau%alpha(1,8) = 0._PR
      ButcherTableau%alpha(2,8) = 0._PR
      ButcherTableau%alpha(3,8) = 0._PR
      ButcherTableau%alpha(4,8) = 0._PR
      ButcherTableau%alpha(5,8) = 0._PR
      ButcherTableau%alpha(6,8) = 0._PR
      ButcherTableau%alpha(7,8) = 1._PR
      ButcherTableau%beta(0,8) = 0._PR
      ButcherTableau%beta(1,8) = 0._PR
      ButcherTableau%beta(2,8) = 0._PR
      ButcherTableau%beta(3,8) = 0._PR
      ButcherTableau%beta(4,8) = 0._PR
      ButcherTableau%beta(5,8) = 0._PR
      ButcherTableau%beta(6,8) = 0._PR
      ButcherTableau%beta(7,8) = 0.195804015330143_PR
#elif defined (_TIME_SSPRK54_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.391752226571890_PR

      ButcherTableau%alpha(0,2) = 0.444370493651235_PR
      ButcherTableau%alpha(1,2) = 0.555629506348765_PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.368410593050371_PR

      ButcherTableau%alpha(0,3) = 0.620101851488403_PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 0.379898148511597_PR
      ButcherTableau%beta(0,3) = 0._PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 0.251891774271694_PR

      ButcherTableau%alpha(0,4) = 0.178079954393132_PR
      ButcherTableau%alpha(1,4) = 0._PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 0.821920045606868_PR
      ButcherTableau%beta(0,4) = 0._PR
      ButcherTableau%beta(1,4) = 0._PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.544974750228521_PR

      ButcherTableau%alpha(0,5) = 0._PR
      ButcherTableau%alpha(1,5) = 0._PR
      ButcherTableau%alpha(2,5) = 0.517231671970585_PR
      ButcherTableau%alpha(3,5) = 0.096059710526147_PR
      ButcherTableau%alpha(4,5) = 0.386708617503269_PR
      ButcherTableau%beta(0,5) = 0._PR
      ButcherTableau%beta(1,5) = 0._PR
      ButcherTableau%beta(2,5) = 0._PR
      ButcherTableau%beta(3,5) = 0.063692468666290_PR
      ButcherTableau%beta(4,5) = 0.226007483236906_PR
!#elif defined (_TIME_SSPRK54_)
      !ButcherTableau%alpha(0,1) = 1._PR
      !ButcherTableau%beta(0,1) = 0.605491839566400_PR

      !ButcherTableau%alpha(0,2) = 0.261216512493821_PR
      !ButcherTableau%alpha(1,2) = 0.738783487506179_PR
      !ButcherTableau%beta(0,2) = 0._PR
      !ButcherTableau%beta(1,2) = 0.447327372891397_PR

      !ButcherTableau%alpha(0,3) = 0.623613752757655_PR
      !ButcherTableau%alpha(1,3) = 0._PR
      !ButcherTableau%alpha(2,3) = 0.376386247242345_PR
      !ButcherTableau%beta(0,3) = 0.000000844149769_PR
      !ButcherTableau%beta(1,3) = 0._PR
      !ButcherTableau%beta(2,3) = 0.227898801230261_PR

      !ButcherTableau%alpha(0,4) = 0.444745181201454_PR
      !ButcherTableau%alpha(1,4) = 0.120932584902288_PR
      !ButcherTableau%alpha(2,4) = 0._PR
      !ButcherTableau%alpha(3,4) = 0.434322233896258_PR
      !ButcherTableau%beta(0,4) = 0.002856233144485_PR
      !ButcherTableau%beta(1,4) = 0.073223693296006_PR
      !ButcherTableau%beta(2,4) = 0._PR
      !ButcherTableau%beta(3,4) = 0.262978568366434_PR

      !ButcherTableau%alpha(0,5) = 0.213357715199957_PR
      !ButcherTableau%alpha(1,5) = 0.209928473023448_PR
      !ButcherTableau%alpha(2,5) = 0.063353148180384_PR
      !ButcherTableau%alpha(3,5) = 0._PR
      !ButcherTableau%alpha(4,5) = 0.513360663596212_PR
      !ButcherTableau%beta(0,5) = 0.002362549760441_PR
      !ButcherTableau%beta(1,5) = 0.127109977308333_PR
      !ButcherTableau%beta(2,5) = 0.038359814234063_PR
      !ButcherTableau%beta(3,5) = 0._PR
      !ButcherTableau%beta(4,5) = 0.310835692561898_PR
#elif defined (_TIME_SSPRK64_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.448860018455995_PR

      ButcherTableau%alpha(0,2) = 0.441581886978406_PR
      ButcherTableau%alpha(1,2) = 0.558418113021594_PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.250651564517035_PR

      ButcherTableau%alpha(0,3) = 0.496140382330059_PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 0.503859617669941_PR
      ButcherTableau%beta(0,3) = 0.004050697317371_PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 0.226162437286560_PR

      ButcherTableau%alpha(0,4) = 0.392013998230666_PR
      ButcherTableau%alpha(1,4) = 0.001687525300458_PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 0.606298476468875_PR
      ButcherTableau%beta(0,4) = 0.000000073512372_PR
      ButcherTableau%beta(1,4) = 0.000757462637509_PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.272143145337661_PR

      ButcherTableau%alpha(0,5) = 0.016884674246355_PR
      ButcherTableau%alpha(1,5) = 0.000000050328214_PR
      ButcherTableau%alpha(2,5) = 0.000018549175549_PR
      ButcherTableau%alpha(3,5) = 0._PR
      ButcherTableau%alpha(4,5) = 0.983096726249882_PR
      ButcherTableau%beta(0,5) = 0.000592927398846_PR
      ButcherTableau%beta(1,5) = 0.000000022590323_PR
      ButcherTableau%beta(2,5) = 0.000008325983279_PR
      ButcherTableau%beta(3,5) = 0._PR
      ButcherTableau%beta(4,5) = 0.441272814688551_PR

      ButcherTableau%alpha(0,6) = 0.128599802059752_PR
      ButcherTableau%alpha(1,6) = 0.150433518466544_PR
      ButcherTableau%alpha(2,6) = 0.179199506866483_PR
      ButcherTableau%alpha(3,6) = 0.173584325551242_PR
      ButcherTableau%alpha(4,6) = 0._PR
      ButcherTableau%alpha(5,6) = 0.368182847055979_PR
      ButcherTableau%beta(0,6) = 0.000000009191468_PR
      ButcherTableau%beta(1,6) = 0.067523591875293_PR
      ButcherTableau%beta(2,6) = 0.080435493959395_PR
      ButcherTableau%beta(3,6) = 0.077915063570602_PR
      ButcherTableau%beta(4,6) = 0._PR
      ButcherTableau%beta(5,6) = 0.165262559524728_PR
#elif defined (_TIME_SSPRK74_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.236998129331275_PR

      ButcherTableau%alpha(0,2) = 0.277584603405600_PR
      ButcherTableau%alpha(1,2) = 0.722415396594400_PR
      ButcherTableau%beta(0,2) = 0.001205136607466_PR
      ButcherTableau%beta(1,2) = 0.310012922173259_PR

      ButcherTableau%alpha(0,3) = 0.528403304637363_PR
      ButcherTableau%alpha(1,3) = 0.018109310473034_PR
      ButcherTableau%alpha(2,3) = 0.453487384889603_PR
      ButcherTableau%beta(0,3) = 0.000000000029361_PR
      ButcherTableau%beta(1,3) = 0.007771318668946_PR
      ButcherTableau%beta(2,3) = 0.194606801046999_PR

      ButcherTableau%alpha(0,4) = 0.363822566916605_PR
      ButcherTableau%alpha(1,4) = 0.025636760093079_PR
      ButcherTableau%alpha(2,4) = 0.000072932527637_PR
      ButcherTableau%alpha(3,4) = 0.610467740462679_PR
      ButcherTableau%beta(0,4) = 0.001612059039346_PR
      ButcherTableau%beta(1,4) = 0.011001602331536_PR
      ButcherTableau%beta(2,4) = 0.000031297818569_PR
      ButcherTableau%beta(3,4) = 0.261972390131100_PR

      ButcherTableau%alpha(0,5) = 0.080433061177282_PR
      ButcherTableau%alpha(1,5) = 0.000000001538366_PR
      ButcherTableau%alpha(2,5) = 0.000000000000020_PR
      ButcherTableau%alpha(3,5) = 0.000000000036824_PR
      ButcherTableau%alpha(4,5) = 0.919566937247508_PR
      ButcherTableau%beta(0,5) = 0.000000000027723_PR
      ButcherTableau%beta(1,5) = 0.000000000660165_PR
      ButcherTableau%beta(2,5) = 0.000000000000009_PR
      ButcherTableau%beta(3,5) = 0.000000000015802_PR
      ButcherTableau%beta(4,5) = 0.394617327778342_PR

      ButcherTableau%alpha(0,6) = 0.305416318145737_PR
      ButcherTableau%alpha(1,6) = 0.017282647045059_PR
      ButcherTableau%alpha(2,6) = 0.214348299745317_PR
      ButcherTableau%alpha(3,6) = 0.001174022148498_PR
      ButcherTableau%alpha(4,6) = 0.003799138070873_PR
      ButcherTableau%alpha(5,6) = 0.457979574844515_PR
      ButcherTableau%beta(0,6) = 0.115125889382648_PR
      ButcherTableau%beta(1,6) = 0.007416569384575_PR
      ButcherTableau%beta(2,6) = 0.091984117559200_PR
      ButcherTableau%beta(3,6) = 0.000503812679890_PR
      ButcherTableau%beta(4,6) = 0.001630338861330_PR
      ButcherTableau%beta(5,6) = 0.196534551952426_PR

      ButcherTableau%alpha(0,7) = 0.112741543203136_PR
      ButcherTableau%alpha(1,7) = 0.042888410429255_PR
      ButcherTableau%alpha(2,7) = 0.185108001868376_PR
      ButcherTableau%alpha(3,7) = 0.000003952121250_PR
      ButcherTableau%alpha(4,7) = 0.230275526732661_PR
      ButcherTableau%alpha(5,7) = 0.110240916986851_PR
      ButcherTableau%alpha(6,7) = 0.318741648658470_PR
      ButcherTableau%beta(0,7) = 0.000102167855778_PR
      ButcherTableau%beta(1,7) = 0.018404869978158_PR
      ButcherTableau%beta(2,7) = 0.079436115076445_PR
      ButcherTableau%beta(3,7) = 0.000001695989127_PR
      ButcherTableau%beta(4,7) = 0.098819030275264_PR
      ButcherTableau%beta(5,7) = 0.047308112450629_PR
      ButcherTableau%beta(6,7) = 0.136782840433305_PR
#elif defined (_TIME_SSPRK84_)
      ButcherTableau%alpha(0,1) = 1._PR
      ButcherTableau%beta(0,1) = 0.282318339066479_PR

      ButcherTableau%alpha(0,2) = 0.538569155333175_PR
      ButcherTableau%alpha(1,2) = 0.461430844666825_PR
      ButcherTableau%beta(0,2) = 0._PR
      ButcherTableau%beta(1,2) = 0.130270389660380_PR

      ButcherTableau%alpha(0,3) = 0.004485387460763_PR
      ButcherTableau%alpha(1,3) = 0._PR
      ButcherTableau%alpha(2,3) = 0.995514612539237_PR
      ButcherTableau%beta(0,3) = 0.003963092203460_PR
      ButcherTableau%beta(1,3) = 0._PR
      ButcherTableau%beta(2,3) = 0.281052031928487_PR

      ButcherTableau%alpha(0,4) = 0.164495299288580_PR
      ButcherTableau%alpha(1,4) = 0.016875060685979_PR
      ButcherTableau%alpha(2,4) = 0._PR
      ButcherTableau%alpha(3,4) = 0.818629640025440_PR
      ButcherTableau%beta(0,4) = 0.000038019518678_PR
      ButcherTableau%beta(1,4) = 0.004764139104512_PR
      ButcherTableau%beta(2,4) = 0._PR
      ButcherTableau%beta(3,4) = 0.231114160282572_PR

      ButcherTableau%alpha(0,5) = 0.426933682982668_PR
      ButcherTableau%alpha(1,5) = 0.157047028197878_PR
      ButcherTableau%alpha(2,5) = 0.023164224070770_PR
      ButcherTableau%alpha(3,5) = 0._PR
      ButcherTableau%alpha(4,5) = 0.392855064748685_PR
      ButcherTableau%beta(0,5) = 0.000019921336144_PR
      ButcherTableau%beta(1,5) = 0.044337256156151_PR
      ButcherTableau%beta(2,5) = 0.006539685265423_PR
      ButcherTableau%beta(3,5) = 0._PR
      ButcherTableau%beta(4,5) = 0.110910189373703_PR

      ButcherTableau%alpha(0,6) = 0.082083400476958_PR
      ButcherTableau%alpha(1,6) = 0.000000039091042_PR
      ButcherTableau%alpha(2,6) = 0.033974171137350_PR
      ButcherTableau%alpha(3,6) = 0.005505195713107_PR
      ButcherTableau%alpha(4,6) = 0._PR
      ButcherTableau%alpha(5,6) = 0.878437193581543_PR
      ButcherTableau%beta(0,6) = 0.000000034006679_PR
      ButcherTableau%beta(1,6) = 0.000000011036118_PR
      ButcherTableau%beta(2,6) = 0.009591531566657_PR
      ButcherTableau%beta(3,6) = 0.001554217709960_PR
      ButcherTableau%beta(4,6) = 0._PR
      ButcherTableau%beta(5,6) = 0.247998929466160_PR

      ButcherTableau%alpha(0,7) = 0.006736365648625_PR
      ButcherTableau%alpha(1,7) = 0.010581829625529_PR
      ButcherTableau%alpha(2,7) = 0.009353386191951_PR
      ButcherTableau%alpha(3,7) = 0.101886062556838_PR
      ButcherTableau%alpha(4,7) = 0.000023428364930_PR
      ButcherTableau%alpha(5,7) = 0._PR
      ButcherTableau%alpha(6,7) = 0.871418927612128_PR
      ButcherTableau%beta(0,7) = 0.013159891155054_PR
      ButcherTableau%beta(1,7) = 0.002987444564164_PR
      ButcherTableau%beta(2,7) = 0.002640632454359_PR
      ButcherTableau%beta(3,7) = 0.028764303955070_PR
      ButcherTableau%beta(4,7) = 0.000006614257074_PR
      ButcherTableau%beta(5,7) = 0._PR
      ButcherTableau%beta(6,7) = 0.246017544274548_PR

      ButcherTableau%alpha(0,8) = 0.071115287415749_PR
      ButcherTableau%alpha(1,8) = 0.018677648343953_PR
      ButcherTableau%alpha(2,8) = 0.007902408660034_PR
      ButcherTableau%alpha(3,8) = 0.319384027162348_PR
      ButcherTableau%alpha(4,8) = 0.007121989995845_PR
      ButcherTableau%alpha(5,8) = 0.001631615692736_PR
      ButcherTableau%alpha(6,8) = 0._PR
      ButcherTableau%alpha(7,8) = 0.574167022729334_PR
      ButcherTableau%beta(0,8) = 0.000000010647874_PR
      ButcherTableau%beta(1,8) = 0.005273042658132_PR
      ButcherTableau%beta(2,8) = 0.002230994887525_PR
      ButcherTableau%beta(3,8) = 0.090167968072837_PR
      ButcherTableau%beta(4,8) = 0.002010668386475_PR
      ButcherTableau%beta(5,8) = 0..000460635032368_PR
      ButcherTableau%beta(6,8) = 0._PR
      ButcherTableau%beta(7,8) = 0.162097880203691_PR
#endif /* _TIME_SCHEME_ */
      !==========================================================================

#if defined(_DEBUG_)
      !==========================================================================
      !                          Basic checking
      !==========================================================================
      do i = 1, NbStages
        sum_alpha = 0._PR
        do k = 0, i - 1
          sum_alpha = sum_alpha + ButcherTableau%alpha(k,i)

          if (ButcherTableau%alpha(k,i) < 0._PR) then
            write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, ' -> negative coefficent'
            write(StdErr,'(A,I4,A,I4,A,ES23.13)') 'i = ', i, ' ; k = ', k, ' ; alpha = ', ButcherTableau%alpha(k,i)
            flush(StdErr)
            STOP
          end if

          if (ButcherTableau%beta(k,i) < 0._PR) then
            write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, ' -> negative coefficent'
            write(StdErr,'(A,I4,A,I4,A,ES23.13)') 'i = ', i, ' ; k = ', k, ' ; beta = ', ButcherTableau%beta(k,i)
            flush(StdErr)
            STOP
          end if

          if (abs(ButcherTableau%alpha(k,i)) < eps .and. abs(ButcherTableau%beta(k,i)) > eps) then
            write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, ' -> alpha is zero where beta is not'
            write(StdErr,'(A,I4,A,I4)') 'i = ', i, ' ; k = ', k
            write(StdErr,'(A,ES23.16,A,ES23.16)') 'alpha = ', ButcherTableau%alpha(k,i), ' ; beta = ', ButcherTableau%beta(k,i)
            flush(StdErr)
            STOP
          end if
        end do

        if (abs(sum_alpha) > 1._PR + 1.E1_PR*eps .or. abs(sum_alpha) < 1._PR - 1.E1_PR*eps) then
          write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, ' -> sum of alpha not equal to one'
          write(StdErr,'(A,I4,A,ES23.16)') 'i = ', i, ' ; sum alpha = ', sum_alpha
          flush(StdErr)
          STOP
        end if

      end do
      !==========================================================================
#endif /* _DEBUG_ */

      !==========================================================================
      !                   Computations of c and d from ShuOsh_1988
      !==========================================================================
      do i = 1, NbStages
        do k = 0, i - 1
          ButcherTableau%c(k,i) = ButcherTableau%beta(k,i)

          do l = k + 1, i - 1
            ButcherTableau%c(k,i) = ButcherTableau%c(k,i) + ButcherTableau%c(k,l)*ButcherTableau%alpha(l,i)
          end do
        end do
      end do

      i = NbStages
      do k = 0, i - 1
        do l = 0, k - 1
          ButcherTableau%d(k) = ButcherTableau%d(k) + ButcherTableau%c(l,k)
        end do
      end do
      !==========================================================================

      !==========================================================================
      !                  Computations of the CFL coefficient
      !==========================================================================
      ButcherTableau%CFL_coeff = InitMin
      do i = 1, NbStages
        do k = 0, i - 1
          if (ButcherTableau%alpha(k,i) > 0 .and. ButcherTableau%beta(k,i) > 0) then
            ButcherTableau%CFL_coeff = min(ButcherTableau%CFL_coeff,ButcherTableau%alpha(k,i)/ButcherTableau%beta(k,i))
          end if
        end do
      end do
      !==========================================================================

#endif /* _TIME_SCHEME_ */

  end subroutine InitButcherTableau

  !--------------------------------------------------------------------------
  !> @brief
  !> Deallocation of the Butcher tableau
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine EndButcherTableau()
#if !defined(_TIME_RK3TVD_)
      write(StdOut,'(A)') 'Freeing Butcher arrays...'

      deallocate(ButcherTableau%alpha)
      deallocate(ButcherTableau%beta)
      deallocate(ButcherTableau%c)
      deallocate(ButcherTableau%d)
#endif /* _TIME_SCHEME_ */
  end subroutine EndButcherTableau

end module mod_Butcher_tableau
