!> Computations of time step
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_CFL

  use mod_numerics, only: PR, eps
  use mod_numerics, only: InitMin, InitMax
  use type_IO, only: StdErr

  use type_mesh, only: pmesh, BndInterfaces, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells, chunk_IntInterfaces
#endif /* _FC_PAR_OMP_ */
  use type_mesh_DLP, only: IntInterfaces, nu_one

  use mod_physics_var, only: NbEq
  use mod_physics, only: compute_greatest_wave_speed

  use mod_bnd_values, only: compute_value_bnd

#if (_HO_MOOD_MAX_DEGREE_ >= 4)
  use type_HO_MOOD, only: MaxOrder
#endif /* _HO_MOOD_MAX_DEGREE_ >= 4 */
#if defined(_FLUX_AT_) && defined(_FLUX_AT_)
  use mod_physics_Euler_AT, only: compute_u_star, compute_speed_relax
  use mod_physics_Euler_AT, only: conservative_to_relaxed
#endif /* defined(_FLUX_AT_) && defined(_FLUX_AT_) */

  use mod_Butcher_Tableau, only: ButcherTableau

  implicit none

  private
  public compute_time_step

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the time step \f$\Delta t\f$, which fulfils the proper CFL condition
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[out] t_n_p1 Next time
  !> @param[out] dt Computed time step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine compute_time_step(U_n,t_n,t_n_p1,dt)
    real(PR), dimension(:,:), intent(in) :: U_n
    real(PR), intent(in)  :: t_n
    real(PR), intent(out) :: t_n_p1, dt

#if defined(_TP_FLUX_)
#if defined(_FLUX_AT_)
    call compute_dt_AT(U_n,t_n,dt)
#else
    call compute_dt_adm(U_n,t_n,dt)
#endif /* defined(_FLUX_AT_) */
#else
    call compute_dt(U_n,t_n,dt)
#endif /* _TP_FLUX_ */

      !$OMP SINGLE
#if (_HO_MOOD_MAX_DEGREE_ >= 4)
#if defined(_TIME_SSPRK33_) || defined(_TIME_RK3TVD_)
      dt = dt**(real(MaxOrder,PR)/3._PR)
#else
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong time scheme used"
      flush(StdErr)
#endif /* _TIME_SCHEME_ */
#endif /* _HO_MOOD_MAX_DEGREE_ >= 4 */

#if !defined(_TIME_RK3TVD_)
      dt = ButcherTableau%CFL_coeff*dt
#endif /* !defined(_TIME_RK3TVD_) */

      if (abs(dt) < eps) then
        write(StdErr,'(2A,I4,A,ES23.16)')__FILE__,':', __LINE__, " -> Time step too small, dt=", dt
        flush(StdErr)
        STOP
      end if
      !$OMP END SINGLE

      t_n_p1 = t_n + dt

  end subroutine compute_time_step

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the time step \f$\Delta t\f$, which fulfils a classical CFL
  !>
  !> \f[\max_{\stackrel{K \in \mathcal{M}}{J \in \overline{{\mathcal E}_K}}}
  !>      \left(b_{KJ} \frac{\Delta t}{\delta_{KJ}}\right)\leq 1\f]
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[out] dt Computed time step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_greatest_wave_speed
  subroutine compute_dt(U_n,t_n,dt)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in)  :: t_n
      real(PR), intent(out) :: dt

#if !defined(_SRC_ENABLE_)
      real(PR), dimension(NbEq) :: sum_d
      integer :: IdxIntInt, i_K
#endif /* .not. _SRC_ENABLE_ */
      real(PR), dimension(dim_) :: eta, tau
      real(PR), dimension(NbEq) :: U_K, U_J
      real(PR), dimension(NbEq) :: delta_K
      real(PR) :: b_KJ
      integer :: CellK, CellJ
#if !defined(_SRC_ENABLE_)
      integer :: idxM
#endif /* !defined(_SRC_ENABLE_) */
      integer :: idxJ
      integer :: IdxBndInt

      !$OMP SINGLE
      dt = InitMax
      !$OMP END SINGLE

      !$OMP DO &
      !$OMP REDUCTION(MAX:dt) &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt: do CellK = 1, pmesh%NbCells
#if !defined(_SRC_ENABLE_)
        sum_d(:) = 0._PR
        do i_K=1, pmesh%elt(CellK)%NbInterfaces
          IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)
          test_bnd: if (IdxIntInt /= 0) then
            idxM = pmesh%elt(CellK)%idxM(i_K)

            loop_S_Ki: do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
              sum_d(:) = sum_d(:) + pmesh%elt(CellK)%le(i_K)*IntInterfaces(IdxIntInt)%nu(:,idxJ,idxM)
            end do loop_S_Ki

          else
            !sum_d(:) = sum_d(:) + pmesh%elt(CellK)%le(i_K)*nu_one(:)
            sum_d(:) = sum_d(:) + pmesh%elt(CellK)%le(i_K)
          end if test_bnd
        end do

        delta_K(:) = pmesh%elt(CellK)%area/sum_d(:)
#else
        ! With source term, the delta coefficients are already computed by mod_delta::compute_delta_all
        delta_K(:) = pmesh%elt(CellK)%delta_K(:)
#endif /* .not. _SRC_ENABLE_ */

        U_K(:) = U_n(:,CellK)
        loop_stencil_DLP: do idxJ=1, pmesh%elt(CellK)%NbSten
          CellJ = pmesh%elt(CellK)%stencil(idxJ)
          eta(:) = pmesh%elt(CellK)%eta(:,idxJ)
          tau(:) = pmesh%elt(CellK)%tau(:,idxJ)

          if (CellJ > 0) then
            U_J(:) = U_n(:,CellJ)
          else
            IdxBndInt = - CellJ
            U_J(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
          endif
          b_KJ = compute_greatest_wave_speed(U_K,U_J,eta,tau)

          dt = max(dt,maxval(b_KJ/delta_K(:)))
        end do loop_stencil_DLP

      end do loop_elt
      !$OMP END DO

      !$OMP SINGLE
      dt = 1._PR/dt
      !$OMP END SINGLE

  end subroutine compute_dt

#if defined(_TP_FLUX_)
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the time step  \f$\Delta t\f$, which fulfils a classical CFL
  !> (for fully admissible meshes)
  !>
  !> \f[\max_{\stackrel{K \in \mathcal{M}}{L \in {\mathcal E}_K}}
  !>      \left(b_{KL} \frac{\Delta t}{\delta_{KL}}\right)\leq 1\f]
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[out] dt Computed time step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_greatest_wave_speed
  subroutine compute_dt_adm(U_n,t_n,dt)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in)  :: t_n
      real(PR), intent(out) :: dt

!#define(_CFL_BY_ELEMENT_)

      real(PR), dimension(NbEq) :: U_K, U_L
#if defined(_CFL_BY_ELEMENT_)
      real(PR) :: delta_K
#endif /* defined(_CFL_BY_ELEMENT_) */
      real(PR) :: b_KL
      integer :: CellK, CellL, IdxIntInt, i_K

      !$OMP SINGLE
      dt = InitMax
      !$OMP END SINGLE

#if defined(_CFL_BY_ELEMENT_)
      ! Loop by elements
      !$OMP DO &
      !$OMP REDUCTION(MAX:dt) &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt: do CellK = 1, pmesh%NbCells
        U_K = U_n(:,CellK)
#if defined(_SRC_ENABLE_)
        pmesh%elt(CellK)%delta_K(:) = pmesh%elt(CellK)%Delta_x
#endif /* _SRC_ENABLE_ */
        loop_interfaces_K: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          CellL = pmesh%elt(CellK)%nghb(i_K)
          test_bnd: if (CellL /= 0) then
            U_L(:) = U_n(:,CellL)
            b_KL = compute_greatest_wave_speed(U_K,U_L,&
                pmesh%elt(CellK)%normal(:,i_K),pmesh%elt(CellK)%tangent(:,i_K))
            dt = max(dt,b_KL/pmesh%elt(CellK)%Delta_x)
          end if test_bnd
        end do loop_interfaces_K
      end do loop_elt
      !$OMP END DO

#else

      ! Loop by interfaces
      !$OMP DO &
      !$OMP REDUCTION(MAX:dt) &
      !$OMP SCHEDULE(STATIC,chunk_IntInterfaces)
      loop_int_interfaces: do IdxIntInt = 1, pmesh%NbIntInterfaces
        CellK = IntInterfaces(IdxIntInt)%nghb(1)
        CellL = IntInterfaces(IdxIntInt)%nghb(2)

        i_K = IntInterfaces(IdxIntInt)%i_Cell(1)

        U_K = U_n(:,CellK)
        U_L = U_n(:,CellL)

        b_KL = compute_greatest_wave_speed(U_K,U_L,&
            pmesh%elt(CellK)%normal(:,i_K),pmesh%elt(CellK)%tangent(:,i_K))
        dt = max(dt,b_KL/pmesh%elt(CellK)%Delta_x,b_KL/pmesh%elt(CellL)%Delta_x)
      end do loop_int_interfaces
      !$OMP END DO
#endif /* defined(_CFL_BY_ELEMENT_) */

      !$OMP SINGLE
      dt = 1._PR/dt
      !$OMP END SINGLE

  end subroutine compute_dt_adm
#endif /* _TP_FLUX_ */

#if defined(_FLUX_AT_)
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the time step  \f$\Delta t\f$, which fulfils a classical CFL
  !> (for fully AT splitting)
  !>
  !> \f[\max_{\stackrel{K \in \mathcal{M}}{L \in {\mathcal E}_K}}
  !>      \left(b_{KL} \frac{\Delta t}{\delta_{KL}}\right)\leq 1\f]
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[out] dt Computed time step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine compute_dt_AT(U_n,t_n,dt)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in)  :: t_n
      real(PR), intent(out) :: dt

      real(PR), dimension(NbEq) :: U_K, U_L
      real(PR), dimension(NbEq) :: W_K, W_L
      integer :: CellK, CellL, i_K
      real(PR), dimension(dim_) :: normal
      real(PR), save :: dt_acoustic, dt_transport
      real(PR) :: a_KL, sum_dt_transport

      !$OMP SINGLE
      dt_transport = eps
      dt_acoustic = eps
      !$OMP END SINGLE

      !$OMP DO &
      !$OMP REDUCTION(MAX:dt_acoustic) &
      !$OMP REDUCTION(MAX:dt_transport) &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt: do CellK = 1, pmesh%NbCells
        U_K(:) = U_n(:,CellK)
        sum_dt_transport = 0._PR
        loop_interfaces_K: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          CellL = pmesh%elt(CellK)%nghb(i_K)
          test_bnd: if (CellL /= 0) then
            U_L(:) = U_n(:,CellL)

            a_KL = compute_speed_relax(U_K,U_L)

#if defined(_FLUX_AT_EXEX_)
            dt_acoustic = max(dt_acoustic,pmesh%elt(CellK)%le(i_K)/pmesh%elt(CellK)%area*a_KL/U_K(1))
#endif /* defined(_FLUX_AT_EXEX_) */
            normal(:) = pmesh%elt(CellK)%normal(:,i_K)
            call conservative_to_relaxed(U_K,W_K)
            call conservative_to_relaxed(U_L,W_L)
            sum_dt_transport = sum_dt_transport + &
                pmesh%elt(CellK)%le(i_K)/pmesh%elt(CellK)%area*abs(compute_u_star(a_KL,W_K,W_L,normal))

          end if test_bnd
          dt_transport = max(dt_transport,sum_dt_transport)
        end do loop_interfaces_K
      end do loop_elt
      !$OMP END DO

      !$OMP SINGLE
      dt = min(0.5_PR/dt_acoustic,1.0_PR/dt_transport)
      !$OMP END SINGLE

  end subroutine compute_dt_AT
#endif /* _TP_FLUX_ */

end module mod_CFL
