!> Perform time iterations
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_iterate

  use mod_numerics, only: PR, MaxLenStr, eps
  use type_IO, only: StdOut, StdErr

  use mod_tools_time, only: GetTime
  use mod_tools_time, only: TimeCounters

  use mod_display_memory, only: DisplayMemoryUsage

  use type_mesh, only: pmesh, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */

  use mod_physics_var, only: NbEq, T_end, T_init
  use mod_physics, only: init_U

  use mod_CFL, only: compute_time_step
#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
  use mod_time_explicit, only: not_last_time_step
#endif /* defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_) */
  use mod_time, only: time_init, time_update, update_coefficeints, time_end

  use mod_write_output, only: write_output_init, write_output

  implicit none

  private
  public iterate

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Perform the different time steps to get the final time
  !>
  !> @param[in] info Informations about the run
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine iterate(info)
      character(len=*), intent(in) :: info

      integer :: CellK
      integer :: NbIter

      real(PR), dimension(:,:), allocatable, save :: U_n, U_n_p1
      real(PR) :: t_n, t_n_p1
      ! "save" to make the reduction in compute_CFL
      real(PR), save :: dt

      !================================================================!
      !                   Allocations
      !================================================================!
      !$OMP SINGLE
      allocate(U_n(NbEq,pmesh%NbCells))
      allocate(U_n_p1(NbEq,pmesh%NbCells))
      call time_init()
      !================================================================!

      call DisplayMemoryUsage()
      write(StdOut,'(A)') "-------------------------------------------------------"
      !$OMP END SINGLE

      !================================================================!
      !                        Initialisation
      !================================================================!
      call Init_U(U_n,NbIter)

      !$OMP SINGLE
      write(StdOut,'(A,ES23.16)') 'Initial time: ', T_init
      write(StdOut,'(A,ES23.16)') 'Final time  : ', T_end
      write(StdOut,'(A)') "-------------------------------------------------------"

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      not_last_time_step = .TRUE.
#endif /* defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_) */
      call write_output_init()
      !$OMP END SINGLE
      t_n = T_init
      !================================================================!

      !================================================================!
      !                         First iteration
      !================================================================!
      call update_coefficeints(U_n,t_n)
      call compute_time_step(U_n,t_n,t_n_p1,dt)
      call write_output(U_n,dt,t_n,NbIter,info)
      !================================================================!

      !================================================================!
      !                 Loop over the time steps
      !================================================================!
      loop_time: do while (t_n_p1 < T_end)
        call time_update(U_n,t_n,t_n_p1,dt,U_n_p1)
        NbIter = NbIter + 1
        call write_output(U_n_p1,dt,t_n_p1,NbIter,info)

        t_n = t_n_p1
        !$OMP DO SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          U_n(:,CellK) = U_n_p1(:,CellK)
        end do
        !$OMP END DO

        call update_coefficeints(U_n,t_n)
        call compute_time_step(U_n,t_n,t_n_p1,dt)
      end do loop_time
      !================================================================!

      !================================================================!
      !          Last time step to have exactly t_n_p1=T_end
      !================================================================!
      !$OMP SINGLE
      dt = T_end - t_n
#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      not_last_time_step = .FALSE.
#endif /* defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_) */
      !$OMP END SINGLE
      t_n_p1 = t_n + dt ! should be equal to T_end
#if defined(_DEBUG_)
      !$OMP SINGLE
      if (abs(T_end - t_n_p1) > eps) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong final time"
        flush(StdErr)
        STOP
      end if
      !$OMP END SINGLE
#endif /* defined(_DEBUG_) */

      call time_update(U_n,t_n,t_n_p1,dt,U_n_p1)
      NbIter = NbIter + 1
      call write_output(U_n_p1,dt,t_n_p1,NbIter,info)
      !================================================================!

      !$OMP SINGLE
      deallocate(U_n_p1)
      deallocate(U_n)
      call time_end()
      !$OMP END SINGLE NOWAIT

  end subroutine iterate
end module mod_iterate
