!> General time stuff
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_time

  ! Test if the Butcher array can be used
#if !defined(_TIME_RK3TVD_)
#define _TIME_BUTCHER_
#endif /* !defined(_TIME_RK3TVD_) */

  use mod_numerics, only: PR
  use type_mesh, only: pmesh
  use mod_physics_var, only: NbEq

#if !defined(_TIME_BUTCHER_)
  use mod_time_explicit, only: Phi
#endif /* !defined(_TIME_BUTCHER_) */
#if defined (_TIME_SSPRK22_)
  use mod_time_explicit, only: U1
#elif defined (_TIME_SSPRK33_)
  use mod_time_explicit, only: U1, U2
#elif defined (_TIME_RK3TVD_)
  use mod_time_explicit, only: U13, U2
#endif /* _TIME_SCHEME_ */

#if defined(_TIME_BUTCHER_)
  use mod_Butcher_Tableau, only: ButcherTableau
  use mod_Butcher_Tableau, only: InitButcherTableau, EndButcherTableau
  use mod_time_explicit_Butcher, only: U_But, Phi_But
  use mod_time_explicit_Butcher, only: compute_U_n_p1_Butcher
#else
  use mod_time_explicit, only: compute_U_n_p1
#endif /* defined(_TIME_BUTCHER_) */

#if defined(_FLUX_AT_)
  use mod_flux_AT, only: AT_init, AT_end
#endif /* _FLUX_AT_*/

#if !defined(_TP_FLUX_)
  use mod_coeff_DLP, only: compute_nu_all
#endif /* _TP_FLUX_ */
#if !defined(_TP_FLUX_) && defined(_SRC_ENABLE_)
  use mod_delta, only: compute_delta_all
#endif /* .not. _TP_FLUX_ && _SRC_ENABLE_ */
#if defined(_HO_MOOD_)
  use mod_HO_MOOD, only: InitMOOD_iter
#endif /* _HO_MOOD_ */
#if !defined(_TP_FLUX_)
  use mod_coeff_DLP, only: compute_nu_all
#endif /* _TP_FLUX_ */
#if !defined(_TP_FLUX_) && defined(_SRC_ENABLE_)
  use mod_delta, only: compute_delta_all
#endif /* .not. _TP_FLUX_ && _SRC_ENABLE_ */
#if defined(_HO_MOOD_)
  use mod_HO_MOOD, only: InitMOOD_iter
#endif /* _HO_MOOD_ */
#if  defined(_HO_MOOD_)
  use mod_tools_time, only: GetTime
  use mod_tools_time, only: TimeCounters
#endif /* defined(_HO_MOOD_) */

  implicit none

  private
  public time_init
  public time_update
  public update_coefficeints
  public time_end

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Initialize arrays relative to time iterations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine time_init()
      !================================================================!
      !                   Allocations
      !================================================================!
#if defined(_TIME_BUTCHER_)
      call InitButcherTableau()
      allocate(U_But(NbEq,pmesh%NbCells,0:ButcherTableau%NbStages))
      allocate(Phi_But(NbEq,pmesh%NbCells,0:ButcherTableau%NbStages-1))
#else
      allocate(Phi(NbEq,pmesh%NbCells))
#if defined (_TIME_SSPRK22_)
      allocate(U1(NbEq,pmesh%NbCells))
#elif defined (_TIME_SSPRK33_)
      allocate(U1(NbEq,pmesh%NbCells))
      allocate(U2(NbEq,pmesh%NbCells))
#elif defined (_TIME_RK3TVD_)
      allocate(U13(NbEq,pmesh%NbCells))
      allocate(U2(NbEq,pmesh%NbCells))
#endif /* _TIME_SCHEME_ */
#endif /* _TIME_BUTCHER_ */
#if defined(_FLUX_AT_)
      call AT_init()
#endif /* _FLUX_AT_*/
      !================================================================!
  end subroutine time_init

  !--------------------------------------------------------------------------
  !> @brief
  !> Performs a time iteration with the right time scheme
  !>
  !> \f[U^{n} \Rightarrow U^{n+1}\f]
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[inout]  t_n_p1, dt Next time and time step
  !> @param[out] U_n_p1 solution at next time step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine time_update(U_n,t_n,t_n_p1,dt,U_n_p1)
    real(PR), dimension(:,:), intent(in) :: U_n
    real(PR), intent(in) :: t_n
    real(PR), intent(inout) :: t_n_p1, dt
    real(PR), dimension(:,:), intent(out) :: U_n_p1

#if defined(_TIME_BUTCHER_)
    call compute_U_n_p1_Butcher(U_n,t_n,t_n_p1,dt,U_n_p1)
#else
    call compute_U_n_p1(U_n,t_n,t_n_p1,dt,U_n_p1)
#endif /* _TIME_BUTCHER_ */

  end subroutine time_update

  !--------------------------------------------------------------------------
  !> @brief
  !> Update the coefficents of the scheme
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine update_coefficeints(U_n,t_n)
    real(PR), dimension(:,:), intent(in) :: U_n
    real(PR), intent(in) :: t_n

#if defined(_HO_MOOD_)
    !$OMP SINGLE
    TimeCounters%ReComp(1) = GetTime()
    !$OMP END SINGLE NOWAIT
    call InitMOOD_iter(U_n)
    !$OMP SINGLE
    TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
    !$OMP END SINGLE NOWAIT
#endif /* _HO_MOOD_ */

#if !defined(_TP_FLUX_)
    call compute_nu_all(U_n,t_n)
#if defined(_SRC_ENABLE_)
    call compute_delta_all()
#endif /* _SRC_ENABLE_ */
#endif /* _TP_FLUX_ */

  end subroutine update_coefficeints

  !--------------------------------------------------------------------------
  !> @brief
  !> Finalize arrays relative to time iterations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine time_end()

#if defined(_TIME_BUTCHER_)
      call EndButcherTableau()
      deallocate(U_But)
      deallocate(Phi_But)
#else
      deallocate(Phi)
#if defined (_TIME_SSPRK22_)
      deallocate(U1)
#elif defined (_TIME_SSPRK33_)
      deallocate(U1)
      deallocate(U2)
#elif defined (_TIME_RK3TVD_)
      deallocate(U13)
      deallocate(U2)
#endif /* _TIME_SCHEME_ */
#endif /* _TIME_BUTCHER_ */
#if defined(_FLUX_AT_)
      call AT_end()
#endif /* _FLUX_AT_*/
  end subroutine time_end

end module mod_time
