!> Some numerics definitions and routines
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_numerics

  !use, INTRINSIC :: ISO_FORTRAN_ENV, only: REAL32, REAL64, REAL128

  implicit none

  !> Simple    precision
  !integer, parameter  :: SP = KIND(0.E0)
  !integer, parameter  :: SP = REAL32
  integer, parameter  :: SP = 4

  !> Double    precision
  !integer, parameter  :: DP = KIND(0.D0)
  !integer, parameter  :: DP = REAL64
  integer, parameter  :: DP = 8

#if (defined(_FC_GNU_) && __GNUC__>=6)
  ! Some quadruple precision are not defined with quadruple precision
  ! https://gcc.gnu.org/bugzilla/show_bug.cgi?id=54656
#if !defined(_ENABLE_LAPACK_)
  ! Quadruple precision routines does not exist
  !> Quadruple precision
  !integer, parameter  :: QP = KIND(0.Q0)
  !integer, parameter  :: QP = REAL128
  integer, parameter  :: QP = 16
#endif /* !defined(_ENABLE_LAPACK_) */
#endif /* defined(_FC_GNU_) && __GNUC__>=6 */

  !> Precision choose for the rest of the code
#if defined(_KIND_) && _KIND_ > 0
  integer, parameter  :: PR = _KIND_
#else
  integer, parameter  :: PR = DP
#endif

  real(PR), parameter :: m_two = - 2._PR
  real(PR), parameter :: m_one = - 1._PR
  real(PR), parameter :: zero  =   0._PR
  real(PR), parameter :: third =  1._PR/3._PR
  real(PR), parameter :: half  =  0.5_PR
  real(PR), parameter :: two_third = 2._PR/3._PR
  real(PR), parameter :: one   =   1._PR
  real(PR), parameter :: two   =   2._PR
  real(PR), parameter :: three =   3._PR
  real(PR), parameter :: four  =   4._PR
  real(PR), parameter :: five  =   5._PR

  real(PR), parameter :: sqrt_2 = sqrt(2._PR)
  real(PR), parameter :: sqrt_3 = sqrt(3._PR)

  !> Definition of \f$\pi\f$
  real(PR), parameter :: pi  = 4._PR*atan(one)

  !> Computing precision \f$\varepsilon\f$
  real(PR), parameter :: eps = epsilon(zero)

  !> Hugest number in the precision
  real(PR), parameter :: huge_pr = huge(zero)
  !> Tiniest number in the precision
  real(PR), parameter :: tiny_pr = tiny(zero)

  !> Initialization for maximum computations
  real(PR), parameter :: InitMax = - huge_pr
  !> Initialization for minimum computations
  real(PR), parameter :: InitMin =   huge_pr

  !> Maximum length authorized for a string
  integer, parameter :: MaxLenStr = 512

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Test if a real is numerically equal to zero
  !>
  !> @remark
  !> It's to avoid the test \f$x==0\f$ by checking
  !> \f$\vert x \vert \leq \varepsilon \f$.
  !> In double precision \f$\varepsilon = 2.220446049250313 \times  10^{-16}\f$.
  !>
  !> @param[in] x Real to test
  !> @return True if \f$\vert x \vert \leq \varepsilon \f$, else false
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function IsZero(x)
    real(PR), intent(in) :: x
    logical :: IsZero

    IsZero = .FALSE.
    if (abs(x) < eps) IsZero = .TRUE.

  end function IsZero

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the negative part of a real: \f$x^{-}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function NegPart(x)
    real(PR), intent(in) :: x
    real(PR) :: NegPart

    NegPart = half*(x - abs(x))

  end function NegPart

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the positive part of a real: \f$x^{+}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function PosPart(x)
    real(PR), intent(in) :: x
    real(PR) :: PosPart

    PosPart = half*(x + abs(x))

  end function PosPart

end module mod_numerics
