!> Tools compute the error
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_tools_error

  use mod_numerics, only: PR, eps
  use type_IO, only: StdOut
  use type_mesh, only: pmesh, dim_

#if defined(_PHY_ADV_)
  use mod_physics_advection, only: ExactSolution
  use mod_physics_advection_var, only: ExactSolutionKnown
#elif defined(_PHY_HYPHEAT_)
  use mod_physics_HypHeat, only: ExactSolution
  use mod_physics_Euler_var, only: ExactSolutionKnown
#elif defined(_PHY_EULER_)
  use mod_physics_Euler, only: ExactSolution
  use mod_physics_Euler_var, only: ExactSolutionKnown
#endif /* _PHY_ */
  use mod_physics_var, only: conservative_name
  use mod_physics_var, only: NbEq, Nbeq_str
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_int, only: compute_mean_f
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

  implicit none

  private
  public PrintError

contains

#if defined(_PHY_ADV_) || defined(_PHY_HYPHEAT_) || defined(_PHY_EULER_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute errors in different norms
  !>
  !> @param[in] t_n_p1 Current time
  !> @param[in] U_h Approximated solution at current time
  !> @param[out] error_L1, error_L2, error_Linf Error in different norms
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_error(t_n_p1,U_h,error_L1,error_L2,error_Linf)
      real(PR), intent(in) :: t_n_p1
      real(PR), dimension(:,:), intent(in) :: U_h
      real(PR), dimension(:), intent(out) :: error_L1, error_L2, error_Linf

      real(PR), dimension(NbEq) :: abs_diff
      real(PR), dimension(dim_) :: xK
      real(PR) :: area
      integer :: CellK, eq

      error_L1(:)    = 0._PR
      error_L2(:)    = 0._PR
      error_Linf(:)  = 0._PR

      do CellK=1, pmesh%NbCells
        area = pmesh%elt(CellK)%area
        xK(:) = pmesh%elt(CellK)%x(:)

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        abs_diff(:) = abs(U_h(:,CellK) - compute_mean_f(CellK,t_n_p1,MaxDegree,ExactSolution,NbEq))
#else
        abs_diff(:) = abs(U_h(:,CellK) - ExactSolution(xK,t_n_p1,NbEq))
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

        error_L1(:)   = error_L1(:)  + area*abs_diff(:)
        error_L2(:)   = error_L2(:)  + area*abs_diff(:)**2
        do eq = 1, NbEq
          error_Linf(eq) = max(error_Linf(eq), abs_diff(eq))
        end do

      end do

      error_L2(:) = sqrt(error_L2(:))

  end subroutine compute_error

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the norm of the exact solution
  !>
  !> @param[in] t_n_p1 Current time
  !> @param[out] norm_L1, norm_L2, norm_Linf Different norms
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_norm_ex(t_n_p1,norm_L1,norm_L2,norm_Linf)
      real(PR), intent(in) :: t_n_p1
      real(PR), dimension(:), intent(out) :: norm_L1, norm_L2, norm_Linf

      real(PR), dimension(NbEq) :: abs_Vec
      real(PR), dimension(dim_) :: xK
      real(PR) :: area
      integer :: CellK, eq

      norm_L1(:)   = 0._PR
      norm_L2(:)   = 0._PR
      norm_Linf(:) = 0._PR

      do CellK=1, pmesh%NbCells
        area = pmesh%elt(CellK)%area
        xK(:) = pmesh%elt(CellK)%x(:)

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        abs_Vec(:) = abs(compute_mean_f(CellK,t_n_p1,MaxDegree,ExactSolution,NbEq))
#else
        abs_Vec(:) = abs(ExactSolution(xK,t_n_p1,NbEq))
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

        norm_L1(:)   = norm_L1(:)  + area*abs_Vec(:)
        norm_L2(:)   = norm_L2(:)  + area*abs_Vec(:)**2
        do eq = 1, NbEq
          norm_Linf(eq) = max(norm_Linf(eq), abs_Vec(eq))
        end do
      end do

      norm_L2(:) = sqrt(norm_L2(:))

  end subroutine compute_norm_ex
#endif /* _PHY_ */

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the norm of a vector
  !>
  !> @param[in] Vec  Vector
  !> @param[out] norm_L1, norm_L2, norm_Linf Different norms
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_norm(Vec,norm_L1,norm_L2,norm_Linf)
      real(PR), dimension(:,:), intent(in) :: Vec
      real(PR), dimension(:), intent(out) :: norm_L1, norm_L2, norm_Linf

      real(PR), dimension(NbEq) :: abs_Vec
      real(PR) :: area
      integer :: CellK, eq

      norm_L1(:)   = 0._PR
      norm_L2(:)   = 0._PR
      norm_Linf(:) = 0._PR

      do CellK = 1, pmesh%NbCells
        area = pmesh%elt(CellK)%area
        abs_Vec(:) = abs(Vec(:,CellK))

        norm_L1(:)   = norm_L1(:)  + area*abs_Vec(:)
        norm_L2(:)   = norm_L2(:)  + area*abs_Vec(:)**2
        do eq = 1, NbEq
          norm_Linf(eq) = max(norm_Linf(eq), abs_Vec(eq))
        end do
      end do

      norm_L2(:) = sqrt(norm_L2(:))

  end subroutine compute_norm

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the relative error
  !>
  !> @param[in] error Norm of the error
  !> @param[in] norm Norm of the exact solution
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function RelativeError(error,norm)
    real(PR), dimension(:), intent(in) :: error, norm
    real(PR), dimension(NbEq) :: RelativeError

    integer :: eq

    do eq = 1, NbEq
      if (norm(eq) > eps) then
        RelativeError(eq) = error(eq)/norm(eq)
      else
        RelativeError(eq) = error(eq)
      end if
    end do

  end function RelativeError

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the mean value of the solution
  !>
  !> @param[in] U_h Solution
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function MeanValue(U_h)
    real(PR), dimension(:,:), intent(in) :: U_h
    real(PR), dimension(NbEq) :: MeanValue

    integer :: CellK

    MeanValue(:) =  0._PR

    do CellK = 1, pmesh%NbCells
      MeanValue(:) = MeanValue(:) + pmesh%elt(CellK)%area*U_h(:,CellK)
    enddo

    MeanValue(:) = MeanValue(:)/pmesh%area

  end function MeanValue

  !--------------------------------------------------------------------!
  !> @brief
  !> Print some stuff regarding to the error
  !>
  !> @param[in] NbIter Number of iterations
  !> @param[in] dt Current time step
  !> @param[in] t_n_p1 Current time
  !> @param[in] U_n_p1 Solution at current time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine PrintError(NbIter,dt,t_n_p1,U_n_p1)
      integer, intent(in) :: NbIter
      real(PR), intent(in) :: t_n_p1, dt
      real(PR), dimension(:,:), intent(in) :: U_n_p1

      real(PR), dimension(NbEq) :: norm_L1, norm_L2, norm_Linf
#if defined(_PHY_ADV_) || defined(_PHY_HYPHEAT_) || defined(_PHY_EULER_)
      real(PR), dimension(NbEq) :: error_L1, error_L2, error_Linf
#endif /* _PHY_ */

      write(StdOut,'(A,I9)')      'Iterations (n+1)         : ', NbIter
      write(StdOut,'(A,ES23.16)') 'Time step (dt)           : ', dt
      write(StdOut,'(A,ES23.16)') 'Physical time (t^{n+1})  : ', t_n_p1
      write(StdOut,'(A)')         '-------'
      write(StdOut,'(A,ES23.16)') 'Delta x                  : ', pmesh%Delta_x
      write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'min(U^{n+1})             : ', minval(U_n_p1(:,:),DIM=2)
      write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'max(U^{n+1})             : ', maxval(U_n_p1(:,:),DIM=2)
      write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'mean(U^{n+1})            : ', MeanValue(U_n_p1)
      call compute_norm(U_n_p1,norm_L1,norm_L2,norm_Linf)
      write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'norm_L1                  : ', norm_L1
      write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'norm_L2                  : ', norm_L2
      write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'norm_Li                  : ', norm_Linf
#if defined(_PHY_ADV_) || defined(_PHY_HYPHEAT_) || defined(_PHY_EULER_)
      if (ExactSolutionKnown) then
        call compute_error(t_n_p1,U_n_p1,error_L1,error_L2,error_Linf)
        call compute_norm_ex(t_n_p1,norm_L1,norm_L2,norm_Linf)
        write(StdOut,'(A)')         '-------'
        write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'Error   (L_1)            : ', error_L1(:)
        write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'R. err. (L_1)            : ', RelativeError(error_L1,norm_L1)
        write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'Error   (L_2)            : ', error_L2(:)
        write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'R. err. (L_2)            : ', RelativeError(error_L2,norm_L2)
        write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'Error   (L_i)            : ', error_Linf(:)
        write(StdOut,'(A,'//NbEq_str//'(ES23.16,1X))') 'R. err. (L_i)            : ', RelativeError(error_Linf,norm_Linf)
      endif
#endif /* _PHY_ */
      write(StdOut,'(A)')'--------------------------'

  end subroutine PrintError

end module mod_tools_error
