!> Sort routines, based on algorithm from \cite CorSteRivLei_2001 and \cite BraGolAda_1996 ,
!>
!> @remark http://www.fortran.com/qsort_c.f95 and http://jblevins.org/mirror/amiller/qsort.f90 .
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_tools_sort

  use mod_numerics, only: PR
  use mod_tools_swap, only: swap

  implicit none
  private
  public qsort

  !> Sort integers or reals
  interface qsort
    module procedure I_qsort, D_qsort
  end interface qsort

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Sort an real array \c list with quick sort algoritm
  !>
  !> @param[in] n Size of the array
  !> @param[in] list Array to sort
  !> @param[out] order Positions of the elements in the original order.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine I_qsort(n,list,order)
      integer, intent(in) :: n
      integer, intent(inout), dimension(:) :: list
      integer, intent(out), dimension(:), optional :: order

      integer :: i

      test_present: if (present(order)) then
        do i = 1, n
          order(i) = i
        end do
      end if test_present

      CALL integer_quick_sort_rec(1,n,list,order)

  end subroutine I_qsort

  !--------------------------------------------------------------------!
  !> @brief
  !> Sort a part of a real array \c list with quick sort algoritm
  !>
  !> @param[in] left_end, right_end Part of the array to sort
  !> @param[in] list Array to sort
  !> @param[out] order Positions of the elements in the original order.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_swap_mp_intswap
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_sort_mp_integer_interchange_sort
  recursive subroutine integer_quick_sort_rec(left_end,right_end,list,order)
      integer, intent(in) :: left_end, right_end
      integer, intent(inout), dimension(:) :: list
      integer, intent(out), dimension(:), optional :: order

      integer            :: i, j
      integer            :: reference
      integer, parameter :: max_simple_sort_size = 6

      IF (right_end < left_end + max_simple_sort_size) THEN
        ! Use interchange sort for small lists
        CALL integer_interchange_sort(left_end, right_end,list,order)

      ELSE
        ! Use partition ("quick") sort
        reference = list((left_end + right_end)/2)
        i = left_end - 1
        j = right_end + 1

        do while (.TRUE.)
          ! Scan list from left end until element >= reference is found
          do while (.TRUE.)
            i = i + 1
            if (list(i) >= reference) exit
          end do
          ! Scan list from right end until element <= reference is found
          do while (.TRUE.)
            j = j - 1
            IF (list(j) <= reference) exit
          end do

          IF (i < j) THEN
            ! Swap two out-of-order elements
            call swap(list(i),list(j))
            test_present: if (present(order)) then
              call swap(order(i),order(j))
            end if test_present
          ELSE IF (i == j) THEN
            i = i + 1
            EXIT
          ELSE
            EXIT
          END IF
        END DO

        IF (left_end < j) CALL integer_quick_sort_rec(left_end, j,list,order)
        IF (i < right_end) CALL integer_quick_sort_rec(i, right_end,list,order)
      END IF

  end subroutine integer_quick_sort_rec

  !--------------------------------------------------------------------!
  !> @brief
  !> Sort a part of a real array \c list with interchange sort algoritm
  !>
  !> @param[in] left_end, right_end Part of the array to sort
  !> @param[in] list Array to sort
  !> @param[out] order Positions of the elements in the original order.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_swap_mp_intswap
  subroutine integer_interchange_sort(left_end,right_end,list,order)
      integer, intent(in) :: left_end, right_end
      integer, intent(inout), dimension(:) :: list
      integer, intent(out), dimension(:), optional :: order

      integer             :: i, j

      do i = left_end, right_end - 1
        do j = i+1, right_end
          if (list(i) > list(j)) then
            call swap(list(i),list(j))
            test_present: if (present(order)) then
              call swap(order(i),order(j))
            end if test_present
          end if
        end do
      end do

  end subroutine integer_interchange_sort

  !--------------------------------------------------------------------!
  !> @brief
  !> Sort an real array \c list with quick sort algoritm
  !>
  !> @param[in] n Size of the array
  !> @param[in] list Array to sort
  !> @param[out] order Positions of the elements in the original order.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine D_qsort(n,list,order)
      integer, intent(in) :: n
      real(PR), intent(inout), dimension(:) :: list
      integer, intent(out), dimension(:), optional :: order

      integer :: i

      test_present: if (present(order)) then
        do i = 1, n
          order(i) = i
        end do
      end if test_present

      CALL double_quick_sort_rec(1,n,list,order)

  end subroutine D_qsort

  !--------------------------------------------------------------------!
  !> @brief
  !> Sort a part of a real array \c list with quick sort algoritm
  !>
  !> @param[in] left_end, right_end Part of the array to sort
  !> @param[in] list Array to sort
  !> @param[out] order Positions of the elements in the original order.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_swap_mp_realswap
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_sort_mp_double_interchange_sort
  recursive subroutine double_quick_sort_rec(left_end,right_end,list,order)
      integer, intent(in) :: left_end, right_end
      real(PR), intent(inout), dimension(:) :: list
      integer, intent(out), dimension(:), optional :: order

      integer            :: i, j
      real(PR)           :: reference
      integer, parameter :: max_simple_sort_size = 6

      IF (right_end < left_end + max_simple_sort_size) THEN
        ! Use interchange sort for small lists
        CALL double_interchange_sort(left_end, right_end,list,order)

      ELSE
        ! Use partition ("quick") sort
        reference = list((left_end + right_end)/2)
        i = left_end - 1
        j = right_end + 1

        do while (.TRUE.)
          ! Scan list from left end until element >= reference is found
          do while (.TRUE.)
            i = i + 1
            if (list(i) >= reference) exit
          end do
          ! Scan list from right end until element <= reference is found
          do while (.TRUE.)
            j = j - 1
            IF (list(j) <= reference) exit
          end do

          IF (i < j) THEN
            ! Swap two out-of-order elements
            call swap(list(i),list(j))
            test_present: if (present(order)) then
              call swap(order(i),order(j))
            end if test_present
          ELSE IF (i == j) THEN
            i = i + 1
            EXIT
          ELSE
            EXIT
          END IF
        END DO

        IF (left_end < j) CALL double_quick_sort_rec(left_end, j,list,order)
        IF (i < right_end) CALL double_quick_sort_rec(i, right_end,list,order)
      END IF

  end subroutine double_quick_sort_rec

  !--------------------------------------------------------------------!
  !> @brief
  !> Sort a part of a real array \c list with interchange sort algoritm
  !>
  !> @param[in] left_end, right_end Part of the array to sort
  !> @param[in] list Array to sort
  !> @param[out] order Positions of the elements in the original order.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_swap_mp_realswap
  subroutine double_interchange_sort(left_end,right_end,list,order)
      integer, intent(in) :: left_end, right_end
      real(PR), intent(inout), dimension(:) :: list
      integer, intent(out), dimension(:), optional :: order

      integer             :: i, j

      do i = left_end, right_end - 1
        do j = i+1, right_end
          if (list(i) > list(j)) then
            call swap(list(i),list(j))
            test_present: if (present(order)) then
              call swap(order(i),order(j))
            end if test_present
          end if
        end do
      end do

  end subroutine double_interchange_sort

end module mod_tools_sort
