!> Stuff related to time computations.
!>
!> @author Florian Blachère
!>
!> @copyright © 2017-2018 Université de technologie de Troyes,
!> @copyright © 2016-2017 Université de Versailles Saint-Quentin-en-Yvelines,
!> @copyright © 2013-2018 CNRS,
!> @copyright © 2013-2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_tools_time

  use mod_numerics, only: PR, MaxLenStr, eps
  use type_IO, only: StdOut, StdErr
  use type_IO, only: Output_Directory
  use mod_physics_var, only: T_init, T_end

  use type_IO, only: out_screen, n_out_screen
#if defined(_IO_ENABLE_VTK_)
  use type_IO, only: out_disk, n_out_disk
#endif /* _IO_ENABLE_VTK_ */
#if defined(_IO_ENABLE_DUMP_)
  use type_IO, only: dump_disk, n_dump_disk
#endif /* _IO_ENABLE_DUMP_ */
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
  use type_recompute, only: TotReComp, TotIterReComp
  use type_recompute, only: MaxInvalidCells, MaxIterReComp
  use type_recompute, only: PrintRecompStats
  use type_mesh, only: pmesh
#endif /* _TP_CORREC_ || _HO_MOOD_ */
#if defined(_TP_CORREC_) && defined(_HO_MOOD_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
  use type_recompute, only: TotReCompTP
  use type_recompute, only: MaxInvalidCellsTP
#endif /* _TP_CORREC_ && _HO_MOOD_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */

#if defined(_FC_PAR_OMP_)
  use OMP_LIB, only: OMP_GET_WTIME, OMP_GET_NUM_THREADS
#elif defined(_FC_PAR_MPI_)
  use mpi, only: MPI_WTIME
#endif /* _PARALLEL_ */

  use mod_execute_command, only: ExecuteCommand
  use mod_tools, only: Int2Str

  implicit none

  !> Type for time counters:
  !> * the first element contains the initial time,
  !> * the second element contains the final time or the total time.
  type type_TimeCounters
    !> Wall time
    real(PR), dimension(2) :: WallTime
    !> CPU Time
    real(PR), dimension(2) :: CPU
    !> Time to create the mesh
    real(PR), dimension(2) :: Mesh
    !> Time to create the additional nodes
    real(PR), dimension(2) :: NodesM
    !> Computational time
    real(PR), dimension(2) :: Calc
#if defined(_IO_ENABLE_VTK_)
    !> I/O time
    real(PR), dimension(2) :: IO
#endif /* _IO_ENABLE_VTK_ */
    !> Printing time
    real(PR), dimension(2) :: Printing
#if defined(_IO_ENABLE_DUMP_)
    !> Dumping time
    real(PR), dimension(2) :: Dump
#endif /* _IO_ENABLE_DUMP_ */
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
    !> Time for the re-computations
    real(PR), dimension(2) :: ReComp
#endif /* _TP_CORREC_ || _HO_MOOD_ */
#if defined(_MESH_RCM_)
    !> Time for the RCM algorithm
    real(PR), dimension(2) :: RCM
#endif /* _MESH_RCM_ */
#if defined(_HO_MOOD_)
    !> Time for the MOOD initialisations
    real(PR), dimension(2) :: MOOD
#endif /* _HO_MOOD_ */
  end type type_TimeCounters

  !> Contains all the time counters
  type(type_TimeCounters), save :: TimeCounters

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Initialise time counters
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine InitTimeCounters()

    write(StdOut,'(A)')'Initialising time counters...'

#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
      TimeCounters%ReComp(:) = 0._PR
#endif /* _TP_CORREC_ || _HO_MOOD_ */
#if defined(_IO_ENABLE_VTK_)
      TimeCounters%IO(:) = 0._PR
#endif /* _IO_ENABLE_VTK_ */
      TimeCounters%Printing(:) = 0._PR
#if defined(_IO_ENABLE_DUMP_)
      TimeCounters%Dump(:) = 0._PR
#endif /* _IO_ENABLE_DUMP_ */
#if defined(_HO_MOOD_)
      TimeCounters%MOOD(:) = 0._PR
#endif /* _HO_MOOD_ */

    write(StdOut,'(A)')'-------------------------------------------------------'

  end subroutine InitTimeCounters

  !--------------------------------------------------------------------------
  !> @brief
  !> Get CPU time according to the way the code is compiled.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function GetTime()
      real(PR) :: GetTime

#if defined(_FC_PAR_OMP_)
      GetTime = OMP_GET_WTIME()
#elif defined(_FC_PAR_MPI_)
      GetTime = MPI_WTIME()
#else
      call CPU_TIME(GetTime)
#endif /* _PARALLEL_ */

  end function GetTime

  !--------------------------------------------------------------------------
  !> @brief
  !> Convert a time in seconds into a more readable time
  !>
  !> @param[in] time_s Time in seconds
  !> @param[out] hours, minutes, seconds Time \c t converted in hours, minutes and seconds
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine HMS(time_s,hours,minutes,seconds)
      real(PR), intent(in)  :: time_s
      integer, intent(out)  :: hours, minutes
      real(PR), intent(out) :: seconds

      real(PR) :: t_work
      t_work = time_s

      hours = floor(t_work/3600.0_PR)
      t_work = mod(t_work,3600._PR)
      minutes = floor(t_work/60.0_PR)
      t_work = mod(t_work,60._PR)
      seconds = t_work

  end subroutine HMS

  !--------------------------------------------------------------------------
  !> @brief
  !> Print time statistics (elapsed and remaining time).
  !>
  !> @param[in] completion Completion of the run
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine Print_Time_Stats(completion)
      real(PR), intent(in) :: completion

      real(PR) :: Current_Time, Time_elapsed, total_Time, Time_remaining
      real(PR) :: seconds
      integer  :: hours, minutes
      integer, dimension(8) :: date_end
      character(len=MaxLenStr) :: str

      call  date_and_time(VALUES=date_end)
      write(StdOut,'(A,I2.2,A,I2.2,A,I2.2,A,I4.4,A,I2.2,A,I2.2)') &
          "Timestamp                : ", date_end(5),':',date_end(6),':',date_end(7), &
          ", ", date_end(1),'/',date_end(2),'/',date_end(3)

      Current_Time  = GetTime()

      Time_elapsed = Current_Time - TimeCounters%Calc(1)
      call HMS(Time_elapsed,hours,minutes,seconds)
      write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')'Time elapsed             : ',&
          & hours, ' h ', minutes, ' min ', seconds, ' s'
      write(StdOut,'(A,F6.2,A)')              'Completion               :        ',&
          & completion, '%'

      if (completion > 0._PR .and. completion < 100._PR) then
        total_Time = Time_elapsed*100._PR/completion
        call HMS(total_Time,hours,minutes,seconds)
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')'Estimated total time     : ',&
            & hours, ' h ', minutes, ' min ', seconds, ' s'

        Time_remaining = total_Time - Time_elapsed
        call HMS(Time_remaining,hours,minutes,seconds)
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')'Estimated remaining time : ',&
            & hours, ' h ', minutes, ' min ', seconds, ' s'

        ! Get the estimated date of the end
        call ExecuteCommand("date '+%T, %Y/%m/%d' --date=@$(($(date '+%s')+"//Int2Str(floor(Time_remaining))//"))",str)
        write(StdOut,'(A,A)')'Estimated end date       : ', trim(str)
      end if
      write(StdOut,'(A)')"-------------------------------------------------------"

  end subroutine Print_Time_Stats

  !--------------------------------------------------------------------------
  !> @brief
  !> Print final time statistics
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine Print_Time_End()
      real(PR) :: seconds
      integer  :: hours, minutes
      integer, dimension(8) :: date_end

      real(PR) :: T_CPU

      write(StdOut,'(A)') "======================================================="
      write(StdOut,'(A)') '                 Time statistics:'
      write(StdOut,'(A)') "-------------------------------------------------------"
      T_CPU = TimeCounters%CPU(2) - TimeCounters%CPU(1)
      write(StdOut,'(A,ES23.16,A)')"CPU time: ", T_CPU, " s"
      call HMS(T_CPU,hours,minutes,seconds)
      write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')"Or      :  ",&
         & hours, ' h ', minutes, ' min ', seconds, ' s'
#if defined(_FC_PAR_OMP_)
      !$OMP PARALLEL
      !$OMP SINGLE
      write(StdOut,'(A,I2,A)')"With    : ", OMP_GET_NUM_THREADS(), " thread(s)"
      !$OMP END SINGLE NOWAIT
      !$OMP END PARALLEL
#endif /* _FC_PAR_OMP_ */

      write(StdOut,'(A)')'--------------------------'
      write(StdOut,'(A,ES23.16,A)')"WallTime: ", TimeCounters%WallTime(2), " s"
      call HMS(TimeCounters%WallTime(2),hours,minutes,seconds)
      write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')"Or      :  ",&
         & hours, ' h ', minutes, ' min ', seconds, ' s'

      write(StdOut,'(A)')'--------------------------'
      call HMS(TimeCounters%Mesh(2),hours,minutes,seconds)
      if (TimeCounters%WallTime(2) > eps) then
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A,F6.2,A)')"Meshes creation          : ",&
          & hours, ' h ', minutes, ' min ', seconds, ' s, ', TimeCounters%Mesh(2)/TimeCounters%WallTime(2)*100._PR, ' %'
      else
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')"Meshes creation          : ",&
          & hours, ' h ', minutes, ' min ', seconds, ' s'
      endif
#if defined(_MESH_RCM_)
      call HMS(TimeCounters%RCM(2),hours,minutes,seconds)
      if (TimeCounters%Mesh(2) > eps) then
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A,F6.2,A)')"-> RCM algorithm         : ",&
          & hours, ' h ', minutes, ' min ', seconds, ' s, ', TimeCounters%RCM(2)/TimeCounters%Mesh(2)*100._PR, ' %'
      else
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')"-> RCM algorithm         : ",&
            & hours, ' h ', minutes, ' min ', seconds, ' s'
      endif
#endif /* _MESH_RCM_ */
      call HMS(TimeCounters%NodesM(2),hours,minutes,seconds)
      if (TimeCounters%WallTime(2) > eps) then
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A,F6.2,A)')"NodesM creation          : ",&
          & hours, ' h ', minutes, ' min ', seconds, ' s, ', TimeCounters%NodesM(2)/TimeCounters%WallTime(2)*100._PR, ' %'
      else
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')"NodesM creation          : ",&
          & hours, ' h ', minutes, ' min ', seconds, ' s'
      endif
      ! Remove the writing time from the computational time
      TimeCounters%Calc(2) = TimeCounters%Calc(2) - TimeCounters%Printing(2)
#if defined(_IO_ENABLE_VTK_)
      TimeCounters%Calc(2) = TimeCounters%Calc(2) - TimeCounters%IO(2)
#endif /* _IO_ENABLE_VTK_ */
#if defined(_IO_ENABLE_DUMP_)
      TimeCounters%Calc(2) = TimeCounters%Calc(2) - TimeCounters%Dump(2)
#endif /* _IO_ENABLE_DUMP_ */
      call HMS(TimeCounters%Calc(2),hours,minutes,seconds)
      write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A,F6.2,A)')"Computational time       : ",&
          & hours, ' h ', minutes, ' min ', seconds, ' s, ', TimeCounters%Calc(2)/TimeCounters%WallTime(2)*100._PR, ' %'

#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
      write(StdOut,'(A)')'--------------------------'
      call HMS(TimeCounters%ReComp(2),hours,minutes,seconds)
      if (TimeCounters%Calc(2) > eps) then
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A,F6.2,A)')'Time spend in re-comp    : ',&
          & hours, ' h ', minutes, ' min ', seconds, ' s, ', TimeCounters%ReComp(2)/TimeCounters%Calc(2)*100._PR, ' %'
      else
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')'Time spend in re-comp    : ',&
          & hours, ' h ', minutes, ' min ', seconds, ' s'
      endif
#endif /* _TP_CORREC_ || _HO_MOOD_ */

#if defined(_HO_MOOD_)
      write(StdOut,'(A)')'--------------------------'
      call HMS(TimeCounters%MOOD(2),hours,minutes,seconds)
      if (TimeCounters%Calc(2) > eps) then
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A,F6.2,A)')'MOOD computations        : ',&
          & hours, ' h ', minutes, ' min ', seconds, ' s, ', TimeCounters%MOOD(2)/TimeCounters%Calc(2)*100._PR, ' %'
      else
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')'MOOD computations        : ',&
          & hours, ' h ', minutes, ' min ', seconds, ' s'
      endif
#endif /* _HO_MOOD_ */

      write(StdOut,'(A)')'--------------------------'
#if defined(_IO_ENABLE_VTK_)
      ! The initial iteration is not in n_out_disk
      call HMS(TimeCounters%IO(2),hours,minutes,seconds)
      write(StdOut,'(A,I9)')      'Total number of I/O      : ', n_out_disk/out_disk + 1
      if (TimeCounters%WallTime(2) > eps) then
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A,F6.2,A)')'Time spend in I/O        : ',&
          & hours, ' h ', minutes, ' min ', seconds, ' s, ', TimeCounters%IO(2)/TimeCounters%WallTime(2)*100._PR, ' %'
      else
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A,F6.2,A)')'Time spend in I/O        : ',&
          & hours, ' h ', minutes, ' min ', seconds, ' s'
      endif
#endif /* _IO_ENABLE_VTK_ */
      ! The initial iteration is not in n_out_screen
      call HMS(TimeCounters%Printing(2),hours,minutes,seconds)
      write(StdOut,'(A,I9)')      'Total number of print    : ', n_out_screen/out_screen + 1
      if (TimeCounters%WallTime(2) > eps) then
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A,F6.2,A)')'Time spend in printing   : ',&
          & hours, ' h ', minutes, ' min ', seconds, ' s, ', TimeCounters%Printing(2)/TimeCounters%WallTime(2)*100._PR, ' %'
      else
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')'Time spend in printing   : ',&
          & hours, ' h ', minutes, ' min ', seconds, ' s'
      endif
#if defined(_IO_ENABLE_DUMP_)
      call HMS(TimeCounters%Dump(2),hours,minutes,seconds)
      write(StdOut,'(A,I9)')      'Total number of dumps    : ', n_dump_disk/dump_disk
      if (TimeCounters%WallTime(2) > eps) then
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A,F6.2,A)')'Time spend in dumps      : ',&
          & hours, ' h ', minutes, ' min ', seconds, ' s, ', TimeCounters%Dump(2)/TimeCounters%WallTime(2)*100._PR, ' %'
      else
        write(StdOut,'(A,I4.4,A,I2.2,A,F5.2,A)')'Time spend in dumps      : ',&
          & hours, ' h ', minutes, ' min ', seconds, ' s'
      endif
#endif /* _IO_ENABLE_DUMP_ */

      call  date_and_time(VALUES=date_end)
      write(StdOut,'(A)') '--------------------------'
      write(StdOut,'(A,I2.2,A,I2.2,A,I2.2,A,I4.4,A,I2.2,A,I2.2)') &
          "Terminated at            : ", date_end(5),':',date_end(6),':',date_end(7), &
          ", on: ", date_end(1),'/',date_end(2),'/',date_end(3)

      write(StdOut,'(A)')"======================================================="
      flush(StdOut)

  end subroutine Print_Time_End

end module mod_tools_time
